/*
    pixilang_utils.cpp
    This file is part of the Pixilang programming language.

    [ MIT license ]

    Copyright (c) 2006 - 2012, Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
    
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to 
    deal in the Software without restriction, including without limitation the 
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in 
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
*/

#include "pixilang.h"

UTF8_CHAR* pix_get_base_path( const UTF8_CHAR* src_name )
{
    int i;
    for( i = (int)mem_strlen( src_name ) - 1; i >= 0; i-- )
    {
	if( src_name[ i ] == '/' ) break;
    }
    if( i <= 0 )
    {
	UTF8_CHAR* rv = (UTF8_CHAR*)MEM_NEW2( 1 * sizeof( UTF8_CHAR ) );
	rv[ 0 ] = 0;
	return rv;
    }
    else 
    {
	UTF8_CHAR* rv = (UTF8_CHAR*)MEM_NEW2( ( i + 2 ) * sizeof( UTF8_CHAR ) );
	mem_copy( rv, src_name, ( i + 1 ) * sizeof( UTF8_CHAR ) );
	rv[ i + 1 ] = 0;
	return rv;
    }
}

UTF8_CHAR* pix_compose_full_path( UTF8_CHAR* base_path, UTF8_CHAR* file_name )
{
    UTF8_CHAR* new_name;
    int name_len = (int)mem_strlen( file_name );
    if( ( name_len > 1 && file_name[ 0 ] == '/' ) ||
       ( name_len > 2 && file_name[ 1 ] == ':' ) )
    {
	//File name with absolute path:
	new_name = (UTF8_CHAR*)MEM_NEW2( name_len + 1 );
	mem_copy( new_name, file_name, name_len + 1 );
    }
    else 
    {
	if( base_path == 0 || base_path[ 0 ] == 0 )
	{
	    //Base path is empty:
	    new_name = (UTF8_CHAR*)MEM_NEW2( name_len + 1 );
	    mem_copy( new_name, file_name, name_len + 1 );
	}
	else 
	{
	    size_t len = mem_strlen( base_path );
	    new_name = (UTF8_CHAR*)MEM_NEW2( len + 1 + name_len + 1 );
	    new_name[ 0 ] = 0;
	    mem_strcat( new_name, base_path );
	    if( base_path[ len - 1 ] != '/' )
		mem_strcat( new_name, "/" );
	    mem_strcat( new_name, file_name );
	}
    }    
    return new_name;
}
