#ifndef __CORE__
#define __CORE__

#define SUNDOG_VERSION "SunDog Engine"
#define SUNDOG_DATE    __DATE__

//#define COLOR8BITS	//8 bit color
//#define COLOR16BITS	//16 bit color
//#define COLOR32BITS	//32 bit color
//#define RGB565	//Default for 16bit color
//#define RGB555	//16bit color in Windows GDI
//#define RGB556	//16bit color in OSX Quartz
//#define GRAYSCALE	//Used with COLOR8BITS for grayscale palette
//#define UNIX		//OS = some UNIX veriation (Linux, FreeBSD...)
//#define FREEBSD	//OS = FreeBSD
//#define LINUX		//OS = Linux
//#define WIN		//OS = Windows
//#define WINCE		//OS = WindowsCE
//#define PALMOS	//OS = PalmOS
//#define OSX		//OS = OSX
//#define IPHONE	//OS = iPhone OS
//#define X11		//X11 support
//#define OPENGL	//OpenGL support
//#define OPENGLES	//OpenGL ES
//#define GDI		//GDI support
//#define DIRECTDRAW	//Win32: DirectDraw; PalmOS: direct write to screen; WinCE: GAPI
//#define FRAMEBUFFER	//Use framebuffer
//#define NOSTORAGE	//PalmOS: do not use the Storage Heap and MemSemaphores
//#define PALMLOWRES	//PalmOS: low-density screen
//#define ONLY44100	//Sampling frequencies other then 44100 not allowed
//#define NODEBUG	//No debug messages
//#define CPUMARK	//0..10: 0 - slow CPU (e.g. 100 MHz ARM without FPU); 10 - fast CPU (e.g. Core2Duo).
//#define HEAPSIZE	//Average memory size available for application (in Mb).

//Variations:
//WIN32:  COLOR32BITS + WIN [ + OPENGL / DIRECTDRAW / GDI + RGB555 / FRAMEBUFFER ]
//LINUX:  COLOR32BITS + LINUX [ + OPENGL / DIRECTDRAW / X11 / FRAMEBUFFER ]
//PALMOS: COLOR8BITS + PALMOS [ + NOSTORAGE / PALMLOWRES / SLOWMODE / DIRECTDRAW ]
//WINCE:  COLOR16BITS + WINCE [ + GDI + RGB555 / FRAMEBUFFER / DIRECTDRAW ]

//Examples:
//WIN32:  ARCH_X86, WIN, COLOR32BITS, GDI
//LINUX:  ARCH_X86, LINUX, COLOR32BITS, X11
//PALMOS: ARCH_ARM, PALMOS, COLOR8BITS, PALMLOWRES
//PALMOS: ARCH_ARM, PALMOS, COLOR8BITS, PALMLOWRES, FRAMEBUFFER
//PALMOS: ARCH_ARM, PALMOS, COLOR8BITS, DIRECTDRAW
//WINCE:  ARCH_ARM, WINCE, COLOR16BITS, GDI, RGB555
//WINCE:  ARCH_ARM, WINCE, COLOR16BITS, DIRECTDRAW

typedef unsigned char   	uchar;
typedef unsigned short  	uint16;
typedef signed short    	int16;
typedef unsigned int    	uint;
typedef unsigned long long 	uint64;
typedef long long		int64;

typedef char			UTF8_CHAR;
typedef uint16	        	UTF16_CHAR;
typedef uint	        	UTF32_CHAR;

#ifdef COLOR16BITS
    #ifndef RGB556
	#ifndef RGB555
	    #ifndef RGB565
		#define RGB565
	    #endif
	#endif
    #endif
#endif

#ifdef DIRECTDRAW
    #define FRAMEBUFFER
#endif

#if defined(LINUX) || defined(FREEBSD) || defined(OSX) || defined(IPHONE)
    #define UNIX
#endif

#ifdef MAEMO
    #define OS_NAME "maemo linux"
    #ifdef OPENGL
	#define OPENGLES
    #endif
    #define HEAPSIZE 16
    #define ONLY44100
#endif

#ifdef MEEGO
    #define OS_NAME "meego linux"
    #ifdef OPENGL
	#define OPENGLES
    #endif
    #define HEAPSIZE 16
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define NO_MATH
#endif

#ifdef DINGUX
    #define OS_NAME "dingux linux"
    #define HEAPSIZE 16
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define NOMIDI
#endif

#ifdef ANDROID
    #define OS_NAME "android linux"
    #ifdef OPENGL
	#define OPENGLES
    #endif
    #define HEAPSIZE 16
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define MULTITOUCH
    #define NOMIDI
#endif

#ifdef LINUX
    #ifndef OS_NAME
	#define OS_NAME "linux"
	#define HEAPSIZE 256
    #endif
#endif

#ifdef FREEBSD
    #define OS_NAME "freebsd"
    #define HEAPSIZE 256
#endif

#ifdef OSX
    #define OS_NAME "osx"
    #define HEAPSIZE 256
    #define HANDLE_THREAD_EVENTS CFRunLoopRunInMode( kCFRunLoopDefaultMode, 0, 1 )
    #if defined(__amd64__) || defined(__amd64) || defined(__x86_64__) || defined(__x86_64)
        #define ARCH_NAME "x86_64"
    #endif
    #if defined(i386) || defined(__i386) || defined(__i386__) || defined(__i486__) || defined(__i586__) || defined(__i686__)
        #define ARCH_NAME "x86"
    #endif
#endif

#ifdef IPHONE
    #define OS_NAME "ios"
    #ifdef OPENGL
	#define OPENGLES
    #endif
    #define HEAPSIZE 16
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define MULTITOUCH
    #define HANDLE_THREAD_EVENTS CFRunLoopRunInMode( kCFRunLoopDefaultMode, 0, 1 )
#endif

#ifdef WIN
    #define OS_NAME "win32"
    #define HEAPSIZE 256
#endif

#ifdef WINCE
    #define OS_NAME "wince"
    #define HEAPSIZE 16
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define NOMIDI
#endif

#ifdef PALMOS
    #define OS_NAME "palmos"
    #ifdef DIRECTDRAW
	#undef PALMLOWRES
    #endif
    #define HEAPSIZE 2
    #define ONLY44100
    #define VIRTUALKEYBOARD
    #define NOMIDI
    #include "palm_functions.h"
#else
    #define NONPALM
    #include <stdio.h>
#endif

#ifndef HANDLE_THREAD_EVENTS
    #define HANDLE_THREAD_EVENTS /**/
#endif

#ifndef OS_NAME
    #error OS_NAME not defined
#endif

#ifndef ARCH_NAME
    #error ARCH_NAME not defined
#endif

#ifndef CPUMARK
    #error CPUMARK not defined
#endif

#ifndef HEAPSIZE
    #error HEAPSIZE not defined
#endif

#endif
