/*
    main.cpp. SunDog Engine main()
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../user_code.h"

//################################
//## DEVICE VARIABLES:          ##
//################################

//PALMOS
#ifdef PALMOS
    #define arm_startup __attribute__ ((section ("arm_startup")))
#endif

//################################
//################################
//################################

//################################
//## APPLICATION MAIN:          ##
//################################

window_manager g_wmanager;
int g_argc = 0;
UTF8_CHAR* g_argv[ 64 ];

MAIN_GLOBALS;

#if defined(WIN) || defined(WINCE)
void make_arguments( UTF8_CHAR* cmd_line )
{
    //Make standart argc and argv[] from windows lpszCmdLine:
    g_argv[ 0 ] = (UTF8_CHAR*)"prog";
    g_argc = 1;
    if( cmd_line && cmd_line[ 0 ] != 0 )
    {
        int str_ptr = 0;
        int space = 1;
	int len = 0;
	for( len = 0; ; len++ ) if( cmd_line[ len ] == 0 ) break;
	int r = 0;
	int w = 0;
	int string = 0;
	while( r <= len )
	{
	    UTF8_CHAR c = cmd_line[ r ];
	    if( c == '"' ) 
	    {
		string ^= 1;
		r++;
		continue;
	    }
	    if( string && c == ' ' ) c = 1;
	    cmd_line[ w ] = c;
	    r++;
	    w++;
	}
        for( int i = 0; i < len; i++ )
        {
    	    if( cmd_line[ i ] != ' ' )
    	    {
    		if( cmd_line[ i ] == 1 ) cmd_line[ i ] = ' ';
	        if( space == 1 )
	        {
	    	    g_argv[ g_argc ] = &cmd_line[ i ];
		    g_argc++;
		}
		space = 0;
	    }
	    else
	    {
		cmd_line[ i ] = 0;
	        space = 1;
	    }
	}
    }
}
#endif

//********************************
//WIN32 MAIN *********************
#ifdef WIN
WCHAR g_cmd_line[ 2048 ];
UTF8_CHAR g_cmd_line_utf8[ 2048 ];
int APIENTRY WinMain( HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPSTR lpszCmdLine, int nCmdShow )
{
    {
	g_wmanager.hCurrentInst = hCurrentInst;
	g_wmanager.hPreviousInst = hPreviousInst; 
	g_wmanager.lpszCmdLine = lpszCmdLine;
	g_wmanager.nCmdShow = nCmdShow;
	mbstowcs( g_cmd_line, lpszCmdLine, 2047 );
	utf16_to_utf8( g_cmd_line_utf8, 2048, (const UTF16_CHAR*)g_cmd_line );
	make_arguments( g_cmd_line_utf8 );
	int argc = g_argc;
	UTF8_CHAR** argv = g_argv;
#endif
//********************************
//********************************

//********************************
//WINCE MAIN *********************
#ifdef WINCE
UTF8_CHAR g_cmd_line_utf8[ 2048 ];
WCHAR g_window_name[ 256 ];
extern WCHAR *className; //defined in window manager (wm_wince.h)
int WINAPI WinMain( HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPWSTR lpszCmdLine, int nCmdShow )
{
    {
	g_wmanager.hCurrentInst = hCurrentInst;
	g_wmanager.hPreviousInst = hPreviousInst;
	g_wmanager.lpszCmdLine = lpszCmdLine;
	g_wmanager.nCmdShow = nCmdShow;
	utf16_to_utf8( g_cmd_line_utf8, 2048, (UTF16_CHAR*)lpszCmdLine );
	make_arguments( g_cmd_line_utf8 );
	utf8_to_utf16( (UTF16_CHAR*)g_window_name, 256, user_window_name );
	HWND wnd = FindWindow( className, g_window_name );
	if( wnd )
	{
	    //Already opened:
	    SetForegroundWindow( wnd ); //Make it foreground
	    return 0;
	}
	int argc = g_argc;
	UTF8_CHAR **argv = g_argv;
#endif
//********************************
//********************************

//********************************
//PALMOS MAIN ********************
#ifdef PALMOS
int ARM_PalmOS_main( const void* emulStateP, void* userData, Call68KFuncType* call68KFuncP ) arm_startup;
int ARM_PalmOS_main( const void* emulStateP, void* userData, Call68KFuncType* call68KFuncP )
{
    {
	volatile void* oldGOT;
	volatile register void* gGOT asm ("r10");
	volatile ARM_INFO* arm_info = (ARM_INFO *)userData;
	oldGOT = (void*)gGOT;
	//gGOT = (void *)arm_info->GOT;
	volatile unsigned int newgot = (unsigned int)arm_info->GOT;
	__asm__ ( "mov r10, %0" : : "r" (newgot) );
	ownID = (unsigned short)arm_info->ID;
	g_form_handler = arm_info->FORMHANDLER; //g_form_handler defined in palm_functions.cpp
	g_new_screen_size = arm_info->new_screen_size;
	CALL_INIT
	dprint( "MAIN: ARM started\n" );
	int autooff_time = SysSetAutoOffTime( 0 );
	g_argc = 1;
	g_argv[ 0 ] = (UTF8_CHAR*)"prog";
	int argc = g_argc;
	UTF8_CHAR** argv = g_argv;
#endif //PALMOS
//********************************
//********************************

//********************************
//UNIX MAIN **********************
#ifdef UNIX
#if defined(IPHONE) || defined(OSX)
extern "C" int sundog_main( int argc, char* argv[] )
{
    {
#else
#include <stdio.h>
#include <pthread.h>
#ifdef ANDROID
int sundog_main( int argc, char* argv[] )
#else
int main( int argc, char* argv[] )
#endif
{
    {
#endif
#endif
//********************************
//********************************
	
	bool restart_request;

main_begin:
	
	restart_request = 0;
	
	MAIN_BEGIN;
	MAIN_SET_DEFAULTS();
	
	dprint( "%s / %s\n", SUNDOG_VERSION, SUNDOG_DATE );

	if( win_init( user_window_name, user_window_xsize, user_window_ysize, user_window_flags, argc, argv, &g_wmanager ) == 0 )
	{
	    sundog_midi_init();
	    if( user_init( &g_wmanager ) == 0 )
	    {
		while( 1 )
		{
		    sundog_event evt;
		    EVENT_LOOP_BEGIN( &evt, &g_wmanager );
		    if( EVENT_LOOP_END( &g_wmanager ) ) break;
		}
		user_close( &g_wmanager );
		if( g_wmanager.restart_request ) restart_request = 1;
	    }
	    else
	    {
		g_err_count++;
	    }
	    sundog_midi_deinit();
	    win_close( &g_wmanager ); //Close window manager
	}
	else
	{
	    g_err_count++;
	}
	
	if( g_err_count > 0 )
	{
	    debug_show_error_report();
	}
	
	MAIN_END;
	
	if( restart_request ) goto main_begin;
	
#ifdef PALMOS
	SysSetAutoOffTime( autooff_time );
	//gGOT = (void*)oldGOT;
	newgot = (unsigned int)oldGOT;
	__asm__ ( "mov r10, %0" : : "r" (newgot) );
	return 1;
#endif
    }

    return g_wmanager.exit_code;
}

//################################
//################################
//################################
