#ifndef __SUNUSER__
#define __SUNUSER__

#include "../core/core.h"
#include "../core/debug.h"
#include "../sound/sound.h"
#include "../window_manager/wmanager.h"
#include "../time/timemanager.h"
#include "../utils/utils.h"

extern const UTF8_CHAR* user_window_name;
extern const UTF8_CHAR* user_profile_names[];
extern const UTF8_CHAR* user_debug_log_file_name;
extern int user_window_xsize;
extern int user_window_ysize;
extern int user_window_flags;

#define MAIN_GLOBALS \
int g_err_count = 0;

#define MAIN_BEGIN \
g_err_count = 0; \
time_ticks(); \
time_ticks_hires(); \
mem_init(); \
v3_init(); \
debug_init(); \
debug_set_output_file( user_debug_log_file_name ); \
debug_reset(); \
profile_new( 0 ); \
profile_load( 0, 0 );

inline void MAIN_SET_DEFAULTS( void )
{
#ifdef MAEMO
    profile_set_int_value( KEY_SCREENX, profile_get_int_value( KEY_SCREENX, 800, 0 ), 0 );
    profile_set_int_value( KEY_SCREENY, profile_get_int_value( KEY_SCREENY, 480, 0 ), 0 );
    profile_set_int_value( KEY_SOUNDBUFFER, profile_get_int_value( KEY_SOUNDBUFFER, 2048, 0 ), 0 );
    profile_set_str_value( KEY_AUDIODRIVER, profile_get_str_value( KEY_AUDIODRIVER, "sdl", 0 ), 0 );
    profile_set_int_value( KEY_FULLSCREEN, profile_get_int_value( KEY_FULLSCREEN, 1, 0 ), 0 );
    profile_set_int_value( KEY_ZOOM, profile_get_int_value( KEY_ZOOM, 1, 0 ), 0 );
    profile_set_int_value( KEY_PPI, profile_get_int_value( KEY_PPI, 240, 0 ), 0 );
    profile_set_int_value( KEY_SCALE, profile_get_int_value( KEY_SCALE, 130, 0 ), 0 );
    profile_set_int_value( KEY_NOCURSOR, profile_get_int_value( KEY_NOCURSOR, 1, 0 ), 0 );
    profile_set_int_value( KEY_NOAUTOLOCK, profile_get_int_value( KEY_NOAUTOLOCK, 1, 0 ), 0 );
#endif
#ifdef MEEGO
    profile_set_int_value( KEY_SCREENX, profile_get_int_value( KEY_SCREENX, 854, 0 ), 0 );
    profile_set_int_value( KEY_SCREENY, profile_get_int_value( KEY_SCREENY, 480, 0 ), 0 );
    profile_set_int_value( KEY_SOUNDBUFFER, profile_get_int_value( KEY_SOUNDBUFFER, 2048, 0 ), 0 );
    profile_set_int_value( KEY_FULLSCREEN, profile_get_int_value( KEY_FULLSCREEN, 1, 0 ), 0 );
    profile_set_int_value( KEY_ZOOM, profile_get_int_value( KEY_ZOOM, 1, 0 ), 0 );
    profile_set_int_value( KEY_PPI, profile_get_int_value( KEY_PPI, 240, 0 ), 0 );
    profile_set_int_value( KEY_NOCURSOR, profile_get_int_value( KEY_NOCURSOR, 1, 0 ), 0 );
    profile_set_int_value( KEY_NOAUTOLOCK, profile_get_int_value( KEY_NOAUTOLOCK, 1, 0 ), 0 );
#endif
}

#define MAIN_END \
profile_close( 0 ); \
v3_deinit(); \
if( mem_free_all() ) g_err_count++; /*Fix memory leaks*/ \
debug_deinit(); \
mem_deinit();

int user_init( window_manager* wm );
int user_event_handler( sundog_event* evt, window_manager* wm );
void user_close( window_manager* wm );

#endif
