## ------------------------------------------ ##
## Main target options:                       ##
## ------------------------------------------ ##

ADD_OBJS = 

##
## Linux ...
##
ifneq (,$(findstring linux,$(TARGET_OS)))
    CC = gcc $(CFLAGS) -Wno-multichar -ffast-math -pipe
    CXX = $(CC)
    TARGET_DEFINES = -DLINUX
##
## Maemo
##
ifneq (,$(findstring maemo,$(TARGET_OS)))
    TARGET_DEFINES += -DMAEMO
endif
##
## Meego
##
ifneq (,$(findstring meego,$(TARGET_OS)))
    TARGET_DEFINES += -DMEEGO
endif
##
## Dingux
##
ifneq (,$(findstring dingux,$(TARGET_OS)))
    TARGET_DEFINES += -DDINGUX
    CC = /opt/mipsel-linux-uclibc/usr/bin/mipsel-linux-gcc $(CFLAGS) -Wno-multichar -ffast-math -pipe
    CXX = $(CC)
endif
##
## Android ...
##
ifneq (,$(findstring android,$(TARGET_OS)))
    NDK = /home/alex/soft/android-ndk
    SYSROOT = $(NDK)/platforms/android-9/arch-$(TARGET_ARCH)
    TARGET_DEFINES += \
	-DANDROID \
	-DGL_GLEXT_PROTOTYPES=1 \
	-I$(NDK)/platforms/android-9/arch-$(TARGET_ARCH)/usr/include \
	-I$(NDK)/sources/android/native_app_glue
ifeq ($(TARGET_ARCH),x86) 
    CC = $(NDK)/toolchains/x86-4.4.3/prebuilt/linux-x86/bin/i686-android-linux-gcc $(CFLAGS) -Wno-multichar -ffast-math -pipe --sysroot=$(SYSROOT)
    OBJDUMP = $(NDK)/toolchains/x86-4.4.3/prebuilt/linux-x86/bin/i686-android-linux-objdump
endif
ifeq ($(TARGET_ARCH),arm)
    CC = $(NDK)/toolchains/arm-linux-androideabi-4.4.3/prebuilt/linux-x86/bin/arm-linux-androideabi-gcc $(CFLAGS) -Wno-multichar -ffast-math -pipe --sysroot=$(SYSROOT)
    OBJDUMP = $(NDK)/toolchains/arm-linux-androideabi-4.4.3/prebuilt/linux-x86/bin/arm-linux-androideabi-objdump
endif
    CXX = $(CC)
    SUNDOG_DLL = true
endif
endif
##
## FreeBSD
##
ifeq ($(TARGET_OS),freebsd)
    CC = gcc $(CFLAGS) -Wno-multichar -ffast-math -pipe
    CXX = $(CC)
    TARGET_DEFINES = -DFREEBSD
endif
##
## OSX
##
ifeq ($(TARGET_OS),osx)
    CC = /Applications/Xcode.app/Contents/Developer/usr/bin/llvm-gcc-4.2 $(CFLAGS) -Wno-multichar -isysroot /Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX10.7.sdk -mmacosx-version-min=10.6 -ffast-math -msse -pipe
    CXX = $(CC)
    TARGET_DEFINES = -DOSX
endif
##
## PalmOS
##
ifeq ($(TARGET_OS),palmos)
    MCC = m68k-palmos-gcc $(CFLAGS) -Wno-multichar -pipe
    CC = arm-elf-gcc $(CFLAGS) -Wno-multichar -pipe
    TARGET_DEFINES = -DPALMOS
endif
##
## Windows
##
ifeq ($(TARGET_OS),windows)
    CC = i586-mingw32msvc-gcc $(CFLAGS) -fno-exceptions -Wno-multichar -ffast-math -pipe
    CXX = $(CC)
    WINDRES = i586-mingw32msvc-windres
    TARGET_DEFINES = -DWIN
endif
##
## WindowsCE
##
ifeq ($(TARGET_OS),windowsce)
    CC = arm-mingw32ce-gcc $(CFLAGS) -fno-exceptions -Wno-multichar -pipe
    CXX = $(CC)
    WINDRES = arm-mingw32ce-windres
    TARGET_DEFINES = -DWINCE
endif

##
## x86
##
ifeq ($(TARGET_ARCH),x86) 
    TARGET_DEFINES += -DARCH_X86 -DARCH_NAME="\"x86\"" -DCPUMARK=10 -march=i686 -malign-double -m32
    ADD_LDFLAGS = -m32
endif
##
## x86_64
##
ifeq ($(TARGET_ARCH),x86_64) 
    TARGET_DEFINES += -DARCH_X86_64 -DARCH_NAME="\"x86_64\"" -DCPUMARK=10 -m64
    ADD_LDFLAGS = -m64
endif
##
## arm
##
ifeq ($(TARGET_ARCH),arm)
    TARGET_DEFINES += -DARCH_ARM -DCPUMARK=0
ifneq (,$(findstring android,$(TARGET_OS)))
ifeq ($(FLOAT_ABI),softfp)
    TARGET_DEFINES += -DARCH_NAME="\"armv7-a\"" -DARM_VFP -march=armv7-a -mfloat-abi=softfp -mfpu=vfp
    ADD_LDFLAGS = -Wl,--fix-cortex-a8
else
    TARGET_DEFINES += -DARCH_NAME="\"armv5te\"" -march=armv5te -mtune=xscale -msoft-float
endif
else
    TARGET_DEFINES += -DARCH_NAME="\"arm\""
endif
endif
##
## mips
##
ifeq ($(TARGET_ARCH),mips)
    TARGET_DEFINES += -DARCH_MIPS -DARCH_NAME="\"mips\"" -DCPUMARK=0
endif

TARGET_DEFINES += -DAPPNAME="\"$(APPNAME)\""

## ------------------------------------------ ##
## Directories:                               ##
## ------------------------------------------ ##

MAIN = $(SUNDOG)main/
CMAIN = $(MAIN)code/
MEM = $(SUNDOG)memory/
CMEM = $(MEM)code/
SOUND = $(SUNDOG)sound/
CSOUND = $(SOUND)code/
CORE = $(SUNDOG)core/
CCORE = $(CORE)code/
FS = $(SUNDOG)filesystem/
CFS = $(FS)code/
TM = $(SUNDOG)time/
CTM = $(TM)code/
WM = $(SUNDOG)window_manager/
CWM = $(WM)code/
UTILS = $(SUNDOG)utils/
CUTILS = $(UTILS)code/

## ------------------------------------------ ##
## Object files and includes:                 ##
## ------------------------------------------ ##

# SunDog engine:
STD_INCS = $(MEM)memory.h $(CORE)core.h $(CORE)debug.h
DEBUG_OBJS = debug.o
MEMORY_OBJS = memory.o
FS_OBJS = v3nus_fs.o
TIME_OBJS = timemanager.o
UTILS_OBJS = utils.o
SOUND_INCS = \
    $(SOUND)sound.h \
    $(CSOUND)sound_common.h \
    $(CSOUND)sound_linux.h \
    $(CSOUND)sound_linux_alsa_midi.h \
    $(CSOUND)sound_palmos.h \
    $(CSOUND)sound_win.h \
    $(CSOUND)sound_win_mmsystem_midi.h \
    $(CSOUND)sound_android.h \
    $(CSOUND)sound_coremidi.h \
    $(CSOUND)sound_osx.h
ifneq ($(SUNDOG_NOSOUND),true)
    SOUND_OBJS = sound_player.o sound.o
endif
WM_INCS = \
    $(CWM)handlers_dialog.h \
    $(CWM)handlers_files.h \
    $(CWM)handlers_list.h \
    $(CWM)handlers_popup.h \
    $(CWM)handlers_virtual_keyboard.h \
    $(CWM)wm_android.h \
    $(CWM)wm_win32.h \
    $(CWM)wm_wince.h \
    $(CWM)wm_opengl.h \
    $(CWM)wm_directx.h \
    $(CWM)wm_framebuffer.h \
    $(CWM)wm_palmos.h \
    $(CWM)wm_unixgraphics.h \
    $(CWM)wm_unixgraphics_sdl.h \
    $(WM)wmanager.h \
    $(WM)struct.h
ifneq ($(SUNDOG_NOWINDOWS),true)
    WM_OBJS = wmanager.o image.o wbd.o handlers.o handlers_preferences.o handlers_color_theme.o handlers_scale.o devrgn.o
endif
ifneq ($(SUNDOG_NOMAIN),true)
    MAIN_OBJS = main.o
endif
OBJECTS = \
    $(MEMORY_OBJS) \
    $(DEBUG_OBJS) \
    $(FS_OBJS) \
    $(TIME_OBJS) \
    $(UTILS_OBJS) \
    $(MAIN_OBJS) \
    $(SOUND_OBJS) \
    $(WM_OBJS)

# Global defines:
DEFINES = $(TARGET_DEFINES) $(SUNDOG_DEFINES) -I$(SUNDOG)
ifeq ($(SUNDOG_NOMAIN),true)
    DEFINES += -DNOMAIN
endif

# PalmOS defines:
ifeq ($(TARGET_OS),palmos)
    ARM_PALM_INCS = -I$(SUNDOG)various/palmos/ -I$(SDK) -I$(SDK)libraries -I$(SDK)Extensions/ExpansionMgr -I$(SDK)Dynamic -I$(SDK)Core -I$(SDK)Core/Hardware -I$(SDK)Core/System -I$(SDK)Core/System/Unix -I$(SDK)Core/ui -I$(SDK)Core/UI -I$(SDK)Libraries
    ARM_PALM_FLAGS = -DNATIVE -mcpu=arm9tdmi -nostartfiles -ffixed-r9 -mpic-register=r10 -msingle-pic-base -fPIC
    DEFINES += $(ARM_PALM_INCS) $(ARM_PALM_FLAGS) $(ARM_PALM_DEFINES)
    ARM_PALM_FILES = $(SUNDOG)various/palmos/palm_functions.cpp $(SUNDOG)various/palmos/PalmTypes.h $(SUNDOG)various/palmos/DebugMgr.h $(SUNDOG)various/palmos/palm_functions.h
    ARM_PALM_OBJS = palm_functions.o
    OBJECTS += $(ARM_PALM_OBJS)
endif

# Win32 defines:
ifeq ($(TARGET_OS),windows)
    DEFINES += -Iresources/
    FILE_EXT = exe
    ADD_LDFLAGS2 = -mwindows
ifeq ($(SUNDOG_DLL),true)
    FILE_EXT = dll
    ADD_LDFLAGS += -shared -Wl,--kill-at
endif
ifeq ($(SHOWCONSOLE),true)
    ADD_LDFLAGS2 =
endif
endif

# WinCE defines:
ifeq ($(TARGET_OS),windowsce)
    DEFINES += -Iresources/ -I$(SUNDOG)various/wince/
endif

# OSX defines:
ifeq ($(TARGET_OS),osx)
ifeq ($(SUNDOG_DLL),true)
    DEFINES += -fPIC
    FILE_EXT = .dylib
    ADD_LDFLAGS += -dynamiclib
endif
endif

# Linux defines:
ifneq (,$(findstring linux,$(TARGET_OS)))
ifeq ($(SUNDOG_DLL),true)
    DEFINES += -fPIC
    FILE_EXT = .so
    ADD_LDFLAGS += -shared -Wl,-soname,$(APPNAME).so
endif
endif

# Android defines:
ifneq (,$(findstring android,$(TARGET_OS)))
ifneq ($(SUNDOG_NOMAIN),true)
    OBJECTS += \
	android_native_app_glue.o \
	sundog_bridge.o
endif
endif

## ------------------------------------------ ##
## Main rules for different targets:          ##
## ------------------------------------------ ##

ifneq (,$(findstring linux,$(TARGET_OS)))
.SECONDEXPANSION:
all: $$(OBJECTS)
	$(CC) $(LDFLAGS) $(ADD_LDFLAGS) -o $(APPNAME)$(FILE_EXT) $(OBJECTS) $(LIBS)
endif

ifeq ($(TARGET_OS),freebsd)
.SECONDEXPANSION:
all: $$(OBJECTS)
	$(CC) $(LDFLAGS) $(ADD_LDFLAGS) -o $(APPNAME)$(FILE_EXT) $(OBJECTS) $(LIBS)
endif

ifeq ($(TARGET_OS),osx)
.SECONDEXPANSION:
all: $$(OBJECTS)
	$(CC) $(LDFLAGS) $(ADD_LDFLAGS) -o $(APPNAME)$(FILE_EXT) $(OBJECTS) $(LIBS)
endif

ifeq ($(TARGET_OS),palmos)
.SECONDEXPANSION:
all: $$(OBJECTS)
#Link main code ( with special ARM script - armscript.ld ) :
	$(CC) $(LDFLAGS) $(DEFINES) -o main_palm.o $(OBJECTS) -Xlinker --script=$(SUNDOG)various/palmos/armscript.ld
#Create BIN files from the program_palm.o :
	arm-elf-objcopy -j .text -O binary main_palm.o armcode
	split -b 64000 armcode armc
	if test -e armcaa; then mv armcaa armc0000.bin; fi
	if test -e armcab; then mv armcab armc0001.bin; fi
	if test -e armcac; then mv armcac armc0002.bin; fi
	if test -e armcad; then mv armcad armc0003.bin; fi
	if test -e armcae; then mv armcae armc0004.bin; fi
	if test -e armcaf; then mv armcaf armc0005.bin; fi
	if test -e armcag; then mv armcag armc0006.bin; fi
	if test -e armcah; then mv armcah armc0007.bin; fi
	if test -e armcai; then mv armcai armc0008.bin; fi
	if test -e armcaj; then mv armcaj armc0009.bin; fi
	if test -e armcak; then mv armcak armc000A.bin; fi
	if test -e armcal; then mv armcal armc000B.bin; fi
	if test -e armcal; then mv armcal armc000C.bin; fi
	if test -e armcal; then mv armcal armc000D.bin; fi
	if test -e armcal; then mv armcal armc000E.bin; fi
	arm-elf-objcopy -j .got -O binary main_palm.o armg0000.bin
	arm-elf-objcopy -j .alldata -O binary main_palm.o armdata
	split -b 64000 armdata armd
	if test -e armdaa; then mv armdaa armd0000.bin; fi
	if test -e armdab; then mv armdab armd0001.bin; fi
	if test -e armdac; then mv armdac armd0002.bin; fi
	if test -e armdad; then mv armdad armd0003.bin; fi
	if test -e armdae; then mv armdae armd0004.bin; fi
	if test -e armdaf; then mv armdaf armd0005.bin; fi
	if test -e armdag; then mv armdag armd0006.bin; fi
	if test -e armdah; then mv armdah armd0007.bin; fi
	if test -e armdai; then mv armdai armd0008.bin; fi
	if test -e armdaj; then mv armdaj armd0009.bin; fi
	if test -e armdak; then mv armdak armd000A.bin; fi
	if test -e armdal; then mv armdal armd000B.bin; fi
	if test -e armdal; then mv armdal armd000C.bin; fi
	if test -e armdal; then mv armdal armd000D.bin; fi
	if test -e armdal; then mv armdal armd000E.bin; fi
	arm-elf-objcopy -j .rel -O binary main_palm.o armr0000.bin
#Create Palm resources ( *.BIN files ) :
	pilrc resources/palmos_res.rcp
#Build PRC file :
	$(MCC) $(SUNDOG)various/palmos/arm_starter.cpp resources/palmos_filetypes.cpp $(SUNDOG_DEFINES) -o arm_starter.o
	m68k-palmos-obj-res arm_starter.o
	build-prc -v 1.0 resources/palmos_info.def -o $(APPNAME).prc *.grc *.bin
palm_functions.o: $(ARM_PALM_FILES) 
	$(CC) $(DEFINES) -c $(SUNDOG)various/palmos/palm_functions.cpp
endif

ifeq ($(TARGET_OS),windows)
.SECONDEXPANSION:
all: $$(OBJECTS)
ifneq ($(SUNDOG_NOWINDOWS),true)
	$(WINDRES) resources/win_res.rc win_res.o
	$(CC) $(LDFLAGS) $(ADD_LDFLAGS) $(ADD_LDFLAGS2) -o $(APPNAME).$(FILE_EXT) $(OBJECTS) win_res.o $(LIBS)
else
	$(CC) $(LDFLAGS) $(ADD_LDFLAGS) $(ADD_LDFLAGS2) -o $(APPNAME).$(FILE_EXT) $(OBJECTS) $(LIBS)
endif
endif

ifeq ($(TARGET_OS),windowsce)
.SECONDEXPANSION:
all: $$(OBJECTS)
	$(WINDRES) resources/win_res.rc win_res.o
	$(CC) $(LDFLAGS) -o $(APPNAME).exe $(OBJECTS) win_res.o $(LIBS)
endif

## ------------------------------------------ ##
## SunDog engine rules: (main)                ##
## ------------------------------------------ ##

main.o: $(CMAIN)main.cpp $(STD_INCS)
	$(CC) $(DEFINES) -c $(CMAIN)main.cpp

## ------------------------------------------ ##
## SunDog engine rules: (sound manager)       ##
## ------------------------------------------ ##

sound_player.o: $(CSOUND)sound_player.cpp $(SOUND)sound.h $(SOUND_INCS)
	$(CC) $(DEFINES) -c $(CSOUND)sound_player.cpp

sound.o: $(CSOUND)sound.cpp $(SOUND)sound.h $(SOUND_INCS)
	$(CC) $(DEFINES) -c $(CSOUND)sound.cpp

## ------------------------------------------ ##
## SunDog engine rules: (...)                 ##
## ------------------------------------------ ##

timemanager.o: $(CTM)timemanager.cpp $(TM)timemanager.h $(STD_INCS)
	$(CC) $(DEFINES) -c $(CTM)timemanager.cpp

v3nus_fs.o: $(CFS)v3nus_fs.cpp $(FS)v3nus_fs.h $(STD_INCS)
	$(CC) $(DEFINES) -c $(CFS)v3nus_fs.cpp

utils.o: $(CUTILS)utils.cpp $(UTILS)utils.h $(STD_INCS)
	$(CC) $(DEFINES) -c $(CUTILS)utils.cpp

debug.o: $(CCORE)debug.cpp
	$(CC) $(DEFINES) -c $(CCORE)debug.cpp

memory.o: $(CMEM)memory.cpp $(MEM)memory.h $(CORE)core.h
	$(CC) $(DEFINES) -c $(CMEM)memory.cpp

## ------------------------------------------ ##
## SunDog engine rules: (window manager)      ##
## ------------------------------------------ ##

wmanager.o: $(CWM)wmanager.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)wmanager.cpp

image.o: $(CWM)image.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)image.cpp

wbd.o: $(CWM)wbd.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)wbd.cpp

handlers.o: $(CWM)handlers.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)handlers.cpp

handlers_preferences.o: $(CWM)handlers_preferences.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)handlers_preferences.cpp

handlers_color_theme.o: $(CWM)handlers_color_theme.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)handlers_color_theme.cpp

handlers_scale.o: $(CWM)handlers_scale.cpp $(WM_INCS) $(STD_INCS)
	$(CC) $(DEFINES) -c $(CWM)handlers_scale.cpp

devrgn.o: $(WM)regions/devrgn.cpp $(STD_INCS)
	$(CC) $(DEFINES) -c $(WM)regions/devrgn.cpp

## ------------------------------------------ ##
## Misc                                       ##
## ------------------------------------------ ##

android_native_app_glue.o: $(NDK)/sources/android/native_app_glue/android_native_app_glue.c
	$(CC) $(DEFINES) -c $(NDK)/sources/android/native_app_glue/android_native_app_glue.c

sundog_bridge.o: $(SUNDOG)various/android/sundog_bridge.cpp
	$(CC) $(DEFINES) -c $(SUNDOG)various/android/sundog_bridge.cpp

clean:
	rm -f *.o *.grc *.bin armcode armdata
