/*
    handlers_preferences.cpp
    This file is part of the SunDog engine.
    Copyright (C) 2010 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/
            
#include "../../main/user_code.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

WINDOWPTR g_colortheme = 0;

int g_grid_xcells = 1;
int g_grid_ycells = 1;

int g_prev_color_theme = 0;

bool g_color_theme_changed = 0;

struct colortheme_data
{
    WINDOWPTR ok;
    WINDOWPTR cancel;
};

int colortheme_ok_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    remove_window( g_colortheme, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    profile_set_int_value( KEY_COLOR_THEME, wm->color_theme, 0 );
    profile_save( 0 );
    if( wm->color_theme == g_prev_color_theme ) return 0;
    if( dialog( "!You must restart the program\nto see the new color theme.\nRestart now?", "Yes;No" ) == 0 )
    {
	wm->exit_request = 1;
	wm->restart_request = 1;
	g_color_theme_changed = 1;
    }
    else 
    {
    }
    return 0;
}

int colortheme_cancel_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    remove_window( g_colortheme, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    wm->color_theme = g_prev_color_theme;
    return 0;
}

int colortheme_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    colortheme_data* data = (colortheme_data*)win->data;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( colortheme_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		g_grid_xcells = 1;
		g_grid_ycells = 1;
		int i = 0;
		while( g_grid_xcells * g_grid_ycells < COLOR_THEMES )
		{
		    if( i & 1 ) 
		    {
			g_grid_ycells++;
		    }
		    else 
		    {
			g_grid_xcells++;
		    }
		    i++;
		}
		
		data->ok = new_window( "OK", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		data->cancel = new_window( "Cancel", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		set_handler( data->ok, colortheme_ok_handler, data, wm );
		set_handler( data->cancel, colortheme_cancel_handler, data, wm );
		int x = INTERELEMENT_SPACE;
		set_window_controller( data->ok, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->ok, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->ok, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->ok, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE, CEND );
		x += 1;
		set_window_controller( data->cancel, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->cancel, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE, CEND );
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		win_draw_lock( win, wm );
		wbd_lock( win );
		
		draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );
		
		int t = 0;
		float grid_xsize = (float)( win->xsize - INTERELEMENT_SPACE * 2 ) / g_grid_xcells;
		float grid_ysize = (float)( win->ysize - INTERELEMENT_SPACE * 3 - BUTTON_YSIZE ) / g_grid_ycells;
		for( int grid_y = 0; grid_y < g_grid_ycells; grid_y++ )
		{
		    for( int grid_x = 0; grid_x < g_grid_xcells; grid_x++ )
		    {
			if( t >= COLOR_THEMES ) break;
			float fx = (float)INTERELEMENT_SPACE + (float)grid_x * grid_xsize;
			float fy = (float)INTERELEMENT_SPACE + (float)grid_y * grid_ysize;
			int x = fx;
			int y = fy;
			int xsize = (int)( ( fx + grid_xsize ) - (float)x );
			int ysize = (int)( ( fy + grid_ysize ) - (float)y );
			xsize--;
			ysize--;
			if( grid_x == g_grid_xcells - 1 ) xsize = win->xsize - INTERELEMENT_SPACE - x;
			if( grid_y == g_grid_ycells - 1 ) ysize = win->ysize - INTERELEMENT_SPACE * 2 - BUTTON_YSIZE - y;
			
			COLOR cc[ 4 ];
			for( int c = 0; c < 4; c++ )
			{
			    uchar rr = g_color_themes[ t * 3 * 4 + c * 3 + 0 ];
			    uchar gg = g_color_themes[ t * 3 * 4 + c * 3 + 1 ];
			    uchar bb = g_color_themes[ t * 3 * 4 + c * 3 + 2 ];
			    cc[ c ] = get_color( rr, gg, bb );
			}
			
			int xsize2 = xsize / 4;
			draw_box( x, y, xsize2, ysize, cc[ 0 ], wm );
			draw_box( x + xsize2, y, xsize2, ysize, cc[ 1 ], wm );
			draw_box( x + xsize2 * 2, y, xsize2, ysize, cc[ 2 ], wm );
			draw_box( x + xsize2 * 3, y, xsize - xsize2 * 3, ysize, cc[ 3 ], wm );
						
			draw_vgradient( x, y, xsize, ysize / 2, wm->color3, 180, 0 );
			draw_vgradient( x, y + ysize / 2, xsize, ysize - ysize / 2, win->color, 0, 180 );
			
			wm->cur_transparency = 128;
			draw_frame( x, y, xsize - 1, ysize - 1, win->color, wm );
			wm->cur_transparency = 255;
			
			if( t == wm->color_theme )
			{
			    wm->cur_transparency = 128;
			    draw_corners( x + 1 + wm->corners_size, y + 1 + wm->corners_size, xsize - 2 - wm->corners_size * 2, ysize - 2 - wm->corners_size * 2, wm->corners_size, wm->corners_len, wm->color0 );
			    wm->cur_transparency = 255;
			    draw_corners( x + wm->corners_size, y + wm->corners_size, xsize - wm->corners_size * 2, ysize - wm->corners_size * 2, wm->corners_size, wm->corners_len, wm->color3 );
			}
			
			t++;
		    }
		    if( t >= COLOR_THEMES ) break;
		}
		
		wbd_draw();
		wbd_unlock();
		win_draw_unlock( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( evt->key == KEY_ENTER )
	    {
		colortheme_ok_handler( data, data->ok, wm );
		retval = 1;
	    }
	    if( evt->key == KEY_LEFT || evt->key == KEY_UP )
	    {
		int t = wm->color_theme;
		if( evt->key == KEY_LEFT ) t--;
		if( evt->key == KEY_UP ) t -= g_grid_xcells;
		if( t >= 0 )
		{
		    wm->color_theme = t;
		    draw_window( win, wm );
		}
		retval = 1;
	    }
	    if( evt->key == KEY_RIGHT || evt->key == KEY_DOWN )
	    {
		int t = wm->color_theme;
		if( evt->key == KEY_RIGHT ) t++;
		if( evt->key == KEY_DOWN ) t += g_grid_xcells;
		if( t < COLOR_THEMES )
		{
		    wm->color_theme = t;
		    draw_window( win, wm );
		}
		retval = 1;
	    }
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    if( evt->key & MOUSE_BUTTON_LEFT )
	    {
		int t = 0;
		float grid_xsize = (float)( win->xsize - INTERELEMENT_SPACE * 2 ) / g_grid_xcells;
		float grid_ysize = (float)( win->ysize - INTERELEMENT_SPACE * 3 - BUTTON_YSIZE ) / g_grid_ycells;
		for( int grid_y = 0; grid_y < g_grid_ycells; grid_y++ )
		{
		    for( int grid_x = 0; grid_x < g_grid_xcells; grid_x++ )
		    {
			if( t >= COLOR_THEMES ) break;
			
			float fx = (float)INTERELEMENT_SPACE + (float)grid_x * grid_xsize;
			float fy = (float)INTERELEMENT_SPACE + (float)grid_y * grid_ysize;
			int x = fx;
			int y = fy;
			int xsize = (int)grid_xsize;
			int ysize = (int)grid_ysize;
			if( rx >= x && ry >= y && rx < x + xsize && ry < y + ysize )
			{
			    wm->color_theme = t;
			    draw_window( win, wm );
			    break;
			}
			
			t++;
		    }
		    if( t >= COLOR_THEMES ) break;
		}
	    }
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    retval = 1;
	    g_colortheme = 0;
	    break;
    }
    return retval;
}
