/*
    handlers_files.h.
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

void (*fdialog_open_wifi)( void ) = 0;

const UTF8_CHAR* FILES_PROPS = 0;
const UTF8_CHAR* FILES_MASK = 0;
const UTF8_CHAR* FILES_PREVIEW = 0;
int (*FILES_PREVIEW_HANDLER)( files_preview_data*, window_manager* ) = 0;
const UTF8_CHAR* FILES_USER_BUTTON = 0;
int (*FILES_USER_BUTTON_HANDLER)( void* user_data, WINDOWPTR win, window_manager* ) = 0;
void* FILES_USER_BUTTON_DATA = 0;
UTF8_CHAR* FILES_RESULTED_FILENAME = 0;
const UTF8_CHAR* FILES_DEF_FILENAME = 0;
bool g_files_preview_flag = 0;

struct files_data
{
    WINDOWPTR this_window;
    WINDOWPTR go_up_button;
    WINDOWPTR disk_buttons[ 32 ];
    WINDOWPTR path;
    WINDOWPTR name;
    WINDOWPTR files;
    WINDOWPTR ok_button;
    WINDOWPTR cancel_button;
    WINDOWPTR del_button;
    WINDOWPTR preview_button;
    int (*preview_handler)( files_preview_data*, window_manager* );
    files_preview_data preview_data;
    const UTF8_CHAR* mask;
    const UTF8_CHAR* props_file; //File with dialog properties
    const UTF8_CHAR* preview;
    UTF8_CHAR* prop_path[ 32 ];
    int prop_cur_file[ 32 ];
    int	prop_cur_disk;
    UTF8_CHAR* resulted_filename;
    int first_time;
};

int files_ok_button_handler( void* user_data, WINDOWPTR win, window_manager* wm );
int files_cancel_button_handler( void* user_data, WINDOWPTR win, window_manager* wm );
int files_name_handler( void* user_data, WINDOWPTR win, window_manager* wm );
void files_refresh_list( WINDOWPTR win, window_manager* wm );
void files_save_props( WINDOWPTR win, window_manager* wm );

UTF8_CHAR* files_make_resulted_filename( files_data* data, window_manager* wm )
{
    UTF8_CHAR* rv = 0;
    if( data->resulted_filename )
    {
	UTF8_CHAR* fname = text_get_text( data->name, wm );
	UTF8_CHAR* res = data->resulted_filename;
	res[ 0 ] = 0;
	if( fname )
	{
	    mem_strcat( res, get_disk_name( data->prop_cur_disk ) );
	    if( data->prop_path[ data->prop_cur_disk ] )
	    {
		mem_strcat( res, data->prop_path[ data->prop_cur_disk ] );
	    }
	    mem_strcat( res, fname );
	    rv = res;
	}
    }
    return rv;
}

int files_list_handler( void* user_data, WINDOWPTR win, window_manager *wm )
{
    files_data* data = (files_data*)user_data;
    list_data* ldata = list_get_data( data->files, wm );
    int last_action = list_get_last_action( win, wm );
    if( ldata && last_action )
    {
	if( last_action == LIST_ACTION_ESCAPE )
	{
	    //ESCAPE KEY:
	    files_cancel_button_handler( data, data->ok_button, wm );
	    return 0;
	}
	UTF8_CHAR* item = list_get_item( list_get_selected_num( ldata ), ldata );
	char attr = list_get_attr( list_get_selected_num( ldata ), ldata );
	int subdir = 0;
	UTF8_CHAR* preview_name = 0;
	if( item )
	{
	    if( last_action == LIST_ACTION_UPDOWN )
	    {
		//UP/DOWN KEY:
		if( attr == 0 )
		{
		    preview_name = item;
		    text_set_text( data->name, item, wm );
		}
	    }
	    if( last_action == LIST_ACTION_ENTER ||
		last_action == LIST_ACTION_DOUBLECLICK )
	    {
		//ENTER/SPACE KEY:
		if( list_get_attr( list_get_selected_num( ldata ), ldata ) == 1 )
		{
		    //It's a dir:
		    subdir = 1;
		}
		else
		{
		    //Select file:
		    if( last_action == LIST_ACTION_DOUBLECLICK )
			text_set_text( data->name, item, wm );
		    files_ok_button_handler( data, data->ok_button, wm );
		    return 0;
		}
	    }
	    if( last_action == LIST_ACTION_CLICK )
	    {
		//MOUSE CLICK:
		if( list_get_attr( list_get_selected_num( ldata ), ldata ) == 1 )
		{
		    //It's a dir:
		    subdir = 1;
		}
		else
		{
		    preview_name = item;
		    text_set_text( data->name, item, wm );
		}
	    }
	    if( subdir )
	    {
		//Go to the subdir:
		UTF8_CHAR *path = data->prop_path[ data->prop_cur_disk ];
		if( path )
		    path = (UTF8_CHAR*)mem_resize( path, mem_strlen( path ) + mem_strlen( item ) + 2 );
		else
		{
		    path = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( item ) + 2 );
		    path[ 0 ] = 0;
		}
		mem_strcat( path, item );
		mem_strcat( path, "/" );
		data->prop_path[ data->prop_cur_disk ] = path;
		data->prop_cur_file[ data->prop_cur_disk ] = -1;
		text_set_text( data->path, path, wm );
		files_refresh_list( data->this_window, wm );
		draw_window( data->this_window, wm );
	    }
	}
	if( preview_name && data->preview && g_files_preview_flag )
	{
	    data->preview_data.name = files_make_resulted_filename( data, wm );
	    if( data->preview_data.name )
	    {
		data->preview_data.status = FPREVIEW_FILE_SELECTED;
		data->preview_handler( &data->preview_data, wm );
		data->preview_data.name[ 0 ] = 0;
	    }
	}
	data->prop_cur_file[ data->prop_cur_disk ] = ldata->selected_item;
    }
    return 0;
}

int files_disk_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data* data = (files_data*)user_data;
    
#ifdef IPHONE
    UTF8_CHAR* cur_path = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, MAX_DIR_LEN );
    if( cur_path )
    {
	cur_path[ 0 ] = 0;
	mem_strcat( cur_path, get_disk_name( data->prop_cur_disk ) );
	if( data->prop_path[ data->prop_cur_disk ] )
	{
	    mem_strcat( cur_path, data->prop_path[ data->prop_cur_disk ] );
	}
	if( mem_strcmp( cur_path, get_current_path() ) == 0 )
	{
	    //We are in the iPhone user directory. Can't go deeper
	    mem_free( cur_path );
	    return 0;
	}
	mem_free( cur_path );
    }
#endif    

    if( win == data->go_up_button )
    {
	//Go to parent dir:
	UTF8_CHAR *path = data->prop_path[ data->prop_cur_disk ];
	if( path )
	{
	    int p;
	    for( p = mem_strlen( path ) - 2; p >= 0; p-- )
	    {
		if( path[ p ] == '/' ) { path[ p + 1 ] = 0; break; }
	    }
	    if( p < 0 ) path[ 0 ] = 0;
	}
	data->prop_cur_file[ data->prop_cur_disk ] = -1;
    }
    else
    {
	for( int a = 0; a < disks; a++ )
	{
	    if( data->disk_buttons[ a ] )
	    {
		data->disk_buttons[ a ]->color = wm->button_color;
		if( data->disk_buttons[ a ] == win )
		    data->prop_cur_disk = a;
	    }
	}
	//Show selected disk:
	if( data->disk_buttons[ data->prop_cur_disk ] )
	{
	    data->disk_buttons[ data->prop_cur_disk ]->color = SELECTED_BUTTON_COLOR;
	}
    }

    //Set new filelist:
    files_refresh_list( data->this_window, wm );
    //Set new path:
    if( data->prop_path[ data->prop_cur_disk ] )
	text_set_text( data->path, data->prop_path[ data->prop_cur_disk ], wm );
    else
	text_set_text( data->path, "", wm );

    //Save props:
    files_save_props( data->this_window, wm );
    draw_window( data->this_window, wm );
    return 0;
}

int files_preview_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    files_data* data = (files_data*)user_data;
    
    g_files_preview_flag ^= 1;
    if( g_files_preview_flag )
    {
	data->preview_button->color = SELECTED_BUTTON_COLOR;
    }
    else 
    {
	data->preview_button->color = wm->button_color;
    }
    
    if( g_files_preview_flag )
	data->preview_data.status = FPREVIEW_OPEN;
    else 
	data->preview_data.status = FPREVIEW_CLOSE;
    data->preview_handler( &data->preview_data, wm );
    
    draw_window( wm->root_win, wm );
    
    return 0;
}

int files_ok_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    files_data* data = (files_data*)user_data;

    files_name_handler( data, data->name, wm ); //Add file extension

    if( data->resulted_filename )
    {
	UTF8_CHAR* res = files_make_resulted_filename( data, wm );
	if( win == data->del_button && res )
	{
	    //Delete:
	    UTF8_CHAR* ts = (UTF8_CHAR*)MEM_NEW2( MAX_DIR_LEN );
	    ts[ 0 ] = 0;
	    mem_strcat( ts, "Delete " );
	    mem_strcat( ts, text_get_text( data->name, wm ) );
	    mem_strcat( ts, "?" );
	    if( data->resulted_filename[ 0 ] != 0 )
	    {
		if( dialog( ts, "Yes;No" ) == 0 )
		{
		    v3_remove( data->resulted_filename );
		    files_refresh_list( data->this_window, wm );
		}
	    }
	    mem_free( ts );
	    data->resulted_filename[ 0 ] = 0;
	}
    }
    if( win != data->del_button )
    {
	remove_window( data->this_window, wm );
	recalc_regions( wm );
    }
    draw_window( wm->root_win, wm );
    return 0;
}

int files_cancel_button_handler( void *user_data, WINDOWPTR win, window_manager *wm )
{
    files_data *data = (files_data*)user_data;
    if( data->resulted_filename )
    {
	UTF8_CHAR *res = data->resulted_filename;
	res[ 0 ] = 0;
    }
    remove_window( data->this_window, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int files_wifi_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    fdialog_open_wifi();
    return 0;
}

int files_name_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    files_data* data = (files_data*)user_data;
    UTF8_CHAR* name = text_get_text( data->name, wm );
    int mask_extensions = 0;
    int mask_cur_ext = 0;
    if( data->mask )
    {
	mask_extensions = 1;
	for( int i = 0; i < mem_strlen( data->mask ); i++ )
	{
	    if( data->mask[ i ] == '/' ) mask_extensions++;
	}
    }
    if( name && data->mask )
    {
	int a;
	bool ext_eq = 0;
	for( a = mem_strlen( name ) - 1; a >= 0; a-- )
	    if( name[ a ] == '.' ) break;
	if( a >= 0 )
	{
	    //Dot found:
	    while( 1 )
	    {
		int p;
		int extnum = 0;
		for( p = 0; p < mem_strlen( data->mask ); p++ )
		{
		    if( extnum == mask_cur_ext ) break;
		    if( data->mask[ p ] == '/' ) 
		    {
			extnum++;
		    }
		}
		ext_eq = 1;
		for( int i = a + 1 ; ; i++, p++ )
		{
		    int c = name[ i ];
		    if( c >= 65 && c <= 90 ) c += 0x20; //Make small leters
		    int m = data->mask[ p ];
		    if( m == '/' ) m = 0;
		    if( c != m )
		    {
			ext_eq = 0;
			break;
		    }
		    if( c == 0 ) break;
		    if( m == 0 ) break;
		}
		if( ext_eq == 0 )
		{
		    //File extension is not equal to extension from mask:
		    if( mask_cur_ext < mask_extensions - 1 )
		    {
			//But there are another extensions in mask. Lets try them too:
			mask_cur_ext++;
			continue;
		    }
		}
		break;
	    }
	}
	if( ext_eq == 0 ) 
	{
	    //Add extension:
	    UTF8_CHAR* new_name = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( name ) + 32 );
	    new_name[ 0 ] = 0;
	    mem_strcat( new_name, name );
	    a = mem_strlen( name );
	    new_name[ a ] = '.';
	    a++;
	    int p = 0;
	    for(;;)
	    {
		new_name[ a ] = data->mask[ p ];
		if( new_name[ a ] == '/' ||
		    new_name[ a ] == 0 )
		{
		    new_name[ a ] = 0;
		    break;
		}
		p++;
		a++;
	    }
	    text_set_text( data->name, new_name, wm );
	    draw_window( data->name, wm );
	    mem_free( new_name );
	}
    }
    return 0;
}

void files_refresh_list( WINDOWPTR win, window_manager* wm )
{
    files_data* data = (files_data*)win->data;
    list_data* ldata = list_get_data( data->files, wm );
    list_close( ldata );
    list_init( ldata );
    find_struct fs;
    UTF8_CHAR* disk_name = get_disk_name( data->prop_cur_disk );
    UTF8_CHAR* path = data->prop_path[ data->prop_cur_disk ];
    UTF8_CHAR* res = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( disk_name ) + mem_strlen( path ) + 1 );
    res[ 0 ] = 0;
    mem_strcat( res, disk_name );
    mem_strcat( res, path );
    fs.start_dir = res;
    fs.mask = data->mask;
    if( find_first( &fs ) )
    {
	if( fs.name[ 0 ] != '.' )
	{
	    list_add_item( fs.name, fs.type, ldata );
	}
	while( find_next( &fs ) )
	{
	    if( fs.name[ 0 ] != '.' )
	    {
		list_add_item( fs.name, fs.type, ldata );
	    }
	}
    }
    find_close( &fs );
    mem_free( res );
    list_sort( ldata );

    //Set file num:
    list_select_item( data->files, data->prop_cur_file[ data->prop_cur_disk ], wm );
}

void files_load_string( V3_FILE f, UTF8_CHAR *dest )
{
    for( int p = 0; p < MAX_DIR_LEN; p++ )
    {
	if( v3_eof( f ) ) { dest[ p ] = 0; break; }
	dest[ p ] = v3_getc( f );
	if( dest[ p ] == 0 ) break;
    }
}

int files_get_disk_num( UTF8_CHAR *name )
{
    for( int a = 0; a < disks; a++ )
    {
	if( mem_strcmp( name, get_disk_name( a ) ) == 0 ) return a;
    }
    return -1;
}

void files_load_props( WINDOWPTR win, window_manager *wm )
{
    UTF8_CHAR* cdir = get_current_path();
    if( cdir && cdir[ 0 ] )
    {
        if( mem_strlen( cdir ) >= 3 )
        {
            if( cdir[ 1 ] == ':' && cdir[ 2 ] == '/' )
            {
                cdir += 3;
            }
            else if( cdir[ 1 ] == ':' && cdir[ 2 ] != '/' )
            {
                cdir += 2;
            }
        }
        if( cdir[ 0 ] == '/' )
        {
            cdir++;
        }
    }
    else
    {
        cdir = (UTF8_CHAR*)"";
    }
    
    if( win )
    {
	files_data* data = (files_data*)win->data;
	if( data->props_file )
	{
	    V3_FILE f = 0;
            bool props_loaded = 0;

	    f = v3_open( data->props_file, "rb" );

	    if( f )
	    {
                props_loaded = 1;
		UTF8_CHAR* temp_str = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, 4096 );

		//Current disk:
		files_load_string( f, temp_str );
		int d = files_get_disk_num( temp_str );
		if( d == -1 ) 
		    data->prop_cur_disk = get_current_disk();
		else
		    data->prop_cur_disk = d;

		//Disks info:
		for( int a = 0; a < 32; a++ )
		{
		    if( v3_eof( f ) ) break;
		    //Disk name:
		    files_load_string( f, temp_str );
		    d = files_get_disk_num( temp_str );
		    if( d != -1 )
		    {
			//Path:
			files_load_string( f, temp_str );
			if( data->prop_path[ d ] )
			{
			    //Remove old path:
			    mem_free( data->prop_path[ d ] );
			    data->prop_path[ d ] = 0;
			}
			if( temp_str[ 0 ] )
			{
			    //Set new path:
			    data->prop_path[ d ] = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( temp_str ) + 1 );
			    mem_copy( data->prop_path[ d ], temp_str, mem_strlen( temp_str ) + 1 );
			}
#ifdef IPHONE
                        if( strstr( data->prop_path[ d ], cdir ) == 0 )
                        {
                            //Broken path
                            props_loaded = 0;
                            break;
                        }
#endif
		    }
		    int file_num;
		    v3_read( &file_num, 4, 1, f );
		    if( d != -1 )
			data->prop_cur_file[ d ] = file_num;
		}
		
		mem_free( temp_str );
		v3_close( f );
	    }
	    
            if( props_loaded == 0 )
	    {
		//File not found
		for( int a = 0; a < 32; a++ )
		{
		    data->prop_path[ a ] = 0;
		    data->prop_cur_file[ a ] = 0;
		}
		//Set current disk:
		data->prop_cur_disk = get_current_disk();
		
		data->prop_path[ data->prop_cur_disk ] = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( cdir ) + 1 );
		mem_copy( data->prop_path[ data->prop_cur_disk ], cdir, mem_strlen( cdir ) + 1 );
	    }
	}
	//Refresh list:
	files_refresh_list( win, wm );
    }
}

void files_save_props( WINDOWPTR win, window_manager *wm )
{
    if( win )
    {
	files_data* data = (files_data*)win->data;
	if( data->props_file )
	{
	    V3_FILE f = 0;
	    f = v3_open( data->props_file, "wb" );
	    if( f )
	    {
		//Save current disk:
		UTF8_CHAR* path;
		UTF8_CHAR* disk_name = get_disk_name( data->prop_cur_disk );
	        v3_write( disk_name, 1, mem_strlen( disk_name ) + 1, f );
		//Save other properties:
		for( int a = 0; a < disks; a++ )
		{
		    //Disk name:
		    disk_name = get_disk_name( a );
		    v3_write( disk_name, 1, mem_strlen( disk_name ) + 1, f );
		    //Current path:
		    if( data->prop_path[ a ] == 0 )
			v3_putc( 0, f );
		    else
		    {
			path = data->prop_path[ a ];
			v3_write( path, 1, mem_strlen( path ) + 1, f );
		    }
		    //Selected file num:
		    v3_write( &data->prop_cur_file[ a ], 4, 1, f );
		}
		v3_close( f );
	    }
	}
    }
}

int files_handler( sundog_event *evt, window_manager *wm )
{
    if( window_handler_check_data( evt, wm ) ) return 0;
    int retval = 0;
    WINDOWPTR win = evt->win;
    files_data* data = (files_data*)win->data;
    int a, b;
    int disks_x;
    int disks_buttons;
    int but_ysize = BUTTON_YSIZE;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( files_data );
	    break;
	case EVT_AFTERCREATE:
	    //DATA INIT:
	    get_disks();
	    for( a = 0; a < 32; a ++ )
	    {
		data->prop_path[ a ] = 0;
		data->prop_cur_file[ a ] = 0;
	    }
	    data->prop_cur_disk = get_current_disk();
	    if( FILES_PROPS )
	    {
		data->props_file = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, mem_strlen( FILES_PROPS ) + 1 );
		mem_copy( (void*)data->props_file, FILES_PROPS, mem_strlen( FILES_PROPS ) + 1 );
	    }
	    else
	    {
		data->props_file = 0;
	    }
	    data->mask = FILES_MASK;
	    data->preview = FILES_PREVIEW;
	    data->preview_handler = FILES_PREVIEW_HANDLER;
	    data->resulted_filename = FILES_RESULTED_FILENAME;
	    data->this_window = win;
	    data->first_time = 1;

	    data->preview_data.win = new_window( "Preview win", 0, INTERELEMENT_SPACE, 0, 0, win->color, win, null_handler, wm );
	    set_window_controller( data->preview_data.win, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
	    set_window_controller( data->preview_data.win, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
	    
	    //DISKS:
	    {
		int disk_button_xsize = string_size( win, "C:/", wm ) * 2;
		if( disk_button_xsize < SCROLLBAR_SIZE ) 
		    disk_button_xsize = SCROLLBAR_SIZE;
		for( a = 0; a < 32; a++ ) data->disk_buttons[ a ] = 0;
		disks_x = INTERELEMENT_SPACE;
		disks_buttons = 1;
		data->go_up_button = new_window( 
		    "..", 
		    disks_x, INTERELEMENT_SPACE, disk_button_xsize, SCROLLBAR_SIZE, 
		    wm->button_color,
		    win,
		    button_handler, 
		    wm );
		disks_x += disk_button_xsize + 1;
		set_handler( data->go_up_button, files_disk_button_handler, data, wm );
		set_window_controller( data->go_up_button, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CEND );
		set_window_controller( data->go_up_button, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)SCROLLBAR_SIZE, CEND );
		if( disks > 1 || get_disk_name( 0 )[ 0 ] != '/' )
		{
		    for( a = 0; a < disks; a++ )
		    {
			data->disk_buttons[ a ] = new_window( 
			    get_disk_name( a ), 
			    disks_x, 
			    INTERELEMENT_SPACE, 
			    disk_button_xsize, 
			    SCROLLBAR_SIZE, 
			    wm->button_color,
			    win,
			    button_handler, 
			    wm );
			set_handler( data->disk_buttons[ a ], files_disk_button_handler, data, wm );
			set_window_controller( data->disk_buttons[ a ], 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CEND );
			set_window_controller( data->disk_buttons[ a ], 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)SCROLLBAR_SIZE, CEND );
			disks_x += disk_button_xsize + 1;
			disks_buttons++;
		    }
		}
	    }

	    //DIRECTORY:
	    a = string_size( win, "Path: ", wm );
	    b = SCROLLBAR_SIZE + INTERELEMENT_SPACE;
#ifdef IPHONE
            TEXT_RO = 1;
	    data->path = new_window( "pathn", 5000, 5000, 1, 1, wm->text_background, win, text_handler, wm );
#else
	    if( disks_buttons > 1 )
	    {
		WINDOWPTR tw = new_window( "Path:", INTERELEMENT_SPACE, b, a, wm->text_ysize, 0, win, label_handler, wm );
		set_window_controller( tw, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b, CEND );
		set_window_controller( tw, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b + wm->text_ysize, CEND );
	    }
	    TEXT_RO = 1;
	    data->path = new_window( "pathn", 0, 0, 1, 1, wm->text_background, win, text_handler, wm );
	    if( disks_buttons > 1 )
	    {
		set_window_controller( data->path, 0, wm, (WCMD)INTERELEMENT_SPACE + a, CEND );
		set_window_controller( data->path, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b, CEND );
		set_window_controller( data->path, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->path, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b + wm->text_ysize, CEND );
        	b += wm->text_ysize + INTERELEMENT_SPACE;
    	    }
    	    else
    	    {
		set_window_controller( data->path, 0, wm, (WCMD)disks_x + INTERELEMENT_SPACE, CEND );
		set_window_controller( data->path, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CEND );
		set_window_controller( data->path, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		set_window_controller( data->path, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)SCROLLBAR_SIZE, CEND );
    	    }
#endif
            
	    //FILENAME:
	    {
		WINDOWPTR tw = new_window( "Name:", INTERELEMENT_SPACE, b, string_size( win, "Name:" , wm ) + 4, wm->text_ysize, 0, win, label_handler, wm );
		set_window_controller( tw, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b, CEND );
		set_window_controller( tw, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b + wm->text_ysize, CEND );
	    }
	    data->name = new_window( "namen", 0, 0, 1, 1, wm->text_background, win, text_handler, wm );
	    set_handler( data->name, files_name_handler, data, wm );
	    set_window_controller( data->name, 0, wm, (WCMD)INTERELEMENT_SPACE + a, CEND );
	    set_window_controller( data->name, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b, CEND );
	    set_window_controller( data->name, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
	    set_window_controller( data->name, 3, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b + wm->text_ysize, CEND );

	    //FILES:
	    b += wm->text_ysize + INTERELEMENT_SPACE;
	    CREATE_SMOOTH_LIST = 1;
	    data->files = new_window( "files", 0, b, 100, 100, wm->list_background, win, list_handler, wm );
	    set_handler( data->files, files_list_handler, data, wm );
	    set_window_controller( data->files, 0, wm, CPERC, (WCMD)0, CADD, (WCMD)INTERELEMENT_SPACE, CEND );
	    set_window_controller( data->files, 1, wm, CWIN, (WCMD)data->preview_data.win, CY2, CADD, (WCMD)b, CEND );
	    set_window_controller( data->files, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
	    set_window_controller( data->files, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE * 2, CEND );

	    //BUTTONS:
	    {
		data->ok_button = new_window( "OK", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		data->cancel_button = new_window( "Cancel", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		data->del_button = new_window( "Delete", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		set_handler( data->ok_button, files_ok_button_handler, data, wm );
		set_handler( data->cancel_button, files_cancel_button_handler, data, wm );
		set_handler( data->del_button, files_ok_button_handler, data, wm );
		int x = INTERELEMENT_SPACE;
		set_window_controller( data->ok_button, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->ok_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->ok_button, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->ok_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
		x += 1;
		set_window_controller( data->cancel_button, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->cancel_button, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
		x += 1;
		set_window_controller( data->del_button, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->del_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->del_button, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->del_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
		if( data->preview )
		{
		    V3_FILE f = v3_open( "2:/.sundog_files_preview", "rb" );
		    if( f )
		    {
			if( v3_getc( f ) == 1 )
			    g_files_preview_flag = 1;
			else 
			    g_files_preview_flag = 0;
			v3_close( f );
		    }
		    COLOR pcolor;
		    if( g_files_preview_flag )
		    {
			pcolor = SELECTED_BUTTON_COLOR;
			data->preview_data.status = FPREVIEW_OPEN;
			data->preview_handler( &data->preview_data, wm );
		    }
		    else
		    {
			pcolor = wm->button_color;
			data->preview_data.status = FPREVIEW_NONE;
		    }
		    data->preview_button = new_window( data->preview, 0, 0, 1, 1, pcolor, win, button_handler, wm );
		    set_handler( data->preview_button, files_preview_button_handler, data, wm );
		    x += 1;
		    set_window_controller( data->preview_button, 0, wm, (WCMD)x, CEND );
		    set_window_controller( data->preview_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		    x += BUTTON_XSIZE;
		    set_window_controller( data->preview_button, 2, wm, (WCMD)x, CEND );
		    set_window_controller( data->preview_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
		}
		else 
		{
		    data->preview_button = 0;
		}
		if( fdialog_open_wifi )
		{
		    WINDOWPTR wifi_button = new_window( "Wi-Fi", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		    set_handler( wifi_button, files_wifi_button_handler, data, wm );
		    x += 1;
		    set_window_controller( wifi_button, 0, wm, (WCMD)x, CEND );
		    set_window_controller( wifi_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		    x += BUTTON_XSIZE;
		    set_window_controller( wifi_button, 2, wm, (WCMD)x, CEND );
		    set_window_controller( wifi_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );
		}
		if( FILES_USER_BUTTON )
		{
		    WINDOWPTR user_button = new_window( FILES_USER_BUTTON, 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		    set_handler( user_button, FILES_USER_BUTTON_HANDLER, FILES_USER_BUTTON_DATA, wm );
		    x += 1;
		    set_window_controller( user_button, 0, wm, (WCMD)x, CEND );
		    set_window_controller( user_button, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		    x += BUTTON_XSIZE;
		    set_window_controller( user_button, 2, wm, (WCMD)x, CEND );
		    set_window_controller( user_button, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)but_ysize + INTERELEMENT_SPACE, CEND );		    
		}
	    }

	    //LOAD LAST PROPERTIES:
	    recalc_regions( wm ); //for getting calculated size of file's list
    	    files_load_props( win, wm );
	    if( data->prop_path[ data->prop_cur_disk ] )
		text_set_text( data->path, data->prop_path[ data->prop_cur_disk ], wm );
	    if( data->prop_cur_file >= 0 )
	    {
		list_data* ldata = list_get_data( data->files, wm );
		int snum = list_get_selected_num( ldata );
		UTF8_CHAR* item = list_get_item( snum, ldata );
		//Set file num:
		list_select_item( data->files, data->prop_cur_file[ data->prop_cur_disk ], wm );
		if( FILES_DEF_FILENAME )
		{
		    text_set_text( data->name, FILES_DEF_FILENAME, wm );
		}
		else
		{
		    if( item )
		    {
			if( list_get_attr( snum, ldata ) == 0 )
			    text_set_text( data->name, item, wm );
		    }
		}
	    }

	    //SHOW CURRENT DISK:
	    if( data->disk_buttons[ data->prop_cur_disk ] )
	    {
		data->disk_buttons[ data->prop_cur_disk ]->color = SELECTED_BUTTON_COLOR;
	    }

	    //SET FOCUS:
	    set_focus_win( data->files, wm );

	    FILES_PROPS = 0;
	    FILES_MASK = 0;
	    FILES_RESULTED_FILENAME = 0;
	    FILES_PREVIEW = 0;
	    FILES_PREVIEW_HANDLER = 0;
	    FILES_USER_BUTTON = 0;
	    FILES_USER_BUTTON_HANDLER = 0;
	    FILES_USER_BUTTON_DATA = 0;
	    FILES_DEF_FILENAME = 0;
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    files_save_props( win, wm );
	    if( data->props_file )
		mem_free( (void*)data->props_file );
	    for( a = 0; a < 32; a++ )
	    {
		if( data->prop_path[ a ] ) mem_free( data->prop_path[ a ] );
	    }
	    if( data->preview )
	    {
		data->preview_data.status = FPREVIEW_CLOSE;
		data->preview_handler( &data->preview_data, wm );
		V3_FILE f = v3_open( "2:/.sundog_files_preview", "wb" );
		if( f )
		{
		    v3_putc( g_files_preview_flag, f );
		    v3_close( f );
		}
	    }
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
	    if( data->first_time )
	    {
		//Set file num:
		list_select_item( data->files, data->prop_cur_file[ data->prop_cur_disk ], wm );
		data->first_time = 0;
	    }
	    win_draw_unlock( win, wm );
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
    }
    return retval;
}
