/*
    handlers_scale.cpp
    This file is part of the SunDog engine.
    Copyright (C) 2011 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../main/user_code.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

WINDOWPTR g_ui_scale = 0;

bool g_ui_scale_changed = 0;

struct ui_scale_data
{
    WINDOWPTR this_window;
    WINDOWPTR ok;
    WINDOWPTR cancel;
    WINDOWPTR auto_win;
    WINDOWPTR ppi;
    WINDOWPTR scale;
    
    int content_ysize;

    bool auto_scale;    
    int cur_ppi;
    int cur_scale;
};

void ui_scale_reinit( WINDOWPTR win )
{
    ui_scale_data* data = (ui_scale_data*)win->data;
    UTF8_CHAR ts[ 32 ];
    const UTF8_CHAR* v;

    scrollbar_set_parameters( data->ppi, data->cur_ppi, 500, 10, 1, g_wm );
    scrollbar_set_parameters( data->scale, data->cur_scale, 1024, 10, 1, g_wm );
    
    if( data->auto_scale )
    {
	v = "ENABLED";    
	data->ppi->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
	data->scale->flags |= WIN_FLAG_ALWAYS_INVISIBLE;
	hide_window( data->ppi, g_wm );
	hide_window( data->scale, g_wm );
    }
    else
    {
	v = "DISABLED";
	data->ppi->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
	data->scale->flags &= ~WIN_FLAG_ALWAYS_INVISIBLE;
	show_window( data->ppi, g_wm );
	show_window( data->scale, g_wm );
    }

    sprintf( ts, "Auto scale = %s", v );
    button_set_text( data->auto_win, ts, g_wm );
}

int ui_scale_ok_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    ui_scale_data* data = (ui_scale_data*)user_data;

    if( g_ui_scale_changed )
    {
	if( data->auto_scale )
	{
	    profile_remove_key( KEY_PPI, 0 );
	    profile_remove_key( KEY_SCALE, 0 );
	}
	else
	{
	    int ppi = data->cur_ppi;
	    int scale = data->cur_scale;
	    if( ppi < 10 ) ppi = 10;
	    if( scale < 10 ) scale = 10;
	    profile_set_int_value( KEY_PPI, ppi, 0 );
	    profile_set_int_value( KEY_SCALE, scale, 0 );
	}
	profile_save( 0 );
	g_prefs_restart_request = 1;
    }
    
    remove_window( g_ui_scale, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    
    return 0;
}

int ui_scale_cancel_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    remove_window( g_ui_scale, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    return 0;
}

int ui_scale_def_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    ui_scale_data* data = (ui_scale_data*)user_data;

    data->auto_scale ^= 1;
    ui_scale_reinit( data->this_window );
    recalc_regions( wm );
    draw_window( data->this_window, wm );
    g_ui_scale_changed = 1;

    return 0;
}

int ui_scale_ppi_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    ui_scale_data* data = (ui_scale_data*)user_data;
    
    data->cur_ppi = scrollbar_get_value( win, wm );
    draw_window( data->this_window, wm );
    g_ui_scale_changed = 1;
    
    return 0;
}

int ui_scale_scale_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    ui_scale_data* data = (ui_scale_data*)user_data;
    
    data->cur_scale = scrollbar_get_value( win, wm );
    draw_window( data->this_window, wm );
    g_ui_scale_changed = 1;
    
    return 0;
}

int ui_scale_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    ui_scale_data* data = (ui_scale_data*)win->data;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( ui_scale_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		
		int y = INTERELEMENT_SPACE;
		data->auto_win = new_window( "Auto scale", 0, y, 1, wm->text_ysize, wm->button_color, win, button_handler, wm );
                set_handler( data->auto_win, ui_scale_def_handler, data, wm );
                set_window_controller( data->auto_win, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
                set_window_controller( data->auto_win, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
                y += wm->text_ysize + INTERELEMENT_SPACE;

		CREATE_COMPACT_SCROLLBAR = 1;                
                data->ppi = new_window( "PPI", 0, y, 1, CONTROLLER_YSIZE, wm->button_color, win, scrollbar_handler, wm );
                scrollbar_set_name( data->ppi, "PPI (Pixels Per Inch)", wm );
                set_handler( data->ppi, ui_scale_ppi_handler, data, wm );
                data->ppi->font = 0;
                set_window_controller( data->ppi, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
                set_window_controller( data->ppi, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
                y += CONTROLLER_YSIZE + INTERELEMENT_SPACE;

		CREATE_COMPACT_SCROLLBAR = 1;                
                data->scale = new_window( "Scale", 0, y, 1, CONTROLLER_YSIZE, wm->button_color, win, scrollbar_handler, wm );
                scrollbar_set_name( data->scale, "Button scale (normal=256)", wm );
                set_handler( data->scale, ui_scale_scale_handler, data, wm );
                data->scale->font = 0;
                set_window_controller( data->scale, 0, wm, (WCMD)INTERELEMENT_SPACE, CEND );
                set_window_controller( data->scale, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
                y += CONTROLLER_YSIZE + INTERELEMENT_SPACE;
                
                data->content_ysize = y;

		data->cur_ppi = wm->screen_ppi * wm->screen_zoom;
		data->cur_scale = (int)( wm->screen_scale * 256 );
		if( profile_get_int_value( KEY_PPI, 0, 0 ) == 0 && profile_get_int_value( KEY_SCALE, 0, 0 ) == 0 )
		    data->auto_scale = 1;
		else
		    data->auto_scale = 0;
		ui_scale_reinit( win );
                
		data->ok = new_window( "OK", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		data->cancel = new_window( "Cancel", 0, 0, 1, 1, wm->button_color, win, button_handler, wm );
		set_handler( data->ok, ui_scale_ok_handler, data, wm );
		set_handler( data->cancel, ui_scale_cancel_handler, data, wm );
		int x = INTERELEMENT_SPACE;
		set_window_controller( data->ok, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->ok, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->ok, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->ok, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE, CEND );
		x += 1;
		set_window_controller( data->cancel, 0, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel, 1, wm, CPERC, (WCMD)100, CSUB, (WCMD)INTERELEMENT_SPACE, CEND );
		x += BUTTON_XSIZE;
		set_window_controller( data->cancel, 2, wm, (WCMD)x, CEND );
		set_window_controller( data->cancel, 3, wm, CPERC, (WCMD)100, CSUB, (WCMD)BUTTON_YSIZE + INTERELEMENT_SPACE, CEND );
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    {
		win_draw_lock( win, wm );
		wbd_lock( win );
		
		draw_box( 0, 0, win->xsize, win->ysize, win->color, wm );
		
		if( data->auto_scale == 0 )
		{
		    wm->cur_font_color = win->color;
		    int i = wbd_char_y_size( wm ) + 2;
		    
		    int ppi = data->cur_ppi;
		    if( wm->screen_zoom > 1 ) ppi /= wm->screen_zoom;

		    wm->cur_transparency = 100;
		    int xs = ( ppi * data->cur_scale ) / 256;
		    draw_box( ( win->xsize - xs ) / 2, data->content_ysize, xs, i * 2, wm->color2, wm );
		    wm->cur_transparency = 255;
		    
		    draw_box( ( win->xsize - ppi ) / 2, data->content_ysize, ppi, i, wm->color2, wm );
		    draw_string( "Inch", ( win->xsize - wbd_string_size( "Inch", wm ) ) / 2, data->content_ysize + ( i - wbd_char_y_size( wm ) ) / 2, wm );
		}		    
		
		wbd_draw();
		wbd_unlock();
		win_draw_unlock( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	case EVT_MOUSEBUTTONUP:
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    retval = 1;
	    g_ui_scale = 0;
	    break;
    }
    return retval;
}
