/*
    wm_iphone.h. Platform-dependent module : framebuffer drawing functions
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#define FB_OFFSET		    g_wm->fb_offset
#define FB_XPITCH		    g_wm->fb_xpitch
#define FB_YPITCH		    g_wm->fb_ypitch

#define LINE_PREC 20

COLORPTR framebuffer = 0;

void device_draw_line( int x1, int y1, int x2, int y2, COLOR color )
{
    if( !g_wm->screen_is_active ) return;
    
#ifdef WINCE
    if( g_wm->vd == VIDEODRIVER_GDI )
    {
	color = ( color & 31 ) | ( ( color & ~63 ) >> 1 ); //RGB565 to RGB555
    }
#endif

    int fb_xpitch = FB_XPITCH;
    int fb_ypitch = FB_YPITCH;

    framebuffer[ FB_OFFSET + y1 * fb_ypitch + x1 * fb_xpitch ] = color;
    framebuffer[ FB_OFFSET + y2 * fb_ypitch + x2 * fb_xpitch ] = color;

    int len_x = x2 - x1; if( len_x < 0 ) len_x = -len_x;
    int len_y = y2 - y1; if( len_y < 0 ) len_y = -len_y;
    if( len_x ) len_x++;
    if( len_y ) len_y++;
    int xdir; if( x2 - x1 >= 0 ) xdir = fb_xpitch; else xdir = -fb_xpitch;
    int ydir; if( y2 - y1 >= 0 ) ydir = fb_ypitch; else ydir = -fb_ypitch;
    COLORPTR ptr = framebuffer + FB_OFFSET + y1 * fb_ypitch + x1 * fb_xpitch;
    int delta;
    int v = 0, old_v = 0;
    if( len_x > len_y )
    {
	//Horisontal:
	if( len_x != 0 )
	    delta = ( len_y << LINE_PREC ) / len_x;
	else
	    delta = 0;
	for( int a = 0; a < len_x; a++ )
	{
	    *ptr = color;
	    old_v = v;
	    v += delta;
	    ptr += xdir;
	    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		ptr += ydir;
	}
    }
    else
    {
	//Vertical:
	if( len_y != 0 ) 
	    delta = ( len_x << LINE_PREC ) / len_y;
	else
	    delta = 0;
	for( int a = 0; a < len_y; a++ )
	{
	    *ptr = color;
	    old_v = v;
	    v += delta;
	    ptr += ydir;
	    if( ( old_v >> LINE_PREC ) != ( v >> LINE_PREC ) ) 
		ptr += xdir;
	}
    }
}

void device_draw_box( int x, int y, int xsize, int ysize, COLOR color )
{
    if( !g_wm->screen_is_active ) return;

#ifdef WINCE
    if( g_wm->vd == VIDEODRIVER_GDI )
    {
	color = ( color & 31 ) | ( ( color & ~63 ) >> 1 ); //RGB565 to RGB555
    }
#endif

    int fb_xpitch = FB_XPITCH;
    int fb_ypitch = FB_YPITCH;

    COLORPTR ptr = framebuffer + FB_OFFSET + y * fb_ypitch + x * fb_xpitch;
    if( fb_xpitch == 1 )
    {
#if defined(COLOR16BITS) || defined(COLOR8BITS)
    /*
	if( ( (uint)ptr & 3 ) == 0 && ( x & 3 ) == 0 && ( xsize & 3 ) == 0 )
	{
	    uint *lptr = (uint*)ptr;
	    uint lxsize = ( xsize * COLORLEN ) / 4;
	    for( int cy = 0; cy < ysize; cy++ )
	    {
		uint *lsize = lptr + lxsize;
		while( lptr < lsize ) *lptr++ = color;
		ptr += fb_ypitch - xsize;
	    }
	}
	else
    */
#endif
	int add = fb_ypitch - xsize;
	for( int cy = 0; cy < ysize; cy++ )
	{
	    COLORPTR size = ptr + xsize;
	    while( ptr < size ) *ptr++ = color;
	    ptr += add;
	}
    }
    else
    {
	int add = fb_ypitch - ( xsize * fb_xpitch );
	for( int cy = 0; cy < ysize; cy++ )
	{
	    for( int cx = 0; cx < xsize; cx++ )
	    {
		*ptr = color;
		ptr += fb_xpitch;
	    }
	    ptr += add;
	}
    }
}

void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    sundog_image* img )
{
    if( !g_wm->screen_is_active ) return;

    int fb_xpitch = FB_XPITCH;
    int fb_ypitch = FB_YPITCH;

    COLORPTR data = (COLORPTR)img->data; //Source
    COLORPTR ptr = framebuffer + FB_OFFSET + dest_y * fb_ypitch + dest_x * fb_xpitch; //Destination

    //Add offset:
    int bp = src_y * img->xsize + src_x;
    data += bp;
    
    //Draw:
    int data_add = img->xsize - dest_xs;
#ifdef WINCE
    if( g_wm->vd == VIDEODRIVER_GDI )
    {
	COLOR color;
	if( fb_xpitch == 1 )
	{
	    int ptr_add = fb_ypitch - ( dest_xs * fb_xpitch );
	    for( int cy = 0; cy < dest_ys; cy++ )
	    {
		COLORPTR size = ptr + dest_xs;
		while( ptr < size ) 
		{
		    color = *data++;
		    color = ( color & 31 ) | ( ( color & ~63 ) >> 1 ); //RGB565 to RGB555
		    *ptr++ = color;
		}
		ptr += ptr_add;
		data += data_add;
	    }
	}
	else
	{
	    for( int cy = 0; cy < dest_ys; cy++ )
	    {
		int ptr_add = fb_ypitch - ( dest_xs * fb_xpitch );
		int cx = dest_xs + 1;
		while( --cx )
		{
		    color = *data;
		    color = ( color & 31 ) | ( ( color & ~63 ) >> 1 ); //RGB565 to RGB555
		    *ptr = color;
		    ptr += fb_xpitch;
		    data++;
		}
		ptr += ptr_add;
		data += data_add;
	    }
	}
    }
    else
#endif
    {
    
    if( fb_xpitch == 1 )
    {
	int ptr_add = fb_ypitch - ( dest_xs * fb_xpitch );
	for( int cy = 0; cy < dest_ys; cy++ )
	{
	    COLORPTR size = ptr + dest_xs;
	    while( ptr < size ) *ptr++ = *data++;
	    ptr += ptr_add;
	    data += data_add;
	}
    }
    else
    {
	for( int cy = 0; cy < dest_ys; cy++ )
	{
	    int ptr_add = fb_ypitch - ( dest_xs * fb_xpitch );
	    int cx = dest_xs + 1;
	    while( --cx )
	    {
		*ptr = *data;
		ptr += fb_xpitch;
		data++;
	    }
	    ptr += ptr_add;
	    data += data_add;
	}
    }
    
    }
}

void device_vsync( bool vsync )
{
}

int g_ccc = 0;
void device_redraw_framebuffer( void )
{
    if( g_wm->screen_changed == 0 ) return;

#ifdef OPENGL
    //OpenGL:
    /* draw polygon with the screen */
    glBindTexture( GL_TEXTURE_2D, g_wm->gl_fb_texture_id );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
#ifdef COLOR8BITS
    glTexSubImage2D( 
	GL_TEXTURE_2D,
        0,
	0, 0,
        g_wm->screen_xsize, g_wm->screen_ysize,
        GL_LUMINANCE,
        GL_UNSIGNED_BYTE,
        framebuffer );
#endif
#ifdef COLOR16BITS
    glTexSubImage2D( 
	GL_TEXTURE_2D,
        0,
	0, 0,
        g_wm->screen_xsize, g_wm->screen_ysize,
	GL_RGB,
        GL_UNSIGNED_SHORT_5_6_5,
        framebuffer );
#endif
#ifdef COLOR32BITS
    glTexSubImage2D( 
	GL_TEXTURE_2D,
        0,
	0, 0,
        g_wm->screen_xsize, g_wm->screen_ysize,
        GL_RGBA,
        GL_UNSIGNED_BYTE,
        framebuffer );
#endif
    int err = glGetError();
    if( err > 0 ) printf( "glTexSubImage2D() err: %x\n", err );
    float v[ 8 ];
    float t[ 8 ];
    float tx = (float)g_wm->screen_xsize / (float)g_wm->fb_xsize;
    float ty = (float)g_wm->screen_ysize / (float)g_wm->fb_ysize;
    v[ 0 ] = -1; v[ 1 ] = 1;
    v[ 2 ] = -1; v[ 3 ] = -1;
    v[ 4 ] = 1; v[ 5 ] = 1;
    v[ 6 ] = 1; v[ 7 ] = -1;
    t[ 0 ] = 0; t[ 1 ] = 0;
    t[ 2 ] = 0; t[ 3 ] = ty;
    t[ 4 ] = tx; t[ 5 ] = 0;
    t[ 6 ] = tx; t[ 7 ] = ty;
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );
    glVertexPointer( 2, GL_FLOAT, 0, v );
    glTexCoordPointer( 2, GL_FLOAT, 0, t );
    glColor4f( 1, 1, 1, 1 );
    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
#ifdef WIN
    glFlush();
    SwapBuffers( g_wm->hdc );
#endif
#ifdef UNIX
    #ifdef X11
	XSync( g_wm->dpy, 0 );
	glFlush();
	if( g_wm->doubleBuffer ) glXSwapBuffers( g_wm->dpy, g_wm->win );
    #endif
#endif //UNIX
    g_wm->screen_changed = 0;
#endif //OPENGL

#if defined(UNIX) && defined(DIRECTDRAW)
    if( g_wm->screen_lock_counter == 0 )
    {
	pthread_mutex_lock( &g_wm->sdl_lock_mutex );
	if( g_wm->screen_zoom == 2 )
	{
	    while( 1 )
	    {
		if( SDL_MUSTLOCK( g_wm->sdl_screen ) ) 
		{
		    if( SDL_LockSurface( g_wm->sdl_screen ) < 0 ) 
		    {
			break;
		    }
		}
		COLORPTR s = (COLORPTR)g_wm->sdl_screen->pixels;
		uint fb_ypitch = g_wm->sdl_screen->pitch / COLORLEN;
		uint fb_xpitch = 1;
		COLORPTR src = (COLORPTR)g_zoom_buffer;
		for( int y = 0; y < g_wm->real_window_height / 2; y++ )
		{
		    COLORPTR s2 = s;
		    for( int x = 0; x < g_wm->real_window_width / 2; x++ )
		    {
			COLOR c = *src;
			s2[ 0 ] = c;
			s2[ 1 ] = c;
			s2[ fb_ypitch ] = c;
			s2[ fb_ypitch + 1 ] = c;
			s2 += fb_xpitch * 2;
			src++;
		    }
		    s += fb_ypitch * 2;
		}
		if( SDL_MUSTLOCK( g_wm->sdl_screen ) ) 
		{
		    SDL_UnlockSurface( g_wm->sdl_screen );
		}
		break;
	    }
	}
	SDL_UpdateRect( g_wm->sdl_screen, 0, 0, 0, 0 );
	g_wm->screen_changed = 0;
	pthread_mutex_unlock( &g_wm->sdl_lock_mutex );
    }
    else
    {
	dprint( "Can't redraw the framebuffer: screen locked!\n" );
    }
#endif

#if defined(IPHONE)
    if( g_wm->screen_lock_counter == 0 )
    {
	iphone_sundog_screen_redraw();
	g_wm->screen_changed = 0;
    }
#endif

#if defined(WINCE) && defined(FRAMEBUFFER)
    //WinCE virtual framebuffer:
    if( framebuffer == 0 ) return;
    if( g_wm->vd == VIDEODRIVER_GDI )
    {
	g_wm->screen_changed = 0;
	int src_xsize = g_wm->screen_xsize;
	int src_ysize = g_wm->screen_ysize;
	if( g_wm->screen_angle == 1 || g_wm->screen_angle == 3 )
	{
	    int ttt = src_xsize;
	    src_xsize = src_ysize;
	    src_ysize = ttt;
	}
	BITMAPINFO bi;
	memset( &bi, 0, sizeof( BITMAPINFO ) );
	bi.bmiHeader.biSize = sizeof( BITMAPINFOHEADER );
	bi.bmiHeader.biWidth = src_xsize;
	bi.bmiHeader.biHeight = -src_ysize;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = COLORBITS;
	bi.bmiHeader.biCompression = BI_RGB;
	StretchDIBits(
	    g_wm->hdc,
	    0, // Destination top left hand corner X Position
	    0, // Destination top left hand corner Y Position
	    GetSystemMetrics( SM_CXSCREEN ), // Destinations Width
	    GetSystemMetrics( SM_CYSCREEN ), // Desitnations height
	    0, // Source low left hand corner's X Position
	    0, // Source low left hand corner's Y Position
	    src_xsize,
	    src_ysize,
	    framebuffer, // Source's data
	    &bi, // Bitmap Info
	    DIB_RGB_COLORS,
	    SRCCOPY );
    }
    if( g_wm->vd == VIDEODRIVER_RAW && g_wm->screen_zoom == 2 )
    {
	RawFrameBufferInfo* rfb = (RawFrameBufferInfo*)g_wm->rfb;
	COLORPTR s = (COLORPTR)rfb->pFramePointer;
	uint fb_ypitch = g_raw_fb_ypitch;
	uint fb_xpitch = g_raw_fb_xpitch;
	COLORPTR src = (COLORPTR)g_zoom_buffer;
	for( int y = 0; y < g_wm->real_window_height / 2; y++ )
	{
	    COLORPTR s2 = s;
	    for( int x = 0; x < g_wm->real_window_width / 2; x++ )
	    {
		COLOR c = *src;
		s2[ 0 ] = c;
		s2[ 1 ] = c;
		s2[ fb_ypitch ] = c;
		s2[ fb_ypitch + 1 ] = c;
		s2 += fb_xpitch * 2;
		src++;
	    }
	    s += fb_ypitch * 2;
	}
    }
    g_wm->screen_changed = 0;
#endif

#ifdef PALMOS
    //PalmOS:
#ifndef DIRECTDRAW
#ifdef PALMLOWRES
    WinDrawBitmap( (BitmapType*)g_screen_low, 0, 0 );
#else
    WinDrawBitmap( (BitmapType*)g_screen, 0, 0 );
#endif
#endif
    g_wm->screen_changed = 0;
#endif
}
