/*
    wm_iphone.h. Platform-dependent module : iPhone
    This file is part of the SunDog engine.
    Copyright (C) 2009 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#ifndef __WINMANAGER_IPHONE__
#define __WINMANAGER_IPHONE__

//#################################
//## DEVICE DEPENDENT FUNCTIONS: ##
//#################################

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>   	//for getenv()
#include <errno.h>
#include <sys/sem.h>    //semaphores ...

//Defined in various/sundog_bridge.m:
extern "C" void iphone_sundog_screen_redraw( void );
extern "C" void iphone_sundog_event_handler( void );
extern "C" void iphone_sundog_screen_angle_handler( int angle );
extern int g_iphone_sundog_screen_xsize;
extern int g_iphone_sundog_screen_ysize;
extern int g_iphone_sundog_screen_angle;
extern void* g_iphone_sundog_framebuffer;
extern GLuint g_view_renderbuffer;
extern GLuint g_view_framebuffer;

int g_prev_screen_angle = -1;

#ifdef OPENGL
#include "wm_opengl.h"
#endif

int g_frame_len;
ticks_t g_ticks_per_frame;
ticks_t g_prev_frame_time = 0;

#define EVENT /**/

extern "C" void iphone_sundog_exit_request( void )
{
    if( g_wm == 0 ) return;
    g_wm->exit_request = 1;
}

int g_event_cnt = 0;
sundog_event g_iphone_events[ 16 ];

extern "C" void iphone_sundog_touches( int x, int y, bool dont_draw, int mode, int touch_id )
{
    if( g_wm == 0 ) return;
    
    if( mode == -1 )
    {
	send_events( g_iphone_events, g_event_cnt, g_wm );
	EVENT;
	g_event_cnt = 0;
	return;
    }

    int temp;
    switch( g_wm->screen_angle )
    {
	case 1:
	    temp = x;
	    x = ( g_wm->real_window_height - 1 ) - y;
	    y = temp;
	    break;
	case 2:
	    x = ( g_wm->real_window_width - 1 ) - x;
	    y = ( g_wm->real_window_height - 1 ) - y;
	    break;
	case 3:
	    temp = x;
	    x = y;
	    y = ( g_wm->real_window_width - 1 ) - temp;
	    break;
    }
    
    if( g_wm->gl_xscale != 1 && g_wm->gl_yscale != 1 )
    {
	x = (int)( (float)x / g_wm->gl_xscale );
	y = (int)( (float)y / g_wm->gl_xscale );
    }
    
    int flags = 0;
    if( dont_draw )
	flags = EVT_FLAG_DONTDRAW;

    if( g_event_cnt >= 16 ) return;    
    sundog_event* evt = &g_iphone_events[ g_event_cnt ];
    g_event_cnt++;
    
    if( x < -16 ) x = -16;
    if( y < -16 ) y = -16;
    if( x > g_wm->screen_xsize + 16 ) x = g_wm->screen_xsize + 16;
    if( y > g_wm->screen_ysize + 16 ) y = g_wm->screen_ysize + 16;
    
    ticks_hr_t t = time_ticks_hires();
    
    if( touch_id == 0 )
    {
	switch( mode )
        {
	    case 0:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_MOUSEBUTTONDOWN;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = 0;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	    case 1:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_MOUSEMOVE;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = 0;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	    case 2:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_MOUSEBUTTONUP;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = 0;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	}
    }
    else
    {
	switch( mode )
	{
	    case 0:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_TOUCHBEGIN;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = touch_id;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	    case 1:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_TOUCHMOVE;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = touch_id;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	    case 2:
		evt->time = t;
		evt->win = 0;
		evt->type = EVT_TOUCHEND;
		evt->flags = flags;
		evt->x = x;
		evt->y = y;
		evt->key = MOUSE_BUTTON_LEFT;
		evt->scancode = touch_id;
		evt->pressure = 1024;
		evt->unicode = 0;
		break;
	}
    }
}

int device_start( const UTF8_CHAR* windowname, int xsize, int ysize, window_manager* wm )
{
    int retval = 0;
    
    g_prev_frame_time = 0;
    g_event_cnt = 0;

    g_frame_len = 1000 / wm->max_fps;
    g_ticks_per_frame = time_ticks_per_second() / wm->max_fps;

    wm->flags |= WIN_INIT_FLAG_FULLSCREEN;
    wm->flags |= WIN_INIT_FLAG_TOUCHCONTROL;
        
    xsize = g_iphone_sundog_screen_xsize;
    ysize = g_iphone_sundog_screen_ysize;
    wm->real_window_width = xsize;
    wm->real_window_height = ysize;

    if( wm->screen_angle & 1 )
    {
	wm->screen_xsize = ysize;
	wm->screen_ysize = xsize;
    }
    else
    {
	wm->screen_xsize = xsize;
	wm->screen_ysize = ysize;
    }
    
    if( wm->screen_ppi == 0 ) wm->screen_ppi = 160;
    win_zoom_init( wm );

#ifdef FRAMEBUFFER
    framebuffer = (COLORPTR)g_iphone_sundog_framebuffer;
    wm->fb_offset = 0;
    wm->fb_ypitch = g_iphone_sundog_screen_xsize;
    wm->fb_xpitch = 1;
#else
    #ifdef GLNORETAIN
	wm->screen_buffer_preserved = 0;
    #else
	wm->screen_buffer_preserved = 1;
    #endif
#endif

#ifdef OPENGL
    gl_init( wm );
    gl_resize( wm );
#endif

    return retval;
}

void device_end( window_manager *wm )
{
#ifdef OPENGL
    gl_deinit( wm );
#endif
}

void device_event_handler( window_manager* wm )
{
    int pause;
    
    iphone_sundog_event_handler();
    if( wm->exit_request ) return;
        
    if( wm->screen_angle != g_iphone_sundog_screen_angle 
        && !( wm->flags & WIN_INIT_FLAG_NOSCREENROTATE ) 
        && !wm->screen_angle_lock )
    {
	bool r = 1;
	if( wm->flags & WIN_INIT_FLAG_LANDSCAPE_ONLY )
	{
	    if( g_iphone_sundog_screen_angle == 1 || g_iphone_sundog_screen_angle == 3 )
		r = 1;
	    else 
		r = 0;
	}
        if( wm->flags & WIN_INIT_FLAG_PORTRAIT_ONLY )
	{
	    if( g_iphone_sundog_screen_angle == 0 || g_iphone_sundog_screen_angle == 2 )
		r = 1;
	    else 
		r = 0;
	}
	if( g_iphone_sundog_screen_angle >= 0 && r )
	{
	    wm->screen_angle = g_iphone_sundog_screen_angle;
	    wm->real_window_width = g_iphone_sundog_screen_xsize;
	    wm->real_window_height = g_iphone_sundog_screen_ysize;
	    if( wm->screen_angle & 1 )
	    {
		wm->screen_xsize = wm->real_window_height / wm->screen_zoom;
		wm->screen_ysize = wm->real_window_width / wm->screen_zoom;
	    }
	    else 
	    {
		wm->screen_xsize = wm->real_window_width / wm->screen_zoom;
		wm->screen_ysize = wm->real_window_height / wm->screen_zoom;
	    }
	    gl_resize( wm );
	    send_event( wm->root_win, EVT_SCREENRESIZE, 0, 0, 0, 0, 0, 0, 0, wm );
	}
    }
    
    if( wm->screen_angle != g_prev_screen_angle )
    {
	iphone_sundog_screen_angle_handler( wm->screen_angle );
	g_prev_screen_angle = wm->screen_angle;
    }

    if( wm->flags & WIN_INIT_FLAG_FULL_CPU_USAGE )
    {
	pause = 1;
    }
    else
    {
	pause = g_frame_len;
    }
    
    if( wm->events_count )
    {
	pause = 0;
    }
    
    if( pause )
    {
	time_sleep( pause );
    }
    
    ticks_t cur_t = time_ticks();
    if( cur_t - g_prev_frame_time >= g_ticks_per_frame )
    {
	wm->frame_event_request = 1;
	g_prev_frame_time = cur_t;
    }
}

void device_screen_lock( WINDOWPTR win, window_manager* wm )
{
    if( wm->screen_lock_counter == 0 )
    {
	gl_lock();
    }
    wm->screen_lock_counter++;

    if( wm->screen_lock_counter > 0 )
	wm->screen_is_active = 1;
    else
	wm->screen_is_active = 0;
}

void device_screen_unlock( WINDOWPTR win, window_manager* wm )
{
    if( wm->screen_lock_counter == 1 )
    {
	gl_unlock();
    }

    if( wm->screen_lock_counter > 0 )
    {
	wm->screen_lock_counter--;
    }

    if( wm->screen_lock_counter > 0 )
	wm->screen_is_active = 1;
    else
	wm->screen_is_active = 0;
}

#ifdef FRAMEBUFFER

#include "wm_framebuffer.h"

#endif

//#################################
//#################################
//#################################

#endif
