/*
    wmanager.cpp. SunDog window manager
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../core/core.h"
#include "../../main/user_code.h"
#include "../../core/debug.h"
#include "../../time/timemanager.h"
#include "../../utils/utils.h"
#include "../wmanager.h"
#include <stdarg.h>

#ifdef PALMOS
    #include <PalmOS.h>
    #include "palm_functions.h"
#endif

//Fonts (WIN1251)
#include "fonts.h"
sundog_image* g_font_img[ FONTS_NUM ];
int g_font_cxsize[ FONTS_NUM ];
int g_font_cysize[ FONTS_NUM ];

window_manager* g_wm = 0;

sundog_event g_sundog_frame_evt;
sundog_event g_empty_sundog_evt; //Only these fields can be changed: win; type; time;

UTF8_CHAR* g_status_message = 0;
WINDOWPTR g_status_window = 0;
int g_status_timer = -1;

bool g_fdialog_open = 0;
int g_dialog_open = 0;

uchar g_color_themes[ 3 * 4 * COLOR_THEMES ] = 
{
    0, 0, 0,
    0, 16, 0,
    120, 200, 120,
    255, 255, 255,
    
    255, 255, 255,
    150, 140, 130,
    60, 50, 40,
    0, 0, 0,

    0, 0, 0,
    20, 40, 20,
    120, 240, 120,
    255, 255, 255,

    0, 0, 0,
    20, 20, 20,
    180, 180, 180,
    255, 255, 255,

    0, 0, 0,
    32, 0, 0,
    240, 120, 120,
    255, 255, 255,

    0, 0, 0,
    0, 0, 40,
    120, 120, 240,
    255, 255, 255,

    255, 255, 255,
    130, 130, 130,
    60, 60, 60,
    0, 0, 0,

    255, 255, 255,
    130, 140, 150,
    40, 50, 60,
    0, 0, 0,

    255, 255, 255,
    130, 110, 100,
    80, 20, 10,
    0, 0, 0,

    0, 0, 0,
    0, 0, 80,
    100, 100, 200,
    255, 255, 255,

    0, 0, 0,
    80, 0, 80,
    200, 100, 200,
    255, 255, 255,

    0, 0, 0,
    40, 40, 0,
    0, 200, 100,
    255, 255, 255,

    0, 0, 0,
    40, 40, 40,
    190, 190, 190,
    255, 255, 255,

    255, 255, 255,
    80, 80, 90,
    16, 16, 32,
    0, 0, 0,

    255, 255, 255,
    80, 90, 80,
    16, 32, 16,
    0, 0, 0,

    255, 255, 255,
    70, 100, 100,
    16, 40, 40,
    0, 0, 0,

    255, 255, 255,
    70, 100, 180,
    16, 40, 40,
    0, 0, 0,

    0, 0, 0,
    16, 40, 40,
    70, 100, 180,
    255, 255, 255,

    0, 0, 0,
    16, 40, 40,
    180, 100, 70,
    255, 255, 255,

    0, 0, 0,
    40, 40, 40,
    200, 200, 0,
    255, 255, 255,
    
    255, 255, 255,
    80, 140, 210,
    16, 40, 80,
    0, 0, 0,

    255, 255, 255,
    80, 210, 210,
    32, 80, 80,
    0, 0, 0,

    255, 255, 255,
    220, 220, 150,
    80, 80, 32,
    0, 0, 0,

    255, 255, 255,
    220, 220, 220,
    80, 80, 80,
    0, 0, 0,

    255, 255, 255,
    220, 210, 200,
    80, 40, 32,
    0, 0, 0,
};

uchar g_texture0[ 4 * 4 ] = 
{
    0xFF, 0xFF, 0xFF, 0xFF,
    0x00, 0x00, 0x00, 0x00,
    0xFF, 0xFF, 0xFF, 0xFF,
    0x00, 0x00, 0x00, 0x00,
};

//################################
//## MAIN FUNCTIONS:            ##
//################################

int win_init(
    const UTF8_CHAR* windowname,
    int xsize,
    int ysize,
    int flags,
    int argc,
    UTF8_CHAR** argv,
    window_manager* wm )
{
    int retval = 0;

    //FIRST STEP: set defaults
    
    g_status_message = 0;
    g_status_window = 0;
    g_status_timer = -1;
    g_fdialog_open = 0;
    g_dialog_open = 0;
    
    wm->wm_initialized = 0;

    wm->argc = argc;
    wm->argv = argv;

    g_wm = wm;

    wm->events_count = 0;
    wm->current_event_num = 0;
    mem_set( &g_sundog_frame_evt, sizeof( g_sundog_frame_evt ), 0 );
    mem_set( &g_empty_sundog_evt, sizeof( g_empty_sundog_evt ), 0 );
    g_sundog_frame_evt.type = EVT_FRAME;
    sundog_mutex_init( &wm->events_mutex, 0 );
    
#ifdef OPENGL
    sundog_mutex_init( &wm->gl_mutex, 0 );
#endif    

    //Load settings from configuration file:
    
    windowname = (const UTF8_CHAR*)profile_get_str_value( KEY_WINDOWNAME, windowname, 0 );
    if( profile_get_int_value( KEY_NOBORDER, -1, 0 ) != -1 ) flags |= WIN_INIT_FLAG_NOBORDER;
    int angle = profile_get_int_value( KEY_ROTATE, 0, 0 );
    if( angle > 0 ) flags |= ( ( angle / 90 ) & 3 ) << WIN_INIT_FLAGOFF_ANGLE;
    int zoom = profile_get_int_value( KEY_ZOOM, 0, 0 );
    if( zoom > 0 ) flags |= ( zoom & 7 ) << WIN_INIT_FLAGOFF_ZOOM;
    if( profile_get_int_value( KEY_NOCURSOR, -1, 0 ) != -1 ) flags |= WIN_INIT_FLAG_NOCURSOR;
    if( profile_get_int_value( KEY_FULLSCREEN, -1, 0 ) != -1 ) flags |= WIN_INIT_FLAG_FULLSCREEN;
    if( profile_get_int_value( KEY_VSYNC, -1, 0 ) != -1 ) flags |= WIN_INIT_FLAG_VSYNC;
#if defined(UNIX)
    flags |= WIN_INIT_FLAG_OPTIMIZE_MOVE_EVENTS;
#endif

    //Save flags:
    wm->flags = flags;

    wm->restart_request = 0;
    wm->exit_request = 0;
    wm->exit_code = 0;

    wm->root_win = 0;

    wm->trash = 0;

    wm->window_counter = 0;

    mem_set( wm->timers, sizeof( wm->timers ), 0 );
    wm->timers_num = 0;
    wm->timers_id_counter = 0;

    wm->screen_buffer_preserved = 1;
    wm->screen_angle = ( flags >> WIN_INIT_FLAGOFF_ANGLE ) & 3;
    wm->screen_angle_lock = 0;
    wm->screen_zoom = ( flags >> WIN_INIT_FLAGOFF_ZOOM ) & 7;
    if( wm->screen_zoom == 0 ) wm->screen_zoom = 1;
    wm->screen_ppi = profile_get_int_value( KEY_PPI, 0, 0 ); //Default - 0 (unknown)
    wm->screen_scale = (float)profile_get_int_value( KEY_SCALE, 0, 0 ) / 256.0F; //Default - 0 (unknown)
    wm->screen_lock_counter = 0;
    wm->screen_is_active = 0;
    wm->screen_changed = 0;
    
    wm->show_virtual_keyboard = 0;
#ifdef VIRTUALKEYBOARD
    wm->show_virtual_keyboard = 1;
#endif
    if( profile_get_int_value( KEY_SHOW_VIRT_KBD, -1, 0 ) != -1 )
	wm->show_virtual_keyboard = 1;
    if( profile_get_int_value( KEY_HIDE_VIRT_KBD, -1, 0 ) != -1 )
	wm->show_virtual_keyboard = 0;
    
#ifdef IPHONE
    {
	V3_FILE f = v3_open( "2:/.sunvox_orient", "rb" );
	if( f )
	{
	    wm->screen_angle = v3_getc( f );
	    wm->screen_angle_lock = 1;
	    v3_close( f );
	}
	f = v3_open( "2:/.sundog_orient", "rb" );
	if( f )
	{
	    wm->screen_angle = v3_getc( f );
	    wm->screen_angle_lock = 1;
	    v3_close( f );
	    v3_remove( "2:/.sunvox_orient" );
	}
    }
#endif

    wm->pen_x = -1;
    wm->pen_y = -1;
    wm->focus_win = 0;

    wm->handler_of_unhandled_events = 0;

    //SECOND STEP: SCREEN SIZE SETTING and device specific init
    dprint( "MAIN: device start\n" );

#if CPUMARK >= 10
    wm->max_fps = 60;
#else
    wm->max_fps = 30;
#endif
#if defined(MAX_FPS)
    wm->max_fps = MAX_FPS;
#endif
    wm->max_fps = profile_get_int_value( KEY_MAXFPS, wm->max_fps, 0 );

    wm->fb_offset = 0;
    wm->fb_xpitch = 1;
    wm->fb_ypitch = 0;
    int err = device_start( windowname, xsize, ysize, wm ); //DEVICE DEPENDENT PART
    if( err )
    {
	dprint( "Error in device_start() %d\n", err );
	return 1; //Error
    }
    
    if( wm->screen_ppi == 0 ) wm->screen_ppi = 110;
    if( wm->screen_zoom > 1 ) wm->screen_ppi /= wm->screen_zoom;
    if( wm->screen_scale == 0 ) wm->screen_scale = 1;

    if( profile_get_int_value( KEY_TOUCHCONTROL, -1, 0 ) != -1 )
	wm->flags |= WIN_INIT_FLAG_TOUCHCONTROL;
    if( profile_get_int_value( KEY_PENCONTROL, -1, 0 ) != -1 )
	wm->flags &= ~WIN_INIT_FLAG_TOUCHCONTROL;
    if( wm->flags & WIN_INIT_FLAG_TOUCHCONTROL )
	wm->control_type = TOUCHCONTROL;
    else
	wm->control_type = PENCONTROL;
    if( wm->flags & WIN_INIT_FLAG_VSYNC )
	win_vsync( 1, wm );
    else
	win_vsync( 0, wm );
	
    dprint( "MAIN: screen_angle = %d\n", wm->screen_angle * 90 );
    dprint( "MAIN: screen_zoom = %d\n", wm->screen_zoom );
    dprint( "MAIN: screen_ppi = %d\n", wm->screen_ppi );
    dprint( "MAIN: screen_scale = %d / 256\n", (int)( wm->screen_scale * 256 ) );
    dprint( "MAIN: screen_xsize = %d\n", wm->screen_xsize );
    dprint( "MAIN: screen_ysize = %d\n", wm->screen_ysize );
    
    wm->g_screen = 0;
    wm->cur_window = 0;

    {
	int size, min;
	
	sundog_window tw;
	tw.font = 1; //Biggest font

	for( int i = 0; i < 2; i++ )
	{
	    min = 18; //smallest icon size = 16
	    if( min < char_y_size( &tw, wm ) + 2 )
		min = char_y_size( &tw, wm ) + 2;
	    float scrollbar_size = (float)wm->screen_ppi * 0.245;
	    size = (int)( scrollbar_size * wm->screen_scale );
	    if( size < min ) size = min;
	    wm->scrollbar_size = size;
	
	    int scrsize = wm->screen_xsize;
	    if( wm->screen_ysize < scrsize ) scrsize = wm->screen_ysize;
	    if( wm->scrollbar_size * 6 > scrsize )
	    {
		dprint( "MAIN: PPI or Scale is too large! They will be reduced temporary.\n" );
		wm->screen_scale = (float)scrsize / ( scrollbar_size * 6 );
	    }
	    else
	    {
		break;
	    }
	}

	min = 4;
    	size = (int)( (float)wm->screen_ppi * 0.07 * wm->screen_scale );
    	if( wm->control_type == TOUCHCONTROL ) size /= 2;
	if( size < min ) size = min;
	wm->decor_border_size = size;

	min = char_y_size( &tw, wm ) + 2;
	size = (int)( (float)wm->screen_ppi * 0.175 * wm->screen_scale );
	if( size < min ) size = min;
	wm->decor_header_size = size;

	min = wm->scrollbar_size * 2;
	size = (int)( (float)string_size( &tw, "Button ||23", wm ) );
	if( size < min ) size = min;
	wm->button_xsize = size;

	min = char_y_size( &tw, wm ) * 2;
	size = (int)( (float)wm->screen_ppi * 0.29 * wm->screen_scale );
	if( size < min ) size = min;
	wm->button_ysize = size;

	wm->normal_window_xsize = (int)( (float)wm->button_xsize * 4.5 );
	wm->normal_window_ysize = ( wm->normal_window_xsize * 11 ) / 16;
	wm->large_window_xsize = (int)( (float)wm->button_xsize * 7 );
	wm->large_window_ysize = ( wm->large_window_xsize * 11 ) / 16;

	min = char_y_size( &tw, wm );
	size = (int)( (float)wm->screen_ppi * 0.14 * wm->screen_scale );
	if( size < min ) size = min;
	wm->list_item_ysize = size;
	
	min = char_y_size( &tw, wm ) + 2;
	size = (int)( (float)wm->screen_ppi * 0.175 * wm->screen_scale );
	if( size < min ) size = min;
	wm->text_ysize = size;

	tw.font = 0; //Smallest font

	min = char_y_size( &tw, wm ) * 2 + 2;
	size = (int)( (float)wm->screen_ppi * 0.2 * wm->screen_scale );
	if( size < min ) size = min;
	wm->popup_button_ysize = size;

	min = char_y_size( &tw, wm ) * 2 + 2;
	size = (int)( (float)wm->screen_ppi * 0.2 * wm->screen_scale );
	if( size < min ) size = min;
	wm->controller_ysize = size;
	
	wm->corners_size = ( 2 * wm->screen_ppi ) / 160;
        wm->corners_len = ( 4 * wm->screen_ppi ) / 160;
        if( wm->corners_size < 2 ) wm->corners_size = 2;
        if( wm->corners_len < 4 ) wm->corners_len = 4;
    }

    wm->default_font = 1;
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    //Create font textures:
    for( int f = 0; f < FONTS_NUM; f++ )
    {
	uchar* font;
	switch( f )
	{
	    case 0: font = g_font0; break;
	    case 1: font = g_font1; break;
	}
	int cxsize = 8;
	int char_ysize = font[ 0 ];
	int cysize = font[ 0 ];
	if( cysize > 8 ) cysize = 16;
	g_font_cxsize[ f ] = cxsize;
	g_font_cysize[ f ] = cysize;
	uchar* tmp = (uchar*)MEM_NEW( HEAP_DYNAMIC, 16 * cxsize * 16 * cysize );
	int img_ptr = 0;
	int font_ptr = 257;
	for( int b = 257 + char_ysize; b < 257 + char_ysize * 2; b++ ) font[ b ] = 0xFF;
	for( int yy = 0; yy < cysize * 16; yy += cysize )
	{
	    for( int xx = 0; xx < cxsize * 16; xx += cxsize )
	    {
		img_ptr = yy * cxsize * 16 + xx;
		for( int fc = 0; fc < char_ysize; fc++ )
		{
		    uchar v = font[ font_ptr ];
		    for( int x = 0; x < cxsize; x++ )
		    {
			if( v & 128 ) 
			    tmp[ img_ptr ] = 255;
			else
			    tmp[ img_ptr ] = 0;
			v <<= 1;
			img_ptr++;
		    }
		    img_ptr += cxsize * 15;
		    font_ptr++;
		}
	    }
	}
	g_font_img[ f ] = new_image( 16 * cxsize, 16 * cysize, tmp, 0, 0, 16 * cxsize, 16 * cysize, IMAGE_ALPHA8 );
	mem_free( tmp );
    }
#else
    for( int f = 0; f < FONTS_NUM; f++ )
    {
	g_font_img[ f ] = 0;
    }
#endif

    wm->color_theme = profile_get_int_value( KEY_COLOR_THEME, 3, 0 );
    win_colors_init( wm );

    wm->texture0 = new_image( 4, 4, g_texture0, 0, 0, 4, 4, IMAGE_ALPHA8 );
    
    dprint( "MAIN: wmanager initialized\n" );

    wm->wm_initialized = 1;
    
    return retval;
}

void win_colors_init( window_manager* wm )
{
    dprint( "MAIN: system palette init\n" );
    
    bool theme = 0;
    if( profile_get_str_value( "theme", 0, 0 ) ) theme = 1;
    
    COLOR cc[ 4 ];
    for( int c = 0; c < 4; c++ )
    {
	uchar rr = g_color_themes[ wm->color_theme * 3 * 4 + c * 3 + 0 ];
	uchar gg = g_color_themes[ wm->color_theme * 3 * 4 + c * 3 + 1 ];
	uchar bb = g_color_themes[ wm->color_theme * 3 * 4 + c * 3 + 2 ];
	cc[ c ] = get_color( rr, gg, bb );
    }
    wm->color0 = cc[ 0 ];
    wm->color1 = cc[ 1 ];
    wm->color2 = cc[ 2 ];
    wm->color3 = cc[ 3 ];
    wm->green = get_color( 0, 255, 0 );
    wm->yellow = get_color( 255, 255, 0 );
    wm->red = get_color( 255, 0, 0 );
    wm->blue = get_color( 0, 0, 255 );
    
    if( theme )
    {
	dprint( "MAIN: color theme loading (base colors)\n" );
	UTF8_CHAR* v;
	v = profile_get_str_value( "c_0", 0, 0 ); if( v ) wm->color0 = get_color_from_string( v );
	v = profile_get_str_value( "c_1", 0, 0 ); if( v ) wm->color1 = get_color_from_string( v );
	v = profile_get_str_value( "c_2", 0, 0 ); if( v ) wm->color2 = get_color_from_string( v );
	v = profile_get_str_value( "c_3", 0, 0 ); if( v ) wm->color3 = get_color_from_string( v );
	v = profile_get_str_value( "c_yellow", 0, 0 ); if( v ) wm->yellow = get_color_from_string( v );
	v = profile_get_str_value( "c_green", 0, 0 ); if( v ) wm->green = get_color_from_string( v );
	v = profile_get_str_value( "c_red", 0, 0 ); if( v ) wm->red = get_color_from_string( v );
	v = profile_get_str_value( "c_blue", 0, 0 ); if( v ) wm->blue = get_color_from_string( v );
    }
    
    wm->header_text_color = wm->color2;
    wm->alternative_text_color = wm->yellow;
    wm->dialog_color = wm->color1;
    wm->decorator_color = wm->color1;
    wm->decorator_border = wm->color2;
    wm->button_color = blend( wm->color1, wm->color3, 12 );
    wm->menu_color = wm->color1;
    wm->selection_color = blend( get_color( 255, 255, 255 ), wm->green, 145 );
    wm->text_background = wm->color1;
    wm->list_background = wm->color1;
    wm->scroll_color = wm->button_color;
    wm->scroll_background_color = wm->color1;
    
    if( theme )
    {
	dprint( "MAIN: color theme loading (other colors)\n" );
	UTF8_CHAR* v;
	v = profile_get_str_value( "c_header", 0, 0 ); if( v ) wm->header_text_color = get_color_from_string( v );
	v = profile_get_str_value( "c_alt", 0, 0 ); if( v ) wm->alternative_text_color = get_color_from_string( v );
	v = profile_get_str_value( "c_dlg", 0, 0 ); if( v ) wm->dialog_color = get_color_from_string( v );
	v = profile_get_str_value( "c_dec", 0, 0 ); if( v ) wm->decorator_color = get_color_from_string( v );
	v = profile_get_str_value( "c_decbord", 0, 0 ); if( v ) wm->decorator_border = get_color_from_string( v );
	v = profile_get_str_value( "c_btn", 0, 0 ); if( v ) wm->button_color = get_color_from_string( v );
	v = profile_get_str_value( "c_menu", 0, 0 ); if( v ) wm->menu_color = get_color_from_string( v );
	v = profile_get_str_value( "c_sel", 0, 0 ); if( v ) wm->selection_color = get_color_from_string( v );
	v = profile_get_str_value( "c_txtback", 0, 0 ); if( v ) wm->text_background = get_color_from_string( v );
	v = profile_get_str_value( "c_lstback", 0, 0 ); if( v ) wm->list_background = get_color_from_string( v );
	v = profile_get_str_value( "c_scroll", 0, 0 ); if( v ) wm->scroll_color = get_color_from_string( v );
	v = profile_get_str_value( "c_scrollback", 0, 0 ); if( v ) wm->scroll_background_color = get_color_from_string( v );
    }
    
    if( red( wm->color3 ) + green( wm->color3 ) + blue( wm->color3 ) >= red( wm->color0 ) + green( wm->color0 ) + blue( wm->color0 ) )
    {
	wm->color3_is_brighter = 1;
    }
    else 
    {
	wm->color3_is_brighter = 0;
    }
    
    int v1 = 255 - red( wm->color1 );
    if( v1 < 0 ) v1 = -v1;
    int v2 = 255 - green( wm->color1 );
    if( v2 < 0 ) v2 = -v2;
    int v3 = 255 - blue( wm->color1 );
    if( v3 < 0 ) v3 = -v3;
    wm->color1_darkness = ( v1 + v2 + v3 ) / 3;
}

void win_reinit( window_manager* wm )
{
    win_colors_init( wm );
    
    mem_set( wm->timers, sizeof( wm->timers ), 0 );
    wm->timers_num = 0;
    wm->timers_id_counter = 0;
    
    wm->handler_of_unhandled_events = 0;
    
    wm->restart_request = 0;
    wm->exit_request = 0;
    wm->exit_code = 0;
}

void win_lock_unlock_orientation( window_manager* wm )
{
#ifdef IPHONE
    if( wm->screen_angle_lock )
    {
	v3_remove( "2:/.sunvox_orient" );
	v3_remove( "2:/.sundog_orient" );
	wm->screen_angle_lock = 0;
    }
    else 
    {
	V3_FILE f = v3_open( "2:/.sundog_orient", "wb" );
	if( f )
	{
	    v3_putc( wm->screen_angle, f );
	    v3_close( f );
	}
	wm->screen_angle_lock = 1;
    }
#endif
}

void win_zoom_init( window_manager* wm )
{
#ifdef FIXEDPPI
#ifdef SCREEN_ZOOM_AVAILABLE
    wm->screen_zoom = wm->screen_ppi / FIXEDPPI;
#endif
#else
    int base_font_ppi = 160;
    int base_font_ppi_limit = 200;
    int dec = base_font_ppi * 2 - base_font_ppi_limit;
    if( wm->screen_ppi / wm->screen_zoom >= base_font_ppi * 2 - dec )
    {
        //HACK. In future versions remove this code and make fonts resizable.
        //SunDog font will be too small on such PPI. Let's increase the pixel size of the screen:
#ifdef SCREEN_ZOOM_AVAILABLE
	if( ( wm->flags & WIN_INIT_FLAG_NOAUTOZOOM ) == 0 )
	{
    	    wm->screen_zoom *= ( wm->screen_ppi / wm->screen_zoom + dec ) / base_font_ppi;
    	}
#endif
    }
#endif
}

void win_vsync( bool vsync, window_manager* wm )
{
    if( vsync )
	wm->flags |= WIN_INIT_FLAG_VSYNC;
    else
	wm->flags &= ~WIN_INIT_FLAG_VSYNC;
    device_vsync( vsync );
}

void win_close( window_manager* wm )
{
    //Clear trash with a deleted windows:
    if( wm->trash )
    {
	for( unsigned int a = 0; a < mem_get_size( wm->trash ) / sizeof( WINDOWPTR ); a++ )
	{
	    if( wm->trash[ a ] ) 
	    {
		mem_free( wm->trash[ a ]->x1com );
		mem_free( wm->trash[ a ]->y1com );
		mem_free( wm->trash[ a ]->x2com );
		mem_free( wm->trash[ a ]->y2com );
		mem_free( wm->trash[ a ] );
	    }
	}
	mem_free( wm->trash );
    }

    for( int f = 0; f < FONTS_NUM; f++ )
    {
	if( g_font_img[ f ] )
	    remove_image( g_font_img[ f ] );
    }
    
    if( wm->screen_lock_counter > 0 )
    {
	dprint( "MAIN: WARNING. Screen is still locked (%d)\n", wm->screen_lock_counter );
	while( wm->screen_lock_counter > 0 ) device_screen_unlock( 0, wm );
    }

    if( wm->g_screen )
	mem_free( wm->g_screen );

    device_end( wm ); //DEVICE DEPENDENT PART (defined in eventloop.h)

    remove_image( wm->texture0 );    
    sundog_mutex_destroy( &wm->events_mutex );
#ifdef OPENGL
    sundog_mutex_destroy( &wm->gl_mutex );
#endif
    
    g_wm = 0;
}

//################################
//## WINDOWS FUNCTIONS:         ##
//################################

WINDOWPTR get_from_trash( window_manager *wm );

WINDOWPTR new_window( 
    const UTF8_CHAR *name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color, 
    WINDOWPTR parent, 
    int (*win_handler)( sundog_event*, window_manager* ),
    window_manager *wm )
{
    //Create window structure:
    WINDOWPTR win = get_from_trash( wm );
    if( win == 0 ) 
    {
	win = (WINDOWPTR)mem_new( HEAP_DYNAMIC, sizeof( WINDOW ), name );
	mem_set( win, sizeof( WINDOW ), 0 );
	win->id = (int16)wm->window_counter;
	wm->window_counter++;
    }

    //Setup properties:
    win->flags = 0;
    win->name = name;
    win->x = x;
    win->y = y;
    win->xsize = xsize;
    win->ysize = ysize;
    win->color = color;
    win->parent = parent;
    win->win_handler = ( int (*)( void*, void* ) )win_handler;
    win->click_time = time_ticks_hires() - time_ticks_per_second_hires() * 10;

    win->controllers_calculated = 0;
    win->controllers_defined = 0;

    win->action_handler = 0;
    win->handler_data = 0;
    win->action_result = 0;
    
    win->font = wm->default_font;

    //Start init:
    if( win_handler )
    {
	sundog_event* evt = &g_empty_sundog_evt;
	evt->win = win;
	evt->type = EVT_GETDATASIZE;
	int datasize = win_handler( evt, wm );
	if( datasize > 0 )
	{
	    win->data = mem_new( HEAP_DYNAMIC, datasize, "win data" );
	}
	evt->type = EVT_AFTERCREATE;
	win_handler( evt, wm );
    }

    //Save it to window manager:
    if( wm->root_win == 0 )
        wm->root_win = win;
    add_child( parent, win, wm );
    return win;
}

void set_window_controller( WINDOWPTR win, int ctrl_num, window_manager *wm, ... )
{
    va_list p;
    va_start( p, wm );
    int ptr = 0;
    win->controllers_defined = 1;
    WCMD* cmds = 0;
    switch( ctrl_num )
    {
	case 0: cmds = win->x1com; break;
	case 1: cmds = win->y1com; break;
	case 2: cmds = win->x2com; break;
	case 3: cmds = win->y2com; break;
    }
    if( cmds == 0 )
	cmds = (WCMD*)MEM_NEW( HEAP_DYNAMIC, sizeof( WCMD ) * 4 );
    while( 1 )
    {
	WCMD command = va_arg( p, WCMD );
	if( mem_get_size( cmds ) / sizeof( WCMD ) <= ptr )
	    cmds = (WCMD*)mem_resize( cmds, sizeof( WCMD ) * ( ptr + 4 ) );
	cmds[ ptr ] = command; 
	if( command == CEND ) break;
	ptr++;
    }
    switch( ctrl_num )
    {
	case 0: win->x1com = cmds; break;
	case 1: win->y1com = cmds; break;
	case 2: win->x2com = cmds; break;
	case 3: win->y2com = cmds; break;
    }
    va_end( p );
}

void move_to_trash( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	win->visible = 0;
	win->flags |= WIN_FLAG_TRASH;
	if( wm->trash == 0 )
	{
	    wm->trash = (WINDOWPTR*)MEM_NEW( HEAP_DYNAMIC, sizeof( WINDOWPTR ) * 16 );
	    mem_set( wm->trash, sizeof( WINDOWPTR ) * 16, 0 );
	    wm->trash[ 0 ] = win;
	}
	else
	{
	    unsigned int w = 0;
	    for( w = 0; w < mem_get_size( wm->trash ) / sizeof( WINDOWPTR ); w++ )
	    {
		if( wm->trash[ w ] == 0 ) break;
	    }
	    if( w < mem_get_size( wm->trash ) / sizeof( WINDOWPTR ) ) 
		wm->trash[ w ] = win;
	    else
	    {
		wm->trash = (WINDOWPTR*)mem_resize( wm->trash, mem_get_size( wm->trash ) + sizeof( WINDOWPTR ) * 16 );
		wm->trash[ w ] = win;
	    }
	}
    }
}

WINDOWPTR get_from_trash( window_manager* wm )
{
    if( wm->trash == 0 ) return 0;
    unsigned int w = 0;
    for( w = 0; w < mem_get_size( wm->trash ) / sizeof( WINDOWPTR ); w++ )
    {
	if( wm->trash[ w ] != 0 ) 
	{
	    WINDOWPTR win = wm->trash[ w ];
	    wm->trash[ w ] = 0;
	    win->id++;
	    win->flags = 0;
	    return win;
	}
    }
    return 0;
}

void remove_window( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	if( win->flags & WIN_FLAG_TRASH )
	{
	    dprint( "ERROR: can't remove already removed window (%s)\n", win->name );
	    return;
	}
	if( win->win_handler )
	{
	    //Sent EVT_BEFORECLOSE to window handler:
	    sundog_event* evt = &g_empty_sundog_evt;
	    evt->win = win;
	    evt->type = EVT_BEFORECLOSE;
	    win->win_handler( evt, wm );
	}
	if( win->childs )
	{
	    //Remove childs:
	    while( win->childs_num )
		remove_window( win->childs[ 0 ], wm );
	    mem_free( win->childs );
	    win->childs = 0;
	}
	//Remove data:
	if( win->data ) 
	    mem_free( win->data );
	win->data = 0;
	//Remove region:
	if( win->reg ) 
	    GdDestroyRegion( win->reg );
	win->reg = 0;
	//Remove commands:
	mem_free( win->x1com ); win->x1com = 0;
	mem_free( win->y1com ); win->y1com = 0;
	mem_free( win->x2com ); win->x2com = 0;
	mem_free( win->y2com ); win->y2com = 0;
	//Remove window:
	if( win == wm->focus_win )
	    wm->focus_win = 0;
	if( win == wm->root_win )
	{
	    //It's root win:
	    move_to_trash( win, wm );
	    wm->root_win = 0;
	}
	else
	{
	    remove_child( win->parent, win, wm );
	    //Full remove:
	    move_to_trash( win, wm );
	}
    }
}

void add_child( WINDOWPTR win, WINDOWPTR child, window_manager* wm )
{
    if( win == 0 || child == 0 ) return;

    if( win->childs == 0 )
    {
	win->childs = (WINDOWPTR*)mem_new( HEAP_DYNAMIC, sizeof( WINDOWPTR ) * 4, "childs" );
    }
    else
    {
	size_t old_size = mem_get_size( win->childs ) / sizeof( WINDOWPTR );
	if( win->childs_num >= old_size )
	    win->childs = (WINDOWPTR*)mem_resize( win->childs, ( old_size + 4 ) * sizeof( WINDOWPTR ) );
    }
    win->childs[ win->childs_num ] = child;
    win->childs_num++;
    child->parent = win;
}

void remove_child( WINDOWPTR win, WINDOWPTR child, window_manager* wm )
{
    if( win == 0 || child == 0 ) return;

    //Remove link from parent window:
    int c;
    for( c = 0; c < win->childs_num; c++ )
    {
	if( win->childs[ c ] == child ) break;
    }
    if( c < win->childs_num )
    {
	for( ; c < win->childs_num - 1; c++ )
	{
	    win->childs[ c ] = win->childs[ c + 1 ];
	}
	win->childs_num--;
	child->parent = 0;
    }
}

void set_handler( WINDOWPTR win, int (*handler)(void*,WINDOWPTR,window_manager*), void* handler_data, window_manager* wm )
{
    if( win )
    {
	win->action_handler = (int (*)(void*,void*,void*)) handler;;
	win->handler_data = handler_data;
    }
}

bool is_window_visible( WINDOWPTR win, window_manager* wm )
{
    if( win && win->visible && win->reg )
    {
	if( win->reg->numRects )
	{
	    return 1;
	}
    }
    return 0;
}

void draw_window( WINDOWPTR win, window_manager* wm )
{
    if( wm->screen_buffer_preserved == 0 ) 
    {
	screen_changed();
	return;
    }
    if( win && win->visible && win->reg )
    {
	win_draw_lock( win, wm );
	if( win->reg->numRects )
	{
	    sundog_event* evt = &g_empty_sundog_evt;
    	    evt->win = win;
	    evt->type = EVT_DRAW;
	    if( win->win_handler && win->win_handler( evt, wm ) )
	    {
		//Draw event was handled
	    }
	    else
	    {
	        win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
	    }
	}
	if( win->childs_num )
	{
	    for( int c = 0; c < win->childs_num; c++ )
	    {
		draw_window( win->childs[ c ], wm );
	    }
	}
	win_draw_unlock( win, wm );
    }
}

void show_window( WINDOWPTR win, window_manager* wm )
{
    if( win && ( win->flags & WIN_FLAG_ALWAYS_INVISIBLE ) == 0 )
    {
	if( win->visible == 0 )
	{
	    sundog_event* evt = &g_empty_sundog_evt;
	    evt->win = win;
	    evt->type = EVT_BEFORESHOW;
	    win->win_handler( evt, wm );
	}
	win->visible = 1;
	for( int c = 0; c < win->childs_num; c++ )
	{
	    show_window( win->childs[ c ], wm );
	}
    }
}

void hide_window( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	if( wm->focus_win == win )
	{
	    //We are focused on hidden window. It is wrong
	    wm->focus_win = 0;
	}
	win->visible = 0;
	for( int c = 0; c < win->childs_num; c++ )
	{
	    hide_window( win->childs[ c ], wm );
	}
    }
}

void recalc_controllers( WINDOWPTR win, window_manager* wm );

void run_controller( WINDOWPTR win, WCMD* c, int* val, int size, window_manager* wm )
{
    if( c == 0 ) return;
    int p = 0;
    WINDOWPTR other_win = 0;
    int a;
    int mode = 0;
    int perc = 0;
    int backval = 0;
    int brk = 0;
    while( !brk )
    {
	switch( c[ p ] )
	{
	    case CWIN: 
		p++;
		other_win = (WINDOWPTR)c[ p ];
		if( other_win->controllers_calculated == 0 ) recalc_controllers( other_win, wm );
		break;
	    case CX1:
		a = other_win->x;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CY1: 
		a = other_win->y;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CX2:
		a = other_win->x + other_win->xsize;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CY2: 
		a = other_win->y + other_win->ysize;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CXSIZE:
		a = other_win->xsize;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CYSIZE: 
		a = other_win->ysize;
		switch( mode )
		{
		    case 0: *val = a; break;
		    case 1: *val -= a; break;
		    case 2: *val += a; break;
		    case 3: if( *val > a ) *val = a; break;
		    case 4: if( *val < a ) *val = a; break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CSUB: mode = 1; break;
	    case CADD: mode = 2; break;
	    case CPERC: perc = 1; break;
	    case CBACKVAL0: backval = 0; break;
	    case CBACKVAL1: backval = 1; break;
	    case CMAXVAL: mode = 3; break;
	    case CMINVAL: mode = 4; break;
	    case CPUTR0: wm->creg0 = *val; break;
	    case CGETR0: *val = wm->creg0; break;
	    case CR0:
		switch( mode )
		{
		    case 0: // =
			*val = wm->creg0;
			break;
		    case 1: // sub
			*val -= wm->creg0;
			break;
		    case 2: // add
			*val += wm->creg0;
			break;
		    case 3: // max
			if( *val > wm->creg0 ) *val = wm->creg0;
			break;
		    case 4: // min
			if( *val < wm->creg0 ) *val = wm->creg0;
			break;
		}
		mode = 0;
		perc = 0;
		break;
	    case CEND: brk = 1; break;
	    default: 
		switch( mode )
		{
		    case 0:
			if( perc )
			{
			    *val = ( (int)c[ p ] * size ) / 100;
			}
			else
			{
			    if( backval )
			    {
				*val = size - (int)c[ p ];
			    }
			    else
			    {
				*val = (int)c[ p ]; //Simple number
			    }
			}
			break;
		
		    case 1:
			//Sub:
			if( perc )
			{
			    *val -= ( (int)c[ p ] * size ) / 100;
			}
			else
			{
			    if( backval )
			    {
				*val -= size - (int)c[ p ];
			    }
			    else
			    {
				*val -= (int)c[ p ]; //Simple number
			    }
			}
			break;

		    case 2:
			//Add:
			if( perc )
			{
			    *val += ( (int)c[ p ] * size ) / 100;
			}
			else
			{
			    if( backval )
			    {
				*val += size - (int)c[ p ];
			    }
			    else
			    {
				*val += (int)c[ p ]; //Simple number
			    }
			}
			break;
		
		    case 3:
			//Max val:
			if( perc )
			{
			    if( *val > ( (int)c[ p ] * size ) / 100 ) *val = ( (int)c[ p ] * size ) / 100;
			}
			else
			{
			    if( backval )
			    {
				if( *val > size - (int)c[ p ] ) *val = size - (int)c[ p ];
			    }
			    else
			    {
				if( *val > (int)c[ p ] ) *val = (int)c[ p ]; //Simple number
			    }
			}
			break;	
		    
		    case 4:
			//Min val:
			if( perc )
			{
			    if( *val < ( (int)c[ p ] * size ) / 100 ) *val = ( (int)c[ p ] * size ) / 100;
			}
			else
			{
			    if( backval )
			    {
				if( *val < size - (int)c[ p ] ) *val = size - (int)c[ p ];
			    }
			    else
			    {
				if( *val < (int)c[ p ] ) *val = (int)c[ p ]; //Simple number
			    }
			}
			break;
		}
		mode = 0;
		perc = 0;
		break;
	}
	p++;
    }
}

void bring_to_front( WINDOWPTR win, window_manager* wm )
{
    if( win == 0 ) return;
    if( win->parent == 0 ) return;
    
    int i;
    for( i = 0; i < win->parent->childs_num; i++ )
    {
	if( win->parent->childs[ i ] == win ) break;
    }
    if( i < win->parent->childs_num - 1 )
    {
	for( int i2 = i; i2 < win->parent->childs_num - 1; i2++ )
	    win->parent->childs[ i2 ] = win->parent->childs[ i2 + 1 ];
	win->parent->childs[ win->parent->childs_num - 1 ] = win;
    }
}

void recalc_controllers( WINDOWPTR win, window_manager* wm )
{
    if( win && win->controllers_calculated == 0 && win->parent )
    {
	if( win->controllers_defined == 0 || ( win->flags & WIN_FLAG_DONT_USE_CONTROLLERS ) ) 
	    win->controllers_calculated = 1;
	else
	{
	    int x1 = win->x;
	    int y1 = win->y;
	    int x2 = win->x + win->xsize;
	    int y2 = win->y + win->ysize;
	    run_controller( win, win->x1com, &x1, win->parent->xsize, wm );
	    run_controller( win, win->x2com, &x2, win->parent->xsize, wm );
	    run_controller( win, win->y1com, &y1, win->parent->ysize, wm );
	    run_controller( win, win->y2com, &y2, win->parent->ysize, wm );
	    int temp;
	    if( x1 > x2 ) { temp = x1; x1 = x2; x2 = temp; }
	    if( y1 > y2 ) { temp = y1; y1 = y2; y2 = temp; }
	    win->x = x1;
	    win->y = y1;
	    win->xsize = x2 - x1;
	    win->ysize = y2 - y1;
	    win->controllers_calculated = 1;
	}
    }
}

void recalc_region( WINDOWPTR win, MWCLIPREGION* reg, int cut_x, int cut_y, int cut_x2, int cut_y2, int px, int py, window_manager* wm )
{
    if( !win->visible )
    {
	if( win->reg ) GdDestroyRegion( win->reg );
	win->reg = 0;
	return;
    }
    if( win->controllers_defined && win->controllers_calculated == 0 )
    {
	recalc_controllers( win, wm );
    }
    win->screen_x = win->x + px;
    win->screen_y = win->y + py;
    int x1 = win->x + px;
    int y1 = win->y + py;
    int x2 = win->x + px + win->xsize;
    int y2 = win->y + py + win->ysize;
    if( cut_x > x1 ) x1 = cut_x;
    if( cut_y > y1 ) y1 = cut_y;
    if( cut_x2 < x2 ) x2 = cut_x2;
    if( cut_y2 < y2 ) y2 = cut_y2;
    if( win->childs_num && !( x1 > x2 || y1 > y2 ) )
    {
	for( int c = win->childs_num - 1; c >= 0; c-- )
	{
	    recalc_region( 
		win->childs[ c ], 
		reg, 
		x1, y1, 
		x2, y2, 
		win->x + px, 
		win->y + py, 
		wm );
	}
    }
    if( win->reg == 0 )
    {
	if( x1 > x2 || y1 > y2 )
	    win->reg = GdAllocRegion();
	else
	    win->reg = GdAllocRectRegion( x1, y1, x2, y2 );
    }
    else
    {
	if( x1 > x2 || y1 > y2 )
	{
	    GdSetRectRegion( win->reg, 0, 0, 0, 0 );
	}
	else
	{
	    GdSetRectRegion( win->reg, x1, y1, x2, y2 );
	}
    }
    //reg - current invisible region.
    //Calc corrected win region:
    GdSubtractRegion( win->reg, win->reg, reg );
    //Calc corrected invisible region:
    GdUnionRegion( reg, reg, win->reg );
}

void clean_regions( WINDOWPTR win, window_manager* wm )
{
    if( win )
    {
	for( int c = 0; c < win->childs_num; c++ )
	    clean_regions( win->childs[ c ], wm );
	if( win->reg ) 
	{
	    GdSetRectRegion( win->reg, 0, 0, 0, 0 );
	}
	win->controllers_calculated = 0;
    }
}

void send_afterresize_event( WINDOWPTR win )
{
    if( win )
    {
	if( win->visible )
	{
	    sundog_event* evt = &g_empty_sundog_evt;
	    evt->win = win;
	    evt->type = EVT_AFTERRESIZE;
	    if( win->win_handler )
		win->win_handler( evt, g_wm );
	    for( int c = 0; c < win->childs_num; c++ )
	    {
		send_afterresize_event( win->childs[ c ] );
	    }
	}
    }

}

void recalc_regions( window_manager* wm )
{
    MWCLIPREGION* reg = GdAllocRegion();
    if( wm->root_win )
    {
	//Control ALWAYS_ON_TOP flag: ===
	int size = wm->root_win->childs_num;
	for( int i = 0; i < size; i++ )
	{
	    WINDOWPTR win = wm->root_win->childs[ i ];
	    if( win && ( win->flags & WIN_FLAG_ALWAYS_ON_TOP ) && i < size - 1 )
	    {
		//Bring this window to front:
		for( int i2 = i; i2 < size - 1; i2++ )
		{
		    wm->root_win->childs[ i2 ] = wm->root_win->childs[ i2 + 1 ];
		}
		wm->root_win->childs[ size - 1 ] = win;
	    }
	}
	//===============================
	clean_regions( wm->root_win, wm );
	recalc_region( wm->root_win, reg, 0, 0, wm->screen_xsize, wm->screen_ysize, 0, 0, wm );
    }
    if( reg ) GdDestroyRegion( reg );
    send_afterresize_event( wm->root_win );
}

void set_focus_win( WINDOWPTR win, window_manager* wm )
{
    if( win && ( win->flags & WIN_FLAG_TRASH ) )
    {
	//This window removed by someone. But we can't focus on removed window.
	//So.. Just focus on NULL window:
	win = 0;
    }

    sundog_event evt;
    mem_set( &evt, sizeof( sundog_event ), 0 );
    evt.time = time_ticks_hires();

    //Unfocus:
    
    WINDOWPTR prev_focus_win = wm->focus_win;
    uint16 prev_focus_win_id = wm->focus_win_id;

    WINDOWPTR new_focus_win = win;
    uint16 new_focus_win_id = 0;
    if( win ) new_focus_win_id = win->id;
    
    if( prev_focus_win != new_focus_win || 
	( prev_focus_win == new_focus_win && prev_focus_win_id != new_focus_win_id ) )
    {
	//Focus changed:
	
	if( prev_focus_win && !( prev_focus_win->flags & WIN_FLAG_TRASH ) )
	{
	    //Send UNFOCUS event:
	    evt.type = EVT_UNFOCUS;
	    evt.win = prev_focus_win;
	    if( !( evt.win->flags & WIN_FLAG_UNFOCUS_HANDLING ) )
	    {
		evt.win->flags |= WIN_FLAG_UNFOCUS_HANDLING;
		handle_event( &evt, wm );
		evt.win->flags &= ~WIN_FLAG_UNFOCUS_HANDLING;
	    }
	}
    }
    
    //Focus on the new window:

    prev_focus_win = wm->focus_win;
    prev_focus_win_id = wm->focus_win_id;

    new_focus_win = win;
    new_focus_win_id = 0;
    if( win ) new_focus_win_id = win->id;

    wm->focus_win = new_focus_win;
    wm->focus_win_id = new_focus_win_id;

    if( prev_focus_win != new_focus_win || 
	( prev_focus_win == new_focus_win && prev_focus_win_id != new_focus_win_id ) )
    {
	//Focus changed:

        wm->prev_focus_win = prev_focus_win;
        wm->prev_focus_win_id = prev_focus_win_id;

	if( new_focus_win )
	{
	    //Send FOCUS event:
	    //In this event's handling user can remember previous focused window (wm->prev_focus_win)
	    evt.type = EVT_FOCUS;
	    evt.win = new_focus_win;
	    if( !( evt.win->flags & WIN_FLAG_FOCUS_HANDLING ) )
	    {
		evt.win->flags |= WIN_FLAG_FOCUS_HANDLING;
		handle_event( &evt, wm );
		evt.win->flags &= ~WIN_FLAG_FOCUS_HANDLING;
	    }
	}
    }
}

int find_focus_window( WINDOWPTR win, WINDOW **focus_win, window_manager *wm )
{
    if( win == 0 ) return 0;
    if( win->reg && GdPtInRegion( win->reg, wm->pen_x, wm->pen_y ) )
    {
	*focus_win = win;
	return 1;
    }
    for( int c = 0; c < win->childs_num; c++ )
    {
	if( find_focus_window( win->childs[ c ], focus_win, wm ) )
	    return 1;
    }
    return 0;
}

int send_events(
    sundog_event* events,
    int events_num,
    window_manager* wm )
{
    int retval = 0;
    
    sundog_mutex_lock( &wm->events_mutex );

    for( int e = 0; e < events_num; e++ )
    {
	sundog_event* evt = &events[ e ];
	if( wm->events_count + 1 <= EVENTS )
	{
	    //Get pointer to new event:
	    int new_ptr = ( wm->current_event_num + wm->events_count ) & ( EVENTS - 1 );
	
	    if( evt->type == EVT_BUTTONDOWN || evt->type == EVT_BUTTONUP )
	    {
		if( evt->key >= 0x41 && evt->key <= 0x5A ) //Capital
		    evt->key += 0x20; //Make it small
	    }
	
	    //Save new event to FIFO buffer:
	    mem_copy( &wm->events[ new_ptr ], evt, sizeof( sundog_event ) );
	
	    //Increment number of unhandled events:
	    volatile int new_event_count = wm->events_count + 1;
	    wm->events_count = new_event_count;
	}
	else
	{
	    retval = 1;
	    break;
	}
    }
    
    sundog_mutex_unlock( &wm->events_mutex );
    
    return retval;
}

int send_event(
    WINDOWPTR win,
    int type,
    int flags,
    int x,
    int y,
    int key,
    int scancode,
    int pressure,
    UTF32_CHAR unicode,
    window_manager* wm )
{
    int retval = 0;
    
    sundog_mutex_lock( &wm->events_mutex );
    
    if( wm->events_count + 1 <= EVENTS )
    {
	//Get pointer to new event:
	int new_ptr = ( wm->current_event_num + wm->events_count ) & ( EVENTS - 1 );

	if( type == EVT_BUTTONDOWN || type == EVT_BUTTONUP )
	{
	    if( key >= 0x41 && key <= 0x5A ) //Capital
		key += 0x20; //Make it small
	}

	//Save new event to FIFO buffer:
	wm->events[ new_ptr ].type = (uint16)type;
	wm->events[ new_ptr ].time = time_ticks_hires();
	wm->events[ new_ptr ].win = win;
	wm->events[ new_ptr ].flags = (uint16)flags;
	wm->events[ new_ptr ].x = (int16)x;
	wm->events[ new_ptr ].y = (int16)y;
	wm->events[ new_ptr ].key = (uint16)key;
	wm->events[ new_ptr ].scancode = (uint16)scancode;
	wm->events[ new_ptr ].pressure = (uint16)pressure;
	wm->events[ new_ptr ].unicode = unicode;

	//Increment number of unhandled events:
	wm->events_count = wm->events_count + 1;
    }
    else
    {
	retval = 1;
    }

    sundog_mutex_unlock( &wm->events_mutex );
    
    return retval;
}

int check_event( sundog_event *evt, window_manager *wm )
{
    if( evt == 0 ) return 1;

    if( evt->win == 0 )
    {
	if( !wm->wm_initialized ) return 1;
	
	if( evt->type == EVT_MOUSEBUTTONDOWN ||
	    evt->type == EVT_MOUSEBUTTONUP ||
	    evt->type == EVT_MOUSEMOVE )
	{
	    wm->pen_x = evt->x;
	    wm->pen_y = evt->y;
	    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	    if( wm->last_unfocused_window )
	    {
		evt->win = wm->last_unfocused_window;
		if( evt->type == EVT_MOUSEBUTTONUP )
		{
		    wm->last_unfocused_window = 0;
		}
		return 0;
	    }
	    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	    if( evt->type == EVT_MOUSEBUTTONDOWN )
	    { //If mouse click:
		if( evt->key & MOUSE_BUTTON_SCROLLUP ||
		    evt->key & MOUSE_BUTTON_SCROLLDOWN )
		{
		    //Mouse scroll up/down...
		    WINDOWPTR scroll_win = 0;
		    if( find_focus_window( wm->root_win, &scroll_win, wm ) )
		    {
			evt->win = scroll_win;
			return 0;
		    }
		    else
		    {
			//Window not found under the pointer:
			return 1;
		    }
		}
		else
		{
		    //Mouse click on some window...
		    WINDOWPTR focus_win = 0;
		    if( find_focus_window( wm->root_win, &focus_win, wm ) )
		    {
			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			if( focus_win->flags & WIN_FLAG_ALWAYS_UNFOCUSED )
			{
			    evt->win = focus_win;
			    wm->last_unfocused_window = focus_win;
			    return 0;
			}
			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			set_focus_win( focus_win, wm );
		    }
		    else
		    {
			//Window not found under the pointer:
			return 1;
		    }
		}
	    }
	}
	if( wm->focus_win )
	{
	    //Set pointer to window:
	    evt->win = wm->focus_win;

	    if( evt->type == EVT_MOUSEBUTTONDOWN )
	    {
		if( ( evt->time - wm->focus_win->click_time ) < ( DOUBLE_CLICK_PERIOD * time_ticks_per_second_hires() ) / 1000 )
		{
		    //Double click detected:
		    wm->focus_win->flags |= WIN_FLAG_DOUBLECLICK;
		    wm->focus_win->click_time = evt->time - time_ticks_per_second_hires() * 10; //Reset click time
		}
		else
		{
		    //One click. Clear double-click flags:
		    wm->focus_win->flags &= ~WIN_FLAG_DOUBLECLICK;
		    wm->focus_win->click_time = evt->time;
		}
	    }
	    if( wm->focus_win->flags & WIN_FLAG_DOUBLECLICK )
	    {
		evt->flags |= EVT_FLAG_DOUBLECLICK;
	    }
	}
    }

    return 0;
}

void handle_event( sundog_event* evt, window_manager *wm )
{
    sundog_event evt2 = *evt;
    if( !user_event_handler( &evt2, wm ) )
    {
	//Event hot handled by first event handler.
	//Send it to window:
	if( handle_event_by_window( &evt2, wm ) == 0 )
	{
	    evt2.win = wm->handler_of_unhandled_events;
	    handle_event_by_window( &evt2, wm );
	}
    }
}

int handle_event_by_window( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    if( win )
    {
	if( win->flags & WIN_FLAG_TRASH )
	{
	    dprint( "ERROR: can't handle event %d by removed window (%s)\n", evt->type, win->name );
	    retval = 1;
	}
	else
	if( evt->type == EVT_DRAW ) 
	{
	    draw_window( win, wm );
	    retval = 1;
	}
	else
	{
	    if( win->win_handler )
	    {
		if( !win->win_handler( evt, wm ) )
		{
		    //Send event to children:
		    for( int c = 0; c < win->childs_num; c++ )
		    {
			evt->win = win->childs[ c ];
			if( handle_event_by_window( evt, wm ) )
			{
			    evt->win = win;
			    retval = 1;
			    goto end_of_handle;
			}
		    }
		    evt->win = win;
		}
		else
		{
		    retval = 1;
		    goto end_of_handle;
		}
	    }
end_of_handle:
	    if( win == wm->root_win )
	    {
		if( evt->type == EVT_SCREENRESIZE )
		{
		    //On screen resize:
		    recalc_regions( wm );
		    draw_window( wm->root_win, wm );
		}
	    }
	}
    }
    return retval;
}

int EVENT_LOOP_BEGIN( sundog_event* evt, window_manager* wm )
{
    int rv = 0;
    device_event_handler( wm );
    if( wm->exit_request ) return 0;
    if( wm->events_count )
    {
	sundog_mutex_lock( &wm->events_mutex );

get_next_event:
	//There are unhandled events:
	//Copy current event to "evt" buffer (prepare it for handling):
	mem_copy( evt, &wm->events[ wm->current_event_num ], sizeof( sundog_event ) );
	//This event will be handled. So decrement count of events:
	wm->events_count--;
	//And increment FIFO pointer:
	wm->current_event_num = ( wm->current_event_num + 1 ) & ( EVENTS - 1 );
	if( evt->type == EVT_NULL && wm->events_count ) goto get_next_event;
#ifdef MULTITOUCH
	if( evt->flags & EVT_FLAG_DONTDRAW )
	{
	    for( int e = 0; e < wm->events_count; e++ )
            {
                sundog_event* next_evt = &wm->events[ ( wm->current_event_num + e ) & ( EVENTS - 1 ) ];
                next_evt->flags |= EVT_FLAG_HANDLING;
                if( !( next_evt->flags & EVT_FLAG_DONTDRAW ) ) break;
            }
	}
#endif

	sundog_mutex_unlock( &wm->events_mutex );
	
	//Check the event and handle it:
	if( check_event( evt, wm ) == 0 )
	{
	    handle_event( evt, wm );
	    rv = 1;
	}
    }
    return rv;
}

int EVENT_LOOP_END( window_manager* wm )
{
    if( wm->exit_request ) return 1;
    if( wm->frame_event_request )
    {
	if( g_mem_error )
	{
	    int size = (int)g_mem_error;
	    g_mem_error = 0;
	    char ts[ 128 ];
	    sprintf( ts, "Memory allocation error\nCan't allocate %d bytes", (int)size );
	    dialog( ts, "Close" );
	}
	if( wm->timers_num )
	{
	    ticks_t cur_time = time_ticks();
	    for( int t = 0; t < wm->timers_num; t++ )
	    {
		sundog_timer* timer = &wm->timers[ t ];
		int timer_id = timer->id;
		if( timer->handler )
		{
		    if( timer->deadline <= cur_time )
		    {
			timer->handler( timer->data, timer, wm ); 
			if( timer->handler && timer->id == timer_id )
			{
			    timer->deadline = time_ticks() + timer->delay;
			}
		    }
		}
	    }
	}
	user_event_handler( &g_sundog_frame_evt, wm );
	wm->frame_event_request = 0;
    }
    if( wm->screen_buffer_preserved == 0 && wm->screen_changed )
    {
	//If buffer is not preserved,
	//gfx can be redrawn in this place only
	//Ways to this place:
	// 1) draw_window();
	// 2) some gfx function called;
	// 3) screen_change().
	wm->screen_buffer_preserved = 1;
	draw_window( wm->root_win, wm );
	wm->screen_buffer_preserved = 0;
    }
    device_redraw_framebuffer(); //Show screen changes, only if they exists.
    if( wm->flags & WIN_INIT_FLAG_OPTIMIZE_MOVE_EVENTS )
    {
	if( wm->events_count )
	{
	    sundog_mutex_lock( &wm->events_mutex );
	    bool move_flag = 1;
#ifdef MULTITOUCH
	    uint mt_move_flags = 0xFFFFFFFF;
#endif
	    for( int i = wm->events_count - 1; i >= 0; i-- )
	    {
		int event_num = ( wm->current_event_num + i ) & ( EVENTS - 1 );
		sundog_event* evt = &wm->events[ event_num ];
		if( evt->type == EVT_MOUSEBUTTONDOWN ||
		    evt->type == EVT_MOUSEBUTTONUP )
		{
		    move_flag = 1;
		}
		else if( evt->type == EVT_MOUSEMOVE )
		{
		    if( move_flag == 0 && !( evt->flags & EVT_FLAG_HANDLING ) ) evt->type = EVT_NULL;
		    if( move_flag ) move_flag = 0;
		}
#ifdef MULTITOUCH
		if( evt->type == EVT_TOUCHBEGIN ||
		    evt->type == EVT_TOUCHEND )
		{
		    mt_move_flags |= 1 << evt->scancode;
		}
		else if( evt->type == EVT_TOUCHMOVE )
		{
		    if( ( mt_move_flags & ( 1 << evt->scancode ) ) == 0 && !( evt->flags & EVT_FLAG_HANDLING ) ) evt->type = EVT_NULL;
		    if( mt_move_flags & ( 1 << evt->scancode ) ) mt_move_flags &= ~( 1 << evt->scancode );
		}
#endif
	    }
	    sundog_mutex_unlock( &wm->events_mutex );
	}
    }
    return 0;
}

//################################
//## STATUS MESSAGES:           ##
//################################

void remove_status_window( void )
{
    remove_window( g_status_window, g_wm );
    remove_timer( g_status_timer, g_wm );
    g_status_window = 0;
    g_status_timer = -1;
    recalc_regions( g_wm );
    draw_window( g_wm->root_win, g_wm );
}

void status_timer( void* user_data, sundog_timer* t, window_manager* wm )
{
    remove_status_window();
}

int status_message_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    switch( evt->type )
    {
	case EVT_DRAW:
	    win_draw_box( win, 0, 1, win->xsize, win->ysize, wm->color0, wm );
	    if( g_status_message )
	    {
		win_draw_string( win, g_status_message, 0, 1, wm->color3, wm->color0, wm );
	    }
	    win_draw_box( win, 0, 0, win->xsize, 1, blend( win->color, BORDER_COLOR_WITHOUT_TRANSP, BORDER_TRANSP ), wm );
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	case EVT_MOUSEMOVE:
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    remove_status_window();
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    mem_free( g_status_message );
	    g_status_message = 0;
	    retval = 1;
	    break;
    }
    return retval;
}

void show_status_message( const UTF8_CHAR* msg, int t )
{
    if( t == 0 ) t = 2;
    int lines = 1;
    for( int i = 0; ; i++ )
    {
	int c = msg[ i ];
	if( c == 0 ) break;
	if( c == 0xA ) lines++;
    }
    mem_free( g_status_message );
    g_status_message = (UTF8_CHAR*)MEM_NEW2( mem_strlen( msg ) + 1 );
    g_status_message[ 0 ] = 0;
    mem_strcat( g_status_message, msg );
    if( g_status_window == 0 )
    {
	g_status_window = new_window( "status", 0, 0, 8, 8, g_wm->color0, g_wm->root_win, status_message_handler, g_wm );
	g_status_timer = add_timer( status_timer, 0, time_ticks_per_second() * t, g_wm );
    }
    else 
    {
	reset_timer( g_status_timer, g_wm );
    }

    set_window_controller( g_status_window, 0, g_wm, (WCMD)0, CEND );
    set_window_controller( g_status_window, 1, g_wm, CPERC, (WCMD)100, CSUB, (WCMD)char_y_size( g_status_window, g_wm ) * lines + 1, CEND );
    set_window_controller( g_status_window, 2, g_wm, CPERC, (WCMD)100, CEND );
    set_window_controller( g_status_window, 3, g_wm, CPERC, (WCMD)100, CEND );
    
    show_window( g_status_window, g_wm );
    recalc_regions( g_wm );
    draw_window( g_wm->root_win, g_wm );
}

//################################
//## TIMERS:                    ##
//################################

void pack_timers( window_manager* wm )
{
    int r = 0;
    int w = 0;
    for( ; r < wm->timers_num; r++ )
    {
	if( wm->timers[ r ].handler )
	{
	    mem_copy( &wm->timers[ w ], &wm->timers[ r ], sizeof( sundog_timer ) );
	    w++;
	}
    }
    wm->timers_num = w;
}

int add_timer( void (*handler)( void*, sundog_timer*, window_manager* ), void* data, ticks_t delay, window_manager* wm )
{
    if( wm->timers_num >= TIMERS )
	return -1;
    int t = wm->timers_num;
    wm->timers[ t ].handler = handler;
    wm->timers[ t ].data = data;
    wm->timers[ t ].deadline = time_ticks() + delay;
    wm->timers[ t ].delay = delay;
    wm->timers_id_counter++;
    wm->timers_id_counter &= 0xFFFFFFF;
    wm->timers[ t ].id = wm->timers_id_counter;
    wm->timers_num++;
    return wm->timers[ t ].id;
}

void reset_timer( int timer, window_manager* wm )
{
    for( int i = 0; i < wm->timers_num; i++ )
    {
	if( wm->timers[ i ].id == timer )
	{
	    if( wm->timers[ i ].handler )
	    {
		wm->timers[ i ].deadline = time_ticks() + wm->timers[ i ].delay;
		pack_timers( wm );
	    }
	    break;
	}
    }
}

void remove_timer( int timer, window_manager* wm )
{
    if( timer == -1 ) return;
    for( int i = 0; i < wm->timers_num; i++ )
    {
	if( wm->timers[ i ].id == timer )
	{
	    if( wm->timers[ i ].handler )
	    {
		wm->timers[ i ].handler = 0;
		pack_timers( wm );
	    }
	    break;
	}
    }
}

//################################
//## WINDOWS DECORATIONS:       ##
//################################

#define DRAG_LEFT	1
#define DRAG_RIGHT	2
#define DRAG_TOP	4
#define DRAG_BOTTOM	8
#define DRAG_MOVE	16
struct decorator_data
{
    WINDOWPTR this_window;
    
    int start_win_x;
    int start_win_y;
    int start_win_xs;
    int start_win_ys;
    int start_pen_x;
    int start_pen_y;
    int drag_mode;
    bool fullscreen;
    bool minimized;
    int prev_x;
    int prev_y;
    int prev_xsize;
    int prev_ysize;
    int flags;
    
    int border;
    int header;
    
    WINDOWPTR close;
    WINDOWPTR minimize;
    
    int initial_x;
    int initial_y;
    int initial_xsize;
    int initial_ysize;
};

int decorator_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    decorator_data* data = (decorator_data*)win->data;
    int dx, dy;
    int rx = evt->x - win->screen_x;
    int ry = evt->y - win->screen_y;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( decorator_data );
	    break;
	case EVT_AFTERCREATE:
	    data->this_window = win;
	    data->fullscreen = 0;
	    data->minimized = 0;
	    data->drag_mode = 0;
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	    if( data->flags & DECOR_FLAG_STATIC )
	    {
		retval = 1;
		break;
	    }
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		bring_to_front( win, wm );
		recalc_regions( wm );
		data->start_pen_x = evt->x;
		data->start_pen_y = evt->y;
		data->start_win_x = win->x;
		data->start_win_y = win->y;
		data->start_win_xs = win->xsize;
		data->start_win_ys = win->ysize;
		data->drag_mode = DRAG_MOVE;
		if( ry >= data->border && !( data->flags & DECOR_FLAG_NOBORDER ) && !( data->flags & DECOR_FLAG_NORESIZE ) )
		{
		    if( rx < data->border + 8 ) data->drag_mode |= DRAG_LEFT;
		    if( rx >= win->xsize - data->border - 8 ) data->drag_mode |= DRAG_RIGHT;
		    if( ry >= win->ysize - data->border - 8 ) data->drag_mode |= DRAG_BOTTOM;
		}
		draw_window( win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    retval = 1;
	    if( data->drag_mode )
	    {
		data->drag_mode = 0;
		if( ( data->flags & DECOR_FLAG_STATIC ) ||
		    ( data->flags & DECOR_FLAG_NOBORDER ) )
		{
		    break;
		}
		if( ( evt->key == MOUSE_BUTTON_LEFT ) &&
		    ( evt->flags & EVT_FLAG_DOUBLECLICK ) )
		{
		    //Make fullscreen (maximize):
		    if( data->fullscreen == 1 )
		    {
			win->x = data->prev_x;
			win->y = data->prev_y;
			win->xsize = data->prev_xsize;
			win->ysize = data->prev_ysize;
		    }
		    else
		    {
			data->prev_x = win->x;
			data->prev_y = win->y;
			data->prev_xsize = win->xsize;
			data->prev_ysize = win->ysize;
			win->x = 0; 
			win->y = 0;
			win->xsize = win->parent->xsize;
			win->ysize = win->parent->ysize;
		    }
		    win->childs[ 0 ]->xsize = win->xsize - data->border * 2;
		    win->childs[ 0 ]->ysize = win->ysize - data->border * 2 - data->header;
		    data->fullscreen ^= 1;
		    recalc_regions( wm );
		    draw_window( wm->root_win, wm );
		}
	    }
	    break;
	case EVT_UNFOCUS:
	    data->drag_mode = 0;
	    retval = 1;
	    break;
	case EVT_MOUSEMOVE:
	    if( data->flags & DECOR_FLAG_STATIC )
	    {
		retval = 1;
		break;
	    }
	    if( evt->key == MOUSE_BUTTON_LEFT )
	    {
		dx = evt->x - data->start_pen_x;
		dy = evt->y - data->start_pen_y;
		if( data->drag_mode == DRAG_MOVE )
		{
		    //Move:
		    int prev_x = win->x;
		    int prev_y = win->y;
		    win->x = data->start_win_x + dx;
		    win->y = data->start_win_y + dy;
		    if( prev_x != win->x || prev_y != win->y )
		    {
			if( data->fullscreen == 1 )
			{
			    win->xsize = data->prev_xsize;
			    win->ysize = data->prev_ysize;
			    data->fullscreen = 0;
			}
		    }
		}
		
		if( !( data->flags & DECOR_FLAG_NOBORDER ) )
		{
		    if( data->drag_mode & DRAG_LEFT )
		    {
			int prev_x = win->x;
			int prev_xsize = win->xsize;
			win->x = data->start_win_x + dx;
			win->xsize = data->start_win_xs - dx;
			if( prev_x != win->x || prev_xsize != win->xsize )
			{
			    if( win->xsize < 16 ) win->xsize = 16;
			    data->fullscreen = 0;
			}
		    }
		    if( data->drag_mode & DRAG_RIGHT )
		    {
			int prev_xsize = win->xsize;
			win->xsize = data->start_win_xs + dx;
			if( prev_xsize != win->xsize )
			{
			    if( win->xsize < 16 ) win->xsize = 16;
			    data->fullscreen = 0;
			}
		    }
		    if( data->drag_mode & DRAG_BOTTOM )
		    {
			int prev_ysize = win->ysize;
			win->ysize = data->start_win_ys + dy;
			if( prev_ysize != win->ysize )
			{
			    if( win->ysize < 16 ) win->ysize = 16;
			    data->fullscreen = 0;
			}
		    }
		    if( win->childs_num )
		    {
			win->childs[ 0 ]->xsize = win->xsize - data->border * 2;
			win->childs[ 0 ]->ysize = win->ysize - data->border * 2 - data->header;
		    }
		} //!NOBORDER
		
		recalc_regions( wm );
		draw_window( wm->root_win, wm );
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    win_draw_lock( win, wm );
	    if( win->childs_num )
	    {
		COLOR c;
		c = win->childs[ 0 ]->color;
		if( data->flags & DECOR_FLAG_NOBORDER )
		{
		    win_draw_box( win, 0, 0, win->xsize, win->ysize, c, wm );
		    if( win->childs[ 0 ]->name )
			win_draw_string( win, win->childs[ 0 ]->name, INTERELEMENT_SPACE, ( data->header - char_y_size( win, wm ) ) / 2, wm->header_text_color, c, wm );
		}
		else
		{
		    win_draw_box( win, data->border, 0, win->xsize - data->border * 2, win->ysize, c, wm );
		    if( win->childs[ 0 ]->name )
			win_draw_string( win, win->childs[ 0 ]->name, data->border + INTERELEMENT_SPACE, ( data->header + data->border - char_y_size( win, wm ) ) / 2, wm->header_text_color, c, wm );
		    win_draw_box( win, 0, 0, data->border, win->ysize, wm->decorator_border, wm );
		    win_draw_box( win, win->xsize - data->border, 0, data->border, win->ysize, wm->decorator_border, wm );
		}
	    }
	    else
	    {
		//No childs more :( Empty decorator. Lets remove it:
		remove_window( win, wm );
		recalc_regions( wm );
		draw_window( wm->root_win, wm );
	    }
	    win_draw_unlock( win, wm );
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	case EVT_TOUCHEND:
	case EVT_TOUCHMOVE:
	    retval = 1;
	    break;
	case EVT_SCREENRESIZE:
	    if( wm->flags & WIN_INIT_FLAG_FULLSCREEN )
		change_decorator( win, data->initial_x, data->initial_y, data->initial_xsize, data->initial_ysize, data->flags, wm );
	    break;
    }
    return retval;
}

void change_decorator(
    WINDOWPTR dec,
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    int flags,
    window_manager* wm )
{
    if( dec == 0 ) return;
    if( dec->childs == 0 ) return;
    WINDOWPTR win = dec->childs[ 0 ];
    
    decorator_data* ddata = (decorator_data*)dec->data;
    
    ddata->fullscreen = 0;
    ddata->flags = flags;
    ddata->initial_x = x;
    ddata->initial_y = y;
    ddata->initial_xsize = xsize;
    ddata->initial_ysize = ysize;
    
    int border = DECOR_BORDER_SIZE;
    int header = DECOR_HEADER_SIZE;
    
    if( flags & DECOR_FLAG_NOBORDER )
	border = 0;

    if( flags & DECOR_FLAG_NOHEADER )
	header = 0;
    
    if( flags & DECOR_FLAG_FULLSCREEN )
    {
	x = 0;
	y = 0;
	xsize = dec->parent->xsize;
	ysize = dec->parent->ysize;
    }
    
    x -= border;
    y -= border + header;
    int dec_xsize = xsize + border * 2;
    int dec_ysize = ysize + border * 2 + header;
    if( flags & DECOR_FLAG_CENTERED )
    {
	x = ( dec->parent->xsize - dec_xsize ) / 2;
	y = ( dec->parent->ysize - dec_ysize ) / 2;
    }
    if( flags & DECOR_FLAG_CHECK_SIZE )
    {
	if( x < 0 )
	{
	    dec_xsize -= ( -x );
	    x = 0;
	}
	if( y < 0 )
	{
	    dec_ysize -= ( -y );
	    y = 0;
	}
	if( x + dec_xsize > dec->parent->xsize )
	{
	    dec_xsize -= ( ( x + dec_xsize ) - dec->parent->xsize );
	}
	if( y + dec_ysize > dec->parent->ysize )
	{
	    dec_ysize -= ( ( y + dec_ysize ) - dec->parent->ysize );
	}
    }
    
    dec->x = x;
    dec->y = y;
    dec->xsize = dec_xsize;
    dec->ysize = dec_ysize;
    
    ddata->border = border;
    ddata->header = header;
        
    if( win )
    {
	win->xsize = dec->xsize - border * 2;
	win->ysize = dec->ysize - ( header + border * 2 );
	win->x = border;
	win->y = header + border;
    }
}

void resize_window_with_decorator( WINDOWPTR win, int xsize, int ysize, window_manager* wm )
{
    WINDOWPTR dec = win->parent;
    decorator_data* ddata = (decorator_data*)dec->data;
    if( xsize > 0 )
    {
	int prev_size = win->xsize;
	win->xsize = xsize;
	int delta = prev_size - win->xsize;
	dec->xsize -= delta;
	if( ddata->flags & DECOR_FLAG_CHECK_SIZE )
	{
	    if( dec->xsize > dec->parent->xsize )
	    {
		int s = dec->xsize - dec->parent->xsize;
		win->xsize -= s;
		dec->xsize -= s;
		delta += s;
	    }
	}
	if( ddata->flags & DECOR_FLAG_CENTERED )
	{
	    dec->x += delta / 2;
	}
    }
    if( ysize > 0 )
    {
	if( ddata->flags & DECOR_FLAG_CHECK_SIZE )
	{
	    if( ysize > dec->parent->ysize )
		ysize = dec->parent->ysize;
	}
	int prev_size = win->ysize;
	win->ysize = ysize;
	int delta = prev_size - win->ysize;
	dec->ysize -= delta;
	if( ddata->flags & DECOR_FLAG_CHECK_SIZE )
	{
	    if( dec->ysize > dec->parent->ysize )
	    {
		int s = dec->ysize - dec->parent->ysize;
		win->ysize -= s;
		dec->ysize -= s;
		delta += s;
	    }
	}
	if( ddata->flags & DECOR_FLAG_CENTERED )
	{
	    dec->y += delta / 2;
	}
    }
}

int decor_close_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    decorator_data* data = (decorator_data*)user_data;
    WINDOWPTR cwin = data->this_window->childs[ 0 ];
    
    if( cwin )
    {
	send_event( cwin, EVT_CLOSEREQUEST, 0, 0, 0, 0, 0, 0, 0, wm );
    }
    
    return 0;
}

int decor_minimize_button_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    decorator_data* data = (decorator_data*)user_data;
    WINDOWPTR dwin = data->this_window;
    WINDOWPTR cwin = dwin->childs[ 0 ];
    if( cwin == 0 ) return 0;
    
    //Minimize / maximize:
    if( data->minimized )
    {
	dwin->xsize = data->prev_xsize;
	dwin->ysize = data->prev_ysize;
	data->minimized = 0;
    }
    else
    {
	data->prev_xsize = dwin->xsize;
	data->prev_ysize = dwin->ysize;
	dwin->xsize = data->border * 2 + SCROLLBAR_SIZE;
	if( data->close ) dwin->xsize += SCROLLBAR_SIZE;
	if( data->minimize ) dwin->xsize += SCROLLBAR_SIZE;
	dwin->ysize = data->border + data->header;
	data->minimized = 1;
    }
    cwin->xsize = dwin->xsize - data->border * 2;
    cwin->ysize = dwin->ysize - data->border * 2 - data->header;
    recalc_regions( wm );
    draw_window( wm->root_win, wm );
    
    return 0;
}

WINDOWPTR new_window_with_decorator( 
    const UTF8_CHAR* name, 
    int x, 
    int y, 
    int xsize, 
    int ysize, 
    COLOR color,
    WINDOWPTR parent, 
    int (*win_handler)( sundog_event*, window_manager* ),
    int flags,
    window_manager* wm )
{
    WINDOWPTR dec = new_window( 
	"decorator", 
	x, y, xsize, ysize,
	wm->decorator_color,
	parent,
	decorator_handler,
	wm );
    new_window( name, x, y, xsize, ysize, color, dec, win_handler, wm );
    change_decorator( dec, x, y, xsize, ysize, flags, wm );
    decorator_data* data = (decorator_data*)dec->data;
    int xx = data->border + INTERELEMENT_SPACE;
    if( flags & DECOR_FLAG_WITH_CLOSE )
    {
	CREATE_FLAT_BUTTON = 1;
	data->close = new_window( "x", 0, 0, SCROLLBAR_SIZE, data->header + data->border, dec->color, dec, button_handler, wm );
	set_handler( data->close, decor_close_button_handler, data, wm );
	set_window_controller( data->close, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)xx + SCROLLBAR_SIZE, CEND );
	set_window_controller( data->close, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)xx, CEND );
	xx += SCROLLBAR_SIZE + 1;
    }
    if( flags & DECOR_FLAG_WITH_MINIMIZE )
    {
	CREATE_FLAT_BUTTON = 1;
	data->minimize = new_window( "_", 0, 0, SCROLLBAR_SIZE, data->header + data->border, dec->color, dec, button_handler, wm );
	set_handler( data->minimize, decor_minimize_button_handler, data, wm );
	set_window_controller( data->minimize, 0, wm, CPERC, (WCMD)100, CSUB, (WCMD)xx + SCROLLBAR_SIZE, CEND );
	set_window_controller( data->minimize, 2, wm, CPERC, (WCMD)100, CSUB, (WCMD)xx, CEND );
	xx += SCROLLBAR_SIZE + 1;
    }
    return dec;
}

//################################
//## DRAWING FUNCTIONS:         ##
//################################

int char_x_size( WINDOWPTR win, UTF32_CHAR c, window_manager* wm )
{
    if( c < 0x10 ) return 0;
    if( win->font == 0 )
    {
        return 8;
    }
    else
    {
        utf32_to_win1251( c, c );
        return g_font1[ 1 + c ];
    }
}

int char_y_size( WINDOWPTR win, window_manager* wm )
{
    if( win->font == 0 )
    {
        return 8;
    }
    else
    {
        return g_font1[ 0 ];
    }
}

int string_size( WINDOWPTR win, const UTF8_CHAR* str, window_manager* wm )
{
    int size = 0;
    while( *str )
    {
	UTF32_CHAR c32;
	str += utf8_to_utf32_char( str, &c32 );
	size += char_x_size( win, c32, wm );
    }    
    return size;
}

void win_draw_lock( WINDOWPTR win, window_manager* wm )
{
    wm->cur_transparency = 255;
    device_screen_lock( win, wm );
}

void win_draw_unlock( WINDOWPTR win, window_manager* wm )
{
    device_screen_unlock( win, wm );
}

void win_draw_box( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager* wm )
{
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x += win->screen_x;
	y += win->screen_y;
	if( win->reg->numRects )
	{
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		int nx = x;
		int ny = y;
		int nxsize = xsize;
		int nysize = ysize;
		if( nx < rx1 ) { nxsize -= ( rx1 - nx ); nx = rx1; }
		if( ny < ry1 ) { nysize -= ( ry1 - ny ); ny = ry1; }
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) nxsize -= nx + nxsize - rx2;
		if( ny + nysize > ry2 ) nysize -= ny + nysize - ry2;
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_box( nx, ny, nxsize, nysize, color );
		wm->screen_changed++;
	    }
	}
    }
}

void win_draw_frame( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, window_manager* wm )
{
    win_draw_line( win, x, y, x + xsize - 1, y, color, wm );
    win_draw_line( win, x + xsize - 1, y, x + xsize - 1, y + ysize - 1, color, wm );
    win_draw_line( win, x + xsize - 1, y + ysize - 1, x, y + ysize - 1, color, wm );
    win_draw_line( win, x, y + ysize - 1, x, y, color, wm );
}

void win_draw_frame3d( WINDOWPTR win, int x, int y, int xsize, int ysize, COLOR color, int inout, window_manager* wm )
{
    inout &= 255;
    int tr = 32;
    int wtr = 100;
    COLOR white = wm->color3;
    COLOR black = wm->color0;
    for( int l = 0; l < 1; l++ )
    {
	if( inout == 0 )
	{
	    win_draw_line( win, x, y, x + xsize - 1, y, blend( color, black, tr ), wm );
	    win_draw_line( win, x + xsize - 1, y, x + xsize - 1, y + ysize - 1, blend( color, black, tr ), wm );
	    win_draw_line( win, x, y, x, y + ysize - 1, blend( color, white, wtr ), wm );
	    win_draw_line( win, x, y + ysize - 1, x + xsize - 1, y + ysize - 1, blend( color, white, wtr ), wm );
	}
	else
	{
	    win_draw_line( win, x, y, x + xsize - 1, y, blend( color, white, wtr ), wm );
	    win_draw_line( win, x + xsize - 1, y, x + xsize - 1, y + ysize - 1, blend( color, white, wtr ), wm );
	    win_draw_line( win, x, y, x, y + ysize - 1, blend( color, black, tr ), wm );
	    win_draw_line( win, x, y + ysize - 1, x + xsize - 1, y + ysize - 1, blend( color, black, tr ), wm );
	}
	x++; 
	y++;
	xsize -= 2;
	ysize -= 2;
	tr >>= 1;
	wtr >>= 1;
    }
}

void win_draw_bitmap_ext( 
    WINDOWPTR win, 
    int x, 
    int y, 
    int dest_xsize, 
    int dest_ysize,
    int source_x,
    int source_y,
    sundog_image* data, 
    window_manager* wm )
{
    if( source_x < 0 ) { dest_xsize += source_x; x -= source_x; source_x = 0; }
    if( source_y < 0 ) { dest_ysize += source_y; y -= source_y; source_y = 0; }
    if( source_x >= data->xsize ) return;
    if( source_y >= data->ysize ) return;
    if( source_x + dest_xsize > data->xsize ) dest_xsize -= ( source_x + dest_xsize ) - data->xsize;
    if( source_y + dest_ysize > data->ysize ) dest_ysize -= ( source_y + dest_ysize ) - data->ysize;
    if( dest_xsize <= 0 ) return;
    if( dest_ysize <= 0 ) return;
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x += win->screen_x;
	y += win->screen_y;
	int xsize = dest_xsize;
	int ysize = dest_ysize;
	if( win->reg->numRects )
	{
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		int src_x = source_x;
		int src_y = source_y;
		int nx = x;
		int ny = y;
		int nxsize = xsize;
		int nysize = ysize;
		if( nx < rx1 ) { nxsize -= ( rx1 - nx ); src_x += ( rx1 - nx ); nx = rx1; }
		if( ny < ry1 ) { nysize -= ( ry1 - ny ); src_y += ( ry1 - ny ); ny = ry1; }
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) nxsize -= nx + nxsize - rx2;
		if( ny + nysize > ry2 ) nysize -= ny + nysize - ry2;
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_bitmap( nx, ny, nxsize, nysize, src_x, src_y, data );
		wm->screen_changed++;
	    }
	}
    }
}

void win_draw_bitmap( 
    WINDOWPTR win, 
    int x, 
    int y, 
    sundog_image* data, 
    window_manager* wm )
{
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x += win->screen_x;
	y += win->screen_y;
	if( win->reg->numRects )
	{
	    for( int r = 0; r < win->reg->numRects; r++ )
	    {
		int rx1 = win->reg->rects[ r ].left;
		int rx2 = win->reg->rects[ r ].right;
		int ry1 = win->reg->rects[ r ].top;
		int ry2 = win->reg->rects[ r ].bottom;

		//Control box size:
		int src_x = 0;
		int src_y = 0;
		int nx = x;
		int ny = y;
		int nxsize = data->xsize;
		int nysize = data->ysize;
		if( nx < rx1 ) { nxsize -= ( rx1 - nx ); src_x += ( rx1 - nx ); nx = rx1; }
		if( ny < ry1 ) { nysize -= ( ry1 - ny ); src_y += ( ry1 - ny ); ny = ry1; }
		if( nx + nxsize <= rx1 ) continue;
		if( ny + nysize <= ry1 ) continue;
		if( nx + nxsize > rx2 ) nxsize -= nx + nxsize - rx2;
		if( ny + nysize > ry2 ) nysize -= ny + nysize - ry2;
		if( nx >= rx2 ) continue;
		if( ny >= ry2 ) continue;
		if( nxsize < 0 ) continue;
		if( nysize < 0 ) continue;
        	
		//Draw it:
		device_draw_bitmap( nx, ny, nxsize, nysize, src_x, src_y, data );
		wm->screen_changed++;
	    }
	}
    }
}

void win_draw_char( WINDOWPTR win, UTF32_CHAR c, int x, int y, COLOR f, COLOR b, window_manager* wm )
{
    if( c < 0x10 ) return; //Control symbols
    int font_num = win->font;
    //Convert from UTF32 to WIN1251:
    utf32_to_win1251( c, c );
    //Draw:
    uchar xsize;
    uchar ysize;
    uchar *font;
    if( font_num == 0 )
    {
	font = g_font0;
	xsize = 8;
	ysize = 8;
    }
    else
    {
	font = g_font1;
	xsize = font[ 1 + c ];
	ysize = font[ 0 ];
    }
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    sundog_image *font_img = g_font_img[ font_num ];
    int cxsize = g_font_cxsize[ font_num ];
    int cysize = g_font_cysize[ font_num ];
    font_img->color = b;
    font_img->transp = 255;
    win_draw_bitmap_ext( 
        win, 
        x, y, 
        xsize, ysize,
        cxsize, 0,
        font_img, 
        wm );
    font_img->color = f;
    win_draw_bitmap_ext( 
        win, 
        x, y, 
        xsize, ysize,
        ( c & 15 ) * cxsize, ( c / 16 ) * cysize,
        font_img, 
        wm );
#else
    sundog_image tmp_img;
    tmp_img.xsize = xsize;
    tmp_img.ysize = ysize;
    tmp_img.flags = IMAGE_NATIVE_RGB | IMAGE_STATIC_SOURCE;
    COLOR bmp[ 8 * 12 ];
    tmp_img.data = bmp;
    c *= ysize;
    int ptr = 0;
    font += 257;
    for( int l = 0; l < ysize; l++ )
    {
        uchar v = font[ c ]; c++;
	for( int x = 0; x < xsize; x++ )
	{
	    if( v & 0x80 )
	        bmp[ ptr ] = f;
	    else
	        bmp[ ptr ] = b;
	    v <<= 1;
	    ptr++;
	}
    }
    win_draw_bitmap( win, x, y, &tmp_img, wm );
#endif
}

void win_draw_string( WINDOWPTR win, const UTF8_CHAR* str, int x, int y, COLOR f, COLOR b, window_manager* wm )
{
    int start_x = x;
    while( *str )
    {
	if( *str == 0xA ) 
	{ 
	    y += char_y_size( win, wm ); 
	    x = start_x; 
	    str++;
	}
	else
	{
	    UTF32_CHAR c32;
	    str += utf8_to_utf32_char( str, &c32 );
	    win_draw_char( win, c32, x, y, f, b, wm );
	    x += char_x_size( win, c32, wm );
	}
    }
}

#define cbottom 1
#define ctop 2
#define cleft 4
#define cright 8
inline int line_clip_make_code( int x, int y, int clip_x1, int clip_y1, int clip_x2, int clip_y2 )
{
    int code = 0;
    if( y >= clip_y2 ) code = cbottom;
    else if( y < clip_y1 ) code = ctop;
    if( x >= clip_x2 ) code += cright;
    else if( x < clip_x1 ) code += cleft;
    return code;
}

bool line_clip( int* x1, int* y1, int* x2, int* y2, int clip_x1, int clip_y1, int clip_x2, int clip_y2 )
{
    //Cohen-Sutherland line clipping algorithm:
    int code0;
    int code1;
    int out_code;
    int x, y;
    code0 = line_clip_make_code( *x1, *y1, clip_x1, clip_y1, clip_x2, clip_y2 );
    code1 = line_clip_make_code( *x2, *y2, clip_x1, clip_y1, clip_x2, clip_y2 );
    while( code0 || code1 )
    {
	if( code0 & code1 ) 
	{
	    //Trivial reject
	    return 0;
	}
	else
	{
	    //Failed both tests, so calculate the line segment to clip
	    if( code0 )
		out_code = code0; //Clip the first point
	    else
		out_code = code1; //Clip the last point
	    
	    if( out_code & cbottom )
	    {
		//Clip the line to the bottom of the viewport
		y = clip_y2 - 1;
		x = *x1 + ( *x2 - *x1 ) * ( y - *y1 ) / ( *y2 - *y1 );
	    }
	    else 
	    {
		if( out_code & ctop )
		{
		    y = clip_y1;
		    x = *x1 + ( *x2 - *x1 ) * ( y - *y1 ) / ( *y2 - *y1 );
		}
		else
		{
		    if( out_code & cright )
		    {
			x = clip_x2 - 1;
			y = *y1 + ( *y2 - *y1 ) * ( x - *x1 ) / ( *x2 - *x1 );
		    }
		    else
		    {
			if( out_code & cleft )
			{
			    x = clip_x1;
			    y = *y1 + ( *y2 - *y1 ) * ( x - *x1 ) / ( *x2 - *x1 );
			}
		    }
		}
	    }
	    
	    if( out_code == code0 )
	    { 
		//Modify the first coordinate 
		*x1 = x; *y1 = y;
		code0 = line_clip_make_code( *x1, *y1, clip_x1, clip_y1, clip_x2, clip_y2 );
	    }
	    else
	    { 
		//Modify the second coordinate
		*x2 = x; *y2 = y;
		code1 = line_clip_make_code( *x2, *y2, clip_x1, clip_y1, clip_x2, clip_y2 );
	    }
	}
    }
    
    return 1;
}

bool line_clip_with_transp( int* x1, int* y1, int* t1, int* x2, int* y2, int* t2, int clip_x1, int clip_y1, int clip_x2, int clip_y2 )
{
    //Cohen-Sutherland line clipping algorithm:
    int code0;
    int code1;
    int out_code;
    int x, y, t;
    code0 = line_clip_make_code( *x1, *y1, clip_x1, clip_y1, clip_x2, clip_y2 );
    code1 = line_clip_make_code( *x2, *y2, clip_x1, clip_y1, clip_x2, clip_y2 );
    while( code0 || code1 )
    {
	if( code0 & code1 ) 
	{
	    //Trivial reject
	    return 0;
	}
	else
	{
	    //Failed both tests, so calculate the line segment to clip
	    if( code0 )
		out_code = code0; //Clip the first point
	    else
		out_code = code1; //Clip the last point
	    
	    if( out_code & cbottom )
	    {
		//Clip the line to the bottom of the viewport
		y = clip_y2 - 1;
		int ylen = *y2 - *y1;
		int ylen2 = y - *y1;
		x = *x1 + ( *x2 - *x1 ) * ylen2 / ylen;
		t = *t1 + ( *t2 - *t1 ) * ylen2 / ylen;
	    }
	    else 
	    {
		if( out_code & ctop )
		{
		    y = clip_y1;
		    int ylen = *y2 - *y1;
		    int ylen2 = y - *y1;
		    x = *x1 + ( *x2 - *x1 ) * ylen2 / ylen;
		    t = *t1 + ( *t2 - *t1 ) * ylen2 / ylen;
		}
		else
		{
		    if( out_code & cright )
		    {
			x = clip_x2 - 1;
			int xlen = *x2 - *x1;
			int xlen2 = x - *x1;
			y = *y1 + ( *y2 - *y1 ) * xlen2 / xlen;
			t = *t1 + ( *t2 - *t1 ) * xlen2 / xlen;
		    }
		    else
		    {
			if( out_code & cleft )
			{
			    x = clip_x1;
			    int xlen = *x2 - *x1;
			    int xlen2 = x - *x1;
			    y = *y1 + ( *y2 - *y1 ) * xlen2 / xlen;
			    t = *t1 + ( *t2 - *t1 ) * xlen2 / xlen;
			}
		    }
		}
	    }
	    
	    if( out_code == code0 )
	    { 
		//Modify the first coordinate 
		*x1 = x; *y1 = y; *t1 = t;
		code0 = line_clip_make_code( *x1, *y1, clip_x1, clip_y1, clip_x2, clip_y2 );
	    }
	    else
	    { 
		//Modify the second coordinate
		*x2 = x; *y2 = y; *t2 = t;
		code1 = line_clip_make_code( *x2, *y2, clip_x1, clip_y1, clip_x2, clip_y2 );
	    }
	}
    }
    
    return 1;
}

void win_draw_line( WINDOWPTR win, int x1, int y1, int x2, int y2, COLOR color, window_manager* wm )
{
    if( win && win->visible && win->reg && win->reg->numRects )
    {
	x1 += win->screen_x;
	y1 += win->screen_y;
	x2 += win->screen_x;
	y2 += win->screen_y;
	for( int r = 0; r < win->reg->numRects; r++ )
	{
	    int rx1 = win->reg->rects[ r ].left;
	    int rx2 = win->reg->rects[ r ].right;
	    int ry1 = win->reg->rects[ r ].top;
	    int ry2 = win->reg->rects[ r ].bottom;

	    int lx1 = x1;
	    int ly1 = y1;
	    int lx2 = x2;
	    int ly2 = y2;
	    
	    if( line_clip( &lx1, &ly1, &lx2, &ly2, rx1, ry1, rx2, ry2 ) )
	    {
		device_draw_line( lx1, ly1, lx2, ly2, color );
		wm->screen_changed++;
	    }
	}
    }
}

void screen_changed( void )
{
    g_wm->screen_changed++;
}

//###################################
//## DIALOGS & BUILT-IN WINDOWS:   ##
//###################################

UTF8_CHAR g_dialog_filename[ MAX_DIR_LEN ];

UTF8_CHAR* dialog_open_file( const UTF8_CHAR* name, const UTF8_CHAR* mask, const UTF8_CHAR* id, const UTF8_CHAR* def_filename )
{
#ifdef DEMO_MODE
    if( strstr( name, "Save" ) ||
        strstr( name, "save" ) ||
        strstr( name, "Export" ) ||
        strstr( name, "export" ) )
    {
	dialog( "!Sorry, but this feature\nis not available\nin demo version", "OK" );
	return 0;
    }
#endif

    if( g_fdialog_open ) return 0;
    if( g_dialog_open ) return 0;
    g_fdialog_open = 1;
  
    WINDOWPTR prev_focus = g_wm->focus_win;

    g_dialog_filename[ 0 ] = 0;

    UTF8_CHAR* path = get_user_path();
    int path_len = mem_strlen( (const UTF8_CHAR*)path );
    int id_len = mem_strlen( id );
    UTF8_CHAR* id2 = (UTF8_CHAR*)MEM_NEW( HEAP_DYNAMIC, path_len + id_len + 1 );
    id2[ 0 ] = 0;
    mem_strcat( id2, (const UTF8_CHAR*)path );
    mem_strcat( id2, id );

    int xsize = g_wm->large_window_xsize;
    int ysize = BUTTON_YSIZE * 2 + g_wm->list_item_ysize * 32 + g_wm->text_ysize * 2;
    int flags = DECOR_FLAG_CENTERED | DECOR_FLAG_CHECK_SIZE;
    
    if( g_wm->control_type == TOUCHCONTROL )
    {
        flags |= DECOR_FLAG_STATIC | DECOR_FLAG_NOBORDER | DECOR_FLAG_FULLSCREEN;
    }
    
    FILES_PROPS = (const UTF8_CHAR*)id2;
    FILES_MASK = mask;
    FILES_RESULTED_FILENAME = g_dialog_filename;
    FILES_DEF_FILENAME = def_filename;
    WINDOWPTR win = new_window_with_decorator( 
	name, 
	0, 0, 
	xsize, ysize, 
	g_wm->dialog_color,
	g_wm->root_win, 
	files_handler,
	flags,
	g_wm );
    show_window( win, g_wm );
    recalc_regions( g_wm );
    draw_window( win, g_wm );

    while( 1 )
    {
	sundog_event evt;
	EVENT_LOOP_BEGIN( &evt, g_wm );
	if( EVENT_LOOP_END( g_wm ) ) break;
	if( win->visible == 0 ) break;
    }

    set_focus_win( prev_focus, g_wm );
    
    mem_free( id2 );
    
    g_fdialog_open = 0;

    if( g_dialog_filename[ 0 ] == 0 ) return 0;
    return g_dialog_filename;
}

int dialog( const UTF8_CHAR* name, const UTF8_CHAR* buttons )
{
    WINDOWPTR prev_focus = g_wm->focus_win;
    int result = 0;
    
    g_dialog_open++;

    dialog_item* ditems = DIALOG_ITEMS;
    
    DIALOG_BUTTONS_TEXT = buttons;
    DIALOG_TEXT = name;
    DIALOG_RESULT = &result;
    int flags = DECOR_FLAG_CENTERED | DECOR_FLAG_CHECK_SIZE;
    const UTF8_CHAR* dialog_name = "";
    if( ditems == 0 || name == 0 || name[ 0 ] == 0 )
    {
	flags |= DECOR_FLAG_NOHEADER;
    }
    else
    {
	dialog_name = name;
    }
    int xsize = g_wm->normal_window_xsize;
    int ysize = 0;
    WINDOWPTR win = new_window_with_decorator( 
	dialog_name,
	0, 0, 
	xsize, 64,
	g_wm->dialog_color, 
	g_wm->root_win, 
	dialog_handler,
	flags,
	g_wm );
    if( flags & DECOR_FLAG_NOHEADER )
    {
	bool new_line = 1;
	int slen = INTERELEMENT_SPACE * 2;
	int p = 0;
	while( 1 )
	{
	    if( slen > xsize ) xsize = slen;
	    UTF8_CHAR c = name[ p ];
	    if( c == 0 ) break;
	    if( c == 0xA )
	    {
		if( new_line )
		{
		    ysize += char_y_size( win, g_wm );
		}
		new_line = 1;
		slen = INTERELEMENT_SPACE * 2;
	    }
	    if( c != 0xA && c != 0xD )
	    {
		slen += char_x_size( win, c, g_wm );
		if( new_line )
		{
		    ysize += char_y_size( win, g_wm );
		    new_line = 0;
		}
	    }
	    p++;
	}
    }
    if( ditems )
    {
	int i = 0;
	while( 1 )
	{
	    int type = ditems[ i ].type;
	    if( type == DIALOG_ITEM_NONE )
	    {
		ysize -= INTERELEMENT_SPACE;
		break;
	    }
	    switch( type )
	    {
		case DIALOG_ITEM_NUMBER:
		case DIALOG_ITEM_NUMBER_HEX:
		    ysize += DIALOG_ITEM_NUMBER_SIZE + INTERELEMENT_SPACE;
		    break;
		case DIALOG_ITEM_TEXT:
		    ysize += DIALOG_ITEM_TEXT_SIZE + INTERELEMENT_SPACE;
		    break;
		case DIALOG_ITEM_LABEL:
		    ysize += DIALOG_ITEM_LABEL_SIZE + INTERELEMENT_SPACE;
		    break;
		case DIALOG_ITEM_POPUP:
		    ysize += DIALOG_ITEM_POPUP_SIZE + INTERELEMENT_SPACE;
		    break;
	    }	    
	    i++;
	}
    }
    change_decorator( win, 0, 0, xsize, ysize + BUTTON_YSIZE + INTERELEMENT_SPACE * 3, flags, g_wm );
    show_window( win, g_wm );
    recalc_regions( g_wm );
    draw_window( win, g_wm );

    while( 1 )
    {
	sundog_event evt;
	EVENT_LOOP_BEGIN( &evt, g_wm );
	if( EVENT_LOOP_END( g_wm ) ) break;
	if( win->visible == 0 ) break;
    }

    set_focus_win( prev_focus, g_wm );
    
    g_dialog_open--;

    return result;
}

int popup_menu( const UTF8_CHAR* name, const UTF8_CHAR* items, int x, int y, COLOR color )
{
    int exit_flag = 0;
    int result = -1;
    CREATE_TEXT_POPUP = items;
    SAVE_POPUP_EXIT_TO = &exit_flag;
    SAVE_POPUP_RESULT_TO = &result;
    WINDOWPTR win = new_window( name, x, y, 1, 1, color, g_wm->root_win, popup_handler, g_wm );
    show_window( win, g_wm );
    recalc_regions( g_wm );
    draw_window( g_wm->root_win, g_wm );
    while( 1 )
    {
	sundog_event evt;
	EVENT_LOOP_BEGIN( &evt, g_wm );
	if( EVENT_LOOP_END( g_wm ) ) break;
	if( exit_flag || win->visible == 0 ) break;
    }
    return result; //win->action_result;
}

void prefs_clear( void )
{
    g_prefs_section_names[ 0 ] = "0";
    g_prefs_sections = 0;
    g_prefs_flags = 0;
}

void prefs_add_sections( const UTF8_CHAR** names, void** handlers, int num )
{
    for( int s = 0; s < num; s++ )
    {
	g_prefs_section_names[ g_prefs_sections ] = names[ s ];
	g_prefs_section_handlers[ g_prefs_sections ] = handlers[ s ];
	g_prefs_sections++;
    }
    g_prefs_section_names[ g_prefs_sections ] = "0";
}

const UTF8_CHAR* g_def_names[ 2 ] = { "Interface", "Audio" };
void* g_def_handlers[ 2 ] = { (void*)prefs_ui_handler, (void*)prefs_audio_handler };

void prefs_add_default_sections( void )
{
    prefs_add_sections( g_def_names, g_def_handlers, 2 );
}

void prefs_open( void )
{
    if( g_prefs == 0 )
    {
        int flags = DECOR_FLAG_CENTERED | DECOR_FLAG_CHECK_SIZE;
        g_prefs = new_window_with_decorator(
            "Preferences",
            0, 0,
            g_wm->large_window_xsize, g_wm->large_window_ysize,
            g_wm->dialog_color,
            g_wm->root_win,
            prefs_handler,
            flags,
            g_wm );
    }
    show_window( g_prefs, g_wm );
    bring_to_front( g_prefs, g_wm );
    recalc_regions( g_wm );
    draw_window( g_wm->root_win, g_wm );

    g_prefs_restart_request = 0;
}

void colortheme_open( void )
{
    if( g_colortheme == 0 )
    {
        g_color_theme_changed = 0;

        g_prev_color_theme = g_wm->color_theme;

        int flags = DECOR_FLAG_CENTERED | DECOR_FLAG_CHECK_SIZE;
        g_colortheme = new_window_with_decorator(
            "Select a color theme",
            0, 0,
            g_wm->large_window_xsize, g_wm->large_window_ysize,
            g_wm->dialog_color,
            g_wm->root_win,
            colortheme_handler,
            flags,
            g_wm );
        show_window( g_colortheme, g_wm );
        bring_to_front( g_colortheme, g_wm );
        recalc_regions( g_wm );
        draw_window( g_wm->root_win, g_wm );
    }
}

void ui_scale_open( void )
{
    if( g_ui_scale == 0 )
    {
        g_ui_scale_changed = 0;

        int flags = DECOR_FLAG_CENTERED | DECOR_FLAG_CHECK_SIZE;
        g_ui_scale = new_window_with_decorator(
            "Set UI scale",
            0, 0,
            g_wm->normal_window_xsize, g_wm->normal_window_ysize,
            g_wm->dialog_color,
            g_wm->root_win,
            ui_scale_handler,
            flags,
            g_wm );
        show_window( g_ui_scale, g_wm );
        bring_to_front( g_ui_scale, g_wm );
        recalc_regions( g_wm );
        draw_window( g_wm->root_win, g_wm );
    }
}

//###################################
//## DEVICE DEPENDENT FUNCTIONS:   ##
//###################################

// device_start(), device_end() and device_event_handler() :

#ifndef FRAMEBUFFER
#endif

#ifndef NONPALM
    #include "wm_palmos.h"
#endif

#ifdef WIN
    #include "wm_win32.h"
#endif

#ifdef WINCE
    #include "wm_wince.h"
#endif

#ifdef UNIX
    #ifdef X11
	#include "wm_unixgraphics.h"
    #else
	#ifdef DIRECTDRAW
	    #include "wm_unixgraphics_sdl.h"
	#endif
	#ifdef IPHONE
	    #include "wm_iphone.h"
	#endif
	#ifdef OSX
	    #include "wm_osx.h"
	#endif
	#ifdef ANDROID
	    #include "wm_android.h"
	#endif
    #endif
#endif

//###################################
//###################################
//###################################
