#include "main/user_code.h"
#include "pixilang.h"
#include "various/iphone/webserver.h"

enum 
{
    pix_status_none = 0,
    pix_status_loading,
    pix_status_working,
    pix_status_disabled
};

pix_data* g_pd = 0; //Global Pixilang structure
UTF8_CHAR* g_pix_prog_name = 0;
bool g_pix_save_code_request = 0;
bool g_pix_no_program_selection_dialog = 0;
int g_pix_open_count = 0;
WINDOWPTR g_pix_win = 0;

#ifdef IPHONE
extern char* g_osx_resources_path;
UTF8_CHAR* g_pix_prog_name2 = 0;
#endif

#ifdef ANDROID
#ifdef FREE_VERSION
#include "android_resources.h"
#else
#include "../hidden/android_resources.h"
#endif
#endif

struct pixilang_window_data
{
    WINDOWPTR this_window;
    pix_vm* vm; //Pixilang virtual machine
    int status;
    sundog_image* screen_image;
    int timer;
};

void pix_vm_draw_screen( WINDOWPTR win, bool draw_changes )
{
    pixilang_window_data* data = (pixilang_window_data*)win->data;
    pix_vm* vm = data->vm;
    if( !vm ) return;
    
    vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i = win->xsize;
    vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i = win->ysize;
    
    int x, y;
    int screen_change_x;
    int screen_change_y;
    int screen_change_xsize;
    int screen_change_ysize;
    pix_vm_container* c = 0;
    
#ifdef OPENGL    
    if( vm->gl_callback != -1 )
    {
	//OpenGL graphics:
	win_draw_lock( win, g_wm );
	
	bool gl_draw = 0;
	if( vm->gl_callback != -1 ) //Double check
	{
	    glPushMatrix();
	    glLoadIdentity();
	    glTranslatef( (int)( win->xsize / 2 ), (int)( win->ysize / 2 ), 0 );
	
	    pix_vm_function fun;
    	    PIX_VAL pp[ 1 ];
    	    char pp_types[ 1 ];
	    fun.p = pp;
    	    fun.p_types = pp_types;
    	    fun.addr = vm->gl_callback;
    	    fun.p[ 0 ] = vm->gl_userdata;
    	    fun.p_types[ 0 ] = vm->gl_userdata_type;
    	    fun.p_num = 1;
    	    pix_vm_run( PIX_VM_THREADS - 2, 0, &fun, PIX_VM_CALL_FUNCTION, vm );
    	
    	    glPopMatrix();
    	    gl_draw = 1;
	}
	
	win_draw_unlock( win, g_wm );
	
	if( gl_draw )
	{
	    screen_changed();
	    goto skip_frame_drawing;
	}
    }
#endif

    //Check for changes:
    if( vm->screen_redraw_request != vm->screen_redraw_answer )
    {
        if( (unsigned)vm->screen < (unsigned)PIX_VM_CONTAINERS && vm->c[ vm->screen ] )
        {
	    c = vm->c[ vm->screen ];
	    screen_change_x = vm->screen_change_x;
	    screen_change_y = vm->screen_change_y;
	    screen_change_xsize = vm->screen_change_xsize;
	    screen_change_ysize = vm->screen_change_ysize;
	}
    }
    if( c && c->data )
    {
	if( data->screen_image ) remove_image( data->screen_image );		
	data->screen_image = new_image( c->xsize, c->ysize, c->data, 0, 0, c->xsize, c->ysize, IMAGE_NATIVE_RGB | IMAGE_STATIC_SOURCE );
    }
    
    if( draw_changes )
    {
	//Draw changed screen region only:
	
	if( !c ) return;
	if( !c->data ) return;

	x = ( win->xsize - c->xsize ) / 2;
	y = ( win->ysize - c->ysize ) / 2;
	
	win_draw_lock( win, g_wm );

	win_draw_bitmap_ext( 
	    win,
    	    x + screen_change_x, y + screen_change_y,
	    screen_change_xsize, screen_change_ysize, 
	    screen_change_x, screen_change_y,
	    data->screen_image, 
	    g_wm );
	
        win_draw_unlock( win, g_wm );        
    }
    else
    {
	//Force screen redraw (last available frame) + border redraw:
	
	win_draw_lock( win, g_wm );

	if( data->screen_image == 0 )
	    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, g_wm );
	else
	{
	    x = ( win->xsize - data->screen_image->xsize ) / 2;
	    y = ( win->ysize - data->screen_image->ysize ) / 2;
	    win_draw_box( win, 0, 0, x, win->ysize, win->color, g_wm );
	    win_draw_box( win, x + data->screen_image->xsize, 0, win->xsize - ( x + data->screen_image->xsize ), win->ysize, win->color, g_wm );
	    win_draw_box( win, x, 0, data->screen_image->xsize, y, win->color, g_wm );
    	    win_draw_box( win, x, y + data->screen_image->ysize, data->screen_image->xsize, win->ysize - ( y + data->screen_image->ysize ), win->color, g_wm );
    	    if( image_can_be_redrawn_safely( data->screen_image ) )
    	    {
    		win_draw_bitmap(
            	    win,
        	    x, y,
            	    data->screen_image,
            	    g_wm );
            }
	}
	
	win_draw_unlock( win, g_wm );        
    }
    
skip_frame_drawing:

    if( vm->screen_redraw_request != vm->screen_redraw_answer )
	vm->screen_redraw_answer = vm->screen_redraw_request;
}

int pixilang_menu_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    switch( win->action_result )
    {
	case 0:
	    //EXIT:
	    send_event( 0, EVT_BUTTONDOWN, 0, 0, 0, KEY_ESCAPE, 0, 1024, 0, wm );
	    break;
    }
    return 0;
}

int pixilang_prefs_handler( void* user_data, WINDOWPTR win, window_manager* wm )
{
    prefs_clear();
    g_prefs_flags |= PREFS_FLAG_NO_COLOR_THEME | PREFS_FLAG_NO_CONTROL_TYPE;
    prefs_add_default_sections();
    prefs_open();
    return 0;
}

void pixilang_timer( void* user_data, sundog_timer* t, window_manager* wm )
{
    pixilang_window_data* data = (pixilang_window_data*)user_data;
    WINDOWPTR win = data->this_window;
    
    switch( data->status )
    {
	case pix_status_none:
	    {
		data->status = pix_status_loading;

		if( g_pix_no_program_selection_dialog && g_pix_open_count > 0 )
		{
		    wm->exit_request = 1;
		    return;
		}

		if( g_pix_prog_name == 0 )
		{
		    FILES_USER_BUTTON = "Prefs";
		    FILES_USER_BUTTON_HANDLER = pixilang_prefs_handler;
		    g_pix_prog_name = dialog_open_file( "Select a program", 0, ".pixilang_f", 0 );
		}
			
		draw_window( win, wm );
                device_redraw_framebuffer();
			
		if( g_pix_prog_name == 0 )
		{
		    wm->exit_request = 1;
		}
		else 
		{
		    g_pix_open_count++;
		    if( data->vm )
		    {
			pix_vm_deinit( data->vm );
			mem_free( data->vm );
		    }
		    data->vm = (pix_vm*)MEM_NEW2( sizeof( pix_vm ) );
		    mem_zero( data->vm );
		    pix_vm_init( data->vm );
		    int cres = pix_compile( g_pix_prog_name, data->vm, g_pd );
		    if( cres == 0 )
		    {
			if( g_pix_save_code_request )
			{
			    UTF8_CHAR* code_name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( g_pix_prog_name ) + 10 );
			    code_name[ 0 ] = 0; mem_strcat( code_name, g_pix_prog_name );
			    int dot_ptr = mem_strlen( code_name );
			    for( int i = mem_strlen( code_name ); i >= 0; i-- )
			    {
				if( code_name[ i ] == '.' )
				{
				    dot_ptr = i;
				    break;
				}
			    }
			    code_name[ dot_ptr ] = 0;
			    mem_strcat( code_name, ".pixicode" );
#ifdef IPHONE
			    pix_vm_save_code( "1:/boot.pixicode", data->vm );
#else
			    pix_vm_save_code( code_name, data->vm );
#endif
			    g_pix_save_code_request = 0;
			    mem_free( code_name );
			}
		    }
		    data->status = pix_status_none;
		    draw_window( win, wm );
            	    device_redraw_framebuffer();
		    if( cres == 0 )
		    {
			data->vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i = win->xsize;
			data->vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i = win->ysize;
			data->vm->vars[ PIX_GVAR_PPI ].i = wm->screen_ppi;
			pix_vm_resize_container( data->vm->screen, win->xsize, win->ysize, -1, data->vm );
			pix_vm_gfx_set_screen( data->vm->screen, data->vm );
			PIX_VAL v;
			v.i = 0;
			set_focus_win( win, wm );
			pix_vm_clean_container( data->vm->screen, 0, v, 0, -1, data->vm );
			pix_vm_run( 0, 1, 0, PIX_VM_CALL_MAIN, data->vm );
			data->status = pix_status_working;
		    }
		    else
		    {
			if( g_pix_no_program_selection_dialog ) wm->exit_request = 1;
		    }
		}
			
		g_pix_prog_name = 0;
	    }
	    break;
	case pix_status_loading:
	    break;
	case pix_status_working:
	    if( data->vm && data->vm->th[ 0 ] && data->vm->th[ 0 ]->active == 0 )
	    {
		data->status = pix_status_none;
	    }
	    else 
	    {
		if( data->vm->screen_redraw_request != data->vm->screen_redraw_answer )
		{
		    if( wm->screen_buffer_preserved )
			pix_vm_draw_screen( win, 1 );
		    else
			screen_changed(); //... and all windows will be redrawn by SunDog engine before device_redraw_framebuffer
		}
	    }
	    if( data->vm->file_dialog_request == 1 )
	    {
		data->vm->file_dialog_request = 2;
#ifdef IPHONE
                fdialog_open_wifi = &webserver_open;
#endif
		data->vm->file_dialog_result = dialog_open_file( data->vm->file_dialog_name, data->vm->file_dialog_mask, data->vm->file_dialog_id, data->vm->file_dialog_def_name );
		data->vm->file_dialog_request = 0;
	    }
	    if( data->vm->prefs_dialog_request == 1 )
	    {
		data->vm->prefs_dialog_request = 2;
		//Show global preferences:
		pixilang_prefs_handler( 0, 0, g_wm );
		data->vm->prefs_dialog_request = 0;
	    }
	    if( data->vm->vsync_request > 0 )
	    {
		win_vsync( data->vm->vsync_request - 1, g_wm );
		data->vm->vsync_request = 0;
	    }
            if( data->vm->webserver_request == 1 )
            {
                data->vm->webserver_request = 2;
#ifdef IPHONE
                webserver_open();
                webserver_wait_for_close();
#endif
                data->vm->webserver_request = 0;
            }
	    break;
    }
}

int pixilang_window_handler( sundog_event* evt, window_manager* wm )
{
    int retval = 0;
    WINDOWPTR win = evt->win;
    pixilang_window_data* data = (pixilang_window_data*)win->data;
    switch( evt->type )
    {
	case EVT_GETDATASIZE:
	    retval = sizeof( pixilang_window_data );
	    break;
	case EVT_AFTERCREATE:
	    {
		data->this_window = win;
		data->vm = 0;
		data->status = pix_status_none;
		data->screen_image = 0;
#ifdef SHOW_PIXILANG_MENU
		int but_size = SCROLLBAR_SIZE;
		CREATE_FLAT_BUTTON = 1;
		CREATE_BUTTON_WITH_TEXT_MENU = "Exit";
		WINDOWPTR menu = new_window( "Pixilang Menu", 0, 0, but_size, but_size, wm->color0, win, button_handler, wm );
		button_set_text( menu, (UTF8_CHAR*)text_down, wm );
		set_window_controller( menu, 0, wm, CPERC, 100, CSUB, but_size, CEND );
		set_window_controller( menu, 2, wm, CPERC, 100, CEND );
		set_handler( menu, pixilang_menu_handler, 0, wm );
#endif
		if( g_pd == 0 )
		{
		    g_pd = (pix_data*)MEM_NEW2( sizeof( pix_data ) );
		    mem_zero( g_pd );
		    pix_init( g_pd );
		}
		data->timer = add_timer( pixilang_timer, (void*)data, 0, wm );
	    }
	    retval = 1;
	    break;
	case EVT_BEFORECLOSE:
	    {
		if( data->vm )
		{
		    if( data->status == pix_status_working )
			pix_vm_send_event( PIX_EVT_QUIT, 0, 0, 0, 0, 0, 0, 0, data->vm );
		    pix_vm_deinit( data->vm );
		    mem_free( data->vm );
		    data->vm = 0;
		}
		remove_image( data->screen_image );
		data->screen_image = 0;	
		remove_timer( data->timer, wm );
		data->timer = -1;
	    }
	    retval = 1;
	    break;
	case EVT_DRAW:
	    switch( data->status )
	    {
		case pix_status_none:
		    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
		    break;
		case pix_status_loading:
		    win_draw_box( win, 0, 0, win->xsize, win->ysize, win->color, wm );
		    win_draw_string( win, "Loading...", 0, 0, wm->color3, win->color, wm );
		    break;
		case pix_status_working:
		    pix_vm_draw_screen( win, 0 );
		    break;
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONDOWN:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_MOUSEBUTTONDOWN, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEBUTTONUP:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_MOUSEBUTTONUP, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_MOUSEMOVE:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_MOUSEMOVE, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_TOUCHBEGIN:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_TOUCHBEGIN, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_TOUCHEND:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_TOUCHEND, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_TOUCHMOVE:
	    if( data->vm )
	    {
		int x = evt->x - win->xsize / 2;
		int y = evt->y - win->ysize / 2;
		pix_vm_send_event( PIX_EVT_TOUCHMOVE, evt->flags, x, y, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_BUTTONDOWN:
	    if( data->vm )
	    {
		pix_vm_send_event( PIX_EVT_BUTTONDOWN, evt->flags, 0, 0, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_BUTTONUP:
	    if( data->vm )
	    {
		pix_vm_send_event( PIX_EVT_BUTTONUP, evt->flags, 0, 0, evt->key, evt->scancode, evt->pressure, evt->unicode, data->vm );
	    }
	    retval = 1;
	    break;
	case EVT_SCREENRESIZE:
	    win->x = 0;
	    win->y = 0;
	    win->xsize = wm->screen_xsize;
	    win->ysize = wm->screen_ysize;
            if( data->vm )
	    {
		pix_vm_send_event( PIX_EVT_SCREENRESIZE, 0, 0, 0, 0, 0, 0, 0, data->vm );
	    }
	    break;
    }
    return retval;
}

int user_init( window_manager* wm )
{
    g_pd = 0;
    g_pix_prog_name = 0;
    g_pix_no_program_selection_dialog = 0;
    g_pix_open_count = 0;

    wm->root_win = new_window( 
	"Desktop", 
	0, 0, 
	wm->screen_xsize, wm->screen_ysize, 
	wm->color0, 
	0, 
	desktop_handler,
	wm );
    show_window( wm->root_win, wm );
    recalc_regions( wm );
    draw_window( wm->root_win, wm );

#ifdef ANDROID
    copy_resources();
#endif
        
    V3_FILE f;
    
    if( wm->argc > 1 )
    {
	if( mem_strcmp( wm->argv[ 1 ], "-c" ) == 0 )
	{
	    if( wm->argc > 2 )
	    {
		g_pix_prog_name = wm->argv[ 2 ];
		g_pix_save_code_request = 1;
	    }
	    else
		dprint( "Error: no source file.\nUsage: pixilang -c source_file_name\n" );
	}
	else
	    g_pix_prog_name = wm->argv[ 1 ];
	f = v3_open( g_pix_prog_name, "rb" );
	if( f )
	    v3_close( f );
	else 
	{
	    g_pix_prog_name = 0;
	}
    }
    
#ifdef IPHONE
    if( g_pix_prog_name == 0 )
    {
	g_pix_prog_name2 = (char*)MEM_NEW2( mem_strlen( g_osx_resources_path ) + 256 );

	sprintf( g_pix_prog_name2, "%s/boot.pixicode", g_osx_resources_path );
	g_pix_prog_name = g_pix_prog_name2;
	f = v3_open( g_pix_prog_name, "rb" );
	if( f )
	    v3_close( f );
	else 
	    g_pix_prog_name = 0;

	if( g_pix_prog_name == 0 )
	{
	    sprintf( g_pix_prog_name2, "%s/boot.pixi", g_osx_resources_path );
	    g_pix_prog_name = g_pix_prog_name2;
	    f = v3_open( g_pix_prog_name, "rb" );
	    if( f )
		v3_close( f );
	    else 
		g_pix_prog_name = 0;
	}
    }
#else    
    if( g_pix_prog_name == 0 )
    {
	g_pix_prog_name = (UTF8_CHAR*)"1:/boot.pixicode";
	f = v3_open( g_pix_prog_name, "rb" );
	if( f )
	    v3_close( f );
	else 
	{
	    g_pix_prog_name = 0;
	}
    }
    
    if( g_pix_prog_name == 0 )
    {
	g_pix_prog_name = (UTF8_CHAR*)"1:/boot.pixi";
	f = v3_open( g_pix_prog_name, "rb" );
	if( f )
	    v3_close( f );
	else 
	{
	    g_pix_prog_name = 0;
	}
    }

    if( g_pix_prog_name == 0 )
    {
	g_pix_prog_name = (UTF8_CHAR*)"1:/boot.txt";
	f = v3_open( g_pix_prog_name, "rb" );
	if( f )
	    v3_close( f );
	else 
	{
	    g_pix_prog_name = 0;
	}
    }
#endif

    if( g_pix_prog_name )
    {
        g_pix_no_program_selection_dialog = 1;
#ifdef PIX_SAVE_BYTECODE
        g_pix_save_code_request = 1;
#endif
    }

    g_pix_win = new_window(
	"Pixilang window", 
	0, 0, 
	wm->screen_xsize, wm->screen_ysize, 
	wm->color0, 
	wm->root_win, 
	pixilang_window_handler,
	wm );
    set_window_controller( g_pix_win, 0, wm, (WCMD)0, CEND );
    set_window_controller( g_pix_win, 1, wm, (WCMD)0, CEND );
    set_window_controller( g_pix_win, 2, wm, CPERC, (WCMD)100, CEND );
    set_window_controller( g_pix_win, 3, wm, CPERC, (WCMD)100, CEND );
    g_pix_win->font = 1;
    show_window( g_pix_win, wm );
    recalc_regions( wm );
    draw_window( g_pix_win, wm );
    
    return 0;
}

int user_event_handler( sundog_event* evt, window_manager* wm )
{
    int handled = 0;
    
    switch( evt->type )
    {
	case EVT_BUTTONDOWN:
	    if( evt->key != KEY_ESCAPE ) break;
	case EVT_QUIT:
	    {
		pixilang_window_data* pdata = 0;
		if( g_pix_win )
		    pdata = (pixilang_window_data*)g_pix_win->data;
		if( g_pix_win && pdata->status != pix_status_working )
		{
		    wm->exit_request = 1;
		}
		else 
		{
		    pix_vm_send_event( PIX_EVT_QUIT, evt->flags, 0, 0, 0, 1, 0, 0, pdata->vm );
		    if( pdata->vm->quit_action == 1 )
		    {
			pix_vm_deinit( pdata->vm );
			mem_free( pdata->vm );
			pdata->vm = 0;
			pdata->status = pix_status_none;
			remove_image( pdata->screen_image );
            		pdata->screen_image = 0;
			draw_window( wm->root_win, wm );
            	    }
		}
		handled = 1;
	    }
	    break;
    }
    
    return handled;
}

void user_close( window_manager* wm )
{
    remove_window( wm->root_win, wm );

    if( g_pd )
    {
	pix_deinit( g_pd );
	mem_free( g_pd );
	g_pd = 0;
    }
#ifdef IPHONE
    mem_free( g_pix_prog_name2 );
    g_pix_prog_name2 = 0;
#endif
}
