#ifndef lint
static const char yysccsid[] = "@(#)yaccpar	1.9 (Berkeley) 02/21/93";
#endif

#include <stdlib.h>
#include <string.h>

#define YYBYACC 1
#define YYMAJOR 1
#define YYMINOR 9
#define YYPATCH 20090221

#define YYEMPTY        (-1)
#define yyclearin      (yychar = YYEMPTY)
#define yyerrok        (yyerrflag = 0)
#define YYRECOVERING() (yyerrflag != 0)

/* compatibility with bison */
#ifdef YYPARSE_PARAM
/* compatibility with FreeBSD */
#ifdef YYPARSE_PARAM_TYPE
#define YYPARSE_DECL() yyparse(YYPARSE_PARAM_TYPE YYPARSE_PARAM)
#else
#define YYPARSE_DECL() yyparse(void *YYPARSE_PARAM)
#endif
#else
#define YYPARSE_DECL() yyparse(void)
#endif /* YYPARSE_PARAM */

extern int YYPARSE_DECL();

static int yygrowstack(void);
#define YYPREFIX "yy"
#line 2 "pixilang_compiler.y"
/*
    pixilang_compiler.y
    This file is part of the Pixilang programming language.
    
    [ MIT license ]
    
    Copyright (c) 2006 - 2012, Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to 
    deal in the Software without restriction, including without limitation the 
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in 
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
*/

#include "pixilang.h"
#include "pixilang_font.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*#define SHOW_DEBUG_MESSAGES*/

#ifdef SHOW_DEBUG_MESSAGES
    #define DPRINT( fmt, ARGS... ) dprint( fmt, ## ARGS )
#else
    #define DPRINT( fmt, ARGS... ) {}
#endif
    
#define ERROR( fmt, ARGS... ) dprint( "ERROR in %s: " fmt "\n", __FUNCTION__, ## ARGS )
#define SHOW_ERROR( fmt, ARGS... ) \
    { \
	UTF8_CHAR* ts = (UTF8_CHAR*)MEM_NEW2( mem_strlen( g_comp->src_name ) + 2048 ); \
	dprint( "ERROR (line %d) in %s: " fmt "\n", g_comp->src_line + 1, g_comp->src_name, ## ARGS ); \
	sprintf( ts, "ERROR (line %d)\nin %s:\n" fmt, g_comp->src_line + 1, g_comp->src_name, ## ARGS ); \
	dialog( ts, "Close" ); \
	mem_free( ts ); \
    }
    
#define NUMERIC( val ) ( val >= '0' && val <= '9' )
#define ABC( val ) ( ( val >= 'a' && val <= 'z' ) || ( val >= 'A' && val <= 'Z' ) || ( (unsigned)val >= 128 && val != -1 ) )

enum lnode_type
{
    lnode_empty,
    
    lnode_statlist, /*statements*/
    
    lnode_int,
    lnode_float,
    lnode_var,
    
    lnode_halt,
    
    lnode_label,
    lnode_function_label_from_node,
    
    lnode_go,
    
    lnode_jmp_to_node,
    lnode_jmp_to_end_of_node,
    
    lnode_if,
    lnode_if_else,
    lnode_while,
    
    lnode_save_to_var,
    
    lnode_save_to_prop,
    lnode_load_from_prop,

    lnode_save_to_mem,
    lnode_load_from_mem,
    
    lnode_save_to_stackframe,
    lnode_load_from_stackframe,
    
    lnode_sub,
    lnode_add,
    lnode_mul,
    lnode_idiv,
    lnode_div,
    lnode_mod,
    lnode_and,
    lnode_or,
    lnode_xor,
    lnode_andand,
    lnode_oror,
    lnode_eq,
    lnode_neq,
    lnode_less,
    lnode_leq,
    lnode_greater,
    lnode_geq,
    lnode_lshift,
    lnode_rshift,
    
    lnode_neg,

    lnode_exprlist,
    lnode_call_builtin_fn,
    lnode_call_builtin_fn_void,
    lnode_call,
    lnode_call_void,
    lnode_ret_int,
    lnode_ret,
    lnode_inc_sp,
};
    
#define LNODE_FLAG_STATLIST_AS_EXPRESSION		1
#define LNODE_FLAG_STATLIST_WITH_JMP_HEADER		2
#define LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER	4
#define LNODE_FLAG_STATLIST_SKIP_NEXT_HEADER		8

struct lnode
{
    lnode_type type;
    uchar flags;
    size_t code_ptr; /*Start of node*/
    size_t code_ptr2; /*End of node (start of the next node)*/
    PIX_VAL val;
    lnode** n; /*Children*/
    int nn; /*Children count*/
};
    
struct pix_lsymtab
{
    pix_symtab* lsym; /*Local symbol table*/
    int lvars_num; /*Number of local variables*/
    int pars_num; /*Number of local function parameters*/
};
    
struct pix_include
{
    UTF8_CHAR* src;
    int src_ptr;
    int src_line;
    int src_size;
    UTF8_CHAR* src_name;
    UTF8_CHAR* base_path;
};

#define VAR_FLAG_LABEL					1
#define VAR_FLAG_FUNCTION				2
#define VAR_FLAG_INITIALIZED				4
#define VAR_FLAG_USED					8
    
struct pix_compiler
{
    lnode* root; /*Root lexical node*/
    
    UTF8_CHAR* src;
    int src_ptr;
    int src_line;
    int src_size;
    UTF8_CHAR* src_name;
    
    pix_symtab sym; /*Global symbol table*/
    pix_lsymtab* lsym; /*Local symbol tables*/
    int lsym_num; /*Number of current local symbol table*/
    UTF8_CHAR temp_sym_name[ 256 + 1 ];
    
    char* var_flags;
    size_t var_flags_size;

    int statlist_header_size; /*Maximum length of JMP instruction*/
    
    bool fn_pars_mode; /*Treat new lvars as function parameters*/
    
    UTF8_CHAR* base_path; /*May be not equal to vm->base_path*/
    
    pix_include* inc; /*Stack for includes*/
    int inc_num;
    
    lnode** while_stack;
    int while_stack_ptr;
    
    lnode** fixup;
    int fixup_num;
    
    pix_vm* vm;
    pix_data* pd;
};

pix_compiler* g_comp = 0;

int yylex( void );
void yyerror( char const* str );
void push_int( PIX_INT v );
lnode* node( lnode_type type, int nn );
void resize_node( lnode* n, int nn );
lnode* clone_tree( lnode* n );
void remove_tree( lnode* n );
void optimize_tree( lnode* n );
void compile_tree( lnode* n );

lnode* make_expr_node_from_var( PIX_INT i )
{
    lnode* n = node( lnode_var, 0 ); 
    n->val.i = i;
    return n;
}

lnode* make_expr_node_from_local_var( PIX_INT i )
{
    lnode* n = node( lnode_load_from_stackframe, 0 );
    n->val.i = i;
    return n;
}

void create_empty_lsym_table( void )
{
    g_comp->lsym_num++;
    if( g_comp->lsym_num >= mem_get_size( g_comp->lsym ) / sizeof( pix_lsymtab ) )
	g_comp->lsym = (pix_lsymtab*)mem_resize( g_comp->lsym, ( g_comp->lsym_num + 8 ) * sizeof( pix_lsymtab ) );
    mem_set( &g_comp->lsym[ g_comp->lsym_num ], sizeof( pix_lsymtab ), 0 );
}
    
lnode* remove_lsym_table( lnode* statlist )
{
    lnode* new_tree;
    
    int lvars_num = g_comp->lsym[ g_comp->lsym_num ].lvars_num;
    if( lvars_num )
    {
	new_tree = node( lnode_statlist, 2 );
	new_tree->n[ 0 ] = node( lnode_inc_sp, 0 );
	new_tree->n[ 0 ]->val.i = -lvars_num;
	new_tree->n[ 1 ] = statlist;
    }
    else 
    {
	new_tree = statlist;
    }
    
    pix_symtab* lsym = g_comp->lsym[ g_comp->lsym_num ].lsym;
    if( lsym )
    {
	pix_symtab_deinit( lsym );
	mem_free( lsym );
	g_comp->lsym[ g_comp->lsym_num ].lsym = 0;
	g_comp->lsym[ g_comp->lsym_num ].lvars_num = 0;
	g_comp->lsym[ g_comp->lsym_num ].pars_num = 0;
    }
    g_comp->lsym_num--;
    
    return new_tree;
}

#line 267 "pixilang_compiler.y"
typedef union /*Possible types for yylval and yyval:*/
{
    PIX_INT i;
    PIX_FLOAT f;
    lnode* n;
} YYSTYPE;
#line 305 "pixilang_compiler.cpp"
#define NUM_I 257
#define NUM_F 258
#define GVAR 259
#define LVAR 260
#define WHILE 261
#define BREAK 262
#define CONTINUE 263
#define IF 264
#define ELSE 265
#define GO 266
#define RET 267
#define FNNUM 268
#define FNDEF 269
#define INCLUDE 270
#define HALT 271
#define OROR 272
#define ANDAND 273
#define OR 274
#define XOR 275
#define AND 276
#define EQ 277
#define NEQ 278
#define LEQ 279
#define GEQ 280
#define LSHIFT 281
#define RSHIFT 282
#define IDIV 283
#define HASH 284
#define NEG 285
#define YYERRCODE 256
static const short yylhs[] = {                           -1,
    0,    0,    2,    2,    2,    3,    3,    5,    5,    5,
    5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
    5,    5,    5,    5,    5,    5,    6,    6,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,   11,    1,    1,    1,   12,
   13,    1,    1,    9,    9,    9,   14,   14,   14,   14,
   14,   10,   10,   10,   10,    8,    8,    8,    8,    7,
    7,    7,    7,    4,    4,    4,    4,   15,    4,    4,
    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
    4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
};
static const short yylen[] = {                            2,
    0,    2,    0,    1,    3,    1,    3,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    0,    2,    1,    2,
    2,    3,    3,    3,    3,    3,    3,    3,    3,    4,
    1,    1,    4,    5,    9,    0,    6,    1,    1,    0,
    0,   10,    2,    0,    1,    3,    1,    1,    1,    1,
    4,    4,    4,    4,    4,    4,    4,    4,    4,    2,
    2,    2,    2,    1,    1,    1,    1,    0,    4,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    2,
};
static const short yydefred[] = {                         1,
    0,   57,   58,    0,    0,    0,   48,   49,    0,    0,
    0,    0,   50,    0,   29,    2,    0,    0,    0,    0,
   18,   17,   15,   16,   14,   19,   20,   21,   23,   22,
   24,   25,   26,    9,    8,   10,   12,   11,   13,   30,
    0,    0,    0,    0,   59,   60,    0,    0,    0,   78,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   53,   73,    0,    0,    0,    0,   72,    0,    0,    0,
    0,   71,    0,    0,   70,    0,    0,    0,    0,    0,
    0,    0,  100,    0,   27,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   46,   27,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   80,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   83,   85,   84,   86,   27,    0,   43,    0,    0,
    0,   65,   69,    0,   64,   68,   63,   67,   62,   66,
   61,   79,   28,    0,    0,    0,    4,    0,    0,   47,
    0,    0,   51,   27,    5,    0,    0,   27,   45,    0,
   52,
};
static const short yydgoto[] = {                          1,
  163,  168,  113,  108,   42,  126,   52,   53,  109,   54,
  146,   60,  176,   55,   85,
};
static const short yysindex[] = {                         0,
 1111,    0,    0,  180, 1590, 1181,    0,    0, 1181, 1181,
  -36,  -33,    0, -254,    0,    0,  146, 1569,  -18,   26,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1181, 1181, 1181, 1181,    0,    0,  -27, 1181, 1181,    0,
 1602,   60,   69,  -18,   26, 1616, 1664, 1181, 1181, -250,
    0,    0, 1181, 1181, 1181, 1181,    0, 1181, 1181, 1181,
 1181,    0, 1181, 1181,    0, 1181, 1181, 1664, 1664, 1664,
 1664, 1181,    0, 1628,    0, 1181, 1181, 1181, 1181, 1181,
 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
 1181, 1181, 1181, 1181,    0,    0, 1640, 1664,  -11,  -25,
 1664,    4,  -79, 1652, 1664, 1664,    6,  -75, 1664,    8,
  -70,   14,  -65,   15,    0,  286,  -26,  -26,  480,  480,
  480,  155,  155,  155,  155,  155,  155,  -37,  -37,  -10,
  -10,    0,    0,    0,    0,    0, 1474,    0, 1181,    0,
 -231,    0,    0, 1181,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1555, -225, 1664,    0,   27, 1664,    0,
  -72, -200,    0,    0,    0,  -46, 1695,    0,    0, 1710,
    0,
};
static const short yyrindex[] = {                         0,
    0,    0,    0,   94,  128,    0,    0,    0,    0,    0,
  694,    0,    0,    0,    0,    0,    0,    0, 1039,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   20,   54,   88,  122,    0, 1094,    0,   28,    0,
    0,    0,    0,   28,    0,    0,    0,    0,   28,    0,
    0,    0,   28,    0,    0,   28,    0, 1129, 1163, 1205,
 1220,   28,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   34,    0,    0,
 1251,    0,    0,   -8, 1276, 1291,    0,    0, 1319,    0,
    0,    0,    0,    0,    0,    0, 1011, 1054,  774,  928,
  974,  532,  566,  724,  813,  848,  882,  440,  474,  372,
  406,    0,    0,    0,    0,    0,    0,    0,    0,    1,
   35,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1334,   40,    0,    0,   -7,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,
};
static const short yygindex[] = {                         0,
   82,    0,  -31, 1707,    7, -104,   52,  413,  -38,  432,
    0,    0,    0,  438,    0,
};
#define YYTABLESIZE 1981
static const short yytable[] = {                        104,
   40,  147,   61,   58,  101,   99,   59,  100,  110,  102,
  104,   44,   82,  153,  151,  101,   99,  156,  100,   77,
  102,   73,  158,   66,   71,  112,  104,  160,  167,  150,
  117,  101,  149,   93,  120,   94,  102,  122,  118,  171,
   61,  164,  121,  124,  152,  123,  155,  149,  157,  149,
  174,  149,   17,   76,  159,  161,   77,  149,  149,  175,
   77,   77,   77,   77,   77,   76,   77,  173,   54,  177,
  172,   54,   74,  180,   55,    3,  178,   55,    3,   77,
   56,   77,   16,   56,    6,    7,    0,   75,    0,    0,
   76,   61,    0,    0,   76,   76,   76,   76,   76,   64,
   76,    0,    0,    0,    0,    0,    0,    0,   69,    0,
    0,    0,   77,   76,    0,   76,   77,    0,    0,    0,
    0,   74,    0,    0,   75,   40,    0,    0,   75,   75,
   75,   75,   75,   59,   75,    0,    0,    0,    0,    0,
    0,    0,   77,    0,   77,    0,   76,   75,    0,   75,
   65,    0,    0,    0,    0,    0,    0,    0,   74,   70,
    0,    0,   74,   74,   74,   74,   74,   60,   74,    0,
    0,    0,    0,    0,    0,    0,   76,   17,   76,    0,
   75,   74,   39,   74,   59,   64,    0,   36,   34,    0,
   35,  104,   37,    0,    0,    0,  101,   99,   17,  100,
    0,  102,    0,    0,    0,   28,   63,   29,    0,    0,
   75,    0,   75,    0,   74,   17,   39,    0,   60,    0,
    0,   36,   34,    0,   35,    0,   37,    0,   17,    0,
    0,   17,    0,    0,    0,    0,   65,   40,    0,   28,
   41,   29,    0,    0,   74,  103,   74,   88,   89,   90,
   91,   92,   95,   96,   97,   98,  103,   40,   40,   40,
   40,   40,   40,   40,   40,   72,   40,   40,   40,   40,
   40,   40,  103,    0,    0,    0,   77,   77,   77,   77,
   77,   77,   77,   77,   61,   77,   77,   77,   77,   77,
   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
   77,   77,   77,    0,    0,    0,    0,    0,    0,   75,
   76,   76,   76,   76,   76,   76,   76,   76,    0,   76,
   76,   76,   76,   76,   76,   76,   76,   76,   76,   76,
   76,   76,   76,   76,   76,   76,   76,    0,    0,    0,
    0,    0,    0,   62,   75,   75,   75,   75,   75,   75,
   75,   75,   67,   75,   75,   75,   75,   75,   75,   75,
   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
   75,   82,    0,    0,    0,    0,    0,   59,   74,   74,
   74,   74,   74,   74,   74,   74,    0,   74,   74,   74,
   74,   74,   74,   74,   74,   74,   74,   74,   74,   74,
   74,   74,   74,   74,   74,   81,    0,    0,    0,    0,
  162,   60,   82,   18,   82,   82,   82,   21,   22,   23,
   24,   25,   26,   27,   30,   31,   32,   33,   38,   62,
    0,   82,   19,   82,    0,   97,   98,  103,   20,   98,
    0,    0,    0,    0,    0,    0,   81,    0,   81,   81,
   81,   21,   22,   23,   24,   25,   26,   27,   30,   31,
   32,   33,   38,    0,   82,   81,    0,   81,    0,    0,
    0,    0,    0,   99,    0,    0,    0,    0,    0,    0,
   98,    0,    0,   98,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   82,    0,   82,    0,   81,   98,
    0,   98,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   99,    0,  104,   99,    0,    0,
    0,  101,   99,    0,  100,    0,  102,    0,   81,    0,
   81,   92,   98,   99,    0,   99,    0,    0,   18,   93,
    0,   94,    2,    3,    4,    5,    6,    7,    8,    9,
    0,   10,   11,   12,   13,   14,   15,   19,    0,   18,
    0,    0,   98,   20,   98,   93,   99,    0,    0,    0,
    0,    0,   92,    0,    0,   92,   18,    0,   19,    0,
    0,    0,    0,    0,   20,    0,    0,    0,    0,   18,
    0,   92,   18,   92,    0,   19,   99,    0,   99,    0,
    0,   20,    0,    0,    0,    0,   93,    0,   19,   93,
    0,   19,    0,    0,   20,    0,    0,   20,    0,    0,
    0,    0,    0,    0,   92,   93,    0,   93,   82,   82,
   82,   82,   82,   82,   82,   82,    0,   82,   82,   82,
   82,   82,   82,   82,   82,   82,   82,   82,   82,   82,
   82,   82,   82,   82,   92,    0,   92,    0,   93,    0,
    0,    0,   81,   81,   81,   81,   81,   81,   81,   81,
    0,   81,   81,   81,   81,   81,   81,   81,   81,   81,
   81,   81,   81,   81,   81,   81,   81,   81,   93,    0,
   93,    0,    0,   42,    0,    0,   98,   98,   98,   98,
   98,   98,   98,   98,    0,   98,   98,   98,   98,   98,
   98,   98,   98,   98,   98,   98,   98,   98,   98,   98,
   98,   98,    0,   94,    0,    0,    0,    0,    0,    0,
   99,   99,   99,   99,   99,   99,   99,   99,    0,   99,
   99,   99,   99,   99,   99,   99,   99,   99,   99,   99,
   99,   99,   99,   99,   99,   99,   91,   92,   95,   96,
   97,   98,  103,    0,   94,    0,    0,   94,    0,    0,
    0,    0,    0,   88,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   94,    0,   94,    0,    0,   92,   92,
   92,   92,   92,   92,   92,   92,    0,   92,   92,   92,
   92,   92,   92,   92,   92,   92,   92,   92,   92,   92,
   92,   92,   96,    0,   88,    0,   94,   88,   42,    0,
    0,    0,   93,   93,   93,   93,   93,   93,   93,   93,
    0,   93,   93,   93,   93,   93,   93,   93,   93,   93,
   93,   93,   93,   93,   93,   93,   94,   95,   94,    0,
    0,    0,    0,   96,    0,    0,   96,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   88,    0,    0,    0,
    0,    0,   96,    0,   96,    0,    0,    0,    0,    0,
    0,   97,    0,    0,    0,    0,    0,    0,   95,    0,
    0,   95,    0,    0,    0,    0,   88,    0,   88,    0,
    0,    0,    0,    0,    0,   96,    0,   95,    0,   95,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   97,    0,    0,   97,    0,   89,    0,    0,
    0,    0,    0,    0,    0,   96,    0,   96,    0,    0,
   95,   97,    0,   97,    0,    0,    0,    0,    0,    0,
   42,   42,   42,   42,   42,   42,   42,   42,    0,   42,
   42,   42,   42,   42,   42,    0,    0,    0,   89,    0,
   95,   89,   95,   87,   97,    0,    0,    0,    0,    0,
   94,   94,   94,   94,   94,   94,   94,   94,    0,   94,
   94,   94,   94,   94,   94,   94,   94,   94,   94,   94,
   94,   94,   94,   94,   97,    0,   97,    0,    0,    0,
   91,    0,    0,    0,   87,    0,    0,   87,    0,    0,
   89,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   88,   88,   88,   88,   88,   88,   88,   88,   41,   88,
   88,   88,   88,   88,   88,   88,   88,   88,   88,   88,
   89,   91,   89,   90,   91,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   87,    0,    0,   96,
   96,   96,   96,   96,   96,   96,   96,    0,   96,   96,
   96,   96,   96,   96,   96,   96,   96,   96,   96,   96,
   96,   96,   96,   31,   90,    0,   87,   90,   87,    0,
    0,    0,    0,   91,   95,   95,   95,   95,   95,   95,
   95,   95,    0,   95,   95,   95,   95,   95,   95,   95,
   95,   95,   95,   95,   95,   95,   95,   95,   32,    0,
    0,    0,    0,   91,    0,   91,    0,    0,   97,   97,
   97,   97,   97,   97,   97,   97,   90,   97,   97,   97,
   97,   97,   97,   97,   97,   97,   97,   97,   97,   97,
   97,   97,   38,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,    0,   90,    0,
    0,    0,    0,    0,   89,   89,   89,   89,   89,   89,
   89,   89,    0,   89,   89,   89,   89,   89,   89,   89,
   89,   89,   89,   89,   33,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   31,   39,
   49,    0,    0,    0,    0,   48,    0,    0,    0,    0,
   87,   87,   87,   87,   87,   87,   87,   87,    0,   87,
   87,   87,   87,   87,   87,   87,   87,   87,   87,   87,
   34,    0,    0,   32,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   91,   91,   91,
   91,   91,   91,   91,   91,   35,   91,   91,   91,   91,
   91,   91,   91,   91,    0,    0,    0,   38,    0,    0,
   36,    0,    0,    0,    0,   41,   41,   41,   41,   41,
   41,   41,   41,   50,   41,   41,   41,   41,   41,   41,
   90,   90,   90,   90,   90,   90,   90,   90,   37,   90,
   90,   90,   90,   90,   90,   90,   90,    0,    0,   33,
    0,    0,    0,   44,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   39,    0,    0,    0,    0,    0,
   31,   31,   31,   31,   31,   31,   31,   31,    0,   31,
   31,   31,   31,   31,   31,    0,    0,    2,    3,    4,
    5,    6,    7,    8,    9,   34,   10,   11,   12,   13,
   14,   15,    0,    0,    0,   32,   32,   32,   32,   32,
   32,   32,   32,    0,   32,   32,   32,   32,   32,   32,
   35,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,    0,    0,   38,
   38,   38,   38,   38,   38,   38,   38,    0,   38,   38,
   38,   38,   38,   38,    0,    0,    0,    2,    3,   45,
   46,    0,    0,   37,    0,    0,    0,    0,   47,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   44,    0,
    0,   33,   33,   33,   33,   33,   33,   33,   33,    0,
   33,   33,   33,   33,   33,   33,   39,   39,   39,   39,
   39,   39,   39,   39,    0,   39,   39,   39,   39,   39,
   39,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   34,   34,   34,
   34,   34,   34,   34,   34,    0,   34,   34,   34,   34,
   34,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   35,   35,   35,   35,   35,   35,   35,   35,
    0,   35,   35,   35,   35,   35,   35,   36,   36,   36,
   36,   36,   36,   36,   36,    0,   36,   36,   36,   36,
   36,   36,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   37,   37,   37,   37,   37,
   37,   37,   37,    0,   37,   37,   37,   37,   37,   37,
   44,   44,   44,   44,   44,   44,   44,   44,  165,   44,
   44,   44,   44,   44,   44,   39,    0,    0,   69,    0,
   36,   34,    0,   35,    0,   37,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   39,    0,   28,   68,
   29,   36,   34,    0,   35,    0,   37,    0,  104,    0,
    0,    0,    0,  101,   99,    0,  100,    0,  102,   28,
   43,   29,  104,    0,    0,    0,    0,  101,   99,   70,
  100,   93,  102,   94,  104,    0,    0,    0,  125,  101,
   99,    0,  100,    0,  102,   93,  104,   94,    0,  170,
  148,  101,   99,    0,  100,    0,  102,   93,  104,   94,
    0,    0,    0,  101,   99,  154,  100,    0,  102,   93,
  104,   94,    0,    0,    0,  101,   99,    0,  100,    0,
  102,   93,   51,   94,    0,   56,   57,    0,    0,    0,
    0,    0,    0,   93,  105,   94,    0,    0,    0,    0,
    2,    3,    4,    5,    6,    7,    8,    9,  106,   10,
   11,   12,   13,   14,   15,    0,    0,   78,   79,   80,
   81,    0,    0,    0,   83,   84,    0,    0,    0,    0,
    0,    0,    0,    0,  107,    0,    0,    0,    0,  111,
    0,  114,  115,    0,  116,    0,  114,  119,    0,    0,
  114,    0,    0,  114,    0,    0,    0,    0,    0,    0,
    0,    0,  127,  128,  129,  130,  131,  132,  133,  134,
  135,  136,  137,  138,  139,  140,  141,  142,  143,  144,
  145,    2,    3,    4,    5,    6,    7,    8,    9,  179,
   10,   11,   12,   13,   14,   15,    0,    0,    0,    0,
    0,    0,    0,    0,  181,    0,    0,    0,    0,    0,
   21,   22,   23,   24,   25,   26,   27,   30,   31,   32,
   33,   38,   67,    0,    0,  166,    0,    0,    0,    0,
  169,   21,   22,   23,   24,   25,   26,   27,   30,   31,
   32,   33,   38,   86,   87,   88,   89,   90,   91,   92,
   95,   96,   97,   98,  103,    0,    0,   86,   87,   88,
   89,   90,   91,   92,   95,   96,   97,   98,  103,   86,
   87,   88,   89,   90,   91,   92,   95,   96,   97,   98,
  103,   86,   87,   88,   89,   90,   91,   92,   95,   96,
   97,   98,  103,   86,   87,   88,   89,   90,   91,   92,
   95,   96,   97,   98,  103,   86,   87,   88,   89,   90,
   91,   92,   95,   96,   97,   98,  103,    0,    0,    0,
    0,    2,    3,    4,    5,    6,    7,    8,    9,    0,
   10,   11,   12,   13,   14,   15,    2,    3,    4,    5,
    6,    7,    8,    9,    0,   10,   11,   12,   13,   14,
   15,
};
static const short yycheck[] = {                         37,
    0,  106,  257,   40,   42,   43,   40,   45,  259,   47,
   37,    5,   40,   93,   40,   42,   43,   93,   45,    0,
   47,   40,   93,   17,   18,   64,   37,   93,  260,   41,
   69,   42,   44,   60,   73,   62,   47,   76,   70,  265,
   40,  146,   74,   82,   41,   77,   41,   44,   41,   44,
  123,   44,    1,    0,   41,   41,   37,   44,   44,  260,
   41,   42,   43,   44,   45,   40,   47,   41,   41,  174,
   44,   44,   91,  178,   41,   41,  123,   44,   44,   60,
   41,   62,    1,   44,   93,   93,   -1,    0,   -1,   -1,
   37,   91,   -1,   -1,   41,   42,   43,   44,   45,   40,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   40,   -1,
   -1,   -1,   93,   60,   -1,   62,   91,   -1,   -1,   -1,
   -1,    0,   -1,   -1,   37,  125,   -1,   -1,   41,   42,
   43,   44,   45,   40,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  123,   -1,  125,   -1,   93,   60,   -1,   62,
   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   37,   91,
   -1,   -1,   41,   42,   43,   44,   45,   40,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  123,  126,  125,   -1,
   93,   60,   37,   62,   91,   40,   -1,   42,   43,   -1,
   45,   37,   47,   -1,   -1,   -1,   42,   43,  147,   45,
   -1,   47,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
  123,   -1,  125,   -1,   93,  164,   37,   -1,   91,   -1,
   -1,   42,   43,   -1,   45,   -1,   47,   -1,  177,   -1,
   -1,  180,   -1,   -1,   -1,   -1,   91,   58,   -1,   60,
   61,   62,   -1,   -1,  123,  283,  125,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  257,  258,  259,
  260,  261,  262,  263,  264,  284,  266,  267,  268,  269,
  270,  271,  283,   -1,   -1,   -1,  257,  258,  259,  260,
  261,  262,  263,  264,  284,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,   -1,   -1,   -1,   -1,   -1,   -1,  284,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,   -1,   -1,   -1,
   -1,   -1,   -1,  284,  257,  258,  259,  260,  261,  262,
  263,  264,  284,  266,  267,  268,  269,  270,  271,  272,
  273,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,    0,   -1,   -1,   -1,   -1,   -1,  284,  257,  258,
  259,  260,  261,  262,  263,  264,   -1,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,    0,   -1,   -1,   -1,   -1,
  125,  284,   41,    1,   43,   44,   45,  272,  273,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
   -1,   60,    1,   62,   -1,  281,  282,  283,    1,    0,
   -1,   -1,   -1,   -1,   -1,   -1,   41,   -1,   43,   44,
   45,  272,  273,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,   -1,   93,   60,   -1,   62,   -1,   -1,
   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,
   41,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  123,   -1,  125,   -1,   93,   60,
   -1,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   41,   -1,   37,   44,   -1,   -1,
   -1,   42,   43,   -1,   45,   -1,   47,   -1,  123,   -1,
  125,    0,   93,   60,   -1,   62,   -1,   -1,  126,   60,
   -1,   62,  257,  258,  259,  260,  261,  262,  263,  264,
   -1,  266,  267,  268,  269,  270,  271,  126,   -1,  147,
   -1,   -1,  123,  126,  125,    0,   93,   -1,   -1,   -1,
   -1,   -1,   41,   -1,   -1,   44,  164,   -1,  147,   -1,
   -1,   -1,   -1,   -1,  147,   -1,   -1,   -1,   -1,  177,
   -1,   60,  180,   62,   -1,  164,  123,   -1,  125,   -1,
   -1,  164,   -1,   -1,   -1,   -1,   41,   -1,  177,   44,
   -1,  180,   -1,   -1,  177,   -1,   -1,  180,   -1,   -1,
   -1,   -1,   -1,   -1,   93,   60,   -1,   62,  257,  258,
  259,  260,  261,  262,  263,  264,   -1,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  123,   -1,  125,   -1,   93,   -1,
   -1,   -1,  257,  258,  259,  260,  261,  262,  263,  264,
   -1,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  123,   -1,
  125,   -1,   -1,    0,   -1,   -1,  257,  258,  259,  260,
  261,  262,  263,  264,   -1,  266,  267,  268,  269,  270,
  271,  272,  273,  274,  275,  276,  277,  278,  279,  280,
  281,  282,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  277,  278,  279,  280,
  281,  282,  283,   -1,   41,   -1,   -1,   44,   -1,   -1,
   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   60,   -1,   62,   -1,   -1,  257,  258,
  259,  260,  261,  262,  263,  264,   -1,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,  278,
  279,  280,    0,   -1,   41,   -1,   93,   44,  125,   -1,
   -1,   -1,  257,  258,  259,  260,  261,  262,  263,  264,
   -1,  266,  267,  268,  269,  270,  271,  272,  273,  274,
  275,  276,  277,  278,  279,  280,  123,    0,  125,   -1,
   -1,   -1,   -1,   41,   -1,   -1,   44,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   60,   -1,   62,   -1,   -1,   -1,   -1,   -1,
   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   41,   -1,
   -1,   44,   -1,   -1,   -1,   -1,  123,   -1,  125,   -1,
   -1,   -1,   -1,   -1,   -1,   93,   -1,   60,   -1,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   41,   -1,   -1,   44,   -1,    0,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  123,   -1,  125,   -1,   -1,
   93,   60,   -1,   62,   -1,   -1,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,   -1,   -1,   -1,   41,   -1,
  123,   44,  125,    0,   93,   -1,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  277,  278,  279,  280,  123,   -1,  125,   -1,   -1,   -1,
    0,   -1,   -1,   -1,   41,   -1,   -1,   44,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,    0,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
  123,   41,  125,    0,   44,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,  257,
  258,  259,  260,  261,  262,  263,  264,   -1,  266,  267,
  268,  269,  270,  271,  272,  273,  274,  275,  276,  277,
  278,  279,  280,    0,   41,   -1,  123,   44,  125,   -1,
   -1,   -1,   -1,   93,  257,  258,  259,  260,  261,  262,
  263,  264,   -1,  266,  267,  268,  269,  270,  271,  272,
  273,  274,  275,  276,  277,  278,  279,  280,    0,   -1,
   -1,   -1,   -1,  123,   -1,  125,   -1,   -1,  257,  258,
  259,  260,  261,  262,  263,  264,   93,  266,  267,  268,
  269,  270,  271,  272,  273,  274,  275,  276,  277,  278,
  279,  280,    0,  125,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  123,   -1,  125,   -1,
   -1,   -1,   -1,   -1,  257,  258,  259,  260,  261,  262,
  263,  264,   -1,  266,  267,  268,  269,  270,  271,  272,
  273,  274,  275,  276,    0,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  125,    0,
   40,   -1,   -1,   -1,   -1,   45,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,  272,  273,  274,  275,  276,
    0,   -1,   -1,  125,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,  258,  259,
  260,  261,  262,  263,  264,    0,  266,  267,  268,  269,
  270,  271,  272,  273,   -1,   -1,   -1,  125,   -1,   -1,
    0,   -1,   -1,   -1,   -1,  257,  258,  259,  260,  261,
  262,  263,  264,  123,  266,  267,  268,  269,  270,  271,
  257,  258,  259,  260,  261,  262,  263,  264,    0,  266,
  267,  268,  269,  270,  271,  272,  273,   -1,   -1,  125,
   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  125,   -1,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,   -1,  266,
  267,  268,  269,  270,  271,   -1,   -1,  257,  258,  259,
  260,  261,  262,  263,  264,  125,  266,  267,  268,  269,
  270,  271,   -1,   -1,   -1,  257,  258,  259,  260,  261,
  262,  263,  264,   -1,  266,  267,  268,  269,  270,  271,
  125,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  125,   -1,   -1,   -1,  257,
  258,  259,  260,  261,  262,  263,  264,   -1,  266,  267,
  268,  269,  270,  271,   -1,   -1,   -1,  257,  258,  259,
  260,   -1,   -1,  125,   -1,   -1,   -1,   -1,  268,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  125,   -1,
   -1,  257,  258,  259,  260,  261,  262,  263,  264,   -1,
  266,  267,  268,  269,  270,  271,  257,  258,  259,  260,
  261,  262,  263,  264,   -1,  266,  267,  268,  269,  270,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,  258,  259,
  260,  261,  262,  263,  264,   -1,  266,  267,  268,  269,
  270,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,  258,  259,  260,  261,  262,  263,  264,
   -1,  266,  267,  268,  269,  270,  271,  257,  258,  259,
  260,  261,  262,  263,  264,   -1,  266,  267,  268,  269,
  270,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,  258,  259,  260,  261,
  262,  263,  264,   -1,  266,  267,  268,  269,  270,  271,
  257,  258,  259,  260,  261,  262,  263,  264,  125,  266,
  267,  268,  269,  270,  271,   37,   -1,   -1,   40,   -1,
   42,   43,   -1,   45,   -1,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   37,   -1,   60,   61,
   62,   42,   43,   -1,   45,   -1,   47,   -1,   37,   -1,
   -1,   -1,   -1,   42,   43,   -1,   45,   -1,   47,   60,
   61,   62,   37,   -1,   -1,   -1,   -1,   42,   43,   91,
   45,   60,   47,   62,   37,   -1,   -1,   -1,   41,   42,
   43,   -1,   45,   -1,   47,   60,   37,   62,   -1,  125,
   41,   42,   43,   -1,   45,   -1,   47,   60,   37,   62,
   -1,   -1,   -1,   42,   43,   44,   45,   -1,   47,   60,
   37,   62,   -1,   -1,   -1,   42,   43,   -1,   45,   -1,
   47,   60,    6,   62,   -1,    9,   10,   -1,   -1,   -1,
   -1,   -1,   -1,   60,  123,   62,   -1,   -1,   -1,   -1,
  257,  258,  259,  260,  261,  262,  263,  264,  123,  266,
  267,  268,  269,  270,  271,   -1,   -1,   41,   42,   43,
   44,   -1,   -1,   -1,   48,   49,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,
   -1,   65,   66,   -1,   68,   -1,   70,   71,   -1,   -1,
   74,   -1,   -1,   77,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   86,   87,   88,   89,   90,   91,   92,   93,
   94,   95,   96,   97,   98,   99,  100,  101,  102,  103,
  104,  257,  258,  259,  260,  261,  262,  263,  264,  125,
  266,  267,  268,  269,  270,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  125,   -1,   -1,   -1,   -1,   -1,
  272,  273,  274,  275,  276,  277,  278,  279,  280,  281,
  282,  283,  284,   -1,   -1,  149,   -1,   -1,   -1,   -1,
  154,  272,  273,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  272,  273,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,   -1,   -1,  272,  273,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  272,
  273,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  272,  273,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  272,  273,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  272,  273,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,   -1,   -1,   -1,
   -1,  257,  258,  259,  260,  261,  262,  263,  264,   -1,
  266,  267,  268,  269,  270,  271,  257,  258,  259,  260,
  261,  262,  263,  264,   -1,  266,  267,  268,  269,  270,
  271,
};
#define YYFINAL 1
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 285
#if YYDEBUG
static const char *yyname[] = {

"end-of-file",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,"'%'",0,0,"'('","')'","'*'","'+'","','","'-'",0,"'/'",0,0,0,0,0,0,0,0,0,0,
"':'",0,"'<'","'='","'>'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,"'['",0,"']'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"'{'",
0,"'}'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"NUM_I","NUM_F","GVAR","LVAR","WHILE","BREAK",
"CONTINUE","IF","ELSE","GO","RET","FNNUM","FNDEF","INCLUDE","HALT","OROR",
"ANDAND","OR","XOR","AND","EQ","NEQ","LEQ","GEQ","LSHIFT","RSHIFT","IDIV",
"HASH","NEG",
};
static const char *yyrule[] = {
"$accept : input",
"input :",
"input : input stat",
"lvarslist :",
"lvarslist : LVAR",
"lvarslist : lvarslist ',' LVAR",
"mem_offset : expr",
"mem_offset : expr ',' expr",
"stat_math_op : '-'",
"stat_math_op : '+'",
"stat_math_op : '*'",
"stat_math_op : IDIV",
"stat_math_op : '/'",
"stat_math_op : '%'",
"stat_math_op : AND",
"stat_math_op : OR",
"stat_math_op : XOR",
"stat_math_op : ANDAND",
"stat_math_op : OROR",
"stat_math_op : EQ",
"stat_math_op : NEQ",
"stat_math_op : '<'",
"stat_math_op : LEQ",
"stat_math_op : '>'",
"stat_math_op : GEQ",
"stat_math_op : LSHIFT",
"stat_math_op : RSHIFT",
"statlist :",
"statlist : statlist stat",
"stat : HALT",
"stat : GVAR ':'",
"stat : GO expr",
"stat : GVAR '=' expr",
"stat : LVAR '=' expr",
"stat : prop_expr '=' expr",
"stat : prop_expr stat_math_op expr",
"stat : mem_expr '=' expr",
"stat : mem_expr stat_math_op expr",
"stat : GVAR stat_math_op expr",
"stat : LVAR stat_math_op expr",
"stat : FNNUM '(' exprlist ')'",
"stat : fn_expr",
"stat : RET",
"stat : RET '(' expr ')'",
"stat : IF expr '{' statlist '}'",
"stat : IF expr '{' statlist '}' ELSE '{' statlist '}'",
"$$1 :",
"stat : WHILE expr '{' $$1 statlist '}'",
"stat : BREAK",
"stat : CONTINUE",
"$$2 :",
"$$3 :",
"stat : FNDEF $$2 GVAR '(' lvarslist ')' $$3 '{' statlist '}'",
"stat : INCLUDE NUM_I",
"exprlist :",
"exprlist : expr",
"exprlist : exprlist ',' expr",
"basic_expr : NUM_I",
"basic_expr : NUM_F",
"basic_expr : GVAR",
"basic_expr : LVAR",
"basic_expr : FNNUM '(' exprlist ')'",
"fn_expr : basic_expr '(' exprlist ')'",
"fn_expr : fn_expr '(' exprlist ')'",
"fn_expr : mem_expr '(' exprlist ')'",
"fn_expr : prop_expr '(' exprlist ')'",
"mem_expr : basic_expr '[' mem_offset ']'",
"mem_expr : fn_expr '[' mem_offset ']'",
"mem_expr : mem_expr '[' mem_offset ']'",
"mem_expr : prop_expr '[' mem_offset ']'",
"prop_expr : basic_expr HASH",
"prop_expr : fn_expr HASH",
"prop_expr : mem_expr HASH",
"prop_expr : prop_expr HASH",
"expr : basic_expr",
"expr : fn_expr",
"expr : mem_expr",
"expr : prop_expr",
"$$4 :",
"expr : '{' $$4 statlist '}'",
"expr : '(' expr ')'",
"expr : expr '-' expr",
"expr : expr '+' expr",
"expr : expr '*' expr",
"expr : expr IDIV expr",
"expr : expr '/' expr",
"expr : expr '%' expr",
"expr : expr AND expr",
"expr : expr OR expr",
"expr : expr XOR expr",
"expr : expr ANDAND expr",
"expr : expr OROR expr",
"expr : expr EQ expr",
"expr : expr NEQ expr",
"expr : expr '<' expr",
"expr : expr LEQ expr",
"expr : expr '>' expr",
"expr : expr GEQ expr",
"expr : expr LSHIFT expr",
"expr : expr RSHIFT expr",
"expr : '-' expr",

};
#endif
#if YYDEBUG
#include <stdio.h>
#endif

/* define the initial stack-sizes */
#ifdef YYSTACKSIZE
#undef YYMAXDEPTH
#define YYMAXDEPTH  YYSTACKSIZE
#else
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 500
#define YYMAXDEPTH  500
#endif
#endif

#define YYINITSTACKSIZE 500

int      yydebug;
int      yynerrs;
int      yyerrflag;
int      yychar;
short   *yyssp;
YYSTYPE *yyvsp;
YYSTYPE  yyval;
YYSTYPE  yylval;

/* variables for the parser stack */
static short   *yyss;
static short   *yysslim;
static YYSTYPE *yyvs;
static unsigned yystacksize;
#line 739 "pixilang_compiler.y"

//Compilation error message:
void yyerror( char const* str )
{
    SHOW_ERROR( "%s", str );
}

size_t handle_control_characters( UTF8_CHAR* str, size_t size )
{
    size_t r = 0;
    size_t w = 0;
    for( ; r < size; r++, w++ )
    {
	int c = str[ r ];
	if( c == '\\' && r < size - 1 )
	{
	    r++;
	    int next_c = str[ r ];
	    switch( next_c )
	    {
		case '0': str[ w ] = 0x00; break;
		case 'a': str[ w ] = 0x07; break; //alert
		case 'b': str[ w ] = 0x08; break; //backspace
		case 't': str[ w ] = 0x09; break; //tab
		case 'r': str[ w ] = 0x0D; break; //carriage-return
		case 'n': str[ w ] = 0x0A; break; //newline
		case 'v': str[ w ] = 0x0B; break; //vertical-tab
		case 'f': str[ w ] = 0x0C; break; //form-feed
		default: str[ w ] = (UTF8_CHAR)next_c; break;
	    }
	}
	else
	{
	    str[ w ] = str[ r ];
	}
    }
    return w;
}

void resize_var_flags( void )
{
    if( g_comp->vm->vars_num > g_comp->var_flags_size )
    {
        size_t new_size = g_comp->vm->vars_num + 64;
	g_comp->var_flags = (char*)mem_resize( g_comp->var_flags, new_size );
        g_comp->var_flags_size = new_size;
    }
    g_comp->var_flags[ g_comp->vm->vars_num - 1 ] = 0;
}

//Parser. yylex() get next token from the source:
int yylex( void )
{
    UTF8_CHAR* src = g_comp->src;

    int c = -1;
    
    int str_symbol = 0;
    int str_start = -1;
    bool str_first_char_is_digit = 0;
    
    while( g_comp->src_ptr < g_comp->src_size + 2 )
    {
	if( g_comp->src_ptr >= g_comp->src_size )
	{
	    if( g_comp->src_ptr == g_comp->src_size )
	    {
		c = ' '; //Last empty character of the file (we need this to finish some strings)
	    }
	    else
	    {
		c = -1; //EOF
		if( g_comp->inc_num > 0 )
		{
		    DPRINT( "Return from include\n" );
		    //Remove current state:
		    mem_free( g_comp->src );
		    mem_free( g_comp->src_name );
		    mem_free( g_comp->base_path );
		    //Restore previous compiler state:
		    g_comp->inc_num--;
		    g_comp->src = g_comp->inc[ g_comp->inc_num ].src;
		    src = g_comp->src;
		    g_comp->src_ptr = g_comp->inc[ g_comp->inc_num ].src_ptr;
		    g_comp->src_line = g_comp->inc[ g_comp->inc_num ].src_line;
		    g_comp->src_size = g_comp->inc[ g_comp->inc_num ].src_size;
		    g_comp->src_name = g_comp->inc[ g_comp->inc_num ].src_name;
		    g_comp->base_path = g_comp->inc[ g_comp->inc_num ].base_path;
		    //Reset parser state:
		    str_symbol = 0;
		    str_start = -1;
		    str_first_char_is_digit = 0;
		    continue;
		}
	    }
	}
	else
	{
	    c = src[ g_comp->src_ptr ];
	}
	g_comp->src_ptr++;
	
	//Ignore text strings (quoted):
        if( c == '"' || c == '\'' )
	{
	    if( !str_symbol )
            {
                //String start:
                str_start = g_comp->src_ptr;
                str_symbol = c;
                continue;
            }
            else
            {
                if( str_symbol == c )
                {
                    //End of string:
                    int str_size = g_comp->src_ptr - str_start - 1;
		    if( c == '"' )
                    {
                        //Return number of text container:
			UTF8_CHAR* data;
			if( str_size > 0 )
			{
			    data = (UTF8_CHAR*)MEM_NEW2( str_size );
			    mem_copy( data, src + str_start, str_size );
			    str_size = handle_control_characters( data, str_size );
			}
			else 
			{
			    data = (UTF8_CHAR*)MEM_NEW2( 1 );
			    data[ 0 ] = 0;
			    str_size = 1;
			}
			yylval.i = pix_vm_new_container( -1, str_size, 1, PIX_CONTAINER_TYPE_INT8, data, 0, g_comp->vm );
			DPRINT( "NEW TEXT CONTAINER %d (%d bytes)\n", (int)yylval.i, str_size );
		    }
		    else
		    {
			//Return string converted to integer:
			UTF8_CHAR ts[ 4 ];
			ts[ 0 ] = src[ str_start ];
			ts[ 1 ] = src[ str_start + 1 ];
			ts[ 2 ] = src[ str_start + 2 ];
			ts[ 3 ] = src[ str_start + 3 ];
			str_size = handle_control_characters( ts, str_size );
			if( str_size == 1 ) 
			    yylval.i = (uchar)ts[ 0 ];
                        if( str_size == 2 ) 
			    yylval.i = (uchar)ts[ 0 ] 
				+ ( (uchar)ts[ 1 ] << 8 );
                        if( str_size == 3 ) 
			    yylval.i = (uchar)ts[ 0 ]
				+ ( (uchar)ts[ 1 ] << 8 )
				+ ( (uchar)ts[ 2 ] << 16 );
                        if( str_size >= 4 ) 
			    yylval.i = (uchar)ts[ 0 ] 
				+ ( (uchar)ts[ 1 ] << 8 )
				+ ( (uchar)ts[ 2 ] << 16 )
				+ ( (uchar)ts[ 3 ] << 24 );
		    }
		    c = NUM_I;
		    break;
		}
		else
		{
		    continue;
		}
	    }
	}
	else 
	{
	    if( str_symbol )
	    {
		//String reading.
		if( c == '\\' )
		{
		    //Skip control character:
		    g_comp->src_ptr++;
		}
		continue;
	    }
	}
	
	//No. It's not a quoted string:
	bool numeric = NUMERIC( c );
	if( numeric || ABC( c ) || c == '_' || c == '#' || c == '.' || c == '$' )
        {
	    if( str_start == -1 )
	    {
		//String begin:
		if( numeric ) str_first_char_is_digit = 1;
		str_start = g_comp->src_ptr - 1;
	    }
	    else
	    {
		if( c == '.' )
		{
		    if( str_first_char_is_digit == 0 )
		    {
			goto string_end;
		    }
		}
	    }
	    continue;
	}
	else
	{
string_end:
	    if( str_start >= 0 )
	    {
		int str_size = g_comp->src_ptr - str_start - 1;
                if( src[ str_start ] == '#' || NUMERIC( src[ str_start ] ) || ( str_size > 1 && src[ str_start ] == '.' && NUMERIC( src[ str_start + 1 ] ) ) )
		{
		    //Number:
		    c = NUM_I;
                    int ni;
                    if( src[ str_start ] == '#' )
                    {
                        //HEX COLOR:
                        yylval.i = 0;
                        for( ni = str_start + 1; ni < str_start + str_size; ni++ )
                        {
                            yylval.i <<= 4;
                            if( src[ ni ] < 58 ) yylval.i += src[ ni ] - '0';
                            else if( src[ ni ] > 64 && src[ ni ] < 91 ) yylval.i += src[ ni ] - 'A' + 10;
                            else yylval.i += src[ ni ] - 'a' + 10;
                        }
                        yylval.i = get_color( 
			    ( yylval.i >> 16 ) & 255,
                            ( yylval.i >> 8  ) & 255,
                            ( yylval.i       ) & 255 );
                    }
                    else
                    {
			if( str_size > 2 && src[ str_start + 1 ] == 'x' )
			{
			    //HEX:
			    yylval.i = 0;
			    for( ni = str_start + 2; ni < str_start + str_size; ni++ )
			    {
				yylval.i <<= 4;
				if( src[ ni ] < 58 ) yylval.i += src[ ni ] - '0';
				else if( src[ ni ] > 64 && src[ ni ] < 91 ) yylval.i += src[ ni ] - 'A' + 10;
				else yylval.i += src[ ni ] - 'a' + 10;
			    }
			}
			else if( str_size > 2 && src[ str_start + 1 ] == 'b' )
			{
			    //BIN:
		            yylval.i = 0;
			    for( ni = str_start + 2; ni < str_start + str_size; ni++ )
			    {
				yylval.i <<= 1;
				yylval.i += src[ ni ] - '0';
			    }
			}
			else
			{
			    bool float_num = 0;
			    for( ni = str_start; ni < str_start + str_size; ni++ )
				if( src[ ni ] == '.' ) { float_num = 1; break; }
			    if( float_num )
			    {
				//FLOATING POINT:
				if( str_size > 256 )
				    str_size = 256;
				mem_copy( g_comp->temp_sym_name, &src[ str_start ], str_size );
				g_comp->temp_sym_name[ str_size ] = 0;
				c = NUM_F;
				yylval.f = atof( g_comp->temp_sym_name );
			    }
			    else 
			    {
				//DEC:
				yylval.i = 0;
				for( ni = str_start; ni < str_start + str_size; ni++ )
				{
				    yylval.i *= 10;
				    yylval.i += src[ ni ] - '0';
				}
			    }
			}
                    }
		}
		else
		{
		    //Name of some symbol:
		    if( str_size == 1 && src[ str_start ] > 0 && src[ str_start ] < 127 )
		    {
			//Standard variable with one char ASCII name:
			c = GVAR;
			yylval.i = src[ str_start ];
			g_comp->var_flags[ yylval.i ] |= VAR_FLAG_USED;
		    }
		    else 
		    {
			//Variable with long name:
			if( str_size > 256 )
			    str_size = 256;
			mem_copy( g_comp->temp_sym_name, &src[ str_start ], str_size );
			g_comp->temp_sym_name[ str_size ] = 0;
			if( src[ str_start ] == '.' )
			{
			    //Container property:
			    c = HASH;
			    //Create global variable with the name and hash of this property:
			    bool created;
                            pix_sym* sym = pix_symtab_lookup( g_comp->temp_sym_name, -1, 1, SYMTYPE_GVAR, 0, 0, &created, &g_comp->sym );
                            if( sym )
                            {
                        	if( created )
                        	{
                            	    //Create new variable:
                            	    sym->val.i = g_comp->vm->vars_num;
                            	    DPRINT( "New global variable: %s (%d)\n", g_comp->temp_sym_name, (int)sym->val.i );
                            	    g_comp->vm->vars_num++;
                            	    pix_vm_resize_variables( g_comp->vm );
                            	    resize_var_flags();
                            	    g_comp->var_flags[ sym->val.i ] |= VAR_FLAG_USED | VAR_FLAG_INITIALIZED;
                            	    //Save variable name:
                            	    UTF8_CHAR* var_name = (UTF8_CHAR*)MEM_NEW2( str_size + 1 );
                            	    if( var_name )
                            	    {
                                	g_comp->vm->var_names[ sym->val.i ] = var_name;
                                	mem_copy( var_name, g_comp->temp_sym_name, str_size + 1 );
                            	    }
                        	}
                        	yylval.i = sym->val.i;
                        	g_comp->vm->vars[ yylval.i ].i = pix_symtab_hash( (const UTF8_CHAR*)( g_comp->temp_sym_name + 1 ), PIX_CONTAINER_SYMTAB_SIZE );
			    }
			    else
			    {
				SHOW_ERROR( "can't create a new symbol for property" );
            			return -1;
			    }
			}
			else
			if( src[ str_start ] == '$' )
			{
			    //Local variable:
			    c = LVAR;
			    if( str_size == 1 )
			    {
				yylval.i = 0;
			    }
			    else 
			    {
				if( NUMERIC( src[ str_start + 1 ] ) )
				{
				    yylval.i = string_to_int( g_comp->temp_sym_name + 1 );
				}
				else 
				{
				    //Named local variable:
				    pix_symtab* lsym = g_comp->lsym[ g_comp->lsym_num ].lsym;
				    if( lsym == 0 )
				    {
					lsym = (pix_symtab*)MEM_NEW2( sizeof( pix_symtab ) );
					g_comp->lsym[ g_comp->lsym_num ].lsym = lsym;
					pix_symtab_init( PIX_COMPILER_SYMTAB_SIZE, lsym );
				    }
				    bool created;
				    pix_sym* sym = pix_symtab_lookup( g_comp->temp_sym_name + 1, -1, 1, SYMTYPE_LVAR, 0, 0, &created, lsym );
				    if( sym )
				    {
					if( created )
					{
					    if( g_comp->fn_pars_mode == 0 )
					    {
						sym->val.i = -3 - g_comp->lsym[ g_comp->lsym_num ].lvars_num;
						g_comp->lsym[ g_comp->lsym_num ].lvars_num++;
						DPRINT( "New local variable: %s (%d)\n", g_comp->temp_sym_name, (int)sym->val.i );
					    }
					    else 
					    {
						sym->val.i = 1 + g_comp->lsym[ g_comp->lsym_num ].pars_num;
						g_comp->lsym[ g_comp->lsym_num ].pars_num++;
						DPRINT( "New local variable (fn parameter): %s (%d)\n", g_comp->temp_sym_name, (int)sym->val.i );
					    }
				        }
					yylval.i = sym->val.i;
				    }
				    else
                        	    {
                            		SHOW_ERROR( "can't create a new symbol for local variable" );
                            		return -1;
                        	    }
				}
			    }
			}
			else
			{
			    //Global variable
			    bool created;
			    pix_sym* sym = pix_symtab_lookup( g_comp->temp_sym_name, -1, 1, SYMTYPE_GVAR, 0, 0, &created, &g_comp->sym );
			    if( sym == 0 )
                            {
                                SHOW_ERROR( "can't create a new symbol for global variable" );
                                return -1;
                            }
			    if( created )
			    {
				//Create new variable:
				sym->val.i = g_comp->vm->vars_num;
				yylval.i = sym->val.i;
				c = GVAR;
				DPRINT( "New global variable: %s (%d)\n", g_comp->temp_sym_name, (int)sym->val.i );
				g_comp->vm->vars_num++;
				pix_vm_resize_variables( g_comp->vm );
				resize_var_flags();
				g_comp->var_flags[ sym->val.i ] |= VAR_FLAG_USED;
				//Save variable name:
				UTF8_CHAR* var_name = (UTF8_CHAR*)MEM_NEW2( str_size + 1 );
				if( var_name )
				{
				    g_comp->vm->var_names[ sym->val.i ] = var_name;
                            	    mem_copy( var_name, g_comp->temp_sym_name, str_size + 1 );
                            	}
			    }
			    else
			    {
				switch( sym->type )
				{
				    case SYMTYPE_GVAR:
					yylval.i = sym->val.i;
					c = GVAR;
					break;
				    case SYMTYPE_LVAR:
					yylval.i = sym->val.i;
					c = LVAR;
					break;
				    case SYMTYPE_NUM_I:
					yylval.i = sym->val.i;
					c = NUM_I;
					break;
				    case SYMTYPE_NUM_F:
					yylval.f = sym->val.f;
					c = NUM_F;
					break;
				    case SYMTYPE_WHILE:
					c = WHILE;
					break;
				    case SYMTYPE_BREAK:
					yylval.i = 1;
					c = BREAK;
					break;
				    case SYMTYPE_BREAK2:
					yylval.i = 2;
					c = BREAK;
					break;
				    case SYMTYPE_BREAK3:
					yylval.i = 3;
					c = BREAK;
					break;
				    case SYMTYPE_BREAK4:
					yylval.i = 4;
					c = BREAK;
					break;
				    case SYMTYPE_BREAKALL:
					yylval.i = -1;
					c = BREAK;
					break;
				    case SYMTYPE_CONTINUE:
					c = CONTINUE;
					break;
				    case SYMTYPE_IF:
					c = IF;
					break;
				    case SYMTYPE_ELSE:
					c = ELSE;
					break;
				    case SYMTYPE_GO:
					c = GO;
					break;
				    case SYMTYPE_RET:
					c = RET;
					break;
				    case SYMTYPE_IDIV:
					c = IDIV;
					break;
				    case SYMTYPE_FNNUM:
					c = FNNUM;
					yylval.i = sym->val.i;
					break;
				    case SYMTYPE_FNDEF:
					c = FNDEF;
					break;
				    case SYMTYPE_INCLUDE:
					c = INCLUDE;
					break;
				    case SYMTYPE_HALT:
					c = HALT;
					break;
				}
			    }
			}
		    }
		}
		g_comp->src_ptr--;
		break;
	    }
        }
	
	//Parse other symbols:
        if( c == '/' )
        {
            if( g_comp->src_ptr < g_comp->src_size && src[ g_comp->src_ptr ] == '/' )
            { 
		//COMMENTS:
                for(;;)
                {
                    g_comp->src_ptr++;
                    if( g_comp->src_ptr >= g_comp->src_size ) break;
                    if( src[ g_comp->src_ptr ] == 0xD || src[ g_comp->src_ptr ] == 0xA ) break;
                }
                continue;
            }
            else
            if( g_comp->src_ptr < g_comp->src_size && src[ g_comp->src_ptr ] == '*' )
            {
                //COMMENTS 2:
                for(;;)
                {
                    g_comp->src_ptr++;
                    if( g_comp->src_ptr >= g_comp->src_size ) break;
                    if( src[ g_comp->src_ptr ] == 0xA ) g_comp->src_line++;
                    if( g_comp->src_ptr + 1 < g_comp->src_size && 
		        src[ g_comp->src_ptr ] == '*' && 
			src[ g_comp->src_ptr + 1 ] == '/' ) 
		    { 
			g_comp->src_ptr += 2; break; 
		    }
                }
                continue;
            }
        }
	
	bool need_to_break = 0;
        switch( c )
        {
	    case 0xA:
		//New line:
		g_comp->src_line++;
		break;
        
            case '-':
            case '+':
            case '*':
            case '/':
            case '%':
            case ':':
            case '(':
            case ')':
            case ',':
            case '[':
            case ']':
            case '{':
            case '}':
	    case '~':
                need_to_break = 1;
                break;

	    case '^':
		c = XOR;
		need_to_break = 1;
		break;
	    case '&':
		if( src[ g_comp->src_ptr ] == '&' )
		{
		    c = ANDAND;
		    g_comp->src_ptr++;
		}
		else
		{
		    c = AND;
		}
		need_to_break = 1;
		break;
	    case '|':
		if( src[ g_comp->src_ptr ] == '|' )
		{
		    c = OROR;
		    g_comp->src_ptr++;
		}
		else
		{
		    c = OR;
		}
		need_to_break = 1;
		break;
	    
	    case '<':
                if( src[ g_comp->src_ptr ] == '=' )
                {
                    c = LEQ;
                    g_comp->src_ptr++;
                }
		else
		if( src[ g_comp->src_ptr ] == '<' )
		{
		    c = LSHIFT;
		    g_comp->src_ptr++;
		}
                need_to_break = 1;
                break;
            case '>':
                if( src[ g_comp->src_ptr ] == '=' )
                {
                    c = GEQ;
		    g_comp->src_ptr++;
                }
		else
		if( src[ g_comp->src_ptr ] == '>' )
		{
		    c = RSHIFT;
		    g_comp->src_ptr++;
		}
                need_to_break = 1;
                break;
            case '!':
                if( src[ g_comp->src_ptr ] == '=' )
                {
                    c = NEQ;
                    g_comp->src_ptr++;
                    need_to_break = 1;
                }
                break;
	    case '=':
		if( src[ g_comp->src_ptr ] == '=' )
		{
		    c = EQ;
		    g_comp->src_ptr++;
		}
		need_to_break = 1;
		break;
	}
	if( need_to_break ) break;
    }
    
    return c;
}

void push_int( PIX_INT v )
{
    if( v < ( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) &&
	v > -( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) )
    {
	DPRINT( "%d: PUSH_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)v );
	pix_vm_put_opcode( OPCODE_PUSH_i | ( (PIX_OPCODE)v << PIX_OPCODE_BITS ), g_comp->vm );
    }
    else
    {
	DPRINT( "%d: PUSH_I %d\n", (int)g_comp->vm->code_ptr, (int)v );
	pix_vm_put_opcode( OPCODE_PUSH_I, g_comp->vm );
	pix_vm_put_int( v, g_comp->vm );
    }
}

void push_jmp( size_t dest_ptr )
{
    size_t code_ptr = g_comp->vm->code_ptr;
    PIX_INT offset = (PIX_INT)dest_ptr - (PIX_INT)code_ptr;
    if( offset < ( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) &&
        offset > -( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) )
    {
        DPRINT( "%d: JMP_i ( %d << OB )\n", (int)code_ptr, (int)offset );
        pix_vm_put_opcode( OPCODE_JMP_i | ( (PIX_OPCODE)offset << PIX_OPCODE_BITS ), g_comp->vm );
    }
    else 
    {
        DPRINT( "%d: JMP_I %d\n", (int)code_ptr, (int)offset );	
        pix_vm_put_opcode( OPCODE_JMP_I, g_comp->vm );
        pix_vm_put_int( offset, g_comp->vm );
    }
    for( int i = 0; i < g_comp->statlist_header_size - ( g_comp->vm->code_ptr - code_ptr ); i++ )
    {
        DPRINT( "%d: NOP\n", (int)g_comp->vm->code_ptr, (int)offset );
        pix_vm_put_opcode( OPCODE_NOP, g_comp->vm );
    }
}

//Add lexical node:
lnode* node( lnode_type type, int nn )
{
    lnode* n = (lnode*)MEM_NEW2( sizeof( lnode ) );
    if( nn > 0 )
    {
        n->n = (lnode**)MEM_NEW2( sizeof( lnode* ) * nn );
    }
    else
    {
	n->n = 0;
    }
    n->type = type;
    n->nn = nn;
    n->flags = 0;
    n->code_ptr = 0;
    n->code_ptr2 = 0;
    return n;
}

void resize_node( lnode* n, int nn )
{
    if( n == 0 ) return;

    if( n->n == 0 )
    {
	n->n = (lnode**)MEM_NEW2( sizeof( lnode* ) * nn );
    }
    else
    {
	if( nn > mem_get_size( n->n ) / sizeof( lnode* ) )
	{
	    n->n = (lnode**)mem_resize( n->n, sizeof( lnode* ) * ( nn + 8 ) );
	}
    }
    n->nn = nn;
}

lnode* clone_tree( lnode* n )
{
    if( n == 0 ) return 0;
    
    lnode* new_n = node( n->type, n->nn );
    
    new_n->flags = n->flags;
    new_n->val = n->val;
    
    for( int i = 0; i < n->nn; i++ )
    {
	new_n->n[ i ] = clone_tree( n->n[ i ] );
    }
    
    return new_n;
}

void remove_tree( lnode* n )
{
    if( n == 0 ) return;
    
    if( n->n )
    {
	for( int i = 0; i < n->nn; i++ )
	{
	    remove_tree( n->n[ i ] );
	}
	mem_free( n->n );
    }
    mem_free( n );
}

void optimize_tree( lnode* n )
{
    if( n == 0 ) return;
    
    if( n->n )
    {
	for( int i = 0; i < n->nn; i++ )
	{
	    optimize_tree( n->n[ i ] );
	}
    }
    
    switch( n->type )
    {
	case lnode_if:
	    {
		int v = 0;
		if( n->n[ 0 ]->type == lnode_int ) 
		{
		    if( n->n[ 0 ]->val.i == 0 )
			v = 1;
		    else
			v = 2;
		}
		else if( n->n[ 0 ]->type == lnode_float )
		{
		    if( n->n[ 0 ]->val.f == 0 )
			v = 1;
		    else
			v = 2;
		}
		if( v )
		{
		    if( v == 1 )
		    {
			//IF 0 { CODE } - IGNORE THIS CODE:
			remove_tree( n->n[ 0 ] );
			remove_tree( n->n[ 1 ] );
			n->nn = 0;
		    }
		    else
		    {
			//IF 1 { CODE } - ALWAYS TRUE:
			remove_tree( n->n[ 0 ] );
			n->n[ 0 ] = 0;
			n->n[ 1 ]->flags = 0; //Just a statlist without any headers
		    }
		}
	    }
	    break;
	case lnode_if_else:
	    {
		int v = 0;
		if( n->n[ 0 ]->type == lnode_int ) 
		{
		    if( n->n[ 0 ]->val.i == 0 )
			v = 1;
		    else
			v = 2;
		}
		else if( n->n[ 0 ]->type == lnode_float )
		{
		    if( n->n[ 0 ]->val.f == 0 )
			v = 1;
		    else
			v = 2;
		}
		if( v )
		{
		    if( v == 1 )
		    {
			//IF 0 { CODE1 } ELSE { CODE2 } - IGNORE CODE1; ALWAYS EXECUTE CODE2:
			remove_tree( n->n[ 0 ] );
			remove_tree( n->n[ 1 ] );
			n->n[ 0 ] = 0;
			n->n[ 1 ] = 0;
			n->n[ 2 ]->flags = 0; //Just a statlist without any headers
		    }
		    else
		    {
			//IF 1 { CODE1 } ELSE { CODE2 } - IGNORE CODE2; ALWAYS EXECUTE CODE1:
			remove_tree( n->n[ 0 ] );
			remove_tree( n->n[ 2 ] );
			n->n[ 0 ] = 0;
			n->n[ 1 ]->flags = 0; //Just a statlist without any headers
			n->n[ 2 ] = 0;
		    }
		}
	    }
	    break;
	case lnode_while:
	    {
		int v = 0;
		if( n->n[ 0 ]->type == lnode_int ) 
		{
		    if( n->n[ 0 ]->val.i == 0 )
			v = 1;
		    else
			v = 2;
		}
		else if( n->n[ 0 ]->type == lnode_float )
		{
		    if( n->n[ 0 ]->val.f == 0 )
			v = 1;
		    else
			v = 2;
		}
		if( v )
		{
		    if( v == 1 )
		    {
			//WHILE 0 { CODE1 } - IGNORE THIS CODE:
			remove_tree( n->n[ 0 ] );
			remove_tree( n->n[ 1 ] );
			remove_tree( n->n[ 2 ] );
			n->nn = 0;
		    }
		    else
		    {
			//WHILE 1 { CODE1 } - INFINITE LOOP:
			remove_tree( n->n[ 0 ] );
			n->n[ 0 ] = 0;
			n->n[ 1 ]->flags = 0; //Just a statlist without any headers
			n->n[ 2 ]->val.p = n->n[ 1 ]; //jmp to CODE1
		    }
		}
	    }
	    break;
	    
	case lnode_sub:
	case lnode_add:
	case lnode_mul:
	case lnode_idiv:
	case lnode_div:
	case lnode_mod:
	case lnode_and:
	case lnode_or:
	case lnode_xor:
	case lnode_andand:
	case lnode_oror:
	case lnode_eq:
	case lnode_neq:
	case lnode_less:
	case lnode_leq:
	case lnode_greater:
	case lnode_geq:
	case lnode_lshift:
	case lnode_rshift:
	    if( ( n->n[ 0 ]->type == lnode_int || n->n[ 0 ]->type == lnode_float ) &&
		( n->n[ 1 ]->type == lnode_int || n->n[ 1 ]->type == lnode_float ) )
	    {
		bool integer_op = 0;
		if( n->n[ 0 ]->type == lnode_int && n->n[ 1 ]->type == lnode_int ) //Both operands are integer
		    integer_op = 1;
		switch( n->type )
		{
		    case lnode_idiv:
		    case lnode_mod:
		    case lnode_and:
		    case lnode_or:
		    case lnode_xor:
		    case lnode_lshift:
		    case lnode_rshift:
			integer_op = 1; //Allways integer
			break;
		    case lnode_div:
			integer_op = 0; //Allways float
			break;
		}
		if( integer_op )
		{
		    PIX_INT i1, i2, res;
		    if( n->n[ 0 ]->type == lnode_float )
			i1 = (PIX_INT)n->n[ 0 ]->val.f;
		    else
			i1 = n->n[ 0 ]->val.i;
		    if( n->n[ 1 ]->type == lnode_float )
			i2 = (PIX_INT)n->n[ 1 ]->val.f;
		    else
			i2 = n->n[ 1 ]->val.i;
		    switch( n->type )
		    {
			case lnode_sub: res = i1 - i2; break;
			case lnode_add: res = i1 + i2; break;
			case lnode_mul: res = i1 * i2; break;
			case lnode_idiv: res = i1 / i2; break;
			case lnode_mod: res = i1 % i2; break;
			case lnode_and: res = i1 & i2; break;
			case lnode_or: res = i1 | i2; break;
			case lnode_xor: res = i1 ^ i2; break;
			case lnode_andand: res = ( i1 && i2 ); break;
			case lnode_oror: res = ( i1 || i2 ); break;
			case lnode_eq: res = ( i1 == i2 ); break;
			case lnode_neq: res = !( i1 == i2 ); break;
			case lnode_less: res = ( i1 < i2 ); break;
			case lnode_leq: res = ( i1 <= i2 ); break;
			case lnode_greater: res = ( i1 > i2 ); break;
			case lnode_geq: res = ( i1 >= i2 ); break;
			case lnode_lshift: res = ( i1 << (unsigned)i2 ); break;
			case lnode_rshift: res = ( i1 >> (unsigned)i2 ); break;
		    }
		    n->type = lnode_int;
		    n->val.i = res;
		}
		else
		{
		    PIX_FLOAT f1, f2, res;
		    if( n->n[ 0 ]->type == lnode_float )
			f1 = n->n[ 0 ]->val.f;
		    else
			f1 = (PIX_FLOAT)n->n[ 0 ]->val.i;
		    if( n->n[ 1 ]->type == lnode_float )
			f2 = n->n[ 1 ]->val.f;
		    else
			f2 = (PIX_FLOAT)n->n[ 1 ]->val.i;
		    bool int_res = 0;
		    switch( n->type )
		    {
			case lnode_sub: res = f1 - f2; break;
			case lnode_add: res = f1 + f2; break;
			case lnode_mul: res = f1 * f2; break;
			case lnode_div: res = f1 / f2; break;
			case lnode_andand: res = ( f1 && f2 ); int_res = 1; break;
			case lnode_oror: res = ( f1 || f2 ); int_res = 1; break;
			case lnode_eq: res = ( f1 == f2 ); int_res = 1; break;
			case lnode_neq: res = !( f1 == f2 ); int_res = 1; break;
			case lnode_less: res = ( f1 < f2 ); int_res = 1; break;
			case lnode_leq: res = ( f1 <= f2 ); int_res = 1; break;
			case lnode_greater: res = ( f1 > f2 ); int_res = 1; break;
			case lnode_geq: res = ( f1 >= f2 ); int_res = 1; break;
		    }
		    if( int_res )
		    {
			n->type = lnode_int;
			n->val.i = (PIX_INT)res;
		    }
		    else
		    {
			n->type = lnode_float;
			n->val.f = res;
		    }
		}
		remove_tree( n->n[ 0 ] );
		remove_tree( n->n[ 1 ] );
		n->nn = 0;
	    }
	    break;
	
	case lnode_neg:
	    switch( n->n[ 0 ]->type )
	    {
		case lnode_int:
		    n->type = lnode_int;
		    n->val.i = -n->n[ 0 ]->val.i;
		    remove_tree( n->n[ 0 ] );
		    n->nn = 0;
		    break;
		case lnode_float:
		    n->type = lnode_float;
		    n->val.f = -n->n[ 0 ]->val.f;
		    remove_tree( n->n[ 0 ] );
		    n->nn = 0;
		    break;
	    }
	    break;
	
	case lnode_call_builtin_fn:
	case lnode_call_builtin_fn_void:
	case lnode_call:
	case lnode_call_void:
	    {
		lnode* pars = n->n[ 0 ];
	        //Flip parameters:
	        for( int i = 0; i < pars->nn / 2; i++ )
	        {
		    lnode* temp_node = pars->n[ i ];
		    pars->n[ i ] = pars->n[ pars->nn - 1 - i ];
		    pars->n[ pars->nn - 1 - i ] = temp_node;
	        }
	    }
	    break;
    }
}

void compile_tree( lnode* n )
{
    if( n == 0 ) return;

    n->code_ptr = g_comp->vm->code_ptr;
    
    switch( n->type )
    {
	case lnode_statlist:
	    if( ( n->flags & LNODE_FLAG_STATLIST_WITH_JMP_HEADER ) ||
	        ( n->flags & LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER ) )
	    {
		DPRINT( "%d: STATLIST_HEADER (%d NOPs)\n", (int)g_comp->vm->code_ptr, g_comp->statlist_header_size );
		for( int i = 0; i < g_comp->statlist_header_size; i++ )
		    pix_vm_put_opcode( OPCODE_NOP, g_comp->vm );
	    }
	    break;
    }
    
    if( n->n )
    {
	for( int i = 0; i < n->nn; i++ )
	{
	    compile_tree( n->n[ i ] );
	}
    }
    
    switch( n->type )
    {
	case lnode_statlist:
	    if( ( n->flags & LNODE_FLAG_STATLIST_WITH_JMP_HEADER ) ||
	        ( n->flags & LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER ) )
	    {
		//Add "skip statlist" code:
		PIX_INT offset = (PIX_INT)g_comp->vm->code_ptr - (PIX_INT)n->code_ptr;
		PIX_OPCODE header_opcode = OPCODE_JMP_i;
		const UTF8_CHAR* header_opcode_name = "?";
		if( n->flags & LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER ) header_opcode = OPCODE_JMP_IF_FALSE_i;
		if( n->flags & LNODE_FLAG_STATLIST_SKIP_NEXT_HEADER ) offset += g_comp->statlist_header_size;
		bool short_int;
		if( offset < ( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) &&
		    offset > -( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) )
		{
		    short_int = 1;
		}
		else 
		{
		    short_int = 0;
		    header_opcode++;
		}
		switch( header_opcode )
		{
		    case OPCODE_JMP_i: header_opcode_name = "JMP_i"; break;
		    case OPCODE_JMP_I: header_opcode_name = "JMP_I"; break;
		    case OPCODE_JMP_IF_FALSE_i: header_opcode_name = "JMP_IF_FALSE_i"; break;
		    case OPCODE_JMP_IF_FALSE_I: header_opcode_name = "JMP_IF_FALSE_I"; break;
		}
		if( short_int )
		{
		    DPRINT( "%d: %s ( %d << OB ) (skip statlist)\n", (int)n->code_ptr, header_opcode_name, (int)offset );
		    size_t prev_code_ptr = g_comp->vm->code_ptr;
		    g_comp->vm->code_ptr = n->code_ptr;
		    pix_vm_put_opcode( header_opcode | ( (PIX_OPCODE)offset << PIX_OPCODE_BITS ), g_comp->vm );
		    g_comp->vm->code_ptr = prev_code_ptr;
		}
		else
		{
		    DPRINT( "%d: %s %d (skip statlist)\n", (int)n->code_ptr, header_opcode_name, (int)offset );
		    size_t prev_code_ptr = g_comp->vm->code_ptr;
		    g_comp->vm->code_ptr = n->code_ptr;
		    pix_vm_put_opcode( header_opcode, g_comp->vm );
		    pix_vm_put_int( offset, g_comp->vm );
		    g_comp->vm->code_ptr = prev_code_ptr;
		}
		if( n->flags & LNODE_FLAG_STATLIST_AS_EXPRESSION )
		{
		    push_int( ( n->code_ptr + g_comp->statlist_header_size ) | PIX_INT_ADDRESS_MARKER );
		}
	    }
	    break;
	    
	case lnode_int:
	    push_int( n->val.i );
	    break;
	case lnode_float:
	    DPRINT( "%d: PUSH_F %d\n", (int)g_comp->vm->code_ptr, (int)n->val.f );
	    pix_vm_put_opcode( OPCODE_PUSH_F, g_comp->vm );
	    pix_vm_put_float( n->val.f, g_comp->vm );
	    break;
	case lnode_var:
	    DPRINT( "%d: PUSH_v %d\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_PUSH_v | ( n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
    
	case lnode_halt:
	    DPRINT( "%d: HALT\n", (int)g_comp->vm->code_ptr );
	    pix_vm_put_opcode( OPCODE_HALT, g_comp->vm );
	    break;
    
        case lnode_label:
	    DPRINT( "LABEL var:%d offset:%d\n", (int)n->val.i, (int)g_comp->vm->code_ptr );
	    g_comp->vm->vars[ n->val.i ].i = g_comp->vm->code_ptr | PIX_INT_ADDRESS_MARKER;
	    g_comp->vm->var_types[ n->val.i ] = 0;
	    break;
	case lnode_function_label_from_node:
	    DPRINT( "FUNCTION var:%d offset:%d\n", (int)n->val.i, (int)((lnode*)n->n[ 0 ]->val.p)->code_ptr + g_comp->statlist_header_size );
	    g_comp->vm->vars[ n->val.i ].i = ( ((lnode*)n->n[ 0 ]->val.p)->code_ptr + g_comp->statlist_header_size ) | PIX_INT_ADDRESS_MARKER;
	    g_comp->vm->var_types[ n->val.i ] = 0;
	    break;
	    
	case lnode_go:
	    DPRINT( "%d: GO\n", (int)g_comp->vm->code_ptr );
	    pix_vm_put_opcode( OPCODE_GO, g_comp->vm );
	    break;
	    
	case lnode_jmp_to_node:
	case lnode_jmp_to_end_of_node:
	    {
		size_t ptr;
		if( n->type == lnode_jmp_to_node )
		    ptr = ((lnode*)n->val.p)->code_ptr;
		else
		    ptr = ((lnode*)n->val.p)->code_ptr2;
		push_jmp( ptr );
		if( ptr == 0 )
		{
		    if( g_comp->fixup == 0 )
			g_comp->fixup = (lnode**)MEM_NEW2( sizeof( lnode* ) );
		    g_comp->fixup[ g_comp->fixup_num ] = n;
		    g_comp->fixup_num++;
		    if( g_comp->fixup_num >= mem_get_size( g_comp->fixup ) / sizeof( lnode* ) )
			g_comp->fixup = (lnode**)mem_resize( g_comp->fixup, mem_get_size( g_comp->fixup ) + 8 * sizeof( lnode* ) );
		}
	    }
	    break;

	case lnode_save_to_var:
	    DPRINT( "%d: SAVE_TO_VAR_v ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_SAVE_TO_VAR_v | ( n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;

	case lnode_save_to_prop:
	    DPRINT( "%d: SAVE_TO_PROP_I %d\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_SAVE_TO_PROP_I, g_comp->vm );
	    pix_vm_put_int( n->val.i, g_comp->vm );
	    break;
	case lnode_load_from_prop:
	    DPRINT( "%d: LOAD_FROM_PROP_I %d\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_LOAD_FROM_PROP_I, g_comp->vm );
	    pix_vm_put_int( n->val.i, g_comp->vm );
	    break;
	    
	case lnode_save_to_mem:
	    DPRINT( "%d: SAVE_TO_MEM\n", (int)g_comp->vm->code_ptr );
	    if( n->n[ 1 ]->type == lnode_empty && n->n[ 1 ]->nn == 2 )
		pix_vm_put_opcode( OPCODE_SAVE_TO_MEM_2D, g_comp->vm );
	    else
		pix_vm_put_opcode( OPCODE_SAVE_TO_MEM, g_comp->vm );
	    break;
	case lnode_load_from_mem:
	    DPRINT( "%d: LOAD_FROM_MEM\n", (int)g_comp->vm->code_ptr );
	    if( n->n[ 1 ]->type == lnode_empty && n->n[ 1 ]->nn == 2 )
		pix_vm_put_opcode( OPCODE_LOAD_FROM_MEM_2D, g_comp->vm );
	    else
		pix_vm_put_opcode( OPCODE_LOAD_FROM_MEM, g_comp->vm );
	    break;
	
	case lnode_save_to_stackframe:
	    DPRINT( "%d: SAVE_TO_STACKFRAME_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_SAVE_TO_STACKFRAME_i | ( (PIX_OPCODE)n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
	case lnode_load_from_stackframe:
	    DPRINT( "%d: LOAD_FROM_STACKFRAME_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_LOAD_FROM_STACKFRAME_i | ( (PIX_OPCODE)n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
    
	case lnode_sub:
	case lnode_add:
	case lnode_mul:
	case lnode_idiv:
	case lnode_div:
	case lnode_mod:
	case lnode_and:
	case lnode_or:
	case lnode_xor:
	case lnode_andand:
	case lnode_oror:
	case lnode_eq:
	case lnode_neq:
	case lnode_less:
	case lnode_leq:
	case lnode_greater:
	case lnode_geq:
	case lnode_lshift:
	case lnode_rshift:
	    DPRINT( "%d: MATH OP %d\n", (int)g_comp->vm->code_ptr, n->type - lnode_sub );
	    pix_vm_put_opcode( n->type - lnode_sub + OPCODE_SUB, g_comp->vm );
	    break;
    
	case lnode_neg:
	    DPRINT( "%d: NEG\n", (int)g_comp->vm->code_ptr );
	    pix_vm_put_opcode( OPCODE_NEG, g_comp->vm );
	    break;
	    
	case lnode_call_builtin_fn:
	    DPRINT( "%d: CALL_BUILTIN_FN ( %d << OB+FB ) ( %d << OB ) (%s)\n", (int)g_comp->vm->code_ptr, (int)n->n[ 0 ]->nn, (int)n->val.i, g_pix_fn_names[ n->val.i ] );
	    pix_vm_put_opcode( OPCODE_CALL_BUILTIN_FN | ( n->n[ 0 ]->nn << ( PIX_OPCODE_BITS + PIX_FN_BITS ) ) | ( n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
	case lnode_call_builtin_fn_void:
	    DPRINT( "%d: CALL_BUILTIN_FN_VOID ( %d << OB+FB ) ( %d << OB ) (%s)\n", (int)g_comp->vm->code_ptr, (int)n->n[ 0 ]->nn, (int)n->val.i, g_pix_fn_names[ n->val.i ] );
	    pix_vm_put_opcode( OPCODE_CALL_BUILTIN_FN_VOID | ( n->n[ 0 ]->nn << ( PIX_OPCODE_BITS + PIX_FN_BITS ) ) | ( n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
	case lnode_call:
	case lnode_call_void:
	    DPRINT( "%d: CALL_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->n[ 0 ]->nn );
	    pix_vm_put_opcode( OPCODE_CALL_i | ( n->n[ 0 ]->nn << PIX_OPCODE_BITS ), g_comp->vm );
	    if( n->type == lnode_call_void )
	    {
		DPRINT( "%d: INC_SP_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, 1 );
		pix_vm_put_opcode( OPCODE_INC_SP_i | ( 1 << PIX_OPCODE_BITS ), g_comp->vm );
	    }
	    break;
	case lnode_ret_int:
	    if( n->val.i < ( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) &&
		n->val.i > -( 1 << ( ( sizeof( PIX_OPCODE ) * 8 ) - ( PIX_OPCODE_BITS + 1 ) ) ) )
	    {
		DPRINT( "%d: RET_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
		pix_vm_put_opcode( OPCODE_RET_i | ( (PIX_OPCODE)n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    }
	    else
	    {
		DPRINT( "%d: RET_I %d\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
		pix_vm_put_opcode( OPCODE_RET_I, g_comp->vm );
		pix_vm_put_int( n->val.i, g_comp->vm );
	    }
	    break;
	case lnode_ret:
	    DPRINT( "%d: RET\n", (int)g_comp->vm->code_ptr );
	    pix_vm_put_opcode( OPCODE_RET, g_comp->vm );
	    break;
	case lnode_inc_sp:
	    DPRINT( "%d: INC_SP_i ( %d << OB )\n", (int)g_comp->vm->code_ptr, (int)n->val.i );
	    pix_vm_put_opcode( OPCODE_INC_SP_i | ( (PIX_OPCODE)n->val.i << PIX_OPCODE_BITS ), g_comp->vm );
	    break;
    }

    n->code_ptr2 = g_comp->vm->code_ptr;
}

void fix_up( void )
{
    if( g_comp->fixup && g_comp->fixup_num )
    {
	for( int i = 0; i < g_comp->fixup_num; i++ )
	{
	    lnode* n = g_comp->fixup[ i ];
	    size_t prev_code_ptr = g_comp->vm->code_ptr;
    	    g_comp->vm->code_ptr = n->code_ptr;
	    switch( n->type )
	    {
		case lnode_jmp_to_node:
            	    push_jmp( ((lnode*)n->val.p)->code_ptr );
		    break;
		case lnode_jmp_to_end_of_node:
            	    push_jmp( ((lnode*)n->val.p)->code_ptr2 );
		    break;
	    }
    	    g_comp->vm->code_ptr = prev_code_ptr;
	}
    }
}

#define ADD_SYMBOL( sname, stype, sval ) if( stype == SYMTYPE_NUM_F ) pix_symtab_lookup( sname, -1, 1, stype, 0, sval, 0, &g_comp->sym ); else pix_symtab_lookup( sname, -1, 1, stype, sval, 0, 0, &g_comp->sym );

int pix_compile_from_memory( UTF8_CHAR* src, int src_size, UTF8_CHAR* src_name, UTF8_CHAR* base_path, pix_vm* vm, pix_data* pd )
{
    int rv = 0;

    DPRINT( "Pixilang compiler started.\n" );

    ticks_t start_time = time_ticks();
    
    if( sundog_mutex_lock( &pd->compiler_mutex ) )
    {
	rv = 1;
	goto compiler_end2;
    }

    DPRINT( "Init...\n" );
    g_comp = (pix_compiler*)MEM_NEW2( sizeof( pix_compiler ) );
    if( g_comp == 0 ) 
    {
	rv = 2;
	goto compiler_end;
    }
    mem_zero( g_comp );
    g_comp->vm = vm;
    g_comp->pd = pd;
    g_comp->src = src;
    g_comp->src_size = src_size;
    g_comp->src_name = src_name;
    if( sizeof( PIX_INT ) >= sizeof( PIX_OPCODE ) )
        g_comp->statlist_header_size = 1 + sizeof( PIX_INT ) / sizeof( PIX_OPCODE );
    else
        g_comp->statlist_header_size = 2;
    //Global symbol table:
    if( pix_symtab_init( PIX_COMPILER_SYMTAB_SIZE, &g_comp->sym ) )
    {
	rv = 3;
	goto compiler_end;
    }
    //Local symbol tables:
    g_comp->lsym = (pix_lsymtab*)MEM_NEW2( 8 * sizeof( pix_lsymtab ) );
    if( g_comp->lsym == 0 )
    {
	rv = 4;
	goto compiler_end;
    }
    mem_zero( g_comp->lsym );
    //Set base path:
    g_comp->base_path = base_path;
    DPRINT( "Base path: %s\n", base_path );
    
    DPRINT( "VM init...\n" );
    vm->vars_num = 128; //standard set of variables with one-char ASCII names
    vm->vars_num += PIX_GVARS - vm->vars_num;
    pix_vm_resize_variables( vm );
    g_comp->var_flags = (char*)MEM_NEW2( vm->vars_num );
    g_comp->var_flags_size = vm->vars_num;
    mem_zero( g_comp->var_flags );
    mem_free( vm->base_path );
    vm->base_path = (UTF8_CHAR*)MEM_NEW2( mem_strlen( base_path ) + 1 );
    if( vm->base_path == 0 )
    {
	rv = 5;
	goto compiler_end;
    }    
    vm->base_path[ 0 ] = 0;
    mem_strcat( vm->base_path, base_path );
    //Screen (container):
    vm->screen = pix_vm_new_container( -1, 16, 16, 32, 0, 0, vm );
    pix_vm_gfx_set_screen( vm->screen, vm );
    //Fonts:
    {
	for( int i = 0; i < PIX_VM_FONTS; i++ )
	    vm->fonts[ i ].font = -1;
	COLORPTR font_data = (COLORPTR)MEM_NEW2( g_font8x8_xsize * g_font8x8_ysize * COLORLEN );
	if( font_data == 0 )
	{
	    rv = 6;
	    goto compiler_end;
	}
	int src_ptr = 0;
	int dest_ptr = 0;
	for( int y = 0; y < g_font8x8_ysize; y++ )
	{
	    for( int x = 0; x < g_font8x8_xsize; x += 8 )
	    {
		int byte = g_font8x8[ src_ptr ];
		for( int i = 0; i < 8; i++ )
		{
		    if( byte & 128 ) font_data[ dest_ptr ] = get_color( 255, 255, 255 ); else font_data[ dest_ptr ] = get_color( 0, 0, 0 );
		    byte <<= 1;
		    dest_ptr++;
		}
		src_ptr++;
	    }
	}
	PIX_CID fc = pix_vm_new_container( -1, g_font8x8_xsize, g_font8x8_ysize, 32, font_data, 0, vm );
	vm->fonts[ 0 ].font = fc;
	vm->fonts[ 0 ].xchars = g_font8x8_xchars;
	vm->fonts[ 0 ].ychars = g_font8x8_ychars;
	vm->fonts[ 0 ].first = 32;
	vm->fonts[ 0 ].last = 32 + g_font8x8_xchars * g_font8x8_ychars - 1;
	pix_vm_set_container_flags( fc, pix_vm_get_container_flags( fc, vm ) | PIX_CONTAINER_FLAG_USES_KEY, vm );
	pix_vm_set_container_key_color( fc, font_data[ 0 ], vm );
    }
    //Current event (container):
    vm->event = pix_vm_new_container( -1, 16, 1, PIX_CONTAINER_TYPE_INT32, 0, 0, vm );
    //System info (containers):
    pix_vm_set_systeminfo_containers( vm );
    //Window size:
    vm->var_types[ PIX_GVAR_WINDOW_XSIZE ] = 0;
    vm->var_types[ PIX_GVAR_WINDOW_YSIZE ] = 0;
    vm->var_types[ PIX_GVAR_FPS ] = 0;
    vm->var_types[ PIX_GVAR_PPI ] = 0;
    vm->vars[ PIX_GVAR_WINDOW_XSIZE ].i = 16;
    vm->vars[ PIX_GVAR_WINDOW_YSIZE ].i = 16;
    vm->vars[ PIX_GVAR_FPS ].i = 0;
    vm->vars[ PIX_GVAR_PPI ].i = 0;
    
    DPRINT( "Adding base symbols...\n" );
    ADD_SYMBOL( "while", SYMTYPE_WHILE, 0 );
    ADD_SYMBOL( "break", SYMTYPE_BREAK, 0 );
    ADD_SYMBOL( "break2", SYMTYPE_BREAK2, 0 );
    ADD_SYMBOL( "break3", SYMTYPE_BREAK3, 0 );
    ADD_SYMBOL( "break4", SYMTYPE_BREAK4, 0 );
    ADD_SYMBOL( "breakall", SYMTYPE_BREAKALL, 0 );
    ADD_SYMBOL( "continue", SYMTYPE_CONTINUE, 0 );
    ADD_SYMBOL( "if", SYMTYPE_IF, 0 );
    ADD_SYMBOL( "else", SYMTYPE_ELSE, 0 );
    ADD_SYMBOL( "go", SYMTYPE_GO, 0 );
    ADD_SYMBOL( "goto", SYMTYPE_GO, 0 );
    ADD_SYMBOL( "ret", SYMTYPE_RET, 0 );
    ADD_SYMBOL( "div", SYMTYPE_IDIV, 0 );
    ADD_SYMBOL( "halt", SYMTYPE_HALT, 0 );
    ADD_SYMBOL( "fn", SYMTYPE_FNDEF, 0 );
    ADD_SYMBOL( "include", SYMTYPE_INCLUDE, 0 );

    DPRINT( "Adding functions...\n" );
    //Memory allocation:
    ADD_SYMBOL( "new", SYMTYPE_FNNUM, FN_NEW_PIXI );
    ADD_SYMBOL( "remove", SYMTYPE_FNNUM, FN_REMOVE_PIXI );
    ADD_SYMBOL( "remove_with_alpha", SYMTYPE_FNNUM, FN_REMOVE_PIXI_WITH_ALPHA );
    ADD_SYMBOL( "resize", SYMTYPE_FNNUM, FN_RESIZE_PIXI );
    ADD_SYMBOL( "clean", SYMTYPE_FNNUM, FN_CLEAN_PIXI );
    ADD_SYMBOL( "clone", SYMTYPE_FNNUM, FN_CLONE_PIXI );
    ADD_SYMBOL( "copy", SYMTYPE_FNNUM, FN_COPY_PIXI );
    ADD_SYMBOL( "get_size", SYMTYPE_FNNUM, FN_GET_PIXI_SIZE );
    ADD_SYMBOL( "get_xsize", SYMTYPE_FNNUM, FN_GET_PIXI_XSIZE );
    ADD_SYMBOL( "get_ysize", SYMTYPE_FNNUM, FN_GET_PIXI_YSIZE );
    ADD_SYMBOL( "get_esize", SYMTYPE_FNNUM, FN_GET_PIXI_ESIZE );
    ADD_SYMBOL( "get_type", SYMTYPE_FNNUM, FN_GET_PIXI_TYPE );
    ADD_SYMBOL( "get_flags", SYMTYPE_FNNUM, FN_GET_PIXI_FLAGS );
    ADD_SYMBOL( "set_flags", SYMTYPE_FNNUM, FN_SET_PIXI_FLAGS );
    ADD_SYMBOL( "reset_flags", SYMTYPE_FNNUM, FN_RESET_PIXI_FLAGS );
    ADD_SYMBOL( "get_prop", SYMTYPE_FNNUM, FN_GET_PIXI_PROP );
    ADD_SYMBOL( "set_prop", SYMTYPE_FNNUM, FN_SET_PIXI_PROP );
    ADD_SYMBOL( "remove_props", SYMTYPE_FNNUM, FN_REMOVE_PIXI_PROPS );
    ADD_SYMBOL( "convert_type", SYMTYPE_FNNUM, FN_CONVERT_PIXI_TYPE );
    //Working with strings:
    ADD_SYMBOL( "num_to_str", SYMTYPE_FNNUM, FN_NUM_TO_STRING );
    ADD_SYMBOL( "num2str", SYMTYPE_FNNUM, FN_NUM_TO_STRING );
    ADD_SYMBOL( "str_to_num", SYMTYPE_FNNUM, FN_STRING_TO_NUM );
    ADD_SYMBOL( "str2num", SYMTYPE_FNNUM, FN_STRING_TO_NUM );
    //Working with files:
    ADD_SYMBOL( "load", SYMTYPE_FNNUM, FN_LOAD );
    ADD_SYMBOL( "save", SYMTYPE_FNNUM, FN_SAVE );
    ADD_SYMBOL( "get_real_path", SYMTYPE_FNNUM, FN_GET_REAL_PATH );
    ADD_SYMBOL( "new_flist", SYMTYPE_FNNUM, FN_NEW_FLIST );
    ADD_SYMBOL( "remove_flist", SYMTYPE_FNNUM, FN_REMOVE_FLIST );
    ADD_SYMBOL( "get_flist_name", SYMTYPE_FNNUM, FN_GET_FLIST_NAME );
    ADD_SYMBOL( "get_flist_type", SYMTYPE_FNNUM, FN_GET_FLIST_TYPE );
    ADD_SYMBOL( "flist_next", SYMTYPE_FNNUM, FN_FLIST_NEXT );
    ADD_SYMBOL( "get_file_size", SYMTYPE_FNNUM, FN_GET_FILE_SIZE );
    ADD_SYMBOL( "remove_file", SYMTYPE_FNNUM, FN_REMOVE_FILE );
    ADD_SYMBOL( "rename_file", SYMTYPE_FNNUM, FN_RENAME_FILE );
    ADD_SYMBOL( "copy_file", SYMTYPE_FNNUM, FN_COPY_FILE );
    ADD_SYMBOL( "create_directory", SYMTYPE_FNNUM, FN_CREATE_DIRECTORY );
    //Graphics:
    ADD_SYMBOL( "frame", SYMTYPE_FNNUM, FN_FRAME );
    ADD_SYMBOL( "vsync", SYMTYPE_FNNUM, FN_VSYNC );
    ADD_SYMBOL( "set_screen", SYMTYPE_FNNUM, FN_SET_SCREEN );
    ADD_SYMBOL( "get_screen", SYMTYPE_FNNUM, FN_GET_SCREEN );
    ADD_SYMBOL( "set_zbuf", SYMTYPE_FNNUM, FN_SET_ZBUF );
    ADD_SYMBOL( "get_zbuf", SYMTYPE_FNNUM, FN_GET_ZBUF );
    ADD_SYMBOL( "clear_zbuf", SYMTYPE_FNNUM, FN_CLEAR_ZBUF );
    ADD_SYMBOL( "get_color", SYMTYPE_FNNUM, FN_GET_COLOR );
    ADD_SYMBOL( "get_red", SYMTYPE_FNNUM, FN_GET_RED );
    ADD_SYMBOL( "get_green", SYMTYPE_FNNUM, FN_GET_GREEN );
    ADD_SYMBOL( "get_blue", SYMTYPE_FNNUM, FN_GET_BLUE );
    ADD_SYMBOL( "get_blend", SYMTYPE_FNNUM, FN_GET_BLEND );
    ADD_SYMBOL( "transp", SYMTYPE_FNNUM, FN_TRANSP );
    ADD_SYMBOL( "get_transp", SYMTYPE_FNNUM, FN_GET_TRANSP );
    ADD_SYMBOL( "clear", SYMTYPE_FNNUM, FN_CLEAR );
    ADD_SYMBOL( "dot", SYMTYPE_FNNUM, FN_DOT );
    ADD_SYMBOL( "dot3d", SYMTYPE_FNNUM, FN_DOT3D );
    ADD_SYMBOL( "get_dot", SYMTYPE_FNNUM, FN_GET_DOT );
    ADD_SYMBOL( "get_dot3d", SYMTYPE_FNNUM, FN_GET_DOT3D );
    ADD_SYMBOL( "line", SYMTYPE_FNNUM, FN_LINE );
    ADD_SYMBOL( "line3d", SYMTYPE_FNNUM, FN_LINE3D );
    ADD_SYMBOL( "box", SYMTYPE_FNNUM, FN_BOX );
    ADD_SYMBOL( "fbox", SYMTYPE_FNNUM, FN_FBOX );
    ADD_SYMBOL( "pixi", SYMTYPE_FNNUM, FN_PIXI );
    ADD_SYMBOL( "triangles3d", SYMTYPE_FNNUM, FN_TRIANGLES );
    ADD_SYMBOL( "sort_triangles3d", SYMTYPE_FNNUM, FN_SORT_TRIANGLES );
    ADD_SYMBOL( "set_key_color", SYMTYPE_FNNUM, FN_SET_KEY_COLOR );
    ADD_SYMBOL( "get_key_color", SYMTYPE_FNNUM, FN_GET_KEY_COLOR );
    ADD_SYMBOL( "set_alpha", SYMTYPE_FNNUM, FN_SET_ALPHA );
    ADD_SYMBOL( "get_alpha", SYMTYPE_FNNUM, FN_GET_ALPHA );
    ADD_SYMBOL( "print", SYMTYPE_FNNUM, FN_PRINT );
    ADD_SYMBOL( "get_text_xsize", SYMTYPE_FNNUM, FN_GET_TEXT_XSIZE );
    ADD_SYMBOL( "get_text_ysize", SYMTYPE_FNNUM, FN_GET_TEXT_YSIZE );
    ADD_SYMBOL( "set_font", SYMTYPE_FNNUM, FN_SET_FONT );
    ADD_SYMBOL( "get_font", SYMTYPE_FNNUM, FN_GET_FONT );
    ADD_SYMBOL( "effector", SYMTYPE_FNNUM, FN_EFFECTOR );
    ADD_SYMBOL( "split_rgb", SYMTYPE_FNNUM, FN_SPLIT_RGB );
    //OpenGL graphics:
    ADD_SYMBOL( "set_gl_callback", SYMTYPE_FNNUM, FN_SET_GL_CALLBACK );
    ADD_SYMBOL( "remove_gl_data", SYMTYPE_FNNUM, FN_REMOVE_GL_DATA );
    //Animation:
    ADD_SYMBOL( "unpack_frame", SYMTYPE_FNNUM, FN_PIXI_UNPACK_FRAME );
    ADD_SYMBOL( "pack_frame", SYMTYPE_FNNUM, FN_PIXI_PACK_FRAME );
    ADD_SYMBOL( "create_anim", SYMTYPE_FNNUM, FN_PIXI_CREATE_ANIM );
    ADD_SYMBOL( "remove_anim", SYMTYPE_FNNUM, FN_PIXI_REMOVE_ANIM );
    ADD_SYMBOL( "clone_frame", SYMTYPE_FNNUM, FN_PIXI_CLONE_FRAME );
    ADD_SYMBOL( "remove_frame", SYMTYPE_FNNUM, FN_PIXI_REMOVE_FRAME );
    ADD_SYMBOL( "play", SYMTYPE_FNNUM, FN_PIXI_PLAY );
    ADD_SYMBOL( "stop", SYMTYPE_FNNUM, FN_PIXI_STOP );
    //Transformation:
    ADD_SYMBOL( "t_reset", SYMTYPE_FNNUM, FN_T_RESET );
    ADD_SYMBOL( "t_rotate", SYMTYPE_FNNUM, FN_T_ROTATE );
    ADD_SYMBOL( "t_translate", SYMTYPE_FNNUM, FN_T_TRANSLATE );
    ADD_SYMBOL( "t_scale", SYMTYPE_FNNUM, FN_T_SCALE );
    ADD_SYMBOL( "t_push_matrix", SYMTYPE_FNNUM, FN_T_PUSH_MATRIX );
    ADD_SYMBOL( "t_pop_matrix", SYMTYPE_FNNUM, FN_T_POP_MATRIX );
    ADD_SYMBOL( "t_get_matrix", SYMTYPE_FNNUM, FN_T_GET_MATRIX );
    ADD_SYMBOL( "t_set_matrix", SYMTYPE_FNNUM, FN_T_SET_MATRIX );
    ADD_SYMBOL( "t_mul_matrix", SYMTYPE_FNNUM, FN_T_MUL_MATRIX );
    ADD_SYMBOL( "t_point", SYMTYPE_FNNUM, FN_T_POINT );
    //Audio:
    ADD_SYMBOL( "set_audio_callback", SYMTYPE_FNNUM, FN_SET_AUDIO_CALLBACK );
    ADD_SYMBOL( "enable_audio_input", SYMTYPE_FNNUM, FN_ENABLE_AUDIO_INPUT );
    ADD_SYMBOL( "get_note_freq", SYMTYPE_FNNUM, FN_GET_NOTE_FREQ );
    //MIDI:
    ADD_SYMBOL( "midi_open_client", SYMTYPE_FNNUM, FN_MIDI_OPEN_CLIENT );
    ADD_SYMBOL( "midi_close_client", SYMTYPE_FNNUM, FN_MIDI_CLOSE_CLIENT );
    ADD_SYMBOL( "midi_get_device", SYMTYPE_FNNUM, FN_MIDI_GET_DEVICE );
    ADD_SYMBOL( "midi_open_port", SYMTYPE_FNNUM, FN_MIDI_OPEN_PORT );
    ADD_SYMBOL( "midi_reopen_port", SYMTYPE_FNNUM, FN_MIDI_REOPEN_PORT );
    ADD_SYMBOL( "midi_close_port", SYMTYPE_FNNUM, FN_MIDI_CLOSE_PORT );
    ADD_SYMBOL( "midi_get_event", SYMTYPE_FNNUM, FN_MIDI_GET_EVENT );
    ADD_SYMBOL( "midi_get_event_time", SYMTYPE_FNNUM, FN_MIDI_GET_EVENT_TIME );
    ADD_SYMBOL( "midi_next_event", SYMTYPE_FNNUM, FN_MIDI_NEXT_EVENT );
    ADD_SYMBOL( "midi_send_event", SYMTYPE_FNNUM, FN_MIDI_SEND_EVENT );
    //Time:
    ADD_SYMBOL( "start_timer", SYMTYPE_FNNUM, FN_START_TIMER );
    ADD_SYMBOL( "get_timer", SYMTYPE_FNNUM, FN_GET_TIMER );
    ADD_SYMBOL( "get_year", SYMTYPE_FNNUM, FN_GET_YEAR );
    ADD_SYMBOL( "get_month", SYMTYPE_FNNUM, FN_GET_MONTH );
    ADD_SYMBOL( "get_day", SYMTYPE_FNNUM, FN_GET_DAY );
    ADD_SYMBOL( "get_hours", SYMTYPE_FNNUM, FN_GET_HOURS );
    ADD_SYMBOL( "get_minutes", SYMTYPE_FNNUM, FN_GET_MINUTES );
    ADD_SYMBOL( "get_seconds", SYMTYPE_FNNUM, FN_GET_SECONDS );
    ADD_SYMBOL( "get_ticks", SYMTYPE_FNNUM, FN_GET_TICKS );
    ADD_SYMBOL( "get_tps", SYMTYPE_FNNUM, FN_GET_TPS );
    ADD_SYMBOL( "sleep", SYMTYPE_FNNUM, FN_SLEEP );
    //Events:
    ADD_SYMBOL( "get_event", SYMTYPE_FNNUM, FN_GET_EVENT );
    ADD_SYMBOL( "set_quit_action", SYMTYPE_FNNUM, FN_SET_QUIT_ACTION );
    //Threads:
    ADD_SYMBOL( "thread_create", SYMTYPE_FNNUM, FN_THREAD_CREATE );
    ADD_SYMBOL( "thread_destroy", SYMTYPE_FNNUM, FN_THREAD_DESTROY );
    ADD_SYMBOL( "mutex_create", SYMTYPE_FNNUM, FN_MUTEX_CREATE );
    ADD_SYMBOL( "mutex_destroy", SYMTYPE_FNNUM, FN_MUTEX_DESTROY );
    ADD_SYMBOL( "mutex_lock", SYMTYPE_FNNUM, FN_MUTEX_LOCK );
    ADD_SYMBOL( "mutex_trylock", SYMTYPE_FNNUM, FN_MUTEX_TRYLOCK );
    ADD_SYMBOL( "mutex_unlock", SYMTYPE_FNNUM, FN_MUTEX_UNLOCK );
    //Mathematical functions:
    ADD_SYMBOL( "acos", SYMTYPE_FNNUM, FN_ACOS );
    ADD_SYMBOL( "acosh", SYMTYPE_FNNUM, FN_ACOSH );
    ADD_SYMBOL( "asin", SYMTYPE_FNNUM, FN_ASIN );
    ADD_SYMBOL( "asinh", SYMTYPE_FNNUM, FN_ASINH );
    ADD_SYMBOL( "atan", SYMTYPE_FNNUM, FN_ATAN );
    ADD_SYMBOL( "atanh", SYMTYPE_FNNUM, FN_ATANH );
    ADD_SYMBOL( "ceil", SYMTYPE_FNNUM, FN_CEIL );
    ADD_SYMBOL( "cos", SYMTYPE_FNNUM, FN_COS );
    ADD_SYMBOL( "cosh", SYMTYPE_FNNUM, FN_COSH );
    ADD_SYMBOL( "exp", SYMTYPE_FNNUM, FN_EXP );
    ADD_SYMBOL( "exp2", SYMTYPE_FNNUM, FN_EXP2 );
    ADD_SYMBOL( "expm1", SYMTYPE_FNNUM, FN_EXPM1 );
    ADD_SYMBOL( "abs", SYMTYPE_FNNUM, FN_ABS );
    ADD_SYMBOL( "floor", SYMTYPE_FNNUM, FN_FLOOR );
    ADD_SYMBOL( "mod", SYMTYPE_FNNUM, FN_MOD );
    ADD_SYMBOL( "log", SYMTYPE_FNNUM, FN_LOG );
    ADD_SYMBOL( "log2", SYMTYPE_FNNUM, FN_LOG2 );
    ADD_SYMBOL( "log10", SYMTYPE_FNNUM, FN_LOG10 );
    ADD_SYMBOL( "pow", SYMTYPE_FNNUM, FN_POW );
    ADD_SYMBOL( "sin", SYMTYPE_FNNUM, FN_SIN );
    ADD_SYMBOL( "sinh", SYMTYPE_FNNUM, FN_SINH );
    ADD_SYMBOL( "sqrt", SYMTYPE_FNNUM, FN_SQRT );
    ADD_SYMBOL( "tan", SYMTYPE_FNNUM, FN_TAN );
    ADD_SYMBOL( "tanh", SYMTYPE_FNNUM, FN_TANH );
    ADD_SYMBOL( "rand", SYMTYPE_FNNUM, FN_RAND );
    ADD_SYMBOL( "rand_seed", SYMTYPE_FNNUM, FN_RAND_SEED );
    //Data processing:
    ADD_SYMBOL( "op_cn", SYMTYPE_FNNUM, FN_OP_CN );
    ADD_SYMBOL( "op_cc", SYMTYPE_FNNUM, FN_OP_CC );
    ADD_SYMBOL( "op_ccn", SYMTYPE_FNNUM, FN_OP_CCN );
    ADD_SYMBOL( "generator", SYMTYPE_FNNUM, FN_GENERATOR );
    ADD_SYMBOL( "wavetable_generator", SYMTYPE_FNNUM, FN_WAVETABLE_GENERATOR );
    ADD_SYMBOL( "sampler", SYMTYPE_FNNUM, FN_SAMPLER );
    ADD_SYMBOL( "envelope2p", SYMTYPE_FNNUM, FN_ENVELOPE2P );
    ADD_SYMBOL( "fft", SYMTYPE_FNNUM, FN_FFT );
    ADD_SYMBOL( "replace_values", SYMTYPE_FNNUM, FN_REPLACE_VALUES );
    //Dialogs:
    ADD_SYMBOL( "file_dialog", SYMTYPE_FNNUM, FN_FILE_DIALOG );
    ADD_SYMBOL( "prefs_dialog", SYMTYPE_FNNUM, FN_PREFS_DIALOG );
    //Network:
    ADD_SYMBOL( "open_url", SYMTYPE_FNNUM, FN_OPEN_URL );    
    //Native code:
    ADD_SYMBOL( "dlopen", SYMTYPE_FNNUM, FN_DLOPEN );
    ADD_SYMBOL( "dlclose", SYMTYPE_FNNUM, FN_DLCLOSE );
    ADD_SYMBOL( "dlsym", SYMTYPE_FNNUM, FN_DLSYM );
    ADD_SYMBOL( "dlcall", SYMTYPE_FNNUM, FN_DLCALL );
    //Posix compatibility:
    ADD_SYMBOL( "fopen", SYMTYPE_FNNUM, FN_FOPEN );
    ADD_SYMBOL( "fclose", SYMTYPE_FNNUM, FN_FCLOSE );
    ADD_SYMBOL( "fputc", SYMTYPE_FNNUM, FN_FPUTC );
    ADD_SYMBOL( "fputs", SYMTYPE_FNNUM, FN_FPUTS );
    ADD_SYMBOL( "fwrite", SYMTYPE_FNNUM, FN_FWRITE );
    ADD_SYMBOL( "fgetc", SYMTYPE_FNNUM, FN_FGETC );
    ADD_SYMBOL( "fgets", SYMTYPE_FNNUM, FN_FGETS );
    ADD_SYMBOL( "fread", SYMTYPE_FNNUM, FN_FREAD );
    ADD_SYMBOL( "feof", SYMTYPE_FNNUM, FN_FEOF );
    ADD_SYMBOL( "fflush", SYMTYPE_FNNUM, FN_FFLUSH );
    ADD_SYMBOL( "fseek", SYMTYPE_FNNUM, FN_FSEEK );
    ADD_SYMBOL( "ftell", SYMTYPE_FNNUM, FN_FTELL );
    ADD_SYMBOL( "strcat", SYMTYPE_FNNUM, FN_STRCAT );
    ADD_SYMBOL( "strcmp", SYMTYPE_FNNUM, FN_STRCMP );
    ADD_SYMBOL( "strlen", SYMTYPE_FNNUM, FN_STRLEN );
    ADD_SYMBOL( "strstr", SYMTYPE_FNNUM, FN_STRSTR );
    ADD_SYMBOL( "sprintf", SYMTYPE_FNNUM, FN_SPRINTF );
    ADD_SYMBOL( "printf", SYMTYPE_FNNUM, FN_PRINTF );
    ADD_SYMBOL( "system", SYMTYPE_FNNUM, FN_SYSTEM );
    ADD_SYMBOL( "argc", SYMTYPE_FNNUM, FN_ARGC );
    ADD_SYMBOL( "argv", SYMTYPE_FNNUM, FN_ARGV );
    ADD_SYMBOL( "exit", SYMTYPE_FNNUM, FN_EXIT );
    //Private API:
    ADD_SYMBOL( "system_copy", SYMTYPE_FNNUM, FN_SYSTEM_COPY );
    ADD_SYMBOL( "system_paste", SYMTYPE_FNNUM, FN_SYSTEM_PASTE );
    ADD_SYMBOL( "open_webserver", SYMTYPE_FNNUM, FN_OPEN_WEBSERVER );

    DPRINT( "Adding constants...\n" );
    //STDIO constants:
    ADD_SYMBOL( "FOPEN_MAX", SYMTYPE_NUM_I, V3_FOPEN_MAX ); //Number of streams which the implementation guarantees can be open simultaneously.
    ADD_SYMBOL( "SEEK_CUR", SYMTYPE_NUM_I, SEEK_CUR ); //Seek relative to current position.
    ADD_SYMBOL( "SEEK_END", SYMTYPE_NUM_I, SEEK_END ); //Seek relative to end-of-file.
    ADD_SYMBOL( "SEEK_SET", SYMTYPE_NUM_I, SEEK_SET ); //Seek relative to start-of-file.
    ADD_SYMBOL( "EOF", SYMTYPE_NUM_I, -1 ); //End-of-file return value.
    ADD_SYMBOL( "STDIN", SYMTYPE_NUM_I, V3_STDIN ); //Standard input stream.
    ADD_SYMBOL( "STDOUT", SYMTYPE_NUM_I, V3_STDOUT ); //Standard output stream.
    ADD_SYMBOL( "STDERR", SYMTYPE_NUM_I, V3_STDERR ); //Standard error output stream.
    //Container flags:
    ADD_SYMBOL( "GL_MIN_LINEAR", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_GL_MIN_LINEAR );
    ADD_SYMBOL( "GL_MAG_LINEAR", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_GL_MAG_LINEAR );
    ADD_SYMBOL( "GL_NO_XREPEAT", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_GL_NO_XREPEAT );
    ADD_SYMBOL( "GL_NO_YREPEAT", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_GL_NO_YREPEAT );
    ADD_SYMBOL( "GL_NICEST", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_GL_NICEST );
    ADD_SYMBOL( "CFLAG_INTERP", SYMTYPE_NUM_I, PIX_CONTAINER_FLAG_INTERP );
    //Colors:
    ADD_SYMBOL( "ORANGE", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 128, 16 ) );
    ADD_SYMBOL( "ORANJ", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 128, 16 ) );
    ADD_SYMBOL( "BLACK", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 0, 0, 0 ) );
    ADD_SYMBOL( "WHITE", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 255, 255 ) );
    ADD_SYMBOL( "SNEG", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 255, 255 ) );
    ADD_SYMBOL( "YELLOW", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 255, 0 ) );
    ADD_SYMBOL( "SUN", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 255, 0 ) );
    ADD_SYMBOL( "RED", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 255, 0, 0 ) );
    ADD_SYMBOL( "GREEN", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 0, 255, 0 ) );
    ADD_SYMBOL( "ZELEN", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 0, 255, 0 ) );
    ADD_SYMBOL( "BLUE", SYMTYPE_NUM_I, (COLORSIGNED)get_color( 0, 0, 255 ) );
    //Alignment:
    ADD_SYMBOL( "TOP", SYMTYPE_NUM_I, 1 );
    ADD_SYMBOL( "BOTTOM", SYMTYPE_NUM_I, 2 );
    ADD_SYMBOL( "LEFT", SYMTYPE_NUM_I, 4 );
    ADD_SYMBOL( "RIGHT", SYMTYPE_NUM_I, 8 );
    //Effects:
    ADD_SYMBOL( "EFF_NOISE", SYMTYPE_NUM_I, PIX_EFFECT_NOISE );
    ADD_SYMBOL( "EFF_SPREAD_LEFT", SYMTYPE_NUM_I, PIX_EFFECT_SPREAD_LEFT );
    ADD_SYMBOL( "EFF_SPREAD_RIGHT", SYMTYPE_NUM_I, PIX_EFFECT_SPREAD_RIGHT );
    ADD_SYMBOL( "EFF_SPREAD_UP", SYMTYPE_NUM_I, PIX_EFFECT_SPREAD_UP );
    ADD_SYMBOL( "EFF_SPREAD_DOWN", SYMTYPE_NUM_I, PIX_EFFECT_SPREAD_DOWN );
    ADD_SYMBOL( "EFF_VBLUR", SYMTYPE_NUM_I, PIX_EFFECT_VBLUR );
    ADD_SYMBOL( "EFF_HBLUR", SYMTYPE_NUM_I, PIX_EFFECT_HBLUR );
    ADD_SYMBOL( "EFF_COLOR", SYMTYPE_NUM_I, PIX_EFFECT_COLOR );
    //File formats:
    ADD_SYMBOL( "FORMAT_RAW", SYMTYPE_NUM_I, PIX_FORMAT_RAW );
    ADD_SYMBOL( "FORMAT_JPEG", SYMTYPE_NUM_I, PIX_FORMAT_JPEG );
    ADD_SYMBOL( "FORMAT_PNG", SYMTYPE_NUM_I, PIX_FORMAT_PNG );
    ADD_SYMBOL( "FORMAT_GIF", SYMTYPE_NUM_I, PIX_FORMAT_GIF );
    ADD_SYMBOL( "FORMAT_WAVE", SYMTYPE_NUM_I, PIX_FORMAT_WAVE );
    ADD_SYMBOL( "FORMAT_PIXICONTAINER", SYMTYPE_NUM_I, PIX_FORMAT_PIXICONTAINER );
    ADD_SYMBOL( "GIF_GRAYSCALE", SYMTYPE_NUM_I, 1 );
    ADD_SYMBOL( "GIF_DITHER", SYMTYPE_NUM_I, 4 );
    //Audio:
    ADD_SYMBOL( "AUDIO_FLAG_INTERP2", SYMTYPE_NUM_I, PIX_AUDIO_FLAG_INTERP2 );
    //MIDI:
    ADD_SYMBOL( "MIDI_PORT_READ", SYMTYPE_NUM_I, MIDI_PORT_READ );
    ADD_SYMBOL( "MIDI_PORT_WRITE", SYMTYPE_NUM_I, MIDI_PORT_WRITE );
    ADD_SYMBOL( "MIDI_NO_DEVICE", SYMTYPE_NUM_I, MIDI_NO_DEVICE );
    //Events:
    ADD_SYMBOL( "EVT", SYMTYPE_NUM_I, vm->event );
    ADD_SYMBOL( "EVT_TYPE", SYMTYPE_NUM_I, 0 );
    ADD_SYMBOL( "EVT_FLAGS", SYMTYPE_NUM_I, 1 );
    ADD_SYMBOL( "EVT_TIME", SYMTYPE_NUM_I, 2 );
    ADD_SYMBOL( "EVT_X", SYMTYPE_NUM_I, 3 );
    ADD_SYMBOL( "EVT_Y", SYMTYPE_NUM_I, 4 );
    ADD_SYMBOL( "EVT_KEY", SYMTYPE_NUM_I, 5 );
    ADD_SYMBOL( "EVT_SCANCODE", SYMTYPE_NUM_I, 6 );
    ADD_SYMBOL( "EVT_PRESSURE", SYMTYPE_NUM_I, 7 );
    ADD_SYMBOL( "EVT_UNICODE", SYMTYPE_NUM_I, 8 );
    ADD_SYMBOL( "EVT_MOUSEBUTTONDOWN", SYMTYPE_NUM_I, PIX_EVT_MOUSEBUTTONDOWN );
    ADD_SYMBOL( "EVT_MOUSEBUTTONUP", SYMTYPE_NUM_I, PIX_EVT_MOUSEBUTTONUP );
    ADD_SYMBOL( "EVT_MOUSEMOVE", SYMTYPE_NUM_I, PIX_EVT_MOUSEMOVE );
    ADD_SYMBOL( "EVT_TOUCHBEGIN", SYMTYPE_NUM_I, PIX_EVT_TOUCHBEGIN );
    ADD_SYMBOL( "EVT_TOUCHEND", SYMTYPE_NUM_I, PIX_EVT_TOUCHEND );
    ADD_SYMBOL( "EVT_TOUCHMOVE", SYMTYPE_NUM_I, PIX_EVT_TOUCHMOVE );
    ADD_SYMBOL( "EVT_BUTTONDOWN", SYMTYPE_NUM_I, PIX_EVT_BUTTONDOWN );
    ADD_SYMBOL( "EVT_BUTTONUP", SYMTYPE_NUM_I, PIX_EVT_BUTTONUP );
    ADD_SYMBOL( "EVT_SCREENRESIZE", SYMTYPE_NUM_I, PIX_EVT_SCREENRESIZE );
    ADD_SYMBOL( "EVT_QUIT", SYMTYPE_NUM_I, PIX_EVT_QUIT );
    ADD_SYMBOL( "EVT_FLAG_SHIFT", SYMTYPE_NUM_I, EVT_FLAG_SHIFT );
    ADD_SYMBOL( "EVT_FLAG_CTRL", SYMTYPE_NUM_I, EVT_FLAG_CTRL );
    ADD_SYMBOL( "EVT_FLAG_ALT", SYMTYPE_NUM_I, EVT_FLAG_ALT );
    ADD_SYMBOL( "EVT_FLAG_MODE", SYMTYPE_NUM_I, EVT_FLAG_MODE );
    ADD_SYMBOL( "EVT_FLAG_MODS", SYMTYPE_NUM_I, EVT_FLAG_MODS );
    ADD_SYMBOL( "EVT_FLAG_DOUBLECLICK", SYMTYPE_NUM_I, EVT_FLAG_DOUBLECLICK );
    ADD_SYMBOL( "KEY_MOUSE_LEFT", SYMTYPE_NUM_I, MOUSE_BUTTON_LEFT );
    ADD_SYMBOL( "KEY_MOUSE_MIDDLE", SYMTYPE_NUM_I, MOUSE_BUTTON_MIDDLE );
    ADD_SYMBOL( "KEY_MOUSE_RIGHT", SYMTYPE_NUM_I, MOUSE_BUTTON_RIGHT );
    ADD_SYMBOL( "KEY_MOUSE_SCROLLUP", SYMTYPE_NUM_I, MOUSE_BUTTON_SCROLLUP );
    ADD_SYMBOL( "KEY_MOUSE_SCROLLDOWN", SYMTYPE_NUM_I, MOUSE_BUTTON_SCROLLDOWN );
    ADD_SYMBOL( "KEY_BACKSPACE", SYMTYPE_NUM_I, KEY_BACKSPACE );
    ADD_SYMBOL( "KEY_TAB", SYMTYPE_NUM_I, KEY_TAB );
    ADD_SYMBOL( "KEY_ENTER", SYMTYPE_NUM_I, KEY_ENTER );
    ADD_SYMBOL( "KEY_ESCAPE", SYMTYPE_NUM_I, KEY_ESCAPE );
    ADD_SYMBOL( "KEY_SPACE", SYMTYPE_NUM_I, KEY_SPACE );
    ADD_SYMBOL( "KEY_F1", SYMTYPE_NUM_I, KEY_F1 );
    ADD_SYMBOL( "KEY_F2", SYMTYPE_NUM_I, KEY_F2 );
    ADD_SYMBOL( "KEY_F3", SYMTYPE_NUM_I, KEY_F3 );
    ADD_SYMBOL( "KEY_F4", SYMTYPE_NUM_I, KEY_F4 );
    ADD_SYMBOL( "KEY_F5", SYMTYPE_NUM_I, KEY_F5 );
    ADD_SYMBOL( "KEY_F6", SYMTYPE_NUM_I, KEY_F6 );
    ADD_SYMBOL( "KEY_F7", SYMTYPE_NUM_I, KEY_F7 );
    ADD_SYMBOL( "KEY_F8", SYMTYPE_NUM_I, KEY_F8 );
    ADD_SYMBOL( "KEY_F9", SYMTYPE_NUM_I, KEY_F9 );
    ADD_SYMBOL( "KEY_F10", SYMTYPE_NUM_I, KEY_F10 );
    ADD_SYMBOL( "KEY_F11", SYMTYPE_NUM_I, KEY_F11 );
    ADD_SYMBOL( "KEY_F12", SYMTYPE_NUM_I, KEY_F12 );
    ADD_SYMBOL( "KEY_UP", SYMTYPE_NUM_I, KEY_UP );
    ADD_SYMBOL( "KEY_DOWN", SYMTYPE_NUM_I, KEY_DOWN );
    ADD_SYMBOL( "KEY_LEFT", SYMTYPE_NUM_I, KEY_LEFT );
    ADD_SYMBOL( "KEY_RIGHT", SYMTYPE_NUM_I, KEY_RIGHT );
    ADD_SYMBOL( "KEY_INSERT", SYMTYPE_NUM_I, KEY_INSERT );
    ADD_SYMBOL( "KEY_DELETE", SYMTYPE_NUM_I, KEY_DELETE );
    ADD_SYMBOL( "KEY_HOME", SYMTYPE_NUM_I, KEY_HOME );
    ADD_SYMBOL( "KEY_END", SYMTYPE_NUM_I, KEY_END );
    ADD_SYMBOL( "KEY_PAGEUP", SYMTYPE_NUM_I, KEY_PAGEUP );
    ADD_SYMBOL( "KEY_PAGEDOWN", SYMTYPE_NUM_I, KEY_PAGEDOWN );
    ADD_SYMBOL( "KEY_CAPS", SYMTYPE_NUM_I, KEY_CAPS );
    ADD_SYMBOL( "KEY_SHIFT", SYMTYPE_NUM_I, KEY_SHIFT );
    ADD_SYMBOL( "KEY_CTRL", SYMTYPE_NUM_I, KEY_CTRL );
    ADD_SYMBOL( "KEY_ALT", SYMTYPE_NUM_I, KEY_ALT );
    ADD_SYMBOL( "KEY_MENU", SYMTYPE_NUM_I, KEY_MENU );
    ADD_SYMBOL( "KEY_UNKNOWN", SYMTYPE_NUM_I, KEY_UNKNOWN );
    ADD_SYMBOL( "QA_NONE", SYMTYPE_NUM_I, 0 );
    ADD_SYMBOL( "QA_CLOSE_VM", SYMTYPE_NUM_I, 1 );
    //Mathematical:
    ADD_SYMBOL( "M_E", SYMTYPE_NUM_F, M_E );
    ADD_SYMBOL( "M_LOG2E", SYMTYPE_NUM_F, M_LOG2E );
    ADD_SYMBOL( "M_LOG10E", SYMTYPE_NUM_F, M_LOG10E );
    ADD_SYMBOL( "M_LN2", SYMTYPE_NUM_F, M_LN2 );
    ADD_SYMBOL( "M_LN10", SYMTYPE_NUM_F, M_LN10 );
    ADD_SYMBOL( "M_PI", SYMTYPE_NUM_F, M_PI );
    ADD_SYMBOL( "M_2_SQRTPI", SYMTYPE_NUM_F, M_2_SQRTPI );
    ADD_SYMBOL( "M_SQRT2", SYMTYPE_NUM_F, M_SQRT2 );
    ADD_SYMBOL( "M_SQRT1_2", SYMTYPE_NUM_F, M_SQRT1_2 );
    //Data processing operations (op_cn):
    ADD_SYMBOL( "OP_MIN", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MIN );
    ADD_SYMBOL( "OP_MAX", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MAX );
    ADD_SYMBOL( "OP_MAXMOD", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MAXABS );
    ADD_SYMBOL( "OP_MAXABS", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MAXABS );
    ADD_SYMBOL( "OP_LIMIT_TOP", SYMTYPE_NUM_I, PIX_DATA_OPCODE_LIMIT_TOP );
    ADD_SYMBOL( "OP_LIMIT_BOTTOM", SYMTYPE_NUM_I, PIX_DATA_OPCODE_LIMIT_BOTTOM );
    ADD_SYMBOL( "OP_ABS", SYMTYPE_NUM_I, PIX_DATA_OPCODE_ABS );
    ADD_SYMBOL( "OP_COLOR_SUB2", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COLOR_SUB2 );
    ADD_SYMBOL( "OP_DIV2", SYMTYPE_NUM_I, PIX_DATA_OPCODE_DIV2 );
    //Data processing operations (op_cn, op_cc):
    ADD_SYMBOL( "OP_ADD", SYMTYPE_NUM_I, PIX_DATA_OPCODE_ADD );
    ADD_SYMBOL( "OP_SADD", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SADD );
    ADD_SYMBOL( "OP_COLOR_ADD", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COLOR_ADD );
    ADD_SYMBOL( "OP_SUB", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SUB );
    ADD_SYMBOL( "OP_SSUB", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SSUB );
    ADD_SYMBOL( "OP_COLOR_SUB", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COLOR_SUB );
    ADD_SYMBOL( "OP_MUL", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MUL );
    ADD_SYMBOL( "OP_SMUL", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SMUL );
    ADD_SYMBOL( "OP_MUL_RSHIFT15", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MUL_RSHIFT15 );
    ADD_SYMBOL( "OP_COLOR_MUL", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COLOR_MUL );
    ADD_SYMBOL( "OP_DIV", SYMTYPE_NUM_I, PIX_DATA_OPCODE_DIV );
    ADD_SYMBOL( "OP_COLOR_DIV", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COLOR_DIV );
    ADD_SYMBOL( "OP_AND", SYMTYPE_NUM_I, PIX_DATA_OPCODE_AND );
    ADD_SYMBOL( "OP_OR", SYMTYPE_NUM_I, PIX_DATA_OPCODE_OR );
    ADD_SYMBOL( "OP_XOR", SYMTYPE_NUM_I, PIX_DATA_OPCODE_XOR );
    ADD_SYMBOL( "OP_LSHIFT", SYMTYPE_NUM_I, PIX_DATA_OPCODE_LSHIFT );
    ADD_SYMBOL( "OP_RSHIFT", SYMTYPE_NUM_I, PIX_DATA_OPCODE_RSHIFT );
    ADD_SYMBOL( "OP_EQUAL", SYMTYPE_NUM_I, PIX_DATA_OPCODE_EQUAL );
    ADD_SYMBOL( "OP_LESS", SYMTYPE_NUM_I, PIX_DATA_OPCODE_LESS );
    ADD_SYMBOL( "OP_GREATER", SYMTYPE_NUM_I, PIX_DATA_OPCODE_GREATER );
    //Data processing operations (op_cc):
    ADD_SYMBOL( "OP_COPY", SYMTYPE_NUM_I, PIX_DATA_OPCODE_COPY );
    //Data processing operations (op_ccn):
    ADD_SYMBOL( "OP_MUL_DIV", SYMTYPE_NUM_I, PIX_DATA_OPCODE_MUL_DIV );
    //Data processing operations (generator):
    ADD_SYMBOL( "OP_SIN", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SIN );
    ADD_SYMBOL( "OP_SIN8", SYMTYPE_NUM_I, PIX_DATA_OPCODE_SIN8 );
    ADD_SYMBOL( "OP_RAND", SYMTYPE_NUM_I, PIX_DATA_OPCODE_RAND );
    //Sampler:
    ADD_SYMBOL( "SMP_DEST", SYMTYPE_NUM_I, PIX_SAMPLER_DEST );
    ADD_SYMBOL( "SMP_DEST_OFF", SYMTYPE_NUM_I, PIX_SAMPLER_DEST_OFF );
    ADD_SYMBOL( "SMP_DEST_LEN", SYMTYPE_NUM_I, PIX_SAMPLER_DEST_LEN );
    ADD_SYMBOL( "SMP_SRC", SYMTYPE_NUM_I, PIX_SAMPLER_SRC );
    ADD_SYMBOL( "SMP_SRC_OFF_H", SYMTYPE_NUM_I, PIX_SAMPLER_SRC_OFF_H );
    ADD_SYMBOL( "SMP_SRC_OFF_L", SYMTYPE_NUM_I, PIX_SAMPLER_SRC_OFF_L );
    ADD_SYMBOL( "SMP_SRC_SIZE", SYMTYPE_NUM_I, PIX_SAMPLER_SRC_SIZE );
    ADD_SYMBOL( "SMP_LOOP", SYMTYPE_NUM_I, PIX_SAMPLER_LOOP );
    ADD_SYMBOL( "SMP_LOOP_LEN", SYMTYPE_NUM_I, PIX_SAMPLER_LOOP_LEN );
    ADD_SYMBOL( "SMP_VOL1", SYMTYPE_NUM_I, PIX_SAMPLER_VOL1 );
    ADD_SYMBOL( "SMP_VOL2", SYMTYPE_NUM_I, PIX_SAMPLER_VOL2 );
    ADD_SYMBOL( "SMP_DELTA", SYMTYPE_NUM_I, PIX_SAMPLER_DELTA );
    ADD_SYMBOL( "SMP_FLAGS", SYMTYPE_NUM_I, PIX_SAMPLER_FLAGS );
    ADD_SYMBOL( "SMP_INFO_SIZE", SYMTYPE_NUM_I, PIX_SAMPLER_PARAMETERS );
    ADD_SYMBOL( "SMP_FLAG_INTERP2", SYMTYPE_NUM_I, PIX_SAMPLER_FLAG_INTERP2 );
    ADD_SYMBOL( "SMP_FLAG_INTERP4", SYMTYPE_NUM_I, PIX_SAMPLER_FLAG_INTERP4 );
    ADD_SYMBOL( "SMP_FLAG_PINGPONG", SYMTYPE_NUM_I, PIX_SAMPLER_FLAG_PINGPONG );
    ADD_SYMBOL( "SMP_FLAG_REVERSE", SYMTYPE_NUM_I, PIX_SAMPLER_FLAG_REVERSE );
    //Native code:
    ADD_SYMBOL( "CCONV_DEFAULT", SYMTYPE_NUM_I, PIX_CCONV_DEFAULT );
    ADD_SYMBOL( "CCONV_CDECL", SYMTYPE_NUM_I, PIX_CCONV_CDECL );
    ADD_SYMBOL( "CCONV_STDCALL", SYMTYPE_NUM_I, PIX_CCONV_STDCALL );
    ADD_SYMBOL( "CCONV_UNIX_AMD64", SYMTYPE_NUM_I, PIX_CCONV_UNIX_AMD64 );
    ADD_SYMBOL( "CCONV_WIN64", SYMTYPE_NUM_I, PIX_CCONV_WIN64 );
    //Another constants:
    ADD_SYMBOL( "INT_SIZE", SYMTYPE_NUM_I, sizeof( PIX_INT ) );
    ADD_SYMBOL( "INT_MAX", SYMTYPE_NUM_I, PIX_INT_MAX_POSITIVE );
    ADD_SYMBOL( "FLOAT_SIZE", SYMTYPE_NUM_I, sizeof( PIX_FLOAT ) );
    ADD_SYMBOL( "COLORBITS", SYMTYPE_NUM_I, COLORBITS );
    ADD_SYMBOL( "INT8", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT8 );
    ADD_SYMBOL( "INT16", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT16 );
    ADD_SYMBOL( "INT32", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT32 );
    ADD_SYMBOL( "INT64", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT64 );
    ADD_SYMBOL( "FLOAT32", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_FLOAT32 );
    ADD_SYMBOL( "FLOAT64", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_FLOAT64 );
    ADD_SYMBOL( "DYNAMIC", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_DYNAMIC );
    if( sizeof( COLOR ) == 1 ) ADD_SYMBOL( "PIXEL", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT8 );
    if( sizeof( COLOR ) == 2 ) ADD_SYMBOL( "PIXEL", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT16 );
    if( sizeof( COLOR ) == 4 ) ADD_SYMBOL( "PIXEL", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT32 );
    if( sizeof( COLOR ) == 8 ) ADD_SYMBOL( "PIXEL", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT64 );
    if( sizeof( PIX_INT ) == 4 ) ADD_SYMBOL( "INT", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT32 );
    if( sizeof( PIX_INT ) == 8 ) ADD_SYMBOL( "INT", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_INT64 );
    if( sizeof( PIX_FLOAT ) == 4 ) ADD_SYMBOL( "FLOAT", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_FLOAT32 );
    if( sizeof( PIX_FLOAT ) == 8 ) ADD_SYMBOL( "FLOAT", SYMTYPE_NUM_I, PIX_CONTAINER_TYPE_FLOAT64 );
    {
	const UTF8_CHAR* major = PIXILANG_VERSION_MAJOR;
	const UTF8_CHAR* minor = PIXILANG_VERSION_MINOR;
	ADD_SYMBOL( "PIXILANG_VERSION", SYMTYPE_NUM_I, string_to_int( major ) * 100000 + atof( minor ) * 100 );
    }
    ADD_SYMBOL( "OS_NAME", SYMTYPE_NUM_I, vm->os_name );
    ADD_SYMBOL( "ARCH_NAME", SYMTYPE_NUM_I, vm->arch_name );
    ADD_SYMBOL( "CURRENT_PATH", SYMTYPE_NUM_I, vm->current_path );
    ADD_SYMBOL( "USER_PATH", SYMTYPE_NUM_I, vm->user_path );
    ADD_SYMBOL( "TEMP_PATH", SYMTYPE_NUM_I, vm->temp_path );
#ifdef OPENGL
    ADD_SYMBOL( "OPENGL", SYMTYPE_NUM_I, 1 );
#else
    ADD_SYMBOL( "OPENGL", SYMTYPE_NUM_I, 0 );
#endif
    ADD_SYMBOL( "GL_SCREEN", SYMTYPE_NUM_I, PIX_GL_SCREEN );
    ADD_SYMBOL( "GL_ZBUF", SYMTYPE_NUM_I, PIX_GL_ZBUF );
    
    DPRINT( "Adding global variables...\n" );
    ADD_SYMBOL( "WINDOW_XSIZE", SYMTYPE_GVAR, PIX_GVAR_WINDOW_XSIZE );
    ADD_SYMBOL( "WINDOW_YSIZE", SYMTYPE_GVAR, PIX_GVAR_WINDOW_YSIZE );
    ADD_SYMBOL( "FPS", SYMTYPE_GVAR, PIX_GVAR_FPS );
    ADD_SYMBOL( "PPI", SYMTYPE_GVAR, PIX_GVAR_PPI );

    DPRINT( "Compilation: lexical tree generation...\n" );
    g_comp->root = node( lnode_statlist, 0 );
    if( g_comp->root == 0 )
    {
	rv = 7;
	goto compiler_end;
    }    
    if( yyparse() )
    {
	rv = 8;
	goto compiler_end;
    }
    //if( yyss ) { free( yyss ); yyss = 0; }
    //if( yyvs ) { free( yyvs ); yyvs = 0; }
    //yystacksize = 0;
    //Close last local symbol table:
    g_comp->root = remove_lsym_table( g_comp->root );
    if( g_comp->root == 0 )
    {
	rv = 9;
	goto compiler_end;
    }
    DPRINT( "Compilation: optimization...\n" );
    optimize_tree( g_comp->root );
    DPRINT( "Compilation: code generation...\n" );
    DPRINT( "0: HALT\n" );
    pix_vm_put_opcode( OPCODE_HALT, vm );
    vm->halt_addr = 0;
    compile_tree( g_comp->root );
    fix_up();
    remove_tree( g_comp->root );
    DPRINT( "%d: RET_i ( 0 << OB )\n", (int)g_comp->vm->code_ptr );
    pix_vm_put_opcode( OPCODE_RET_i, vm );
    
    for( size_t i = 0; i < g_comp->vm->vars_num; i++ )
    {
	int flags = g_comp->var_flags[ i ];
	if( flags & VAR_FLAG_USED )
	{
	    if( ( g_comp->var_flags[ i ] & VAR_FLAG_INITIALIZED ) == 0 )
	    {
	        dprint( "Variable %s is not initialized. Default value = 0.\n", pix_vm_get_variable_name( g_comp->vm, i ) );
	    }
	}
    }

    DPRINT( "Deinit...\n" );
    pix_symtab_deinit( &g_comp->sym );
    mem_free( g_comp->var_flags );
    mem_free( g_comp->lsym );
    mem_free( g_comp->inc );
    mem_free( g_comp->while_stack );
    mem_free( g_comp->fixup );
    mem_free( g_comp );
    g_comp = 0;

compiler_end:
	
    sundog_mutex_unlock( &pd->compiler_mutex );
        
compiler_end2:

    ticks_t end_time = time_ticks();
    
    DPRINT( "Pixilang compiler finished. %d ms.\n", ( ( end_time - start_time ) * 1000 ) / time_ticks_per_second() );
    
    if( rv )
    {
	ERROR( "%d", rv );
    }

    return rv;
}

#ifdef PIX_ENCODED_SOURCE
extern void pix_decode_source( void* src, size_t size );
#endif

int pix_compile( const UTF8_CHAR* name, pix_vm* vm, pix_data* pd )
{
    int rv = 0;
    
    UTF8_CHAR* src = 0;
    UTF8_CHAR* base_path = 0;
    
    size_t fsize = v3_get_file_size( name );
    if( fsize >= 8 )
    {
	V3_FILE f = v3_open( name, "rb" );
	if( f )
	{
	    char sign[ 9 ];
	    sign[ 8 ] = 0;
	    v3_read( &sign, 1, 8, f );
	    v3_close( f );
	    if( mem_strcmp( (const char*)sign, "PIXICODE" ) == 0 )
	    {
		//Binary code:
		base_path = pix_get_base_path( name );
		pix_vm_load_code( name, base_path, vm );
		goto pix_compile_end;
	    }
	}
    }
    if( fsize )
    {
	src = (UTF8_CHAR*)MEM_NEW2( fsize );
	if( src == 0 ) 
	{
	    rv = 1;
	    goto pix_compile_end;
	}
	V3_FILE f = v3_open( name, "rb" );
	if( f == 0 )
	{
	    rv = 2;
	    ERROR( "can't open file %s", name );
	    goto pix_compile_end;
	}
	v3_read( src, 1, fsize, f );	
	v3_close( f );
	base_path = pix_get_base_path( name );
#ifdef PIX_ENCODED_SOURCE
	pix_decode_source( src, fsize );
#endif
	if( pix_compile_from_memory( src, fsize, (UTF8_CHAR*)name, base_path, vm, pd ) )
	{
	    rv = 3;
	    goto pix_compile_end;
	}
    }
    else 
    {
	ERROR( "%s not found (or it's empty)", name );
	rv = 4;
    }
    
pix_compile_end:
    
    mem_free( src );
    mem_free( base_path );
    
    if( rv )
    {
	ERROR( "%d", rv );
    }
    
    return rv;
}
#line 3044 "pixilang_compiler.cpp"
/* allocate initial stack or double stack size, up to YYMAXDEPTH */
static int yygrowstack(void)
{
    int i;
    unsigned newsize;
    short *newss;
    YYSTYPE *newvs;

    if ((newsize = yystacksize) == 0)
        newsize = YYINITSTACKSIZE;
    else if (newsize >= YYMAXDEPTH)
        return -1;
    else if ((newsize *= 2) > YYMAXDEPTH)
        newsize = YYMAXDEPTH;

    i = yyssp - yyss;
    newss = (yyss != 0)
          ? (short *)realloc(yyss, newsize * sizeof(*newss))
          : (short *)malloc(newsize * sizeof(*newss));
    if (newss == 0)
        return -1;

    yyss  = newss;
    yyssp = newss + i;
    newvs = (yyvs != 0)
          ? (YYSTYPE *)realloc(yyvs, newsize * sizeof(*newvs))
          : (YYSTYPE *)malloc(newsize * sizeof(*newvs));
    if (newvs == 0)
        return -1;

    yyvs = newvs;
    yyvsp = newvs + i;
    yystacksize = newsize;
    yysslim = yyss + newsize - 1;
    return 0;
}

#define YYABORT  goto yyabort
#define YYREJECT goto yyabort
#define YYACCEPT goto yyaccept
#define YYERROR  goto yyerrlab

int
YYPARSE_DECL()
{
    int yym, yyn, yystate;
#if YYDEBUG
    const char *yys;

    if ((yys = getenv("YYDEBUG")) != 0)
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = YYEMPTY;
    yystate = 0;

    if (yyss == NULL && yygrowstack()) goto yyoverflow;
    yyssp = yyss;
    yyvsp = yyvs;
    yystate = 0;
    *yyssp = 0;

yyloop:
    if ((yyn = yydefred[yystate]) != 0) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, reading %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: state %d, shifting to state %d\n",
                    YYPREFIX, yystate, yytable[yyn]);
#endif
        if (yyssp >= yysslim && yygrowstack())
        {
            goto yyoverflow;
        }
        yystate = yytable[yyn];
        *++yyssp = yytable[yyn];
        *++yyvsp = yylval;
        yychar = YYEMPTY;
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;

    yyerror("syntax error");

    goto yyerrlab;

yyerrlab:
    ++yynerrs;

yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yyssp]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: state %d, error recovery shifting\
 to state %d\n", YYPREFIX, *yyssp, yytable[yyn]);
#endif
                if (yyssp >= yysslim && yygrowstack())
                {
                    goto yyoverflow;
                }
                yystate = yytable[yyn];
                *++yyssp = yytable[yyn];
                *++yyvsp = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: error recovery discarding state %d\n",
                            YYPREFIX, *yyssp);
#endif
                if (yyssp <= yyss) goto yyabort;
                --yyssp;
                --yyvsp;
            }
        }
    }
    else
    {
        if (yychar == 0) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, error recovery discards token %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
        yychar = YYEMPTY;
        goto yyloop;
    }

yyreduce:
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: state %d, reducing by rule %d (%s)\n",
                YYPREFIX, yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    if (yym)
        yyval = yyvsp[1-yym];
    else
        memset(&yyval, 0, sizeof yyval);
    switch (yyn)
    {
case 2:
#line 293 "pixilang_compiler.y"
	{
            DPRINT( "input stat\n" );
            resize_node( g_comp->root, g_comp->root->nn + 1 );
            g_comp->root->n[ g_comp->root->nn - 1 ] = yyvsp[0].n;
        }
break;
case 6:
#line 305 "pixilang_compiler.y"
	{ yyval.n = yyvsp[0].n; }
break;
case 7:
#line 307 "pixilang_compiler.y"
	{
	    yyval.n = node( lnode_empty, 2 );
	    yyval.n->n[ 0 ] = yyvsp[-2].n;
	    yyval.n->n[ 1 ] = yyvsp[0].n;
	}
break;
case 8:
#line 314 "pixilang_compiler.y"
	{ yyval.i = 0; }
break;
case 9:
#line 315 "pixilang_compiler.y"
	{ yyval.i = 1; }
break;
case 10:
#line 316 "pixilang_compiler.y"
	{ yyval.i = 2; }
break;
case 11:
#line 317 "pixilang_compiler.y"
	{ yyval.i = 3; }
break;
case 12:
#line 318 "pixilang_compiler.y"
	{ yyval.i = 4; }
break;
case 13:
#line 319 "pixilang_compiler.y"
	{ yyval.i = 5; }
break;
case 14:
#line 320 "pixilang_compiler.y"
	{ yyval.i = 6; }
break;
case 15:
#line 321 "pixilang_compiler.y"
	{ yyval.i = 7; }
break;
case 16:
#line 322 "pixilang_compiler.y"
	{ yyval.i = 8; }
break;
case 17:
#line 323 "pixilang_compiler.y"
	{ yyval.i = 9; }
break;
case 18:
#line 324 "pixilang_compiler.y"
	{ yyval.i = 10; }
break;
case 19:
#line 325 "pixilang_compiler.y"
	{ yyval.i = 11; }
break;
case 20:
#line 326 "pixilang_compiler.y"
	{ yyval.i = 12; }
break;
case 21:
#line 327 "pixilang_compiler.y"
	{ yyval.i = 13; }
break;
case 22:
#line 328 "pixilang_compiler.y"
	{ yyval.i = 14; }
break;
case 23:
#line 329 "pixilang_compiler.y"
	{ yyval.i = 15; }
break;
case 24:
#line 330 "pixilang_compiler.y"
	{ yyval.i = 16; }
break;
case 25:
#line 331 "pixilang_compiler.y"
	{ yyval.i = 17; }
break;
case 26:
#line 332 "pixilang_compiler.y"
	{ yyval.i = 18; }
break;
case 27:
#line 335 "pixilang_compiler.y"
	{ yyval.n = node( lnode_statlist, 0 ); }
break;
case 28:
#line 337 "pixilang_compiler.y"
	{
	    DPRINT( "statlist stat\n" );
            resize_node( yyvsp[-1].n, yyvsp[-1].n->nn + 1 );
	    yyvsp[-1].n->n[ yyvsp[-1].n->nn - 1 ] = yyvsp[0].n;
	    yyval.n = yyvsp[-1].n;
        }
break;
case 29:
#line 345 "pixilang_compiler.y"
	{ DPRINT( "HALT\n" ); yyval.n = node( lnode_halt, 0 ); }
break;
case 30:
#line 347 "pixilang_compiler.y"
	{ 
	    DPRINT( "VAR(%d) :\n", (int)yyvsp[-1].i ); 
	    if( g_comp->var_flags[ yyvsp[-1].i ] & VAR_FLAG_LABEL )
	    {
		SHOW_ERROR( "label %s is already defined", pix_vm_get_variable_name( g_comp->vm, yyvsp[-1].i ) );
                YYERROR;
	    }
	    if( g_comp->var_flags[ yyvsp[-1].i ] & VAR_FLAG_FUNCTION )
	    {
		SHOW_ERROR( "label %s is already defined as function", pix_vm_get_variable_name( g_comp->vm, yyvsp[-1].i ) );
                YYERROR;
	    }
	    yyval.n = node( lnode_label, 0 ); 
	    yyval.n->val.i = yyvsp[-1].i;
	    g_comp->var_flags[ yyvsp[-1].i ] |= VAR_FLAG_LABEL | VAR_FLAG_INITIALIZED;
	}
break;
case 31:
#line 363 "pixilang_compiler.y"
	{ DPRINT( "GO expr\n" ); yyval.n = node( lnode_go, 1 ); yyval.n->n[ 0 ] = yyvsp[0].n; }
break;
case 32:
#line 365 "pixilang_compiler.y"
	{ 
	    DPRINT( "GVAR(%d) = expr\n", (int)yyvsp[-2].i ); 
            yyval.n = node( lnode_save_to_var, 1 );
	    yyval.n->val.i = yyvsp[-2].i;
	    yyval.n->n[ 0 ] = yyvsp[0].n;
	    g_comp->var_flags[ yyvsp[-2].i ] |= VAR_FLAG_INITIALIZED;
        }
break;
case 33:
#line 373 "pixilang_compiler.y"
	{ 
	    DPRINT( "LVAR(%d) = expr\n", (int)yyvsp[-2].i );
	    yyval.n = node( lnode_save_to_stackframe, 1 );
	    yyval.n->val.i = yyvsp[-2].i;
	    yyval.n->n[ 0 ] = yyvsp[0].n;
        }
break;
case 34:
#line 380 "pixilang_compiler.y"
	{
	    DPRINT( "prop_expr = expr\n" );
	    yyval.n = yyvsp[-2].n;
	    yyval.n->type = lnode_save_to_prop;
	    resize_node( yyval.n, yyval.n->nn + 1 );
	    yyval.n->n[ 1 ] = yyvsp[0].n;
	}
break;
case 35:
#line 388 "pixilang_compiler.y"
	{
	    DPRINT( "prop_expr stat_math_op(%d) expr\n", (int)yyvsp[-1].i );
	    /*Create math operation:*/
	    lnode* op = node( (lnode_type)( lnode_sub + yyvsp[-1].i ), 2 ); 
	    op->n[ 0 ] = yyvsp[-2].n;
	    op->n[ 1 ] = yyvsp[0].n;
	    /*Result:*/
	    yyval.n = clone_tree( yyvsp[-2].n );
	    yyval.n->type = lnode_save_to_prop;
	    resize_node( yyval.n, yyval.n->nn + 1 );
	    yyval.n->n[ 1 ] = op;
	}
break;
case 36:
#line 401 "pixilang_compiler.y"
	{
	    DPRINT( "mem_expr = expr\n" );
	    yyval.n = yyvsp[-2].n;
	    yyval.n->type = lnode_save_to_mem;
	    resize_node( yyval.n, yyval.n->nn + 1 );
	    yyval.n->n[ 2 ] = yyvsp[0].n;
        }
break;
case 37:
#line 409 "pixilang_compiler.y"
	{
	    DPRINT( "mem_expr stat_math_op(%d) expr\n", (int)yyvsp[-1].i );
	    /*Create math operation:*/
	    lnode* op = node( (lnode_type)( lnode_sub + yyvsp[-1].i ), 2 ); 
	    op->n[ 0 ] = yyvsp[-2].n;
	    op->n[ 1 ] = yyvsp[0].n;
	    /*Result:*/
	    yyval.n = clone_tree( yyvsp[-2].n );
	    yyval.n->type = lnode_save_to_mem;
	    resize_node( yyval.n, yyval.n->nn + 1 );
	    yyval.n->n[ 2 ] = op;
	}
break;
case 38:
#line 422 "pixilang_compiler.y"
	{
	    DPRINT( "GVAR(%d) stat_math_op(%d) expr\n", (int)yyvsp[-2].i, (int)yyvsp[-1].i );
	    /*Create first operand:*/
	    lnode* n1 = make_expr_node_from_var( yyvsp[-2].i );
	    /*Create math operation:*/
	    lnode* n2 = node( (lnode_type)( lnode_sub + yyvsp[-1].i ), 2 ); 
	    n2->n[ 0 ] = n1; 
	    n2->n[ 1 ] = yyvsp[0].n;
	    /*Save result:*/
            yyval.n = node( lnode_save_to_var, 1 );
	    yyval.n->val.i = yyvsp[-2].i;
	    yyval.n->n[ 0 ] = n2;
	}
break;
case 39:
#line 436 "pixilang_compiler.y"
	{
	    DPRINT( "LVAR(%d) stat_math_op(%d) expr\n", (int)yyvsp[-2].i, (int)yyvsp[-1].i );
	    /*Create first operand:*/
	    lnode* n1 = make_expr_node_from_local_var( yyvsp[-2].i );
	    /*Create math operation:*/
	    lnode* n2 = node( (lnode_type)( lnode_sub + yyvsp[-1].i ), 2 ); 
	    n2->n[ 0 ] = n1; 
	    n2->n[ 1 ] = yyvsp[0].n;
	    /*Save result:*/
	    yyval.n = node( lnode_save_to_stackframe, 1 );
	    yyval.n->val.i = yyvsp[-2].i;
	    yyval.n->n[ 0 ] = n2;
	}
break;
case 40:
#line 450 "pixilang_compiler.y"
	{
	    DPRINT( "FNNUM(%d) ( exprlist )\n", (int)yyvsp[-3].i );
	    yyval.n = node( lnode_call_builtin_fn_void, 1 );
	    yyval.n->val.i = yyvsp[-3].i;
	    yyval.n->n[ 0 ] = yyvsp[-1].n;
        }
break;
case 41:
#line 457 "pixilang_compiler.y"
	{
	    DPRINT( "fn_expr (call void function. statement)\n" );
	    yyval.n = yyvsp[0].n;
	    yyval.n->type = lnode_call_void;
	}
break;
case 42:
#line 463 "pixilang_compiler.y"
	{
	    DPRINT( "RET\n" );
	    yyval.n = node( lnode_ret_int, 0 );
	    yyval.n->val.i = 0;
        }
break;
case 43:
#line 469 "pixilang_compiler.y"
	{
	    DPRINT( "RET ( expr )\n" );
	    yyval.n = node( lnode_ret, 1 );
	    yyval.n->n[ 0 ] = yyvsp[-1].n;
        }
break;
case 44:
#line 475 "pixilang_compiler.y"
	{ 
	    DPRINT( "IF expr statlist\n" );
	    yyval.n = node( lnode_if, 2 );
	    yyval.n->n[ 0 ] = yyvsp[-3].n;
	    yyval.n->n[ 1 ] = yyvsp[-1].n;
	    yyvsp[-1].n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER;
	}
break;
case 45:
#line 483 "pixilang_compiler.y"
	{ 
	    DPRINT( "IF expr statlist ELSE statlist\n" );
	    yyval.n = node( lnode_if_else, 3 );
	    yyval.n->n[ 0 ] = yyvsp[-7].n;
	    yyval.n->n[ 1 ] = yyvsp[-5].n;
	    yyval.n->n[ 2 ] = yyvsp[-1].n;
	    yyvsp[-5].n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER | LNODE_FLAG_STATLIST_SKIP_NEXT_HEADER;
	    yyvsp[-1].n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_HEADER;
	}
break;
case 46:
#line 493 "pixilang_compiler.y"
	{
	    DPRINT( "WHILE expr {\n" );
	    yyval.n = node( lnode_while, 3 );
	    yyval.n->n[ 0 ] = yyvsp[-1].n;
	    yyval.n->n[ 2 ] = node( lnode_jmp_to_node, 0 );
	    yyval.n->n[ 2 ]->val.p = yyvsp[-1].n;
	    /*Push it to stack:*/
	    if( g_comp->while_stack == 0 )
		g_comp->while_stack = (lnode**)MEM_NEW2( sizeof( lnode* ) );
	    g_comp->while_stack[ g_comp->while_stack_ptr ] = yyval.n;
	    g_comp->while_stack_ptr++;
	    if( g_comp->while_stack_ptr >= mem_get_size( g_comp->while_stack ) / sizeof( lnode* ) )
		g_comp->while_stack = (lnode**)mem_resize( g_comp->while_stack, mem_get_size( g_comp->while_stack ) + sizeof( lnode* ) );
	}
break;
case 47:
#line 508 "pixilang_compiler.y"
	{ 
	    DPRINT( "statlist } (while)\n" );
	    yyval.n = yyvsp[-2].n;
	    yyval.n->n[ 1 ] = yyvsp[-1].n;
	    yyvsp[-1].n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_IF_FALSE_HEADER | LNODE_FLAG_STATLIST_SKIP_NEXT_HEADER;
	    /*Pop from stack:*/
	    g_comp->while_stack_ptr--;
	}
break;
case 48:
#line 517 "pixilang_compiler.y"
	{ 
	    DPRINT( "BREAK (level %d)\n", (int)yyvsp[0].i );
	    if( g_comp->while_stack_ptr > 0 )
	    {
        	yyval.n = node( lnode_jmp_to_end_of_node, 0 );
		if( yyvsp[0].i == -1 )
        	    yyval.n->val.p = g_comp->while_stack[ 0 ]->n[ 2 ];
        	else
        	{
        	    if( g_comp->while_stack_ptr - yyvsp[0].i >= 0 )
        		yyval.n->val.p = g_comp->while_stack[ g_comp->while_stack_ptr - yyvsp[0].i ]->n[ 2 ];
        	    else
        	    {
            		SHOW_ERROR( "wrong level number %d for 'break' operator", (int)yyvsp[0].i );
            		YYERROR;
        	    }
        	}
    	    }
    	    else
    	    {
                SHOW_ERROR( "operator 'break' can not be outside of while loop" );
                YYERROR;
            }
        }
break;
case 49:
#line 542 "pixilang_compiler.y"
	{ 
	    DPRINT( "CONTINUE\n" );
	    if( g_comp->while_stack_ptr > 0 )
	    {
        	yyval.n = node( lnode_jmp_to_node, 0 );
        	yyval.n->val.p = g_comp->while_stack[ g_comp->while_stack_ptr - 1 ]->n[ 2 ];
    	    }
    	    else
    	    {
                SHOW_ERROR( "operator 'continue' can not be outside of while loop" );
                YYERROR;
            }
        }
break;
case 50:
#line 556 "pixilang_compiler.y"
	{
	    DPRINT( "function begin\n" );
	    /*Create new empty local symbol table:*/
	    create_empty_lsym_table();
	    g_comp->fn_pars_mode = 1;
	}
break;
case 51:
#line 563 "pixilang_compiler.y"
	{
	    g_comp->fn_pars_mode = 0;
	    if( g_comp->var_flags[ yyvsp[-3].i ] & VAR_FLAG_FUNCTION )
	    {
		SHOW_ERROR( "function %s is already defined", pix_vm_get_variable_name( g_comp->vm, yyvsp[-3].i ) );
                YYERROR;
	    }
	    if( g_comp->var_flags[ yyvsp[-3].i ] & VAR_FLAG_LABEL )
	    {
		SHOW_ERROR( "function %s is already defined as label", pix_vm_get_variable_name( g_comp->vm, yyvsp[-3].i ) );
                YYERROR;
	    }
	    g_comp->var_flags[ yyvsp[-3].i ] |= VAR_FLAG_FUNCTION | VAR_FLAG_INITIALIZED;
	}
break;
case 52:
#line 578 "pixilang_compiler.y"
	{
	    DPRINT( "function\n" );
	    /*Remove local symbol table:*/
	    yyval.n = remove_lsym_table( yyvsp[-1].n );
	    /*Add the header:*/
	    yyval.n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_HEADER;
            /*Add ret instruction to this statlist, because it is the function now:*/
            resize_node( yyval.n, yyval.n->nn + 2 );
	    yyval.n->n[ yyval.n->nn - 2 ] = node( lnode_ret_int, 0 );
	    yyval.n->n[ yyval.n->nn - 2 ]->val.i = 0;
	    /*Save address of this function to global variable:*/
	    yyval.n->n[ yyval.n->nn - 1 ] = node( lnode_function_label_from_node, 1 );
	    yyval.n->n[ yyval.n->nn - 1 ]->val.i = yyvsp[-7].i;
	    yyval.n->n[ yyval.n->nn - 1 ]->n[ 0 ] = node( lnode_empty, 0 );
	    yyval.n->n[ yyval.n->nn - 1 ]->n[ 0 ]->val.p = yyval.n;
	}
break;
case 53:
#line 595 "pixilang_compiler.y"
	{
	    if( (unsigned)yyvsp[0].i < (unsigned)PIX_VM_CONTAINERS )
	    {
		int name_size = g_comp->vm->c[ yyvsp[0].i ]->size;
		UTF8_CHAR* name = (UTF8_CHAR*)MEM_NEW2( name_size + 1 );
		mem_copy( name, g_comp->vm->c[ yyvsp[0].i ]->data, name_size );
		name[ name_size ] = 0;
		UTF8_CHAR* new_name = pix_compose_full_path( g_comp->base_path, name );
		mem_free( name );
		pix_vm_remove_container( yyvsp[0].i, g_comp->vm );
		DPRINT( "include \"%s\"\n", new_name );
		
		int fsize = v3_get_file_size( new_name );
		if( fsize == 0 )
		{
		    mem_free( new_name );
		    SHOW_ERROR( "%s not found", new_name );
		    YYERROR;
		}
		
		/*Save previous compiler state:*/
		if( g_comp->inc == 0 )
		{
		    g_comp->inc = (pix_include*)MEM_NEW2( 2 * sizeof( pix_include ) );
		}
		if( g_comp->inc_num >= mem_get_size( g_comp->inc ) / sizeof( pix_include ) )
		{
		    g_comp->inc = (pix_include*)mem_resize( g_comp->inc, ( g_comp->inc_num + 2 ) * sizeof( pix_include ) );
		}
		g_comp->inc[ g_comp->inc_num ].src = g_comp->src;
		g_comp->inc[ g_comp->inc_num ].src_ptr = g_comp->src_ptr;
		g_comp->inc[ g_comp->inc_num ].src_line = g_comp->src_line;
		g_comp->inc[ g_comp->inc_num ].src_size = g_comp->src_size;
		g_comp->inc[ g_comp->inc_num ].src_name = g_comp->src_name;
		g_comp->inc[ g_comp->inc_num ].base_path = g_comp->base_path;
		g_comp->inc_num++;
		
		/*Set new compiler state:*/
		g_comp->src = (UTF8_CHAR*)MEM_NEW2( fsize );
		V3_FILE f = v3_open( new_name, "rb" );
		v3_read( g_comp->src, 1, fsize, f );	
		v3_close( f );
		g_comp->src_ptr = 0;
		g_comp->src_line = 0;
		g_comp->src_size = fsize;
		g_comp->src_name = new_name;
		g_comp->base_path = pix_get_base_path( new_name );
		DPRINT( "New base path: %s\n", g_comp->base_path );
	    }
	    else 
	    {
		YYERROR;
	    }
	    yyval.n = node( lnode_empty, 0 );
	}
break;
case 54:
#line 652 "pixilang_compiler.y"
	{ yyval.n = node( lnode_exprlist, 0 ); }
break;
case 55:
#line 653 "pixilang_compiler.y"
	{ yyval.n = node( lnode_exprlist, 1 ); yyval.n->n[ 0 ] = yyvsp[0].n; }
break;
case 56:
#line 655 "pixilang_compiler.y"
	{
	    /*Add new node (expr) to list of parameters (exprlist):*/
	    resize_node( yyvsp[-2].n, yyvsp[-2].n->nn + 1 );
            yyvsp[-2].n->n[ yyvsp[-2].n->nn - 1 ] = yyvsp[0].n;
	    yyval.n = yyvsp[-2].n;
        }
break;
case 57:
#line 663 "pixilang_compiler.y"
	{ DPRINT( "NUM_I(%d)\n", (int)yyvsp[0].i ); yyval.n = node( lnode_int, 0 ); yyval.n->val.i = yyvsp[0].i; }
break;
case 58:
#line 664 "pixilang_compiler.y"
	{ DPRINT( "NUM_F(%d)\n", (int)yyvsp[0].f ); yyval.n = node( lnode_float, 0 ); yyval.n->val.f = yyvsp[0].f; }
break;
case 59:
#line 665 "pixilang_compiler.y"
	{ DPRINT( "GVAR(%d)\n", (int)yyvsp[0].i ); yyval.n = make_expr_node_from_var( yyvsp[0].i ); }
break;
case 60:
#line 666 "pixilang_compiler.y"
	{ DPRINT( "LVAR(%d)\n", (int)yyvsp[0].i ); yyval.n = make_expr_node_from_local_var( yyvsp[0].i ); }
break;
case 61:
#line 668 "pixilang_compiler.y"
	{
	    DPRINT( "FNNUM(%d) ( exprlist )\n", (int)yyvsp[-3].i );
	    yyval.n = node( lnode_call_builtin_fn, 1 );
	    yyval.n->val.i = yyvsp[-3].i;
	    yyval.n->n[ 0 ] = yyvsp[-1].n;
        }
break;
case 62:
#line 676 "pixilang_compiler.y"
	{ DPRINT( "basic_expr ( exprlist )\n" ); yyval.n = node( lnode_call, 2 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->n[ 1 ] = yyvsp[-3].n; }
break;
case 63:
#line 677 "pixilang_compiler.y"
	{ DPRINT( "fn_expr ( exprlist )\n" ); yyval.n = node( lnode_call, 2 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->n[ 1 ] = yyvsp[-3].n; }
break;
case 64:
#line 678 "pixilang_compiler.y"
	{ DPRINT( "mem_expr ( exprlist )\n" ); yyval.n = node( lnode_call, 2 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->n[ 1 ] = yyvsp[-3].n; }
break;
case 65:
#line 679 "pixilang_compiler.y"
	{ DPRINT( "prop_expr ( exprlist )\n" ); yyval.n = node( lnode_call, 2 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->n[ 1 ] = yyvsp[-3].n; }
break;
case 66:
#line 682 "pixilang_compiler.y"
	{ DPRINT( "basic_expr [ mem_offset ]\n" ); yyval.n = node( lnode_load_from_mem, 2 ); yyval.n->n[ 0 ] = yyvsp[-3].n; yyval.n->n[ 1 ] = yyvsp[-1].n; }
break;
case 67:
#line 683 "pixilang_compiler.y"
	{ DPRINT( "fn_expr [ mem_offset ]\n" ); yyval.n = node( lnode_load_from_mem, 2 ); yyval.n->n[ 0 ] = yyvsp[-3].n; yyval.n->n[ 1 ] = yyvsp[-1].n; }
break;
case 68:
#line 684 "pixilang_compiler.y"
	{ DPRINT( "mem_expr [ mem_offset ]\n" ); yyval.n = node( lnode_load_from_mem, 2 ); yyval.n->n[ 0 ] = yyvsp[-3].n; yyval.n->n[ 1 ] = yyvsp[-1].n; }
break;
case 69:
#line 685 "pixilang_compiler.y"
	{ DPRINT( "prop_expr [ mem_offset ]\n" ); yyval.n = node( lnode_load_from_mem, 2 ); yyval.n->n[ 0 ] = yyvsp[-3].n; yyval.n->n[ 1 ] = yyvsp[-1].n; }
break;
case 70:
#line 688 "pixilang_compiler.y"
	{ DPRINT( "basic_expr.%d\n", (int)yyvsp[0].i ); yyval.n = node( lnode_load_from_prop, 1 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->val.i = yyvsp[0].i; }
break;
case 71:
#line 689 "pixilang_compiler.y"
	{ DPRINT( "fn_expr.%d\n", (int)yyvsp[0].i ); yyval.n = node( lnode_load_from_prop, 1 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->val.i = yyvsp[0].i; }
break;
case 72:
#line 690 "pixilang_compiler.y"
	{ DPRINT( "mem_expr.%d\n", (int)yyvsp[0].i ); yyval.n = node( lnode_load_from_prop, 1 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->val.i = yyvsp[0].i; }
break;
case 73:
#line 691 "pixilang_compiler.y"
	{ DPRINT( "prop_expr.%d\n", (int)yyvsp[0].i ); yyval.n = node( lnode_load_from_prop, 1 ); yyval.n->n[ 0 ] = yyvsp[-1].n; yyval.n->val.i = yyvsp[0].i; }
break;
case 74:
#line 694 "pixilang_compiler.y"
	{ DPRINT( "basic expression\n" ); yyval.n = yyvsp[0].n; }
break;
case 75:
#line 695 "pixilang_compiler.y"
	{ DPRINT( "fn_expr\n" ); yyval.n = yyvsp[0].n; }
break;
case 76:
#line 696 "pixilang_compiler.y"
	{ DPRINT( "mem_expr\n" ); yyval.n = yyvsp[0].n; }
break;
case 77:
#line 697 "pixilang_compiler.y"
	{ DPRINT( "prop_expr\n" ); yyval.n = yyvsp[0].n; }
break;
case 78:
#line 699 "pixilang_compiler.y"
	{ 
	    DPRINT( "statlist begin (expr)\n" );
	    /*Create new empty local symbol table:*/
	    create_empty_lsym_table();
	}
break;
case 79:
#line 705 "pixilang_compiler.y"
	{ 
	    DPRINT( "statlist (expr)\n" );
	    /*Remove local symbol table:*/
	    yyval.n = remove_lsym_table( yyvsp[-1].n );
	    /*Add the header:*/
	    yyval.n->flags |= LNODE_FLAG_STATLIST_WITH_JMP_HEADER | LNODE_FLAG_STATLIST_AS_EXPRESSION;
            /*Add ret instruction to this statlist, because it is the function now:*/
            resize_node( yyval.n, yyval.n->nn + 1 );
	    yyval.n->n[ yyval.n->nn - 1 ] = node( lnode_ret_int, 0 );
	    yyval.n->n[ yyval.n->nn - 1 ]->val.i = 0;
        }
break;
case 80:
#line 716 "pixilang_compiler.y"
	{ yyval.n = yyvsp[-1].n; }
break;
case 81:
#line 717 "pixilang_compiler.y"
	{ DPRINT( "SUB\n" ); yyval.n = node( lnode_sub, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 82:
#line 718 "pixilang_compiler.y"
	{ DPRINT( "ADD\n" ); yyval.n = node( lnode_add, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 83:
#line 719 "pixilang_compiler.y"
	{ DPRINT( "MUL\n" ); yyval.n = node( lnode_mul, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 84:
#line 720 "pixilang_compiler.y"
	{ DPRINT( "IDIV\n" ); yyval.n = node( lnode_idiv, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 85:
#line 721 "pixilang_compiler.y"
	{ DPRINT( "DIV\n" ); yyval.n = node( lnode_div, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 86:
#line 722 "pixilang_compiler.y"
	{ DPRINT( "MOD\n" ); yyval.n = node( lnode_mod, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 87:
#line 723 "pixilang_compiler.y"
	{ DPRINT( "AND\n" ); yyval.n = node( lnode_and, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 88:
#line 724 "pixilang_compiler.y"
	{ DPRINT( "OR\n" ); yyval.n = node( lnode_or, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 89:
#line 725 "pixilang_compiler.y"
	{ DPRINT( "XOR\n" ); yyval.n = node( lnode_xor, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 90:
#line 726 "pixilang_compiler.y"
	{ DPRINT( "ANDAND\n" ); yyval.n = node( lnode_andand, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 91:
#line 727 "pixilang_compiler.y"
	{ DPRINT( "OROR\n" ); yyval.n = node( lnode_oror, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 92:
#line 728 "pixilang_compiler.y"
	{ DPRINT( "EQ\n" ); yyval.n = node( lnode_eq, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 93:
#line 729 "pixilang_compiler.y"
	{ DPRINT( "NEQ\n" ); yyval.n = node( lnode_neq, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 94:
#line 730 "pixilang_compiler.y"
	{ DPRINT( "LESS\n" ); yyval.n = node( lnode_less, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 95:
#line 731 "pixilang_compiler.y"
	{ DPRINT( "LEQ\n" ); yyval.n = node( lnode_leq, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 96:
#line 732 "pixilang_compiler.y"
	{ DPRINT( "GREATER\n" ); yyval.n = node( lnode_greater, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 97:
#line 733 "pixilang_compiler.y"
	{ DPRINT( "GEQ\n" ); yyval.n = node( lnode_geq, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 98:
#line 734 "pixilang_compiler.y"
	{ DPRINT( "LSHIFT\n" ); yyval.n = node( lnode_lshift, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 99:
#line 735 "pixilang_compiler.y"
	{ DPRINT( "RSHIFT\n" ); yyval.n = node( lnode_rshift, 2 ); yyval.n->n[ 0 ] = yyvsp[-2].n; yyval.n->n[ 1 ] = yyvsp[0].n; }
break;
case 100:
#line 736 "pixilang_compiler.y"
	{ DPRINT( "NEG\n" ); yyval.n = node( lnode_neg, 1 ); yyval.n->n[ 0 ] = yyvsp[0].n; }
break;
#line 3907 "pixilang_compiler.cpp"
    }
    yyssp -= yym;
    yystate = *yyssp;
    yyvsp -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: after reduction, shifting from state 0 to\
 state %d\n", YYPREFIX, YYFINAL);
#endif
        yystate = YYFINAL;
        *++yyssp = YYFINAL;
        *++yyvsp = yyval;
        if (yychar < 0)
        {
            if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
            if (yydebug)
            {
                yys = 0;
                if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
                if (!yys) yys = "illegal-symbol";
                printf("%sdebug: state %d, reading %d (%s)\n",
                        YYPREFIX, YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == 0) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: after reduction, shifting from state %d \
to state %d\n", YYPREFIX, *yyssp, yystate);
#endif
    if (yyssp >= yysslim && yygrowstack())
    {
        goto yyoverflow;
    }
    *++yyssp = (short) yystate;
    *++yyvsp = yyval;
    goto yyloop;

yyoverflow:
    yyerror("yacc stack overflow");

yyabort:
    return (1);

yyaccept:
    return (0);
}
